Type: Package
Package: sprtt
Title: Sequential Probability Ratio Tests Toolbox
Version: 0.3.1
Date: 2026-04-13
Authors@R: 
    c(person(given = "Meike",
             family = "Snijder-Steinhilber",
             role = c("aut", "cre"),
             email = "Meike.Steinhilber@aol.com",
             comment = c(ORCID = "0000-0002-7144-2100")),
      person(given = "Martin",
             family = "Schnuerch",
             role = c("aut", "ths"),
             comment = c(ORCID = "0000-0001-6531-2265")),
      person(given = "Anna-Lena",
             family = "Schubert",
             role = c("aut", "ths"),
             comment = c(ORCID = "0000-0001-7248-0662"))
      )
Maintainer: Meike Snijder-Steinhilber <Meike.Steinhilber@aol.com>
Description: A toolbox for Sequential Probability Ratio Tests (SPRT) based on
    Wald (1945) <doi:10.2134/agronj1947.00021962003900070011x>.
    SPRTs are applied during the sampling process, ideally after each
    observation, and at every stage return a decision to either continue
    sampling or terminate and accept one of the specified hypotheses.
    The `seq_ttest()` function performs one-sample, two-sample, and paired
    t-tests for one- and two-sided hypotheses (Schnuerch & Erdfelder (2019)
    <doi:10.1037/met0000234>). The `seq_anova()` function performs a sequential
    one-way fixed effects ANOVA (Steinhilber et al. (2024)
    <doi:10.1037/met0000677>). The `plan_sample_size()` function helps plan
    sequential studies by simulating required sample sizes across a range of
    effect sizes. For more information, see the vignettes
    browseVignettes(package = "sprtt") or the package website
    <https://meikesteinhilber.github.io/sprtt/>.
License: AGPL (>= 3)
URL: https://meikesteinhilber.github.io/sprtt/
BugReports: https://github.com/MeikeSteinhilber/sprtt/issues
Depends: R (>= 3.5.0)
Imports: rappdirs, methods, stats, dplyr, MBESS, purrr, glue, ggplot2,
        lifecycle, knitr, rmarkdown, gt, ggtext, scales, piggyback
Suggests: testthat (>= 3.0.0), testthis, withr, xml2, effsize,
        effectsize, vdiffr, diffviewer
VignetteBuilder: knitr
Encoding: UTF-8
Language: en-US
LazyData: yes
LazyDataCompression: xz
RoxygenNote: 7.3.3
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-05-06 11:58:02 UTC; msteinhi
Author: Meike Snijder-Steinhilber [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-7144-2100>),
  Martin Schnuerch [aut, ths] (ORCID:
    <https://orcid.org/0000-0001-6531-2265>),
  Anna-Lena Schubert [aut, ths] (ORCID:
    <https://orcid.org/0000-0001-7248-0662>)
Repository: CRAN
Date/Publication: 2026-05-06 12:20:02 UTC
Built: R 4.6.0; ; 2026-05-06 16:44:06 UTC; unix

# Generated by roxygen2: do not edit by hand

export(cache_clear)
export(cache_info)
export(download_sample_size_data)
export(draw_sample_mixture)
export(draw_sample_normal)
export(load_sample_size_data)
export(plan_sample_size)
export(plot_anova)
export(seq_anova)
export(seq_ttest)
exportMethods("[")
import(dplyr)
import(ggplot2)
import(glue)
import(gt)
import(purrr)
importFrom(MBESS,conf.limits.ncf)
importFrom(ggtext,element_markdown)
importFrom(glue,glue)
importFrom(knitr,opts_chunk)
importFrom(lifecycle,deprecated)
importFrom(methods,callNextMethod)
importFrom(methods,new)
importFrom(methods,validObject)
importFrom(scales,percent)
importFrom(scales,percent_format)
importFrom(stats,as.formula)
importFrom(stats,df)
importFrom(stats,dt)
importFrom(stats,model.frame)
importFrom(stats,rnorm)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,t.test)

<!-- ATTENTION! see below -->

# sprtt

<!-- badges: start -->

[![CRAN_Status_Badge](http://www.r-pkg.org/badges/version/sprtt?color=green)](https://cran.r-project.org/package=sprtt)
[![total](https://cranlogs.r-pkg.org/badges/grand-total/sprtt)](https://cran.r-project.org/package=sprtt)
[![monthly](https://cranlogs.r-pkg.org/badges/sprtt)](https://cran.r-project.org/package=sprtt)
[![codecov](https://codecov.io/gh/MeikeSteinhilber/sprtt/branch/main/graph/badge.svg?token=IQHTDTRBAW)](https://app.codecov.io/gh/MeikeSteinhilber/sprtt)
[![pkgdown](https://github.com/MeikeSteinhilber/sprtt/actions/workflows/pkgdown-pak.yaml/badge.svg)](https://github.com/MeikeSteinhilber/sprtt/actions/workflows/pkgdown-pak.yaml)
[![R-CMD-check](https://github.com/MeikeSteinhilber/sprtt/actions/workflows/R-CMD-check-windows-macOs.yaml/badge.svg)](https://github.com/MeikeSteinhilber/sprtt/actions/workflows/R-CMD-check-windows-macOs.yaml)

<!-- badges: end -->

> ⚠️⚠️⚠️ **The documentation and website are currently under
> construction for the next release.** ⚠️⚠️⚠️

## Overview

The `sprtt` package is a **s**equential **p**robability **r**atio
**t**ests **t**oolbox (**sprtt**).

The package contains:

- `seq_ttest()`, `seq_anova()` calculates sequential *t*-test and
  sequential one-way ANOVAs

- `plot_anova()` plots results of sequential ANOVAs

- `plan_sample_size()` creates comprehensive HTML reports with sample
  size recommendations based on pre-computed SPRT simulations

- `draw_sample_normal()`, `draw_sample_mixture()` simulation of data
  sets

- three data sets (`df_income`, `df_stress`, `df_cancer`) to run the
  examples in the *t*-test documentation

## Installation

### Release version from CRAN

This is the recommended version for a normal user.

``` r
# installs the package
install.packages("sprtt")
```

### Development version from GitHub

To get a bug fix or to use a feature from the development version, you
can install the latest version from GitHub.

``` r
# the installation requires the "devtools" package
# install.packages("devtools")
# stable GitHub version
devtools::install_github("MeikeSteinhilber/sprtt")

# development version, may not be stable
devtools::install_github("MeikeSteinhilber/sprtt", ref="develop")
```

## Documentation

Detailed documentation can be found on the [home
page](https://meikesteinhilber.github.io/sprtt/index.html). There are
several articles (called vignettes in R) covering the usage of the
package and its functions.

Short examples can be found in the following paragraph.

### Quick Examples

> **Note**
>
> In the R code sections:
>
> `# comment`: is a comment
>
> `function()`: is R code
>
> `#> results of function()`: is console output

``` r
# set seed --------------------------------------------------------------------
set.seed(333)

# load library ----------------------------------------------------------------
library(sprtt)
#> Note: Sample size planning functions require simulation data (~150 MB).
#> Data will download automatically on first use.
#> See ?download_sample_size_data for more information.

# t-TEST ----------------------------------------------------------------------
# one sample: numeric input ---------------------------------------------------
treatment_group <- rnorm(20, mean = 0, sd = 1)
results <- seq_ttest(treatment_group, mu = 1, d = 0.6)

# get access to the slots -----------------------------------------------------
# @ Operator
results@likelihood_ratio
#> [1] 323.9051

# [] Operator
results["likelihood_ratio"]
#> [1] 323.9051

# ANOVA -----------------------------------------------------------------------
# simulate data ---------------------------------------------------------------
set.seed(333)
data <- sprtt::draw_sample_normal(k_groups = 3,
                                  f = 0.15,
                                  sd = c(1, 1, 1),
                                  max_n = 46)
# look at the data
head(data)
#>            y x
#> 1  0.2762330 1
#> 2 -1.3415005 2
#> 3 -0.4521177 3
#> 4  1.2315479 1
#> 5  0.8180678 2
#> 6  0.1697634 3

# calculate sequential ANOVA --------------------------------------------------
results <- sprtt::seq_anova(y ~ x, f = 0.15, data = data, plot = TRUE)
# test decision
results@decision
#> [1] "accept H1"
# test results
results
#> 
#> *****  Sequential ANOVA *****
#> 
#> formula: y ~ x
#> test statistic:
#>  log-likelihood ratio = 3.061, decision = accept H1
#> SPRT thresholds:
#>  lower log(B) = -2.944, upper log(A) = 2.944
#> Log-Likelihood of the:
#>  alternative hypothesis = -3.839
#>  null hypothesis = -6.9
#> alternative hypothesis: true difference in means is not equal to 0.
#> specified effect size: Cohen's f = 0.15
#> empirical Cohen's f = 0.3255325, 95% CI[0.1324216, 0.4835088]
#> Cohen's f adjusted = 0.296
#> degrees of freedom: df1 = 2, df2 = 135
#> SS effect = 12.83987, SS residual = 121.1635, SS total = 134.0034
#> *Note: to get access to the object of the results use the @ or [] instead of the $ operator.

# plot results -----------------------------------------------------------------
sprtt::plot_anova(results)
```

<img src="man/figures/readme_example.png" width="100%" />

``` r

# save the plot ----------------------------------------------------------------
ggplot2::ggsave(
  "man/figures/readme_example.png",       # define the path
   units = "cm", height = 10, width = 16, # define the image size
   dpi = 300                              # define resolution
  )
```

<!--
&#10;ATTENTION!
in the .md file must be changed by hand. Otherwise the picture is not found on the website and CRAN.
&#10;<img src="man/figures/readme_example.png" width="100%" />
-->

### Sample Size Planning

The sample size planning function requires simulation data (~150 MB). On
first use, this data will be downloaded and chached automatically:

``` r
sprtt::plan_sample_size(f_expected = 0.25,
                        k_groups = 3,
                        beta = 0.05,
                        decision_rate = 0.85)
```

The package contains functions to help managing the cached data, if
necessary.

**Managing cached data**

``` r
# Check cache status
sprtt::cache_info()

# Force re-download (if data was updated)
sprtt::download_sample_size_data(force = TRUE)

# Clear cache
sprtt::cache_clear()
```

If you need to use the package offline, download the data while
connected beforehand.

# simulate data for the example ------------------------------------------------
set.seed(3)
data <- sprtt::draw_sample_normal(3, f = 0.25, max_n = 50)

# calculate the SPRT -----------------------------------------------------------
anova_results <- sprtt::seq_anova(y~x, f = 0.25, data = data, plot = TRUE)

# plot the results -------------------------------------------------------------
# default settings
sprtt::plot_anova(anova_results)
# variant 1
sprtt::plot_anova(anova_results,
                 labels = TRUE,
                 position_labels_x = 0.05,
                 position_lr_x = 150,
                 position_lr_y = 0,
                 highlight_color = "green"
                 )
# variant 2
sprtt::plot_anova(anova_results,
                  labels = TRUE,
                  position_labels_x = 0.15,
                  position_labels_y = 0.2,
                  position_lr_x = 60,
                  position_lr_y = 1,
                  font_size = 25,
                  line_size = 2,
                  highlight_color = "darkred"
)
# no labels
sprtt::plot_anova(anova_results,
                 labels = FALSE
                 )
# custom additions
sprtt::plot_anova(anova_results) +
  ggplot2::geom_vline(xintercept = 66, linewidth = 1, linetype = "dashed")

# further information ----------------------------------------------------------
# run this code:
vignette("one_way_anova", package = "sprtt")

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  echo = TRUE,
  eval = FALSE,
  message = TRUE,
  warning = FALSE,
  collapse = TRUE,
  comment = "#>"
)

## ----echo=FALSE, results='asis'-----------------------------------------------
# cat(sprintf(
#   'Source code of the simulation database: <a href="https://github.com/MeikeSteinhilber/sprtt_plan_sample_size" target="_blank" rel="noopener"><img src="https://img.shields.io/badge/GitHub-MeikeSteinhilber/sprtt__plan__sample__size-blue?logo=github" alt="GitHub" style="vertical-align:middle;"></a></span></p>'
# ))

## ----first-report-------------------------------------------------------------
# plan_sample_size(
#   f_expected = 0.25,   # Expected effect size
#   k_groups = 3,        # Number of groups
#   beta = 0.05,         # beta error rate
#   decision_rate = 0.85 # desired percentage of decisions
# )

## ----error-example, eval=FALSE------------------------------------------------
# # This will produce an error:
# plan_sample_size(f_expected = 0.22, k_groups = 3)
# #> Error: `f_expected` = 0.22 is not available.
# #> Please choose one of 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, or 0.4

## ----example-effect-sizes-----------------------------------------------------
# # report 1
# plan_sample_size(f_expected = 0.15, k_groups = 3, beta = 0.05)
# 
# # report 2
# plan_sample_size(f_expected = 0.35, k_groups = 3, beta = 0.05)

## ----example-custom-location--------------------------------------------------
# plan_sample_size(
#   f_expected = 0.25,
#   k_groups = 4,
#   output_dir = "~/Documents/research/sample_size_planning",
#   output_file = "study1_anova.html",
#   open = TRUE
# )

## ----workflow-scenarios-------------------------------------------------------
# # Define scenarios to compare
# scenarios <- data.frame(
#   effect = c(0.15, 0.20, 0.25),
#   label = c("conservative", "expected", "optimistic")
# )
# 
# # Generate reports for each scenario
# for (i in 1:(nrow(scenarios))) {
#   plan_sample_size(
#     f_expected = scenarios$effect[i],
#     k_groups = 3,
#     beta = 0.10,
#     output_dir = "sample_size_reports",
#     output_file = sprintf("plan_sample_size_%s.html", scenarios$label[i]),
#     open = FALSE,  # Don't open each one
#     overwrite = TRUE
#   )
# }
# 
# message("Generated ", nrow(scenarios), " sample size reports")

## ----download-explicit--------------------------------------------------------
# # Download simulation data manually
# download_sample_size_data()

## ----download-force-----------------------------------------------------------
# download_sample_size_data(force = TRUE)

## ----cache-status-------------------------------------------------------------
# cache_info()

## ----cache-clear--------------------------------------------------------------
# cache_clear()

## ----load-data-advanced-------------------------------------------------------
# # Load the complete simulation dataset (downloads automatically if not yet cached)
# loaded <- load_sample_size_data()
# 
# # Access the simulation data frame
# df_all <- loaded$data
# 
# # Check which dataset version this report is based on
# loaded$description  # short description
# loaded$version      # e.g. "v0.1.0-data"
# loaded$created      # date the dataset was created
# loaded$n_rep        # number of simulation iterations per condition

## ----echo = FALSE, message = FALSE--------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE
)

## ----example-0.1--------------------------------------------------------------
set.seed(333)
# Generate data with a medium effect
data <- sprtt::draw_sample_normal(
  k = 3,           # number of groups
  f = 0.25,        # effect size (Cohen's f)
  max_n = 22       # maximum sample size per group
)

## ----example-0.1b-------------------------------------------------------------
# View the first few rows
head(data)

# Check the sample sizes per group for the first 6 (2*k) data points
table(data$x[1:6])

## ----example-0.2--------------------------------------------------------------
# Calculate the sequential ANOVA
anova_results <- sprtt::seq_anova(
  y ~ x,
  f = 0.25,
  data = data[1:6, ],
  verbose = FALSE
)

# View results
anova_results

# Access the decision
anova_results@decision

## ----example-0.3--------------------------------------------------------------
# Calculate sequential ANOVA with larger sample
anova_results <- sprtt::seq_anova(
  y ~ x,
  f = 0.25,
  data = data[1:20, ],
  verbose = FALSE
)

# View results
anova_results

# Check decision
anova_results@decision

## ----example-0.4--------------------------------------------------------------
# Calculate sequential ANOVA with complete dataset
anova_results <- sprtt::seq_anova(
  y ~ x,
  f = 0.25,
  data = data,
  verbose = TRUE
)

# View full results
anova_results

## ----example-0.5--------------------------------------------------------------
# Access the decision
anova_results@decision

# Access the likelihood ratio
anova_results@likelihood_ratio

# Access the total sample size
anova_results@total_sample_size

## ----example-1----------------------------------------------------------------
set.seed(333)
data <- sprtt::draw_sample_normal(3, f = 0.25, max_n = 22)

# calculate the SPRT -----------------------------------------------------------
# Default: plot = TRUE with seq_steps = "single"
anova_results <- sprtt::seq_anova(y~x, f = 0.25,
                                  data = data, plot = TRUE)
# Explicitly specify seq_steps = "single"
anova_results <- sprtt::seq_anova(y~x, f = 0.25,
                                  data = data, plot = TRUE,
                                  seq_steps = "single")
# Use balanced sequential steps
anova_results <- sprtt::seq_anova(y~x, f = 0.25,
                                  data = data, plot = TRUE,
                                  seq_steps = "balanced")

# plot the results -------------------------------------------------------------
sprtt::plot_anova(anova_results)

## ----example-2----------------------------------------------------------------
set.seed(333)
# Generate unbalanced data with a 1:1:2 sampling ratio -------------------------
data <- sprtt::draw_sample_normal(3, f = 0.25, max_n = 37, sample_ratio = c(1,1,2))
# Randomize the order to get a more realistic data collection
data <- data[sample(nrow(data)),] 

# Calculate the SPRT with custom sequential steps ------------------------------
anova_results <- sprtt::seq_anova(
                          y~x,
                          f = 0.25,
                          data = data,
                          plot = TRUE,
                          # Start at n=12, then test after each observation
                          seq_steps = 12:nrow(data)) 

# Plot the results with custom styling -----------------------------------------
sprtt::plot_anova(anova_results,
                 labels = TRUE,
                 position_labels_x = 0.2,
                 position_labels_y = 0.2,
                 position_lr_x = 100,
                 position_lr_y = 1.8,
                 font_size = 20,
                 line_size = 1,
                 highlight_color = "steelblue"
                 )

set.seed(333)

data <- sprtt::draw_sample_normal(
  k_groups = 2,
  f = 0.20,
  max_n = 2
)
data

data <- sprtt::draw_sample_normal(
  k_groups = 4,
  f = 0,
  max_n = 2,
  sd = c(1, 2, 1, 8)
)
data

data <- sprtt::draw_sample_normal(
  k_groups = 3,
  f = 0.40,
  max_n = 2,
  sd = c(1, 0.8, 1),
  sample_ratio = c(1, 2, 3)
)
data

set.seed(333)

data <- sprtt::draw_sample_mixture(
  k_groups = 2,
  f = 0.40,
  max_n = 2
)
data

data <- sprtt::draw_sample_mixture(
  k_groups = 4,
  f = 1.2, # very large effect size
  max_n = 4,
  counter_n = 1000, # increase of counter is necessary
  verbose = TRUE # prints more information to the console
)
data

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(sprtt)

## ----echo=TRUE----------------------------------------------------------------
d <- 0.2

## ----echo=TRUE----------------------------------------------------------------
alpha <- 0.05
power <- 0.95

## ----echo=TRUE----------------------------------------------------------------
paired <- TRUE

## ----echo=TRUE----------------------------------------------------------------
alternative <- "greater"

## ----echo=TRUE----------------------------------------------------------------
# first data from the Human Resources department ---
# current sample size
n_person <- 2

# get data
df <- df_stress[1:n_person,]

# print data
df

# sequential t-test
results <- seq_ttest(df$one_year_stress,
                     df$baseline_stress,
                     alpha = alpha,
                     power = power,
                     d = d,
                     paired = paired,
                     alternative = alternative,
                     verbose = FALSE)

# print results: console output
results

## ----echo=TRUE----------------------------------------------------------------
results@decision

## ----echo=TRUE----------------------------------------------------------------
# new data from the Human Resources department ---
# get one more person
n_person <- n_person + 1
df <- df_stress[1:n_person,]

# print new data
df

# sequential t-test
results <- seq_ttest(df$one_year_stress,
                     df$baseline_stress,
                     alpha = alpha,
                     power = power,
                     d = d,
                     paired = paired,
                     alternative = alternative,
                     verbose = FALSE)

# print results
results@decision

## ----echo=TRUE----------------------------------------------------------------
# define the starting point
decision <- "continue sampling"
n_person <- 3

# simulation of the sequential procedure
while(decision == "continue sampling") {
  # get the current data
  df <- df_stress[1:n_person,]
  
  # run the sequential test and save the results
  results <- seq_ttest(df$one_year_stress,
                       df$baseline_stress,
                       alpha = alpha,
                       power = power,
                       d = d,
                       paired = paired,
                       alternative = alternative)
  # save the current desicion
  decision <- results@decision
  
  # add a new person
  n_person <- n_person + 1
  
  # break if the maximum of the data is reached
  if (n_person > nrow(df_stress)) {
    break
  }
}

# console output
results

## ----echo=TRUE----------------------------------------------------------------
# Required results for the report

# likelihood ratio (LR)
LR <- round(results@likelihood_ratio, digits = 2)
LR

# sample size (N) = degrees of freedom +2 (two-samples) or +1 (one-sample & paired)
N <- results@df + 1
N

# baseline stress (M and SD)
mean_t1 <- round(mean(df$baseline_stress), digits = 2)
mean_t1
sd_t1 <- round(sd(df$baseline_stress), digits = 2)
sd_t1

# after one year stress (M and SD)
mean_t2 <- round(mean(df$one_year_stress), digits = 2)
mean_t2
sd_t2 <- round(sd(df$one_year_stress), digits = 2)
sd_t2

# NOT INCLUDED IN THE PACKAGE 

# calculate effect size: Cohen´s d
d_results <- effsize::cohen.d(df$one_year_stress,
                 df$baseline_stress,
                 paired = TRUE)
d <- round(d_results$estimate, digits = 2)
d

# confidence intervall
d_lower <- round(d_results$conf.int[[1]], digits = 2)
d_lower
d_upper <- round(d_results$conf.int[[2]], digits = 2)
d_upper

# Add this after the monthly calculation:
monthly_complete <- monthly %>%
  filter(month < floor_date(Sys.Date(), "month"))  # exclude current incomplete month

ggplot(monthly, aes(x = month, y = monthly_downloads)) +
  geom_col(fill = "#92B4D3", alpha = 0.7, width = 25) +
  # Trend line only on complete months
  geom_smooth(
    data     = monthly_complete,
    method   = "loess",
    span     = 0.75,
    se       = TRUE,
    color    = "#1E4D7B",
    fill     = "#1E4D7B",
    alpha    = 0.15,
    linewidth = 0.8
  ) +
  geom_vline(
    data     = releases,
    aes(xintercept = date),
    linetype = "dashed",
    color    = "#555555",
    linewidth = 0.4
  ) +
  geom_label(
    data       = releases,
    aes(x = date, y = max(monthly$monthly_downloads) * 1.05, label = version),
    size       = 3,
    color      = "#555555",
    label.size = 0.2,
    hjust      = 0.5
  ) +
  annotate(
    "text",
    x     = as.Date("2025-01-01"),
    y     = max(monthly$monthly_downloads) * 1.05,
    label = paste0("Total downloads: ", format(avg_monthly$total, big.mark = ","), "\n",
                   "12-month avg: ", round(avg_recent$mean_last12), "/month"),
    hjust = 0,
    vjust = 3,
    size  = 3,
    color = "#1E4D7B"
  ) +
  scale_x_date(
    date_breaks = "6 months",
    date_labels = "%b %Y",
    limits      = c(as.Date("2021-08-01"), as.Date("2026-03-31"))
  ) +
  scale_y_continuous(expand = expansion(mult = c(0, 0.15))) +
  labs(
    x       = NULL,
    y       = "Monthly Downloads",
    title   = "CRAN Downloads of the sprtt Package"
    # caption = "Source: CRAN logs via cranlogs."
  ) +
  theme_minimal(base_size = 12) +
  theme(
    axis.text.x        = element_text(angle = 45, hjust = 1),
    panel.grid.minor   = element_blank(),
    panel.grid.major.x = element_blank(),
    plot.title         = element_text(face = "bold", size = 13),
    plot.caption       = element_text(color = "grey50", size = 8)
  )

ggsave("./inst/joss/sprtt_downloads.png", width = 8, height = 4.5, dpi = 300)

library(showtext)
## Loading Google fonts (http://www.google.com/fonts)
font_add_google("Gochi Hand", "gochi")
## Automatically use showtext to render text for future devices
showtext_auto()

## use the ggplot2 example
library(hexSticker)

library(showtext)
## Loading Google fonts (http://www.google.com/fonts)
# font_add_google("RocknRoll One", "rockn")
font_add_google("Alfa Slab One", "alfa")
# font_add_google("Ultra", "ultra")
# font_add_google("Paytone One", "paytone")
# font_add_google("Courgette", "courgette")
# font_add_google("Love Ya Like A Sister", "lovesister")
# font_add_google("Sue Ellen Francisco", "sue")
# font_add_google("Limelight", "limelight")
# font_add_google("Wendy One", "wendy")
# font_add_google("Arbutus", "arbutus")
# font_add_google("Londrina Shadow", "shadow")
# font_add_google("", "")
## Automatically use showtext to render text for future devices
showtext_auto()

sticker("man/figures/LogLikeSpike.png",
        package = "sprtt",
        filename = "man/figures/logo_.png",
        # Schrift
        p_color = "#c9144e",
        p_size = 140,    #70
        p_family = "alfa",
        # p_fontface = "bold",
        #subplot
        s_width = 1.5,
        s_height = 0,
        s_x = 0.99,
        #URL
        # url = "  https://meike-steinhilber.
        # shinyapps.io/spirit/",
        u_color = "#474f5c",      #    "#8c0e36",
        u_size = 20,
        u_x = 1,
        u_y = 0.10,
        # Rahmen
        h_color = "#FFFFFF", #"#cedbf0", #"#474f5c",
        h_fill =  "#cedbf0",        #"#cedbf0", "#b0a5a9",
        h_size = 2,
        white_around_sticker = FALSE,
        dpi = 1200,
        #spotlight
        spotlight = TRUE,
        l_alpha = 0.999,
        l_x = 1.011, # 1 ist mitte
        l_y = 1.2, # 1 ist mitte
        l_width = 8
)


# Run to build & detect logo
pkgdown::build_favicons(pkg = ".", overwrite = TRUE)

#' Test data to run the examples
#'
#'A dataset that includes 120 individuals with sex gender and monthly income.
#'
#' @format  A data frame with 2 variables:
#' \describe{
#'   \item{monthly_income}{ }
# #'   \item{in_debt}{0 = no and 1 = yes}
#'   \item{sex}{ }
# #'   \item{age}{ }
#' }
"df_income"


#' Test data to run the examples
#'
#'A dataset that includes 120 individuals.
#'
#' @format  A data frame with 2 variables:
#' \describe{
#'   \item{baseline_stress}{ }
#'   \item{one_year_stress}{ }
#' }
"df_stress"

#' Test data to run the examples
#'
#'A dataset that includes 120 individuals.
#'
#' @format  A data frame with 2 variables:
#' \describe{
#'   \item{treatment_group}{ }
#'   \item{control_group}{ }
#' }
"df_cancer"

# t-test -----------------------------------------------------------------------
extract_formula_ttest <- function(formula, data, wanted = "both") {

  data_matrix_formula <- model.frame(formula, data, na.action = NULL)
  x <- data_matrix_formula[, 1]

  if (formula[[3]] == 1) {
    # one sample
    y <- 1
  } else{
    # two samples
    y <- data_matrix_formula[, 2]
    # transform the structure in to the standard x, y
    x_values <- x[y == levels(y)[1]]
    y_values <- x[y == levels(y)[2]]
    x <- x_values
    y <- y_values
  }

  #get wanted output
  if (wanted == "x") {
    x
  } else if (wanted == "y") {
    y
  } else {
    list(x, y)
  }
}

# ANOVA ------------------------------------------------------------------------
extract_formula_anova <- function(formula, data) {

  data_matrix_formula <- model.frame(formula, data, na.action = NULL)
  names <- c("y",
             "factor_A", # one-way ANOVA only uses factor_A
             "factor_B", # two-way ANOVA
             "factor_C", # tree-way ANOVA
             "factor_D",
             "factor_E"
  )
  colnames(data_matrix_formula) <- names[1:ncol(data_matrix_formula)]
  data_matrix_formula
}

#* @testing extract_formula

# t-test -----------------------------------------------------------------------
test_that("extract_formula_ttest: effect in data: data are correct", {
  x_1 <- rnorm(5)
  x_2 <- rnorm(5)
  x <- c(x_1, x_2)
  y <- as.factor(c(rep(1,5),rep(2,5)))
  z <- x
  data <- data.frame(x, z, y)
  formula <- x ~ y

  x_test <- extract_formula_ttest(formula, data, wanted = "x")
  y_test <- extract_formula_ttest(formula, data, wanted = "y")
  x_y <- extract_formula_ttest(formula, data)
  expect_equal(x_test,x_1)
  expect_equal(y_test, x_2)
  expect_equal(x_y, list(x_1, x_2))

  data <- data.frame(a = x,c = z,b = y)
  formula <- a ~ 1
  x_test <- extract_formula_ttest(formula, data, wanted = "x")
  y_test <- extract_formula_ttest(formula, data, wanted = "y")
  expect_equal(x_test, x)
  expect_equal(y_test, 1)
})

# ANOVA ------------------------------------------------------------------------
test_that("extract_formula_anova: effect in data: data are correct", {
  # one-way ANOVA
  formula <- y ~ factor_A
  data <- draw_sample_normal(k_groups = 4, f = 0, max_n = 30)
  colnames(data) <- c("y", "factor_A")
  data_test <- cbind(data, test = rnorm(nrow(data)))
  results <- extract_formula_anova(formula, data_test)
  expect_equal(data, results, ignore_attr = TRUE)

  formula <- "y ~ factor_A"
  data <- draw_sample_normal(k_groups = 4, f = 0, max_n = 30)
  colnames(data) <- c("y", "factor_A")
  data_test <- cbind(data, test = rnorm(nrow(data)))
  results <- extract_formula_anova(formula, data_test)
  expect_equal(data, results, ignore_attr = TRUE)

  # two-way ANOVA
  formula <- y ~ factor_A+factor_B
  data <- draw_sample_normal(k_groups = 4, f = 0, max_n = 30)
  colnames(data) <- c("y", "factor_A")
  data_test <- cbind(data,
                     factor_B = as.factor(rep(c(1,0), 30)))

  results <- extract_formula_anova(formula, data_test)
  expect_equal(data_test, results, ignore_attr = TRUE)


})

#' @title Generates HTML reports for sample size planning for sequential ANOVAs.
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Renders a parameterized R Markdown report that helps plan sample size for the sequential ANOVA.
#' The function takes expected effect size (`f_expected`), number of groups (`k_groups`),
#' the power, and decision rate, then generates a reproducible HTML report summarizing the simulation-based
#' sample size recommendations. The alpha level is always 0.05.
#'
#' The template is located under:
#' `inst/rmarkdown/templates/report_sample_size/skeleton/skeleton.Rmd`.
#'
#' @param f_expected Numeric scalar. The expected standardized effect size (e.g., Cohen's f).
#'   Must be between 0.1 and 0.4 (increments of 0.05).
#' @param k_groups Integer scalar. The number of groups to compare. Must be between 2 and 4.
#' @param beta Numeric scalar (default = 0.05). Desired beta error rate (Type II error).
#'   Possible values are 0.20, 0.10, and 0.05.
#' @param decision_rate Numeric scalar (default = 0.85). Desired chance to reach a decision.
#'   Must be between 0.75 and 0.95 (increments of 0.05).
#' @param output_dir Character string. Directory in which to save the rendered HTML report.
#'   Defaults to a temporary directory (`tempdir()`).
#' @param output_file Character string. File name of the generated HTML report.
#'   Defaults to `"sprtt-report-sample-size-planning.html"`.
#' @param open Logical (default = `interactive()`). If `TRUE`, the generated report is opened
#'   in the system's default web browser after rendering.
#' @param overwrite Logical (default = `FALSE`). If `FALSE` and the target file already exists,
#'   the user is prompted interactively whether to overwrite it. In non-interactive sessions,
#'   an error is raised unless `overwrite = TRUE`.
#'
#' @details
#' This function is a front-end utility for rendering a pre-defined R Markdown report using
#' `rmarkdown::render()`.
#'
#' @return
#' Invisibly returns the path to the rendered HTML file (character string).
#' The report is optionally opened in the default browser.
#'
#' @section File Overwrite Behavior:
#' - If the specified output file already exists:
#'   - and `overwrite = FALSE`, the user is asked whether to overwrite (in interactive sessions);
#'     otherwise, an error is thrown.
#'   - If `overwrite = TRUE`, the file is replaced silently.
#'
#' @export
#'
#' @examples
#' \dontrun{
#' # Generate and open an SPRT sample size planning report:
#' plan_sample_size(
#'   f_expected = 0.25,
#'   k_groups = 3,
#'   decision_rate = 0.9
#' )
#'
#' # Prevent overwriting an existing file:
#' plan_sample_size(0.25, 3, overwrite = FALSE)
#' }
plan_sample_size <- function(f_expected,
                             k_groups,
                             beta = 0.05,
                             decision_rate = 0.85,
                             output_dir = tempdir(),
                             output_file = "sprtt-report-sample-size-planning.html",
                             open = interactive(),
                             overwrite = FALSE) {
  # Basic validation
  stopifnot(length(f_expected) == 1, is.numeric(f_expected))
  stopifnot(length(beta) == 1, is.numeric(beta), beta > 0, beta < 1)
  stopifnot(length(k_groups) == 1, is.numeric(k_groups), k_groups >= 2)

  power <- NULL  # suppress R CMD check note

  # check input parameters
  if (!decision_rate %in% c(0.75, 0.80, 0.85, 0.90, 0.95)) {
    stop(
      glue("`decision_rate` = {decision_rate} is not available. Please choose one of {glue_collapse(shQuote(c(0.75,0.80,0.85,0.90,0.95)), ', ', last = ' or ')}")
    )
  }

  # Load data — may trigger download prompt
  loaded <- load_sample_size_data()
  if (is.null(loaded)) {
    stop(
      "The simulation data is required to generate the report.\n",
      "Please run `download_sample_size_data()` and accept the download.",
      call. = FALSE
    )
  }
  df_all       <- loaded$data
  data_version <- loaded$version
  data_created <- loaded$created

  df <- df_all %>%
    distinct(f_expected, power, k_groups)

  if (!f_expected %in% df$f_expected) {
    stop(
      glue("`f_expected` = {f_expected} is not available. Please choose one of {glue_collapse(shQuote(sort(unique(df$f_expected))), ', ', last = ' or ')}")
    )
  }
  if (!beta %in% c(0.20, 0.10, 0.05)) {
    stop(
      glue("`beta` = {beta} is not available. Please choose one of {glue_collapse(shQuote(c(0.20,0.10,0.05)), ', ', last = ' or ')}")
    )
  }
  if (!k_groups %in% df$k_groups) {
    stop(
      glue("`k_groups` = {k_groups} is not available. Please choose one of {glue_collapse(shQuote(sort(unique(df$k_groups))), ', ', last = ' or ')}")
    )
  }

  # Construct full output path
  output_path <- file.path(output_dir, output_file)

  # Check for existing file
  # Skip overwrite check if the report is meant to be opened immediately
  if (!open && file.exists(output_path)) {
    if (!overwrite) {
      if (interactive()) {
        answer <- utils::menu(c("Yes", "No"),
                              title = sprintf("File '%s' already exists. Overwrite?", output_file)
        )
        if (answer != 1) {
          message("Aborted - file not overwritten.")
          return(invisible(NULL))
        }
      } else {
        stop(sprintf("Output file '%s' already exists. Use overwrite = TRUE to replace it.",
                     output_file),
             call. = FALSE)
      }
    }
  }

  # Locate the Rmd template
  rmd_path <- system.file(
    "rmarkdown", "templates", "report_sample_size", "skeleton", "skeleton.Rmd",
    package = "sprtt", mustWork = TRUE
  )

  # Render the report
  output <- rmarkdown::render(
    rmd_path,
    params = list(
      f_expected   = f_expected,
      beta         = beta,
      k_groups     = k_groups,
      df_all       = df_all,
      decision_rate = decision_rate,
      data_version = data_version,
      data_created = data_created
    ),
    output_file = output_file,
    output_dir  = output_dir,
    envir = new.env(parent = globalenv())
  )

  if (open) utils::browseURL(output)
  invisible(output)
}

#* @testing plan_sample_size


# helpers ----------------------------------------------------------------------

canon_html <- function(path) {
  stopifnot(file.exists(path))
  doc <- xml2::read_html(path)
  xml2::xml_find_all(doc, "//meta[@name='generator']") |> xml2::xml_remove()
  xml2::xml_find_all(doc, "//comment()") |> xml2::xml_remove()
  txt <- as.character(doc)
  txt <- gsub("[0-9]{4}-[0-9]{2}-[0-9]{2}[ T][0-9]{2}:[0-9]{2}:[0-9]{2}Z?", "<TIME>", txt)
  txt <- gsub("id=\"[A-Za-z0-9_-]{6,}\"", "id=\"<ID>\"", txt)
  txt <- gsub("[ \t]+", " ", txt)
  txt <- gsub("\r\n?", "\n", txt, perl = TRUE)
  txt
}

# error messages ---------------------------------------------------------------

test_that("plan_sample_size: paremters out of scope", {
  testthat::skip_on_cran()
  expect_error(
    plan_sample_size(f_expected = 0.24, k_groups = 3, overwrite = FALSE),
    "is not available"
  )
  expect_error(
    plan_sample_size(f_expected = 0.25, k_groups = 7, overwrite = FALSE),
    "is not available"
  )
  expect_error(
    plan_sample_size(f_expected = 0.25, k_groups = 3,
                     beta = 0.40, overwrite = FALSE),
    "is not available"
  )
  expect_error(
    plan_sample_size(f_expected = 0.25, k_groups = 3, decision_rate = 0.20),
    "is not available"
  )
})


test_that("plan_sample_size: paremters wrong data type", {
  testthat::skip_on_cran()
  expect_error(
    plan_sample_size(f_expected = "0.25", k_groups = 3, overwrite = FALSE),
    "is not TRUE"
  )
  expect_error(
    plan_sample_size(f_expected = 0.25, k_groups = "3", overwrite = FALSE),
    "is not TRUE"
  )
  expect_error(
    plan_sample_size(f_expected = 0.25, k_groups = 3,
                     beta = "0.05",
                     overwrite = FALSE),
    "is not TRUE"
  )
})


# general functioning - creating html output -----------------------------------

test_that("template is shipped", {
  testthat::skip_on_cran()
  p <- system.file("rmarkdown", "templates", "report_sample_size", "skeleton",
                   "skeleton.Rmd", package = "sprtt")
  expect_true(nzchar(p), info = "system.file returned empty path")
  expect_true(file.exists(p), info = "template skeleton.Rmd not found in inst/")
})

test_that("renders HTML to the specified directory", {
  # testthat::skip_if_not_installed("rmarkdown")
  # testthat::skip_if_not(rmarkdown::pandoc_available())
  # Optional: skip on CRAN if render is slow/heavy
  testthat::skip_on_cran()

  withr::local_tempdir() -> tmp
  out <- plan_sample_size(
    f_expected = 0.25,
    k_groups   = 4,
    beta      = 0.05,
    output_dir = tmp,
    output_file = "report.html",
    open = FALSE,
    overwrite = TRUE
  )

  # Function returns the output file path (invisible)
  expect_true(is.character(out) && length(out) == 1)
  expect_true(file.exists(out))
  expect_gt(file.info(out)$size, 0)

  # Quick sanity on contents (don’t rely on exact formatting)
  head <- readLines(out, n = 100L, warn = FALSE)
  expect_true(any(grepl("<html", head, ignore.case = TRUE)))
  # If your Rmd has a stable title/headline, check for it too:
  # expect_true(any(grepl("Sample-size report", head, fixed = TRUE)))
})

test_that("renders HTM too", {
  # testthat::skip_if_not_installed("rmarkdown")
  # testthat::skip_if_not(rmarkdown::pandoc_available())
  testthat::skip_on_cran()

  withr::local_tempdir() -> tmp
  out <- plan_sample_size(
    f_expected = 0.25,
    k_groups   = 4,
    beta      = 0.05,
    output_dir = tmp,
    output_file = "report.htm",
    open = FALSE,
    overwrite = TRUE
  )
  expect_true(file.exists(out))
  expect_match(out, "\\.htm$", perl = TRUE)
})


test_that("overwrite protection works when open = FALSE", {
  testthat::skip_on_cran()
  withr::local_tempdir() -> tmp
  path <- file.path(tmp, "report.html")
  cat("<html></html>", file = path)

  expect_error(
    plan_sample_size(
      f_expected = 0.25,
      k_groups   = 4,
      beta      = 0.05,
      output_dir = tmp,
      output_file = basename(path),
      open = FALSE,
      overwrite = FALSE
    ),
    regexp = "already exists.*overwrite = TRUE",
    ignore.case = TRUE
  )

  # Now with overwrite = TRUE it should succeed
  testthat::skip_if_not_installed("rmarkdown")
  testthat::skip_if_not(rmarkdown::pandoc_available())
  out <- plan_sample_size(
    f_expected = 0.25,
    k_groups   = 4,
    beta      = 0.05,
    output_dir = tmp,
    output_file = basename(path),
    open = FALSE,
    overwrite = TRUE
  )
  expect_true(file.exists(out))
})


# snapshots of output ----------------------------------------------------------

# too frickle

# t-test -----------------------------------------------------------------------
calc_non_centrality_parameter_ttest <- function(seq_ttest_arguments) {
  x <- seq_ttest_arguments@x
  y <- seq_ttest_arguments@y
  d <- seq_ttest_arguments@d

  if (seq_ttest_arguments@one_sample == TRUE ||
      seq_ttest_arguments@paired == TRUE
  ) {
    d * sqrt(length(x))
  } else{
    d / sqrt(1 / length(x) + 1 / length(y))
  }
}

# ANOVA ------------------------------------------------------------------------

calc_non_centrality_parameter_anova <- function(seq_anova_arguments) {
  # seq_anova_arguments@f^2 * seq_anova_arguments@n
  seq_anova_arguments@f^2 * seq_anova_arguments@total_sample_size
}

#* @testing calc_likelihoods_ttest

# t-test -----------------------------------------------------------------------

test_that("calc_likelihoods_ttest: check log", {
  seq_ttest_arguments <-
    build_prototype_seq_ttest_arguments()
  t_statistic <-
    calc_t_statistic(seq_ttest_arguments)
  non_centrality_parameter <-
    calc_non_centrality_parameter_ttest(seq_ttest_arguments)
  df <- 18
  likelihoods <-
    calc_likelihoods_ttest(
      seq_ttest_arguments,
      t_statistic,
      df,
      non_centrality_parameter,
      wanted = NULL
  )
  expect_true(log(likelihoods$ratio) - likelihoods$ratio_log < 1e-5)

})

test_that("calc_likelihoods_ttest: Check warnings & messages", {
  x <- rnorm(2000000)
  d <- 5
  expect_warning(seq_ttest(x, d = d),
                 "At least one likelihood is equal to 0")
})


test_that("calc_likelihoods_ttest: Check warnings & messages", {
  set.seed(3)
  seq_ttest_arguments <-
    build_prototype_seq_ttest_arguments()

  seq_ttest_arguments@x <- rnorm(40, mean = 0)
  seq_ttest_arguments@y <- rnorm(40, mean = 5)
  seq_ttest_arguments@alternative <- "less"

  t_statistic <-
    calc_t_statistic(seq_ttest_arguments)
  non_centrality_parameter <-
    calc_non_centrality_parameter_ttest(seq_ttest_arguments)
  df <- 78

  warning_text <- capture_warnings(
    calc_likelihoods_ttest(
      seq_ttest_arguments,
      t_statistic,
      df,
      non_centrality_parameter,
      wanted = NULL
    )
  )

  # at least one warning occurred
  expect_true(length(warning_text) > 0)

  # check for your specific warning (substring match)
  expect_true(any(grepl(
    "At least one log-likelihood reached infinity",
    warning_text,
    fixed = TRUE
  )))
})

# ANOVA ------------------------------------------------------------------------

test_that("calc_likelihoods_anova: Check warnings & messages", {
  set.seed(333)

  data <- draw_sample_normal(f = 1000, k_groups = 3, max_n = 100)
  warning_text <- capture_warnings(seq_anova(y~x, f = 0.1, data = data))
  # at least one warning occurred
  expect_true(length(warning_text) > 0)
  # check for your specific warning (substring match)
  expect_true(any(grepl(
    "At least one likelihood is equal to 0",
    warning_text,
    fixed = TRUE
  )))

  data <- draw_sample_normal(f = 1000000000, k_groups = 3, max_n = 100)
  warning_text <- capture_warnings(seq_anova(y~x, f = 0.000001, data = data))
  # at least one warning occurred
  expect_true(length(warning_text) > 0)
  # check for your specific warning (substring match)
  expect_true(any(grepl(
    "At least one log-likelihood reached infinity.",
    warning_text,
    fixed = TRUE
  )))

  data <- draw_sample_normal(f = 0, k_groups = 3, max_n = 100)
  warning_text <- capture_warnings(seq_anova(y~x, f = 10000, data = data))
  # at least one warning occurred
  expect_true(length(warning_text) > 0)
  # check for your specific warning (substring match)
  expect_true(any(grepl(
    "At least one likelihood is equal to 0",
    warning_text,
    fixed = TRUE
  )))

})


test_that("calc_likelihoods_anova: Check warnings & messages", {
  set.seed(333)

  seq_anova_arguments <- build_prototype_seq_anova_arguments()
  non_centrality_parameter <-
    calc_non_centrality_parameter_anova(
      seq_anova_arguments
    )
  seq_anova_arguments@data <-
    calc_group_means(
      seq_anova_arguments
    )
  ss_effect <-
    calc_ss_effect(
      seq_anova_arguments
    )
  ss_residual <-
    calc_ss_residual(
      seq_anova_arguments
    )
  F_statistic <-
    calc_F_statistic_(
      seq_anova_arguments,
      ss_effect,
      ss_residual
    )

  # 3.ed edition necessary for expect_snapshot
  testthat::local_edition(3)

  expect_snapshot(
    calc_likelihoods_anova(
      seq_anova_arguments,
      non_centrality_parameter,
      F_statistic
    )
  )
})

test_that("calc_non_centrality_parameter", {
  # 3.ed edition necessary for expect_snapshot
  testthat::local_edition(3)
  set.seed(4657)
  seq_ttest_arguments <- build_prototype_seq_ttest_arguments()
  seq_anova_arguments <- build_prototype_seq_anova_arguments()

  expect_snapshot(
    show(
      calc_non_centrality_parameter_ttest(seq_ttest_arguments)
    )
  )
  expect_snapshot(
    show(
      calc_non_centrality_parameter_anova(seq_anova_arguments)
    )
  )
})

# set.seed(3333)
#
# # null effect ------------------------------------------------------------------
# k_groups <- 4
# f <- 0
# sd <- c(1,1,1,1)
# sample_ratio <-  c(1,1,1,1)
# max_n <- 200
# rep_n <- 100
#
# data <- matrix(
#   data = 0,
#   nrow = max_n*sum(sample_ratio),
#   ncol = rep_n*2
# )
# f_estimated <- double(rep_n)
#
# a = 1
# b = a + 1
# for (i in 1:rep_n) {
#   if (i == 1) {
#     sample <- draw_sample_normal(k_groups = k_groups,
#                                   f = f,
#                                   sd = sd,
#                                   sample_ratio = sample_ratio,
#                                   max_n = max_n)
#     data[, a] <- as.numeric(sample$x)
#     data[, b] <- sample$y
#     f_estimated[i] <- effect_sizes(y~x, sample)$cohens_f
#     a = b + 1
#     b = a + 1
#   } else{
#     sample <- draw_sample_normal(k_groups = k_groups,
#                                   f = f,
#                                   sd = sd,
#                                   sample_ratio = sample_ratio,
#                                   max_n = max_n)
#     f_estimated[i] <- effect_sizes(y~x, sample)$cohens_f
#     data[, a] <- sample$y
#     a = a + 1
#   }
# }
#
# mean_f_est <- mean(f_estimated)
# mean_f_est
#
#
# # small effect -----------------------------------------------------------------
#
# k_groups <- 4
# f <- 0.10
# sd <- c(1,1,1,1)
# sample_ratio <-  c(1,1,1,1)
# max_n <- 200
# rep_n <- 100
#
# data <- matrix(
#   data = 0,
#   nrow = max_n*sum(sample_ratio),
#   ncol = rep_n*2
# )
# f_estimated <- double(rep_n)
#
# a = 1
# b = a + 1
# for (i in 1:rep_n) {
#   if (i == 1) {
#     sample <- draw_sample_normal(k_groups = k_groups,
#                                  f = f,
#                                  sd = sd,
#                                  sample_ratio = sample_ratio,
#                                  max_n = max_n)
#     data[, a] <- as.numeric(sample$x)
#     data[, b] <- sample$y
#     f_estimated[i] <- effect_sizes(y~x, sample)$cohens_f
#     a = b + 1
#     b = a + 1
#   } else{
#     sample <- draw_sample_normal(k_groups = k_groups,
#                                  f = f,
#                                  sd = sd,
#                                  sample_ratio = sample_ratio,
#                                  max_n = max_n)
#     f_estimated[i] <- effect_sizes(y~x, sample)$cohens_f
#     data[, a] <- sample$y
#     a = a + 1
#   }
# }
#
# mean_f_est <- mean(f_estimated)
# mean_f_est
#
#
# # large effect -----------------------------------------------------------------
#
# k_groups <- 4
# f <- 0.80
# sd <- c(1,1,1,1)
# sample_ratio <-  c(1,1,1,1)
# max_n <- 200
# rep_n <- 100
#
# data <- matrix(
#   data = 0,
#   nrow = max_n*sum(sample_ratio),
#   ncol = rep_n*2
# )
# f_estimated <- double(rep_n)
#
# a = 1
# b = a + 1
# for (i in 1:rep_n) {
#   if (i == 1) {
#     sample <- draw_sample_normal(k_groups = k_groups,
#                                  f = f,
#                                  sd = sd,
#                                  sample_ratio = sample_ratio,
#                                  max_n = max_n)
#     data[, a] <- as.numeric(sample$x)
#     data[, b] <- sample$y
#     f_estimated[i] <- effect_sizes(y~x, sample)$cohens_f
#     a = b + 1
#     b = a + 1
#   } else{
#     sample <- draw_sample_normal(k_groups = k_groups,
#                                  f = f,
#                                  sd = sd,
#                                  sample_ratio = sample_ratio,
#                                  max_n = max_n)
#     f_estimated[i] <- effect_sizes(y~x, sample)$cohens_f
#     data[, a] <- sample$y
#     a = a + 1
#   }
# }
#
# mean_f_est <- mean(f_estimated)
# mean_f_est
#
#
# # MIXED ------------------------------------------------------------------------
#
# # large effect -----------------------------------------------------------------
# set.seed(333)
# k_groups <- 4
# f <- 0.70
# sd <- c(1,1,1,1)
# # sd <- c(40,40,40,40)
# sample_ratio <-  c(1,1,1,1)
# max_n <- 200
# rep_n <- 100
#
# data <- matrix(
#   data = 0,
#   nrow = max_n*sum(sample_ratio),
#   ncol = rep_n*2
# )
# f_estimated <- double(rep_n)
#
# a = 1
# b = a + 1
# for (i in 1:rep_n) {
#   if (i == 1) {
#     sample <- draw_sample_mixture(k_groups = k_groups,
#                                  f = f,
#                                  max_n = max_n)
#     data[, a] <- as.numeric(sample$x)
#     data[, b] <- sample$y
#     f_estimated[i] <- effect_sizes(y~x, sample)$cohens_f
#     a = b + 1
#     b = a + 1
#   } else{
#     sample <- draw_sample_mixture(k_groups = k_groups,
#                                  f = f,
#                                  max_n = max_n)
#     f_estimated[i] <- effect_sizes(y~x, sample)$cohens_f
#     data[, a] <- sample$y
#     a = a + 1
#   }
# }
#
# mean_f_est <- mean(f_estimated)
# mean_f_est
# sd_f_est

check_data_ttest <- function(x, y, paired) {

  if (!is.numeric(x))
    stop(paste("Invalid argument: x must be numeric."))

  if (!is.null(y)) {
    if (is.factor(y))
      stop("Is y a grouping factor? Use formula interface x ~ y.")
    if (!is.numeric(y))
      stop(paste("Invalid argument:y must be numeric."))
    if (!paired && (length(x) + length(y) < 3))
      stop("SPRT for two independent samples requires at least 3 observations.")
    if (!is.logical(paired))
      stop("Invalid argument <paired>: Must be logical.")
  }

  check_constant_data_ttest(x, y)
}

check_data_anova <- function(data) {

  if (!is.numeric(data$y))
    stop(paste("Invalid argument: y must be numeric (y~factor_A)."))
  if (length(levels(data$factor_A)) == 2 && (length(data$y) < 3))
    stop("Requires at least 3 observations.")
  if (length(unique(data$factor_A)) < 2)
    stop("A factor needs at least 2 levels.")
  if (any(as.numeric(table(data$factor_A)) < 2))
    stop("Every group needs at least two observations.")
  # if (!is.logical(paired))
  #   stop("Invalid argument <paired>: Must be logical.")

  check_constant_data_anova(data)
}

#* @testing check_data

# t-test -----------------------------------------------------------------------

test_that("check_data_ttest: numeric: Correct activation of errors", {

  x <- list(1:5, 9:5)
  y <- c(1:5, 9:5)
  expect_error(
    check_data_ttest(x, y, paired = FALSE),
    "must be numeric"
  )
  x <- c(1:5, 9:5)
  y <- list(1:5, 9:5)
  expect_error(
    check_data_ttest(x, y, paired = FALSE),
    "must be numeric"
  )
  x <- rnorm(10)
  y <- as.factor(sample(c(1,2), 10, replace = TRUE))
  expect_error(
    check_data_ttest(x, y, paired = FALSE),
    "grouping factor"
  )
  x <- rnorm(1)
  y <- rnorm(1)
  expect_error(
    check_data_ttest(x, y, paired = FALSE),
    "at least 3 observations"
  )
  x <- rnorm(5)
  y <- rnorm(5)
  expect_error(
    check_data_ttest(x, y, paired = 5),
    " Must be logical"
  )
})

# ANOVA ------------------------------------------------------------------------


test_that("check_data_ttest: numeric: Correct activation of errors", {
  data <- draw_sample_normal(k_groups = 2, f = 0, max_n = 10)
  colnames(data) <- c("y", "factor_A")
  formula <- y~factor_A
  data$y <- as.character(data$y)
  expect_error(
    check_data_anova(data),
    "Invalid argument: y must be numeric "
  )

  data <- draw_sample_normal(k_groups = 2, f = 0, max_n = 2)[1:2,]
  colnames(data) <- c("y", "factor_A")
  formula <- y~factor_A
  expect_error(
    check_data_anova(data),
    "Requires at least 3 observations"
  )

  data <- draw_sample_normal(k_groups = 2, f = 0, sample_ratio = c(3, 1), max_n = 2)[1:3,]
  colnames(data) <- c("y", "factor_A")
  formula <- y~factor_A
  expect_error(
    check_data_anova(data),
    "A factor needs at least 2 levels."
  )

  data <- draw_sample_normal(k_groups = 2, f = 0, sd = c(1, 1), max_n = 2)
  colnames(data) <- c("y", "factor_A")
  formula <- y~factor_A
  expect_silent(
    check_data_anova(data)
  )

  data <- draw_sample_normal(k_groups = 2, f = 0, sample_ratio = c(3, 1), max_n = 2)[1:4,]
  colnames(data) <- c("y", "factor_A")
  formula <- y~factor_A
  expect_error(
    check_data_anova(data),
    "Every group needs at least two observations"
  )
  data <- draw_sample_normal(k_groups = 3, f = 0, sample_ratio = c(3, 2, 1), max_n = 2)[1:6,]
  colnames(data) <- c("y", "factor_A")
  formula <- y~factor_A
  expect_error(
    check_data_anova(data),
    "Every group needs at least two observations"
  )

})

#' Plot Sequential ANOVA Results
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Creates a visualization of the sequential probability ratio test (SPRT) for
#' ANOVA results, showing the log-likelihood ratio trajectory across sample sizes
#' and decision boundaries.
#'
#' @param anova_results A `seq_anova_results` object from [seq_anova()].
#'   **Important:** The `seq_anova()` function must be called with `plot = TRUE`
#'   to generate the necessary data for plotting.
#' @param labels Logical. If `TRUE` (default), display decision labels
#'   ("Accept H0" / "Accept H1") and the likelihood ratio at the decision point.
#' @param position_labels_x Numeric value between 0 and 1 controlling the
#'   horizontal position of decision labels as a proportion of maximum sample
#'   size. Default is `0.15` (left side); `0.5` centers the labels.
#' @param position_labels_y Numeric value controlling the vertical spacing
#'   between decision boundaries and their labels. The value is multiplied by
#'   `max(|log-likelihood ratio|)` to determine spacing. Larger values move
#'   labels further from boundaries. Default is `0.075`.
#' @param position_lr_x Optional numeric value for the x-coordinate (sample size)
#'   of the likelihood ratio label. If `NULL` (default), positioned at the
#'   decision point or final sample size.
#' @param position_lr_y Optional numeric value for the y-coordinate
#'   (log-likelihood ratio) of the likelihood ratio label. If `NULL` (default),
#'   positioned at `y = 0` for early decisions, or slightly offset for
#'   continuing sampling scenarios.
#' @param font_size Numeric. Base font size for plot text. Default is `20`.
#' @param line_size Numeric. Line width for the trajectory and boundaries.
#'   Default is `1.5`.
#' @param highlight_color Character string. Color for highlighting the decision
#'   point or final sample. Default is `"#CD2626"` (red).
#'
#' @return A [ggplot2::ggplot()] object showing:
#'   \itemize{
#'     \item Log-likelihood ratio trajectory across sample sizes
#'     \item Dashed horizontal lines indicating decision boundaries
#'     \item Highlighted point showing where decision was reached (or final sample)
#'     \item Optional labels for decision regions and likelihood ratio value
#'   }
#'
#' @import ggplot2 glue purrr
#'
#' @export
#'
#' @example inst/examples/plot_anova.R
#'
plot_anova <- function(anova_results,
                      labels = TRUE,
                      position_labels_x = 0.15,
                      position_labels_y = 0.1,
                      position_lr_x = NULL,
                      position_lr_y = NULL,
                      font_size = 15,
                      line_size = 1,
                      highlight_color = "#CD2626"
                      ) {

  if (inherits(anova_results, c("seq_ttest_results"))) {
    stop("The plot_anova() function only works for sequential anovas (anova_results argument must be of class seq_anova_results).")
  }
  if (!inherits(anova_results, c("seq_anova_results"))) {
    stop("anova_results argument must be of class seq_anova_results.")
  }

  if (is.null(anova_results@plot)) {
    stop("The anova_results@plot is NULL. Solution: The function argument `plot` must be set to TRUE in the seq_anova() function.")
  }

  A_boundary_log <- anova_results@A_boundary_log
  B_boundary_log <- anova_results@B_boundary_log

  results <- data.frame(lr_log = anova_results@plot$lr_log,
                        sample_size = anova_results@plot$sample_size) %>%
    mutate(decision = case_when(
      .data$lr_log >= A_boundary_log ~ "H1",
      .data$lr_log <= B_boundary_log ~ "H0",
      TRUE ~ "CS"
    ))

  N_steps <- nrow(results)
  N <- results$sample_size[N_steps]

  if (any(results$decision == "H1" | results$decision == "H0")) {
    decision_sample_position <- which(results$decision != "CS")[1]
  } else{
    decision_sample_position <- 0
  }

  max_lr <- max(abs(results$lr_log))
  lr_factor <- 0.05


  # set defaults ---------------------------------------------------------------
  # palette <- "Dark2"
  # green <- "#69b3a2"
  # blue <- "#404080"
  # red <- 	"#CD2626" #"#8B0000"
  theme_set(theme_bw(base_size = font_size))

  # create plot ----------------------------------------------------------------

  plot <- results %>%
    # slice(1:decision_sample_position) %>%
    ggplot(aes(x = .data$sample_size, y = .data$lr_log)) +
    xlim(0 , results$sample_size[N_steps] + results$sample_size[N_steps]*(lr_factor+0.1)) +
    geom_line(linewidth = line_size) +
    geom_hline(yintercept = A_boundary_log, linetype = "dashed", linewidth = line_size) +
    geom_hline(yintercept = B_boundary_log, linetype = "dashed", linewidth = line_size) +
    labs(#title = glue("f expected = {f_expected}, f simulated = {f_simulated}"),
         x = "Sample Size N",
         y = "Log-Likelihood Ratio") +
    theme_bw(base_size = font_size)

    if (decision_sample_position > 0) {
      x_decision <- results$sample_size[decision_sample_position]
      y_decision <- results$lr_log[decision_sample_position]

      plot <- plot +
        annotate("point",
                 x = x_decision,
                 y = y_decision,
                 colour = highlight_color,
                 size = line_size * 4)

      if (labels == TRUE) {
        LR <- round(exp(results$lr_log[decision_sample_position]), 2)
        nLR <- results$sample_size[decision_sample_position]

        if (is.null(position_lr_y)) {
          position_lr_y <- 0
        }

        if (is.null(position_lr_x)) {
          position_lr_x <- results$sample_size[decision_sample_position] + results$sample_size[decision_sample_position]*lr_factor
        }

        plot <- plot +
          annotate(
            geom = "text",
            x = position_lr_x,
            y = position_lr_y,
            label = glue("LR[{nLR}] ==~ {LR}"),
            parse = TRUE,
            size = font_size/.pt, color = highlight_color
          )
      }
    } else{
      x_last <- results$sample_size[N_steps]
      y_last <- results$lr_log[N_steps]

      plot <- plot +
        annotate("point",
                 x = x_last,
                 y = y_last,
                 colour = highlight_color,
                 size = line_size * 4)

      if (labels == TRUE) {
        LR <- round(exp(results$lr_log[N_steps]), 2)
        nLR <- results$sample_size[N_steps]

        if (is.null(position_lr_y)) {
          if (results$lr_log[N_steps] > 0) {position_lr_y = -0.5} else {position_lr_y = 0.05}
        }

        if (is.null(position_lr_x)) {
          position_lr_x <- results$sample_size[N_steps] + results$sample_size[N_steps]*lr_factor
        }

        plot <- plot +
          annotate(geom = "text",
                   x = position_lr_x,
                   y = position_lr_y,
                   label = glue("LR[{nLR}] ==~ {LR}"),
                   parse = TRUE,
                   size = font_size/.pt, color = highlight_color)
      }
    }

    if (labels == TRUE) {
      if (decision_sample_position == 0) {
        h0_col <- "black"
        h1_col <- "black"
      } else if (results$decision[decision_sample_position] == "H1") {
        h0_col <- "black"
        h1_col <- highlight_color
      } else if (results$decision[decision_sample_position] == "H0") {
        h0_col <- highlight_color
        h1_col <- "black"
      }
      distance_h <- max_lr*position_labels_y
      plot <- plot +
        annotate(geom = "text", x = N*position_labels_x, y = A_boundary_log + distance_h,
                label = "Accept~ H[1]", size = font_size/.pt, parse = TRUE, color = h1_col) +
        annotate(geom = "text", x = N*position_labels_x, y = B_boundary_log - distance_h,
                 label = "Accept~ H[0]", size = font_size/.pt, parse = TRUE, color = h0_col)
        # annotate(geom = "text", x = 10, y = B_boundary_log + distance_h/2,
        #         label = "Continue Sampling")
    }
  plot
}

# labels = TRUE
# position_labels_x = 0.15
# position_labels_y = 0.075
# font_size = 25
# line_size = 1.5
# highlight_color = "#CD2626"
# position_lr_x = 0.05

test_that("plot with ANOVA", {
  set.seed(333)
  seq_anova_arguments <- build_prototype_seq_anova_arguments()

  # check seq_steps ------------------------------------------------------------
  results <- calc_plot_anova(seq_anova_arguments, seq_steps = c(8,20,15))
  expect_true(nrow(results@plot) == 3)

  results <- calc_plot_anova(seq_anova_arguments, seq_steps = "single")
  expect_equal(results@plot$sample_size, seq(8,200,1))

  results <- calc_plot_anova(seq_anova_arguments, seq_steps = "balanced")
  expect_equal(results@plot$sample_size, seq(8,200,4))


  # wrong first data points - unequal sample sizes
  set.seed(333)
  data <- sprtt::draw_sample_normal(3, f = 0.25, max_n = 30, sample_ratio = c(1,2,2))
  seq_anova_arguments@data <- data[sample(nrow(data)),] # destroy the perfect order of the data
  expect_error(calc_plot_anova(seq_anova_arguments, seq_steps = "single"),
               "Every group needs two data points.")

})


# # ttest ------------------------------------------------------------------------
#
# test_that("plot ttest", {
#   set.seed(3333)
#   data <- draw_sample_normal(k_groups = 2, f = 0.20, max_n = 150)
#   formula <- y ~ x
#
#   sprt_results <- seq_ttest(formula, d = 0.10, data = data, plot = TRUE)
#   plot_anova(sprt_results)
#
#   set.seed(3333)
#   d <- 0.7
#   x <- rnorm(30)
#   y <- rnorm(15)
#   sprt_results <- seq_ttest(x, y, d = d, plot = TRUE)
#   plot_anova(sprt_results)
#
# })

test_that("plot with ANOVA", {
  skip_on_cran()
  testthat::skip_on_ci()

  set.seed(333)
  testthat::local_edition(3)

  data <- draw_sample_normal(k_groups = 4, f = 0.20, max_n = 150)
  formula <- y ~ x
  results <- seq_anova(formula, f = 0.10, data = data)

  expect_identical(results@plot, NULL)

  sprt_results <- seq_anova(formula, f = 0.10, data = data, plot = TRUE)
  p <- plot_anova(sprt_results)
  vdiffr::expect_doppelganger("general", p)

  k_groups = 4
  sprt_results <- seq_anova(formula, f = 0.10, data = data, plot = TRUE, seq_steps = c(k_groups*2, k_groups*5, k_groups*7))
  p <- plot_anova(sprt_results)
  vdiffr::expect_doppelganger("custom steps", p)

  data <- draw_sample_normal(k_groups = 4, f = 0, max_n = 90)
  anova_results <- seq_anova(formula, f = 0.20, data = data, plot = TRUE)
  p <- plot_anova(anova_results)
  vdiffr::expect_doppelganger("larger sample", p)

  p <- plot_anova(anova_results, position_lr_x = 200)
  vdiffr::expect_doppelganger("change lr position", p)

  sprt_results <- seq_anova(formula, f = 0.20, data = data, plot = TRUE, seq_steps = "balanced")
  p <- plot_anova(sprt_results)
  vdiffr::expect_doppelganger("balanced", p)

  sprt_results <- seq_anova(formula, f = 0.20, data = data, plot = TRUE, seq_steps = c(8, 20, 300))
  p <- plot_anova(sprt_results)
  vdiffr::expect_doppelganger("custom steps 2", p)

  sprt_results <- seq_anova(formula, f = 0.20, data = data[1:20,], plot = TRUE, seq_steps = c(8, 20))
  p <- plot_anova(sprt_results)
  vdiffr::expect_doppelganger("early stage", p)

  data <- draw_sample_normal(k_groups = 4, f = 0, max_n = 90)
  sprt_results <- seq_anova(formula, f = 0.20, data = data, plot = TRUE)
  p <- plot_anova(sprt_results, highlight_color = "green")
  vdiffr::expect_doppelganger("color change", p)

  sprt_results <- seq_anova(formula, f = 0.20, data = data, plot = TRUE, seq_steps = "balanced")
  p <- plot_anova(sprt_results, labels = FALSE)
  vdiffr::expect_doppelganger("no labels", p)

  sprt_results <- seq_anova(formula, f = 0.20, data = data, plot = TRUE, seq_steps = "balanced")
  p <- plot_anova(sprt_results, position_labels_x = 1)
  vdiffr::expect_doppelganger("position_labels_x", p)

  sprt_results <- seq_anova(formula, f = 0.20, data = data, plot = TRUE, seq_steps = "balanced")
  p <- plot_anova(sprt_results, position_labels_y = -5)
  vdiffr::expect_doppelganger("position_labels_y", p)


  sprt_results <- seq_anova(formula, f = 0.20, data = data, plot = TRUE, seq_steps = "balanced")
  p <- plot_anova(sprt_results, position_lr_x  = 300)
  vdiffr::expect_doppelganger("position_lr_x", p)

  sprt_results <- seq_anova(formula, f = 0.20, data = data, plot = TRUE, seq_steps = "balanced")
  p <- plot_anova(sprt_results, position_lr_y  = 6)
  vdiffr::expect_doppelganger("position_lr_y", p)

  sprt_results <- seq_anova(formula, f = 0.20, data = data, plot = TRUE, seq_steps = "balanced")
  p <- plot_anova(sprt_results, font_size = 6)
  vdiffr::expect_doppelganger("font_size", p)

  sprt_results <- seq_anova(formula, f = 0.20, data = data, plot = TRUE, seq_steps = "balanced")
  p <- plot_anova(sprt_results, line_size = 0.1)
  vdiffr::expect_doppelganger("line_size", p)


  data <- draw_sample_normal(k_groups = 4, f = 0, max_n = 150)
  sprt_results <- seq_anova(formula, f = 0.20, data = data, plot = TRUE, seq_steps = "balanced")
  p <- plot_anova(sprt_results, highlight_color = "darkgreen")
  vdiffr::expect_doppelganger("highlight_color", p)


  set.seed(333)
  data <- sprtt::draw_sample_normal(3, f = 0.25, max_n = 30)
  anova_results <- sprtt::seq_anova(y~x, f = 0.25, data = data, plot = TRUE)
  p <- plot_anova(anova_results,
                  labels = TRUE,
                  position_labels_x = 0.5,
                  position_labels_y = 0.1,
                  position_lr_x = 75,
                  font_size = 25,
                  line_size = 2,
                  highlight_color = "green")
  vdiffr::expect_doppelganger("exmple case all arguments", p)

  # check errors ---------------------------------------------------------------
  expect_error(sprtt::seq_anova(y~x, f = 0.25, data = data, plot = TRUE, seq_steps = "X"),
               "seq_steps")

  # wrong first data points - unequal sample sizes
  set.seed(333)
  data <- sprtt::draw_sample_normal(3, f = 0.25, max_n = 30, sample_ratio = c(1,2,2))
  data <- data[sample(nrow(data)),] # destroy the perfect order of the data
  expect_error(sprtt::seq_anova(y~x, f = 0.25, data = data, plot = TRUE),
               "Every group needs two data points.")

  data <- sprtt::draw_sample_normal(2, f = 0.25, max_n = 30)
  ttest_results <- sprtt::seq_ttest(y~x, d = 0.25, data = data)
  expect_error(plot_anova(ttest_results),
               "only works for sequential anovas")

  expect_error(plot_anova("character"),
               "must be of class seq_anova_results")

  expect_error(plot_anova(sprtt::seq_anova(y~x, f = 0.25, data = data)),
               "The anova_results@plot is NULL.")
})

.onAttach <- function(libname, pkgname) {
  cache_dir <- get_sprtt_cache_dir()
  data_file <- file.path(cache_dir, "df_all.rds")

  if (!file.exists(data_file)) {
    packageStartupMessage(
      "Note: Sample size planning functions require simulation data (~150 MB).\n",
      "Data will download automatically on first use.\n",
      "See ?download_sample_size_data for more information."
    )
  }
}

calc_seq_ttest <- function(seq_ttest_arguments, verbose = TRUE){

  t_statistic <-
    calc_t_statistic(
      seq_ttest_arguments
    )
  df <-
    as.vector(
      t_statistic$parameter
    )
  non_centrality_parameter <-
    calc_non_centrality_parameter_ttest(
      seq_ttest_arguments
    )
  likelihoods <-
    calc_likelihoods_ttest(
      seq_ttest_arguments,
      t_statistic,
      df,
      non_centrality_parameter
    )
  boundaries <-
    calc_boundaries(
      power = seq_ttest_arguments@power,
      alpha = seq_ttest_arguments@alpha,
      log = TRUE
    )
  decision <-
    get_seq_decision(
      likelihood_ratio = likelihoods$ratio_log,
      boundaries = boundaries
    )
  seq_ttest_results <-
    build_seq_ttest_results(
      seq_ttest_arguments,
      t_statistic,
      df,
      non_centrality_parameter,
      likelihoods,
      boundaries,
      decision,
      verbose
    )

  seq_ttest_results
}

# t-TEST -----------------------------------------------------------------------
test_that("show: print output ttest?", {
  # 3.ed edition necessary for expect_snapshot
  testthat::local_edition(3)
  set.seed(4657)

  expect_snapshot(
    show(
      seq_ttest(
        rnorm(20),
        d = 0.8)
    )
  )

  x_special_name <- rnorm(20)
  expect_snapshot(
    show(
      seq_ttest(x_special_name, d = 0.8, alternative = "less", mu = 2)
    )
  )

  x <- rnorm(20)
  result <-
  expect_snapshot(
    show(
      seq_ttest(x, d = 0.8, alternative = "greater")
    )
  )

  x_special_name <- rnorm(20)
  y_secial_name <- rnorm(20)
  expect_snapshot(
    show(
      seq_ttest(x_special_name, y_secial_name, d = 0.8)
    )
  )

  x <- rnorm(20)
  y <- rnorm(20)
  expect_snapshot(
    show(
      seq_ttest(x, y, d = 0.8, alternative = "less")
    )
  )

  x <- rnorm(20)
  y <- as.factor(rep(c(0, 1), 10))
  expect_snapshot(
    show(
      seq_ttest(
        x ~ y,
        d = 0.8)
    )
  )

  data <- sprtt::draw_sample_normal(2, 0.8, 40)
  expect_snapshot(
    show(
      seq_ttest(
        y ~ x,
        d = 0.60,
        data = data,
        alternative = "less",
        verbose = TRUE)
    )
  )

  data <- sprtt::draw_sample_normal(2, 0.8, 40)
  expect_snapshot(
    show(
      seq_ttest(
        y ~ x,
        d = 0.60,
        data = data,
        alternative = "greater",
        verbose = TRUE)
    )
  )

  set.seed(333)
  expect_snapshot(
    show(
      seq_ttest(
        rnorm(50),
        mu = 1,
        d = 0.4,
        verbose = TRUE,
        alternative = "greater")
    )
  )
})

test_that("show: verbose", {
  # 3.ed edition necessary for expect_snapshot
  testthat::local_edition(3)

  expect_snapshot(
    show(
      calc_seq_ttest(
        build_prototype_seq_ttest_arguments()
      )
    )
  )

  expect_snapshot(
    show(
      calc_seq_ttest(
        build_prototype_seq_ttest_arguments(),
        verbose = FALSE
      )
    )
  )


})

# ANOVA ------------------------------------------------------------------------
test_that("show: print output anova?", {
  # 3.ed edition necessary for expect_snapshot
  testthat::local_edition(3)
  set.seed(4657)

  data <- draw_sample_normal(5, 0.25, 66)
  expect_snapshot(
    show(
      seq_anova(
        y ~ x,
        f = 0.25,
        data = data)
    )
  )

  df_job <- draw_sample_normal(2, 0.6, 24)
  colnames(df_job) <- c("happiness", "job_satisfaction")
  expect_snapshot(
    show(
      seq_anova(
        happiness ~ job_satisfaction,
        f = 0.25,
        data = df_job)
    )
  )

  data <- draw_sample_normal(k_groups = 4, f = 0.1, sd = c(1,1,1,1), max_n = 20)
  expect_snapshot(
    show(
      seq_anova(
        y ~ x,
        f = 0.1,
        data = data)
    )
  )

  data <- draw_sample_normal(k_groups = 2, f = 0.5, sd = c(1, 1), max_n = 60)
  expect_snapshot(
    show(
      seq_anova(
        y ~ x,
        f = 0.25,
        alpha = .30,
        power = .95,
        data = data)
    )
  )

  data <- draw_sample_normal(k_groups = 2, f = 0.5, sd = c(1, 1), max_n = 60)
  expect_snapshot(
    show(
      seq_anova(
        y ~ x,
        f = 0.25,
        data = data,
        verbose = FALSE)
    )
  )
})

#---- GENERAL SETTINGS ---------------------------------------------------------
#' @importFrom stats dt df model.frame as.formula runif
#' @importFrom methods callNextMethod new validObject
#' @importFrom MBESS conf.limits.ncf
#' @importFrom glue glue
#' @importFrom knitr opts_chunk
#' @importFrom ggtext element_markdown
#' @importFrom scales percent_format percent
#' @import dplyr ggplot2 gt
#'
#'
#---- MAIN FUNCTION DOCUMENTATION ----------------------------------------------
#' @title Sequential Analysis of Variance
#' @description
#'
#' Performs a sequential one-way fixed effects ANOVA, which is a variant of a Sequential
#' Probability Ratio Test (SPRT). The test allows for continuous
#' monitoring of data collection and provides stopping boundaries based on
#' likelihood ratios, offering efficiency gains over traditional fixed-N designs.
#'
#' The sequential ANOVA continuously evaluates the likelihood ratio after each
#' observation (or group of observations), stopping when sufficient evidence
#' accumulates for either H0 or H1.
#'
#' For methodological details, see Steinhilber et al. (2024)
#' <doi:10.1037/met0000677>. For practical guidance, see
#' `vignette("one_way_anova", package = "sprtt")`.
#'
#' @section Limitations:
#' * Only one-way fixed effects ANOVA is currently supported
#' * Repeated measures ANOVA is not yet implemented
#'
#' @param formula A formula specifying the model (e.g., `outcome ~ group`).
#'   The response variable should be on the left side and the grouping factor
#'   on the right side. Currently only supports one-way designs.
#' @param data A data frame containing the variables specified in the formula.
#'   Missing values (NA) will be removed with a warning
#' @param f Cohen's f (expected minimal effect size or effect size of interest), that defines the H1.
#' @param alpha Type I error rate (alpha level). The probability of rejecting H0
#'   when it is true. Default is 0.05. Must be between 0 and 1.
#' @param power Statistical power (1 - beta), where beta is the Type II error rate.
#'   The probability of correctly rejecting H0 when H1 is true with effect size f.
#'   Default is 0.95. Must be between 0 and 1. Higher values lead to wider
#'   stopping boundaries and potentially larger sample sizes.
#' @param verbose a logical value whether you want a verbose output or not.
#' @param plot Logical. If `TRUE`, stores sequential test statistics at each sequential step
#'   for visualization with `plot_anova()`. This enables retrospective examination
#'   of the decision process but increases computation time for large datasets.
#'   Default is `FALSE`.
#' @param seq_steps Specifies when to calculate test statistics during sequential
#'   testing (only relevant when `plot = TRUE`). Options:
#'   * Vector of integers: Calculate at specific sample sizes (e.g., `c(10, 20, 30)`)
#'   * `"single"`: Calculate after each observation (step size = 1).
#'     Note: Starts at `k_groups * 2` to ensure sufficient data
#'   * `"balanced"`: Calculate when each group gains one observation
#'     (step size = number of groups). Note: Starts at `k_groups * 2`
#'
#'   For unbalanced designs or non-standard sequences, specify custom steps as a vector.
#'
#'@return An object of the S4 class [`seq_anova_results-class`]. Click on the
#' class link to see the full description of the slots.
#' To get access to the object use the
#' `@`-operator or `[]`-brackets instead of `$`.
#' See the examples below.
#'
#' @seealso
#' * [`plot_anova()`] for visualizing sequential ANOVA results
#' * [`plan_sample_size()`] for sample size planning
#' * [`seq_ttest()`] for sequential t-tests
#' * `vignette("one_way_anova", package = "sprtt")` for detailed tutorial
#' * `vignette("plan_sample_size", package = "sprtt")` for planning guidance
#' * Steinhilber et al. (2023) for theoretical background
#'
#' @export
#'
#' @example inst/examples/seq_anova.R


#---- MAIN FUNCTION ------------------------------------------------------------
seq_anova <- function(
  formula,
  f,
  alpha = 0.05,
  power = 0.95,
  data,
  verbose = TRUE,
  plot = FALSE,
  seq_steps = "single"
) {
  # get the original names of the variables
  data_name <- deparse(substitute(formula))

  # calculate the sequential anova
  seq_anova_arguments <-
    build_seq_anova_arguments(
      formula,
      data,
      f,
      alpha,
      power,
      data_name,
      verbose
    )
  seq_anova_results <-
    calc_seq_anova(
      seq_anova_arguments
    )
  if (plot) {
    seq_anova_results <- calc_plot_anova(
      seq_anova_arguments,
      seq_steps
    )
  }
  seq_anova_results
}

test_that("calc_effect_sizes", {
  # 3.ed edition necessary for expect_snapshot
  testthat::local_edition(3)
  set.seed(333)

  expect_error(
    effect_sizes(NULL, 1:10),
    "formula"
  )

  expect_error(
    effect_sizes(x~k, NULL),
    "data"
  )


  formula <- y ~ x
  df_no_effect <- draw_sample_normal(k_groups = 4, f = 0, max_n = 20)
  df_0.01_effect <- draw_sample_normal(k_groups = 4, f = 0.01, max_n = 500)
  df_0.10_effect <- draw_sample_normal(k_groups = 4, f = 0.10, max_n = 200)
  df_0.25_effect <- draw_sample_normal(k_groups = 3, f = 0.25, max_n = 50)
  df_0.40_effect <- draw_sample_normal(k_groups = 5, f = 0.40, max_n = 10)
  df_2_effect <- draw_sample_normal(k_groups = 2, f = 2, max_n = 10)

  expect_equal(ignore_attr = TRUE,
    effect_sizes(formula, df_no_effect)$cohens_f,
    effectsize::cohens_f(verbose = FALSE,
                         lm(formula, df_no_effect))$Cohens_f
  )

  expect_equal(ignore_attr = TRUE,
    effect_sizes(formula, df_0.01_effect)$cohens_f,
    effectsize::cohens_f(verbose = FALSE,
                         lm(formula, df_0.01_effect))$Cohens_f
  )
  expect_equal(ignore_attr = TRUE,
    effect_sizes(formula, df_0.10_effect)$cohens_f,
    effectsize::cohens_f(verbose = FALSE,
                         lm(formula, df_0.10_effect))$Cohens_f
  )
  expect_equal(ignore_attr = TRUE,
    effect_sizes(formula, df_0.25_effect)$cohens_f,
    effectsize::cohens_f(verbose = FALSE,
                         lm(formula, df_0.25_effect))$Cohens_f
  )
  expect_equal(ignore_attr = TRUE,
    effect_sizes(formula, df_0.40_effect)$cohens_f,
    effectsize::cohens_f(verbose = FALSE,
                         lm(formula, df_0.40_effect))$Cohens_f
  )
  expect_equal(ignore_attr = TRUE,
    effect_sizes(formula, df_2_effect)$cohens_f,
    effectsize::cohens_f(verbose = FALSE,
                         lm(formula, df_2_effect))$Cohens_f
  )

})


test_that("snaphot: effect_size", {
  # 3.ed edition necessary for expect_snapshot
  testthat::local_edition(3)
  set.seed(333)

  df_no_effect <- draw_sample_normal(k_groups = 4, f = 0, max_n = 2000)
  df_no_effect_extreme <- draw_sample_normal(k_groups = 4, f = 0, max_n = 20000)
  df_0.01_effect <- draw_sample_normal(k_groups = 4, f = 0.01, max_n = 500)
  df_0.10_effect <- draw_sample_normal(k_groups = 4, f = 0.10, max_n = 200)
  df_0.25_effect <- draw_sample_normal(k_groups = 3, f = 0.25, max_n = 50)
  df_0.40_effect <- draw_sample_normal(k_groups = 5, f = 0.40, max_n = 10)
  df_2_effect <- draw_sample_normal(k_groups = 2, f = 2, max_n = 10)
  formula <- "y ~ x"

  expect_snapshot(effect_sizes(formula, df_no_effect))
  expect_snapshot(effect_sizes(formula, df_no_effect_extreme))
  expect_snapshot(seq_anova(formula, 1, data = df_no_effect_extreme))
  expect_snapshot(seq_anova(formula, 0.01, data = df_0.01_effect))
  expect_snapshot(effect_sizes(formula, df_0.01_effect))
  expect_snapshot(effect_sizes(formula, df_0.10_effect))
  expect_snapshot(effect_sizes(formula, df_0.25_effect))
  expect_snapshot(effect_sizes(formula, df_0.40_effect))
  expect_snapshot(effect_sizes(formula, df_2_effect))

  formula <- y ~ x

  expect_snapshot(effect_sizes(formula, df_no_effect))
  expect_snapshot(effect_sizes(formula, df_0.01_effect))
  expect_snapshot(effect_sizes(formula, df_0.10_effect))
  expect_snapshot(effect_sizes(formula, df_0.25_effect))
  expect_snapshot(effect_sizes(formula, df_0.40_effect))
  expect_snapshot(effect_sizes(formula, df_2_effect))

  set.seed(7)
  df_no_effect <- draw_sample_normal(k_groups = 4, f = 0, max_n = 2000,
                              sd = c(1,2,4,4),
                              sample_ratio = c(1,2,1,2))
  # effect_sizes(formula, df_no_effect)$eta
  results <- seq_anova(formula, 0.10, data =  df_no_effect)
  expect_equal(results@effect_sizes$ci_ncp_upper, 0)
})

#* @testing seq_anova

test_that("Compare ttest with anova results (two groups)", {
  testthat::skip_if_not_installed("testthis")
  # load packages
  # library(testthis)
  # library(dplyr)

  # skip the tests in special situations
  skip_on_cran() # skip on CRAN checks: necessary!
  # skip_on_ci() # skip on continuous platforms like GitHub Actions
  # skip("seq_anova simulation: not written yet") # skip test

  # test setup: simulate the data
  testthat::source_dir(
    "testdata-raw/",
    pattern = "simulation_anova\\.[rR]$",
    env = test_env(),
    chdir = TRUE,
    wrap = TRUE
  )

  results_anova_simulation <- testthis::read_testdata("results_anova_simulation")
  # test calculation anova function vs martins implementation
  expect_equal(
    results_anova_simulation$likelihood_ratio_anova,
    results_anova_simulation$likelihood_ratio
  )
  expect_equal(
    results_anova_simulation$decision_anova,
    results_anova_simulation$decision
  )
  # test data reduction: seq_anova must be more efficient than fixed anova
  # expect less than
  expect_lte(
    mean(results_anova_simulation$sample_size_seq) /
    mean(results_anova_simulation$sample_size_fixed),
    1
  )
  # at least in 50% of the cases the seq_anova is more efficient
  percent_smaller_samples <- mean(results_anova_simulation$sample_smaller)
  expect_gte(percent_smaller_samples, 0.50)

})


test_that("seq_anova: snapshots", {
  # 3.ed edition necessary for expect_snapshot
  testthat::local_edition(3)
  set.seed(4657)

  data <- draw_sample_normal(4, 0.4, 20, sd = c(1,2,3,4), sample_ratio = c(1,2,1,3))
  expect_snapshot(
    seq_anova(y~x, f = 0.40, data = data)
  )

  df_age <- draw_sample_normal(4, 0.4, 20, sd = c(1,2,3,4), sample_ratio = c(1,2,1,3))
  colnames(df_age) <- c("age", "sex")
  expect_snapshot(
    seq_anova(age~sex, f = 0.40, data = df_age)
  )

})

test_that("seq_anova: ttest vs anova", {
  # 3.ed edition necessary for expect_snapshot
  testthat::local_edition(3)
  set.seed(4657)
  data <- draw_sample_normal(2, 0.4, 20)
  results_anova <- seq_anova(y~x, f = 0.40, data = data)
  results_ttest <- seq_ttest(y~x, d = 0.80, data = data)

  expect_equal(results_anova@likelihood_ratio, results_ttest@likelihood_ratio)
  expect_equal(results_anova@likelihood_ratio_log, results_ttest@likelihood_ratio_log)
  expect_equal(results_anova@decision, results_ttest@decision)
  expect_equal(results_anova@A_boundary_log, results_ttest@A_boundary_log)

})

test_that("CLASS_seq_anova_results: Check access to class", {
  data <- draw_sample_normal(4, 0.23, 35)
  results <- seq_anova(y~x, f = 0.25, data = data)

  expect_equal(results["likelihood_ratio_log"], results@likelihood_ratio_log)
  expect_equal(results["decision"], results@decision)
  expect_equal(results["A_boundary_log"], results@A_boundary_log)
  expect_equal(results["B_boundary_log"], results@B_boundary_log)
  expect_equal(results["f"], results@f)
  expect_equal(results["effect_sizes"], results@effect_sizes)
  expect_equal(results["alpha"], results@alpha)
  expect_equal(results["power"], results@power)
  expect_equal(results["likelihood_ratio"], results@likelihood_ratio)
  expect_equal(results["likelihood_1"], results@likelihood_1)
  expect_equal(results["likelihood_0"], results@likelihood_0)
  expect_equal(results["likelihood_1_log"], results@likelihood_1_log)
  expect_equal(results["likelihood_0_log"], results@likelihood_0_log)
  expect_equal(results["non_centrality_parameter"], results@non_centrality_parameter)
  expect_equal(results["F_value"], results@F_value)
  expect_equal(results["df_1"], results@df_1)
  expect_equal(results["df_2"], results@df_2)
  expect_equal(results["ss_effect"], results@ss_effect)
  expect_equal(results["ss_residual"], results@ss_residual)
  expect_equal(results["ss_total"], results@ss_total)
  expect_equal(results["total_sample_size"], results@total_sample_size)
  expect_equal(results["data_name"], results@data_name)
  expect_equal(results["verbose"], results@verbose)

  expect_error(results["gdjhfgd"], "Wrong slot name:")


})

# seq_anova_arguments <- build_prototype_seq_anova_arguments(seed = 56, max_n = 1000, f = 0.40)
# seq_anova_arguments <- build_prototype_seq_anova_arguments(seed = 31, max_n = 90, f = 0)

calc_effect_sizes <- function(seq_anova_arguments, ss_effect, ss_total, F_statistic, decision = "NULL") {
  k_group <- length(unique(seq_anova_arguments@data$factor_A))

  eta_squared <- ss_effect/ss_total
  partial_eta_squared <- (F_statistic$F_value * F_statistic$df_1) / (F_statistic$F_value * F_statistic$df_1 + F_statistic$df_2)
  adjusted_eta_squared = partial_eta_squared - (1 - partial_eta_squared) * F_statistic$df_1 / F_statistic$df_2

  cohens_f <- sqrt(eta_squared/(1-eta_squared))

  # # adjusted f using Correction: Grissom Effect Size for Research 2005
  # cohens_f_unbiased <- ((k_group-1)/seq_anova_arguments@total_sample_size)*(F_statistic$F_value-1)
  # if (cohens_f_unbiased < 0) {
  #   cohens_f_unbiased <- 0
  # }else{
  #   cohens_f_unbiased <- sqrt(cohens_f_unbiased)
  # }

  # adjusted f using Correction: Maxwell et al. 2017
  F_adust <- (seq_anova_arguments@total_sample_size - k_group - 2) / (seq_anova_arguments@total_sample_size - k_group)
  cohens_f_adj <- ((k_group-1)/seq_anova_arguments@total_sample_size)*(F_adust * F_statistic$F_value-1)
  if (cohens_f_adj < 0) {
    cohens_f_adj <- 0
  }else{
    cohens_f_adj <- sqrt(cohens_f_adj)
  }

  cohens_f_median <- MBESS::ci.srsnr(F.value = F_statistic$F_value,
                                     df.1 = F_statistic$df_1,
                                     df.2 = F_statistic$df_2,
                                     N = seq_anova_arguments@total_sample_size,
                                     alpha.lower = 0.5,
                                     alpha.upper = 0
                                     )$Lower
  # fix NA in MBESS only if Maxwell unbiased estimator is 0
  if (is.na(cohens_f_median) & cohens_f_adj == 0) {
    cohens_f_median <- 0
  }



  # # calculate Cohen's f manually
  #
  # n_group <- seq_anova_arguments@data %>%
  #   group_by(factor_A) %>%
  #   mutate(n_group = length(factor_A)) %>%
  #   group_by(n_group, factor_A) %>%
  #   summarise(.groups = "drop") %>%
  #   pull(n_group)
  # sd_group <- seq_anova_arguments@data %>%
  #   group_by(factor_A) %>%
  #   mutate(sd_group = sd(y)) %>%
  #   group_by(sd_group, factor_A) %>%
  #   summarise(.groups = "drop") %>%
  #   pull(sd_group)
  # mean_group <- seq_anova_arguments@data %>%
  #   group_by(factor_A, group_mean_A) %>%
  #   summarise(.groups = "drop") %>%
  #   pull(group_mean_A)
  # pooled_sd_group <- sqrt(
  #   sum(sd_group^2*(n_group - 1)) /
  #     (sum(n_group) - k_group)
  # )
  # sd_means <- sqrt(sum((mean_group - mean(mean_group))^2)/k_group)
  # cohens_f_manual <- sd_means/pooled_sd_group

  # Confidence Interval for the Non Centrality Parameter
  # Using a non central F distribution
  ci_non_centrality_parameter <- MBESS::conf.limits.ncf(
    F_statistic$F_value,
    conf.level = 0.95,
    F_statistic$df_1,
    F_statistic$df_2
  )
  ci_ncp_lower <- ci_non_centrality_parameter$Lower.Limit
  ci_ncp_upper <- ci_non_centrality_parameter$Upper.Limit

  if(!is.null(decision)) {
  # MBESS Package returns NA if the CIs are 0 -> transform them to 0
    if (is.na(ci_ncp_lower) & decision == "accept H0") {ci_ncp_lower = 0}
    if (is.na(ci_ncp_upper) &
       cohens_f < 0.005 &
       decision == "accept H0" &
       ci_ncp_lower == 0
    ) {
      ci_ncp_upper = 0
    }
  }

  ci_cohens_f_lower <- sqrt(ci_ncp_lower/seq_anova_arguments@total_sample_size)
  ci_cohens_f_upper <- sqrt(ci_ncp_upper/seq_anova_arguments@total_sample_size)


  effect_sizes = list(
    "cohens_f" = cohens_f,
    "ci_cohens_f_lower" = ci_cohens_f_lower,
    "ci_cohens_f_upper" = ci_cohens_f_upper,
    "cohens_f_adj" = cohens_f_adj,
    "cohens_f_median" = cohens_f_median,
    # "cohens_f_manual" = cohens_f_manual,
    # "cohens_f_unbiased" = cohens_f_unbiased, # Grissom 2005
    "eta_squared" = eta_squared,
    "partial_eta_squared" = partial_eta_squared,
    "adjusted_eta_squared" = adjusted_eta_squared,
    "ci_ncp_lower" = ci_ncp_lower,
    "ci_ncp_upper" = ci_ncp_upper
  )
  effect_sizes
}


# potentially a public function ------------------------------------------------

effect_sizes <- function(formula, data) {
  if (is.null(formula)) {stop("formula argument is missing")}
  if (is.null(data)) {stop("data argument is missing")}

  formula <- as.formula(formula)

  seq_anova_arguments <-
    build_seq_anova_arguments(
      formula,
      data,
      f = 0.5,
      alpha = 0.1,
      power = 0.1,
      data_name = "test_name",
      verbose = FALSE
    )
  seq_anova_arguments@data <-
    calc_group_means(
      seq_anova_arguments
    )
  ss_effect <-
    calc_ss_effect(
      seq_anova_arguments
    )
  ss_residual <-
    calc_ss_residual(
      seq_anova_arguments
    )
  ss_total <-
    calc_ss_total(
      seq_anova_arguments
    )
  F_statistic <-
    calc_F_statistic_(
      seq_anova_arguments,
      ss_effect,
      ss_residual
    )

  calc_effect_sizes(
    seq_anova_arguments,
    ss_effect,
    ss_total,
    F_statistic)

}
# data = draw_sample_normal(k_groups = 3, f = 0.25, sd = c(1, 1, 1), max_n = 50)
# formula = y~x
# effect_sizes(formula, data)
# effectsize::cohens_f(lm(formula, data))$Cohens_f


# m <- lm(formula, data)
# parameters::model_parameters(anova(m))

setClassUnion("numeric_data.frame", c("numeric","data.frame"))


setClass(
  Class = "seq_anova_arguments",
  slots = c(
    data = "data.frame",
    f = "numeric",
    alpha = "numeric",
    power = "numeric",
    total_sample_size = "numeric",
    data_name = "character",
    verbose = "logical"
  )
)
build_prototype_seq_anova_arguments <- function(seed = 333, max_n = 50, f_sim = 0.4, f_exp = 0.4, k_groups = 4, alpha = 0.05, power = 0.95) {
  set.seed(seed)
  data <- draw_sample_normal(k_groups = k_groups, f = f_sim, max_n = max_n)
  formula <- y~x
  build_seq_anova_arguments(formula, data, f_exp, alpha, power, "test name", TRUE)
}

setValidity(
  Class = "seq_anova_arguments",
  function(object) {
    # correct input arguments
    if(object@f <= 0){stop("f must be greater than 0.")}
    if (!(object@alpha > 0 &&
          object@alpha < 1
    ))
      stop("Invalid argument <alpha>: Probabilities must be in ]0;1[.")
    if (!(object@power > 0 &&
          object@power < 1
    ))
      stop("Invalid argument <power>: Probabilities must be in ]0;1[.")

    TRUE
  })
setMethod(
  f = "initialize",
  signature = "seq_anova_arguments",
  function(.Object, ...) { # '.Object' is necessary and can not replaced by 'x'
    .Object <- callNextMethod() # necessary line
    validObject(.Object)
    .Object
  }
)

#' Method to retrieve the contents of a slot of an object of the
#'  seq_anova_arguments class.
#'
#' @description This method is only used internally to process
#' the input arguments of the [`seq_anova`] function. As a normal user,
#' you can ignore this specific documentation.
#'
#' @param seq_anova_arguments the corresponding class to this method.
#' @param x the seq_anova_arguments object.
#' @param i indices indicating elements to extract.
#' @param j not used.
#' @param drop not used.
#'
#' @return Returns the contents of the specified slot. For more information,
#' see the documentation for the seq_anova_arguments class.
#' @export
#'
#' @keywords internal
# #' @examples
setMethod(
  f = "[",
  signature = "seq_anova_arguments",
  function(x, i, j, drop){ # must be this names!

    if (i == "data") {return(x@data)}
    if (i == "f") {return(x@f)}
    if (i == "alpha") {return(x@alpha)}
    if (i == "power") {return(x@power)}
    if (i == "total_sample_size") {return(x@total_sample_size)}
    if (i == "data_name") {return(x@data_name)}
    if (i == "verbose") {return(x@verbose)}
    stop(paste("Wrong slot name: '", i, "' is not a slot name of the class 'seq_anova_arguments'"))
  }
)

test_that("CLASS_seq_anova_arguments: Check getters", {
  arguments <- build_prototype_seq_anova_arguments()

  expect_equal(arguments["data"], arguments@data)
  expect_equal(arguments["f"], arguments@f)
  expect_equal(arguments["alpha"], arguments@alpha)
  expect_equal(arguments["power"], arguments@power)
  expect_equal(arguments["total_sample_size"], arguments@total_sample_size)
  expect_equal(arguments["data_name"], arguments@data_name)
  expect_equal(arguments["verbose"], arguments@verbose)
  # expect_equal(arguments["na.rm"], arguments@na.rm)

  expect_error(arguments["jewhf"], "Wrong slot name:")
})

# ANOVA ------------------------------------------------------------------------
test_that("calc_group_means", {
  # 3.ed edition necessary for expect_snapshot
  testthat::local_edition(3)
  set.seed(4657)

  seq_anova_arguments <- build_prototype_seq_anova_arguments(33, 100)

  expect_snapshot(
    show(calc_group_means(seq_anova_arguments))
  )
})

# ANOVA ------------------------------------------------------------------------
test_that("calc_ss_residual", {
  # 3.ed edition necessary for expect_snapshot
  testthat::local_edition(3)
  set.seed(333)

  seq_anova_arguments <- build_prototype_seq_anova_arguments()
  seq_anova_arguments@data <-
    calc_group_means(
      seq_anova_arguments
    )

  expect_snapshot(
    show(
      calc_ss_residual(seq_anova_arguments)
    )
  )
})

# seq_anova_arguments <- build_prototype_seq_anova_arguments(max_n = 15, seed = 22)
# seq_anova_arguments <- build_prototype_seq_anova_arguments(
#                         seed = 333, max_n = 50, f_sim = 0.4, f_exp = 0.4,
#                         k_groups = 4,
#                         alpha = 0.05, power = 0.95)
# seq_anova_results <- calc_seq_anova(seq_anova_arguments)
# seq_steps <- "single"
# seq_steps <- "balanced"
# seq_steps <- c(8, 20, 35, 60)

calc_plot_anova <- function(seq_anova_arguments, seq_steps) {
  k_groups <- length(unique(seq_anova_arguments@data$factor_A))
  N <- seq_anova_arguments@total_sample_size
  temp_arguments <- seq_anova_arguments

  if (is.numeric(seq_steps)) {
    seq_steps <- seq_steps
  } else if (seq_steps == "balanced") {
    seq_steps <- seq(k_groups*2, N, k_groups)
  } else if (seq_steps == "single") {
    seq_steps <- (k_groups*2):N
  } else{
    stop("wrong input for seq_steps argument.")
  }

  lr_log <- double(length(seq_steps))
  sample_size <- double(length(seq_steps))
  decision <- character(length(seq_steps))
  i = 1

  k_groups_start <- table(seq_anova_arguments@data[1:seq_steps[1], 2])

  if (any(k_groups_start < 2)) {
    stop("The first 2*k_groups data points are not balanced. Every group needs two data points.
         Solution: Use the argument 'plot' to define a customized vector with the sample size steps in the seq_anova() function")
  }

  for (position in seq_steps) {
    temp_arguments@data <- seq_anova_arguments@data[1:position, ]
    temp_arguments@total_sample_size <- position
    seq_anova_results <- calc_seq_anova(temp_arguments)
    lr_log[i] <- seq_anova_results@likelihood_ratio_log
    decision[i] <- seq_anova_results@decision
    sample_size[i] <- position
    i <- i + 1
  }

  # plot_sprt(sample_size, lr_log,
  #           seq_anova_results@A_boundary_log,
  #           seq_anova_results@B_boundary_log
  # )

  seq_anova_results@plot <- data.frame(
    lr_log = lr_log,
    sample_size = sample_size,
    decision = decision)
  seq_anova_results
}


# ttest ------------------------------------------------------------------------
# seq_ttest_arguments <- build_prototype_seq_ttest_arguments()
# seq_steps <- "balanced"

# calc_plot_ttest <- function(seq_ttest_arguments, seq_steps = "balanced") {
#   N_x <- length(seq_ttest_arguments@x)
#   x <- seq_ttest_arguments@x
#   # c(rbind(seq_ttest_arguments@x, seq_ttest_arguments@y))
#   #
#   # a = c(1,2,3)
#   # b = c(5,5,5,5)
#   #
#   # c(rbind(a,b))
#
#   if (!seq_ttest_arguments@one_sample) {
#     N_y <- length(seq_ttest_arguments@y)
#     if (N_x == N_y) {
#       y <- seq_ttest_arguments@y
#     } else if (N_x < N_y) {
#       x <- rep(NA, N_y)
#       x <- seq_ttest_arguments@x
#       y <- seq_ttest_arguments@y
#     } else if (N_x > N_y) {
#       y <- rep(NA, N_x)
#       y <- seq_ttest_arguments@y
#     }
#   }
#
#   N <- length(x)
#   temp_arguments <- seq_ttest_arguments
#
#   if (is.numeric(seq_steps)) {
#     seq_steps <- seq_steps
#   } else if (seq_steps == "balanced") {
#     seq_steps <- seq(2, N, 1)
#   # } else if (seq_steps == "single") {
#   #   seq_steps <- (4):N
#   } else{
#     stop("XXX")
#   }
#
#
#   lr_log <- double(length(seq_steps))
#   sample_size <- double(length(seq_steps))
#   i = 1
#
#   for (position in seq_steps) {
#     temp_arguments@x <- x[1:position]
#     if (!seq_ttest_arguments@one_sample) {temp_arguments@y <- y[1:position]}
#     seq_ttest_results <- calc_seq_ttest(temp_arguments)
#     lr_log[i] <- seq_ttest_results@likelihood_ratio_log
#     sample_size[i] <- position
#     i <- i + 1
#   }
#
#   seq_ttest_results@plot <- list(lr_log = lr_log,
#                                  sample_size = sample_size,
#                                  A_boundary_log = seq_ttest_results@A_boundary_log,
#                                  B_boundary_log = seq_ttest_results@B_boundary_log)
#   seq_ttest_results
# }

build_seq_anova_results <- function(
  seq_anova_arguments,
  likelihoods,
  boundaries,
  decision,
  non_centrality_parameter,
  ss_effect,
  ss_residual,
  ss_total,
  F_statistic,
  effect_sizes
){

  seq_anova_results <-
    new(
      "seq_anova_results",
      likelihood_ratio = likelihoods["ratio"][[1]],
      likelihood_ratio_log = likelihoods["ratio_log"][[1]],
      decision = decision,
      A_boundary_log = boundaries["A"][[1]],
      B_boundary_log = boundaries["B"][[1]],
      f = seq_anova_arguments["f"],
      effect_sizes = effect_sizes,
      alpha = seq_anova_arguments["alpha"],
      power = seq_anova_arguments["power"],
      likelihood_1 = likelihoods["likelihood_1"][[1]],
      likelihood_0 = likelihoods["likelihood_0"][[1]],
      likelihood_1_log = likelihoods["likelihood_1_log"][[1]],
      likelihood_0_log = likelihoods["likelihood_0_log"][[1]],
      non_centrality_parameter = non_centrality_parameter,
      F_value = F_statistic["F_value"][[1]],
      df_1 = F_statistic["df_1"][[1]],
      df_2 = F_statistic["df_2"][[1]],
      ss_effect = ss_effect,
      ss_residual = ss_residual,
      ss_total = ss_total,
      total_sample_size = seq_anova_arguments["total_sample_size"],
      data_name = seq_anova_arguments["data_name"],
      verbose = seq_anova_arguments["verbose"]
    )
  seq_anova_results
}

rnorm_mix <- function(n, mu1, sig1, mu2, sig2, gamma) {
  x <- sample(c(1,0), n, replace = T, prob = c(gamma, 1 - gamma))
  x <- ifelse(x,
              rnorm(sum(x), mean = mu1, sd = sig1),
              rnorm(sum(!x), mean = mu2, sd = sig2))
  return(x)
}



#' @title Draw Samples from a Gaussian Mixture Distribution
#' @description
#'
#' Draws exemplary samples with a certain effect size for the sequential one-oway ANOVA or the sequential t-test, see Steinhilber et al. (2023) <doi:10.31234/osf.io/m64ne>
#' @param k_groups number of groups (levels of factor_A)
#' @param f Cohen's f. The simulated effect size.
#' @param max_n sample size for the groups (total sample size = max_n*k_groups)
#' @param counter_n number of times the function tries to find a possible parameter combination for the distribution. Default value is set to 100.
#' @param verbose `TRUE` or `FALSE.` Print out more information about the internal process of sampling the parameters
#' (the internal counter that was reached, some additional hints and the drawn parameters for the Gaussian Mixture distributions.)
#'
#' @return returns a data.frame with the columns y (observations) and x (factor_A).
#'
#' @export
#'
#' @example inst/examples/draw_sample_mixture.R



draw_sample_mixture <- function(
  k_groups,
  f,
  max_n,
  counter_n = 100,
  verbose = FALSE
) {
  gamma <- 0.5
  sd <- rep(1, k_groups)
  sample_ratio <- rep(1, k_groups)

  if (!is.numeric(k_groups)) {stop("argument k_groups must be numeric.")}
  if (k_groups <= 1) {stop("argument k_groups must be larger than 1.")}

  if (!is.numeric(f)) {stop("argument f must be numeric.")}
  if (f < 0) {stop("argument f must be equal to or larger than 0.")}

  if (!is.numeric(max_n)) {stop("argument max_n must be numeric.")}
  if (max_n <= 0) {stop("argument max_n must be larger than 1.")}

  if (!is.numeric(counter_n)) {stop("argument counter_n must be numeric.")}
  if (counter_n < 1) {stop("argument counter_n must be equal to or larger than 1.")}

  sample_sizes_max <- max_n * sample_ratio
  total_sample_size <- sum(max_n*sample_ratio)
  if (total_sample_size <= k_groups) {stop("total sample size must be greater than k_groups. Try to increase max_n argument.")}

  seq_step_increase <- sum(sample_ratio)
  seq_steps <- seq((seq_step_increase), total_sample_size, seq_step_increase)
  # seq_steps <- seq_steps[2:length(seq_steps)]


  test <- TRUE
  counter <- 0
  while (any(test) & counter < counter_n) {
    raw_means <- rnorm(k_groups)
    pop_means = (raw_means - mean(raw_means)) / sd(raw_means) * sqrt(k_groups / (k_groups - 1)) * f

    # draw factor randomly & calulate means
    factor <- runif(k_groups, min = 0.8, max = 1)
    mean1 = pop_means*2*factor
    mean2 = pop_means*2*(1-factor)

    test <- (sd^2 - 0.25*(mean1 - mean2)^2) <= 0
    counter <- counter + 1
  }
  if (counter == counter_n) {message("Maximum of counter_n was reached.")}
  if (verbose == TRUE) {message(paste(c("Internal counter reached = ", counter)))}
  if (verbose == TRUE & f > 1.5) {message(paste(c("f values larger than 1.5 should not be used in this function.")))}
  if (any(test)) {stop("Internal calulation failed: try to increase counter_n or decrease f and set verbose to TRUE.")}

  # calculate variance
  factor <- runif(k_groups, min = 0.8, max = 1)
  variance_12 = 2*sd^2-0.5*(mean1-mean2)^2
  sigma1 = sqrt(variance_12 * factor)
  sigma2 = sqrt(variance_12 * (1-factor))

  if (verbose) {message(glue::glue("{rep('\n',k_groups)}group{1:k_groups}:\nmean1 = {mean1}, mean2 = {mean2},\nsigma1 = {sigma1}, sigma2 = {sigma2}"))}

  position <- 1
  data <- matrix(double(total_sample_size*2), ncol = 2)

  for (steps in seq_steps) {

    y <- 1:k_groups %>%
      # purrr::map(~EnvStats::rnormMix(n = sample_ratio[.x], mean1 = mean1[.x], sd1 = sd[.x], mean2 = mean2[.x], sd2 = sd[.x], p.mix = gamma)) %>%
      purrr::map(~rnorm_mix(n = sample_ratio[.x], mu1 = mean1[.x], sig1 = sigma1[.x], mu2 = mean2[.x], sig2 = sigma2[.x], gamma = gamma)) %>%
      unlist()

    indice <- 1:100
    x <- 1:k_groups %>%
      purrr::map(~rep(indice[.x], sample_ratio[.x])) %>%
      unlist()

    data[position:steps, 1] <- y
    data[position:steps, 2] <- x
    position <- position + seq_step_increase
  }
  data.frame(y = data[, 1], x = as.factor(data[, 2]))
}

# data <- draw_sample_mixture(
#     k_groups = 2,
#     f = 0.4,
#     max_n = 10,
#     counter_n = 100,
#     verbose = TRUE
# )


# k_groups = 4
# f = 0.4
# # f = 1.51
# max_n = 50
# counter_n = 100
# verbose = TRUE

#* @testing draw_sample_normal


test_that("I draw_sample_normal: check correct behaviour oh show", {
  set.seed(333)
  # 3.ed edition necessary for expect_snapshot
  testthat::local_edition(3)
  expect_snapshot(
    show(
      draw_sample_normal(k_groups = 2, f = .25, max_n = 30)
    )
  )
  expect_snapshot(
    show(
      draw_sample_normal(k_groups = 4, f = 1, sd = c(1, 0.8, 1, 1), max_n = 30)
    )
  )
  expect_snapshot(
    show(
      draw_sample_normal(k_groups = 3,
                  f = 0.12,
                  sd = c(1, 0.8, 2),
                  sample_ratio = c(1,2,1),
                  max_n = 50)
    )
  )
})

test_that("draw_sample_normal: check error messages", {
  set.seed(333)
  expect_error(
    draw_sample_normal(),
    'argument "k_groups" is missing, with no default'
  )
  expect_error(
    draw_sample_normal(k_groups = 4),
    'argument "f" is missing'
  )
  expect_error(
    draw_sample_normal(k_groups = 3, f = 2),
    'argument "max_n" is missing'
  )

  expect_error(
    draw_sample_normal(k_groups = "3", f = 2, max_n = 10),
    'k_groups must be numeric.'
  )
  expect_error(
    draw_sample_normal(k_groups = 3, f = "2", max_n = 10),
    'f must be numeric.'
  )
  expect_error(
    draw_sample_normal(k_groups = 3, f = 2, max_n = "10"),
    'max_n must be numeric.'
  )

  expect_error(
    draw_sample_normal(k_groups = 0, f = 2, max_n = 10),
    'k_groups must be larger than 1.'
  )
  expect_error(
    draw_sample_normal(k_groups = 1, f = 2, max_n = 10),
    'k_groups must be larger than 1.'
  )
  expect_error(
    draw_sample_normal(k_groups = 2, f = -2, max_n = 10),
    'argument f must be equal to or larger than 0.'
  )
  expect_error(
    draw_sample_normal(k_groups = 3, f = 2, max_n = 0),
    'argument max_n must be larger than 1.'
  )


  expect_error(
    draw_sample_normal(k_groups = 3, f = 10, max_n = 1),
    'total sample size must be greater than k_groups.'
  )
  expect_error(
    draw_sample_normal(k_groups = 3, f = .4, max_n = 40, sd = c(1,1,1,1)),
    'sd'
  )
  expect_error(
    draw_sample_normal(k_groups = 3, f = .4, max_n = 40, sd = c("1,1,1,1"))
    , 'numeric'
  )
  expect_error(
    draw_sample_normal(k_groups = 3, f = .4, max_n = 40, sample_ratio = c(1,1)),
    'sample_ratio'
  )
  expect_error(
    draw_sample_normal(k_groups = 3, f = .4, max_n = 40, sample_ratio = c("1","1")),
    'numeric'
  )
})

test_that("II draw_sample_normal: check correct behaviour of show", {
  set.seed(333)
  # 3.ed edition necessary for expect_snapshot
  testthat::local_edition(3)
  expect_snapshot(
    show(
      draw_sample_normal(k_groups = 2, f = .25, max_n = 30)
    )
  )
  expect_snapshot(
    show(
      draw_sample_normal(k_groups = 4, f = 1, sd = c(1, 0.8, 1, 1), max_n = 30)
    )
  )
  expect_snapshot(
    show(
      draw_sample_normal(k_groups = 3,
                  f = 0.12,
                  sd = c(1, 0.8, 2),
                  sample_ratio = c(1,2,1),
                  max_n = 50)
    )
  )
})

test_that("draw_sample_normal: check error messages", {
  set.seed(333)
  expect_error(
    draw_sample_normal(),
    'argument "k_groups" is missing, with no default'
  )
  expect_error(
    draw_sample_normal(k_groups = 4),
    'argument "f" is missing'
  )
  expect_error(
    draw_sample_normal(k_groups = 3, f = 2),
    'argument "max_n" is missing'
  )

  expect_error(
    draw_sample_normal(k_groups = "3", f = 2, max_n = 10),
    'k_groups must be numeric.'
  )
  expect_error(
    draw_sample_normal(k_groups = 3, f = "2", max_n = 10),
    'f must be numeric.'
  )
  expect_error(
    draw_sample_normal(k_groups = 3, f = 2, max_n = "10"),
    'max_n must be numeric.'
  )

  expect_error(
    draw_sample_normal(k_groups = 0, f = 2, max_n = 10),
    'k_groups must be larger than 1.'
  )
  expect_error(
    draw_sample_normal(k_groups = 1, f = 2, max_n = 10),
    'k_groups must be larger than 1.'
  )
  expect_error(
    draw_sample_normal(k_groups = 2, f = -2, max_n = 10),
    'argument f must be equal to or larger than 0.'
  )
  expect_error(
    draw_sample_normal(k_groups = 3, f = 2, max_n = 0),
    'argument max_n must be larger than 1.'
  )


  expect_error(
    draw_sample_normal(k_groups = 3, f = 10, max_n = 1),
    'total sample size must be greater than k_groups.'
  )
  expect_error(
    draw_sample_normal(k_groups = 3, f = .4, max_n = 40, sd = c(1,1,1,1)),
    'sd'
  )
  expect_error(
    draw_sample_normal(k_groups = 3, f = .4, max_n = 40, sd = c("1,1,1,1"))
    , 'numeric'
  )
  expect_error(
    draw_sample_normal(k_groups = 3, f = .4, max_n = 40, sample_ratio = c(1,1)),
    'sample_ratio'
  )
  expect_error(
    draw_sample_normal(k_groups = 3, f = .4, max_n = 40, sample_ratio = c("1","1")),
    'numeric'
  )
})

# mixture distribution ---------------------------------------------------------
test_that("III draw_sample_mixture: check correct behaviour of show", {
  set.seed(333)
  testthat::local_edition(3)

  round_nums <- function(lines) {
    # keep 3 decimal places, drop the rest (very rough but works here)
    gsub("(-?[0-9]+\\.[0-9]{4})[0-9]*", "\\1", lines)
  }

  expect_snapshot(
    show(
      draw_sample_mixture(k_groups = 2, f = .25, max_n = 30)
    ),
    transform = round_nums
  )

  expect_snapshot(
    suppressMessages(
      show(
        draw_sample_mixture(k_groups = 4, f = 1, max_n = 30, verbose = TRUE)
      )
    ),
    transform = round_nums
  )
  expect_snapshot(
    suppressMessages(
      show(
        draw_sample_mixture(k_groups = 3, f = 0.12, max_n = 50, verbose = TRUE)
      )
    ),
    transform = round_nums
  )
})


test_that("draw_sample_mixture: check error messages", {
  set.seed(333)
  expect_error(
    draw_sample_mixture(),
    'argument "k_groups" is missing, with no default'
  )
  expect_error(
    draw_sample_mixture(k_groups = 4),
    'argument "f" is missing'
  )
  expect_error(
    draw_sample_mixture(k_groups = 3, f = 0.2),
    'argument "max_n" is missing'
  )

  expect_error(
    draw_sample_mixture(k_groups = "3", f = 0.2, max_n = 10),
    'k_groups must be numeric.'
  )
  expect_error(
    draw_sample_mixture(k_groups = 3, f = "0.2", max_n = 10),
    'f must be numeric.'
  )
  expect_error(
    draw_sample_mixture(k_groups = 3, f = 0.4, max_n = "10"),
    'max_n must be numeric.'
  )
  expect_error(
    draw_sample_mixture(k_groups = 3, f = 0.4, max_n = 10, counter_n = "1000"),
    'counter_n must be numeric.'
  )
  expect_error(
    draw_sample_mixture(k_groups = 3, f = 0.4, max_n = 10, counter_n = -2),
    'argument counter_n must be equal to or larger than 1'
  )

  expect_error(
    draw_sample_mixture(k_groups = 0, f = 0.2, max_n = 10),
    'k_groups must be larger than 1.'
  )
  expect_error(
    draw_sample_mixture(k_groups = 1, f = 0.2, max_n = 10),
    'k_groups must be larger than 1.'
  )
  expect_error(
    draw_sample_mixture(k_groups = 2, f = -0.2, max_n = 10),
    'argument f must be equal to or larger than 0.'
  )
  expect_error(
    draw_sample_mixture(k_groups = 3, f = 0.2, max_n = 0),
    'argument max_n must be larger than 1.'
  )
  expect_error(
    draw_sample_mixture(k_groups = 3, f = 0.2, max_n = 1),
    'total sample size must be greater than k_groups.'
  )
  expect_message(
    draw_sample_mixture(k_groups = 4, f = 1.51, max_n = 5,
                        counter_n = 1000000, verbose = TRUE),
  "f values larger than 1.5 should not be used in this function."
  )
  expect_error(
    draw_sample_mixture(k_groups = 4, f = 4, max_n = 5,
                        counter_n = 100, verbose = TRUE),
  "Internal calulation failed:"
  )

})

# check effect size structure od the simulated data ----------------------------

get_seq_decision <- function(likelihood_ratio, boundaries){
  if (likelihood_ratio >= boundaries$A) {
    decision <- "accept H1"
  } else if (likelihood_ratio <= boundaries$B) {
    decision <- "accept H0"
  } else{
    decision <- "continue sampling"
  }
  decision
}

# * @testing get_seq_decision
test_that("get_seq_decision", {
  likelihood_ratio <- 4
  A <- 17
  B <- 1
  A_log <- log(A)
  B_log <- log(B)
  boundaries <- list(A = A,
                    B = B,
                    A_log = A_log,
                    B_log = B_log
                    )
  decision <- get_seq_decision(likelihood_ratio, boundaries)
  expect_equal(decision, "continue sampling")

  likelihood_ratio <- 17
  decision <- get_seq_decision(likelihood_ratio, boundaries)
  expect_equal(decision, "accept H1")


  likelihood_ratio <- 1
  decision <- get_seq_decision(likelihood_ratio, boundaries)
  expect_equal(decision, "accept H0")


})

build_seq_anova_arguments <- function(
  formula,
  data,
  f,
  alpha,
  power,
  data_name,
  verbose
) {
  formula <- formula(formula)
  check_formula_anova(formula = formula, data = data)
  data <- extract_formula_anova(formula = formula, data = data)

  check_data_anova(data)
  seq_anova_arguments <-
    new(
      "seq_anova_arguments",
      data = data,
      f = f,
      alpha = alpha,
      power = power,
      total_sample_size = length(data$y),
      data_name = data_name,
      verbose = verbose
    )
  seq_anova_arguments
}

#* @testing build_seq_anova_arguments

test_that("build_seq_anova_arguments: silent", {
  formula = y ~ x
  data = draw_sample_normal(4, 0.4, 20)
  f = 0.25
  alpha = 0.05
  power = 0.95
  data_name = "test_data"
  verbose = TRUE

  expect_silent(
    build_seq_anova_arguments(
      formula = formula,
      data = data,
      f = f,
      alpha = alpha,
      power = power,
      data_name = data_name,
      verbose = verbose
    )
  )
})

test_that("build_seq_anova_arguments: correct build", {
  formula = y ~ x
  data = draw_sample_normal(4, 0.4, 20)
  f = 0.25
  alpha = 0.05
  power = 0.95
  data_name = "test_data"
  verbose = TRUE

  test <- build_seq_anova_arguments(
    formula = formula,
    data = data,
    f = f,
    alpha = alpha,
    power = power,
    data_name = data_name,
    verbose = verbose
  )

  expect_equal(test@data$y, data$y)
  expect_equal(test@data$factor_A, data$x)
  expect_equal(test@f, f)
  expect_equal(test@alpha, alpha)
  expect_equal(test@power, power)
  expect_equal(test@data_name, data_name)
  expect_equal(test@verbose, verbose)
})


test_that("build_seq_anova_arguments: correct build", {
  formula = 5
  data = draw_sample_normal(4, 0.4, 20)
  f = 0.25
  alpha = 0.05
  power = 0.95
  data_name = "test_data"
  verbose = TRUE

  expect_error(
    build_seq_anova_arguments(
      formula = formula,
      data = data,
      f = f,
      alpha = alpha,
      power = power,
      data_name = data_name,
      verbose = verbose
    ),
    "invalid formula"
  )
})

check_constant_data_ttest <- function(x, y) {
  message <- "Can't perform sequential t-test on constant data."

  if (!is.null(y) && !is.factor(y)) {
    sd.check <- c(sd(x), sd(y))
    sd.check <- ifelse(is.na(sd.check), 0, sd.check)
    if (!(max(sd.check) > 0))
      stop(message)

  } else if (!is.null(y) && length(y) != 1) {
    sd.check <- tapply(x, INDEX = y, FUN = sd)
    sd.check <- ifelse(is.na(sd.check), 0, sd.check)
    if (max(sd.check) == 0)
      stop(message)

  } else if (is.null(y)) {
    sd.check <- ifelse(is.na(sd(x)), 0, sd(x))
    if (!(sd.check > 0))
      stop(message)
  }
}

# ANOVA ------------------------------------------------------------------------
check_constant_data_anova <- function(data) {
  message <- "Can't perform sequential ANOVA on constant data."
  sd.check <- tapply(data$y, INDEX = data$factor_A, FUN = sd)
  sd.check <- ifelse(is.na(sd.check), 0, sd.check)
  if (!(max(sd.check) > 0))
    stop(message)
}

#* @testing check_constant_data_ttest

# t-test -----------------------------------------------------------------------

test_that("check_constant_data_ttest: check error messages", {
  message <- "Can't perform sequential t-test on constant data."

  x <- rep(2, 10)
  y <- NULL
  expect_error(
    check_constant_data_ttest(x, y),
    message
  )

  x <- rep(2, 10)
  y <- rep(2, 10)
  expect_error(
    check_constant_data_ttest(x, y),
    message
  )

  x <- rep(3, 10)
  y <- as.factor(rep(c(1,2), 5))
  expect_error(
    check_constant_data_ttest(x, y),
    message
  )

  x <- rep(2, 10)
  expect_error(
    check_constant_data_ttest(x, 1)
  )

})

test_that("check_constant_data_ttest: silent behaviour: no errors occur", {

  x <- rnorm(10)
  y <- as.factor(sample(c(1,2), 10, replace = TRUE))
  x <- ifelse(y == 1, x + 0.8, x)
  expect_silent(
    check_constant_data_ttest(x, y)
  )

  x <- rnorm(10)
    expect_silent(
    check_constant_data_ttest(x, 1)
  )
})

# ANOVA ------------------------------------------------------------------------

test_that("check_constant_data_anova: check error messages", {
  message <- "Can't perform sequential ANOVA on constant data."

  y <- rep(c(1,2,3), 10)
  factor_A <- as.factor(rep(c(1,2,3), 10))
  data <- data.frame(y, factor_A)
  expect_error(
    check_constant_data_anova(data),
    message
  )
})

build_seq_ttest_arguments <- function(
  input1,
  y = NULL,
  data = NULL,
  mu, d, alpha, power, alternative, paired, data_name, na.rm
){

  if (inherits(input1,"formula")) {
    check_formula_ttest(formula = input1, data = data, paired = paired)
    x <- extract_formula_ttest(formula = input1, data = data, wanted = "x")
    y <- extract_formula_ttest(formula = input1, data = data, wanted = "y")
  } else if (is.numeric(input1)) {
    x <- input1
  } else {
    stop(
      "The class of the input1 argument hast to be either 'formula' or 'numeric'."
      )
  }

  one_sample <- get_one_sample(y)
  x <- delete_na(x, y, one_sample, paired, na.rm, wanted = "x")
  y <- delete_na(x, y, one_sample, paired, na.rm, wanted = "y")
  check_data_ttest(x, y, paired)

  seq_ttest_arguments <-
    new(
      "seq_ttest_arguments",
      x = x,
      y = y,
      mu = mu,
      d = d,
      alpha = alpha,
      power = power,
      alternative = alternative,
      paired = paired,
      one_sample = one_sample,
      data_name = data_name,
      na.rm = na.rm
    )
  seq_ttest_arguments
}

#* @testing build_seq_ttest_arguments

test_that("build_seq_ttest_arguments: errors", {
  input1 <- TRUE
  mu <- 0
  d <- 0.8
  alpha <- .05
  power <- .80
  alternative <- "two.sided"
  paired <- FALSE
  data_name <- "testname"
  expect_error(
    build_seq_ttest_arguments(input1, y = NULL, data = NULL,
                          mu, d, alpha, power, alternative, paired, data_name, na.rm),
    "argument hast to be either 'formula' or 'numeric'"
  )
})

test_that("build_seq_ttest_arguments: numeric", {
  input1 <- rnorm(10)
  mu <- 0
  d <- 0.8
  alpha <- .05
  power <- .80
  alternative <- "two.sided"
  paired <- FALSE
  data_name <- "test name"
  na.rm <- TRUE
  test <- build_seq_ttest_arguments(input1, y = NULL, data = NULL,
                                    mu, d, alpha, power, alternative, paired, data_name, na.rm)
  expect_equal(test@x, input1)
  expect_equal(test@y, NULL)
  expect_equal(test@mu, mu)
  expect_equal(test@d, d)
  expect_equal(test@alpha, alpha)
  expect_equal(test@power, power)
  expect_equal(test@alternative, alternative)
  expect_equal(test@paired, paired)
  expect_equal(test@one_sample, TRUE)
  # expect_equal(test@one_sided, FALSE)


  y <- rnorm(10)
  alternative <- "less"
  paired <- FALSE
  test <- build_seq_ttest_arguments(input1, y, data = NULL,
                                mu, d, alpha, power, alternative, paired, data_name, na.rm)
  expect_equal(test@x, input1)
  expect_equal(test@y, y)
  expect_equal(test@mu, mu)
  expect_equal(test@d, d)
  expect_equal(test@alpha, alpha)
  expect_equal(test@power, power)
  expect_equal(test@alternative, alternative)
  expect_equal(test@paired, paired)
  expect_equal(test@one_sample, FALSE)
  # expect_equal(test@one_sided, TRUE)

  y <- rnorm(10)
  alternative <- "greater"
  paired <- TRUE
  test <- build_seq_ttest_arguments(input1, y = NULL, data = NULL,
                                mu, d, alpha, power, alternative, paired, data_name, na.rm)

  expect_equal(test@alternative, alternative)
  expect_equal(test@one_sample, TRUE)
  # expect_equal(test@one_sided, TRUE)

})

test_that("build_seq_ttest_arguments: formula", {
  input1 <- a~b
  a_1 <- rnorm(5)
  a_2 <- rnorm(5)
  a <- c(a_1, a_2)
  b <- as.factor(c(rep(1,5),rep(2,5)))

  mu <- 0
  d <- 0.8
  alpha <- .05
  power <- .80
  alternative <- "two.sided"
  paired <- FALSE
  data_name <- "test name"
  na.rm <- TRUE
  test <- build_seq_ttest_arguments(input1, y = NULL, data = NULL,
                        mu, d, alpha, power, alternative, paired, data_name, na.rm)
  expect_equal(test@x, a_1)
  expect_equal(test@y, a_2)
  expect_equal(test@mu, mu)
  expect_equal(test@d, d)
  expect_equal(test@alpha, alpha)
  expect_equal(test@power, power)
  expect_equal(test@alternative, alternative)
  expect_equal(test@paired, paired)
  expect_equal(test@one_sample, FALSE)
  # expect_equal(test@one_sided, FALSE)



  c <- rnorm(10)
  data <- data.frame(a,c,b)
  test <- build_seq_ttest_arguments(input1, y = NULL, data,
                        mu, d, alpha, power, alternative, paired, data_name, na.rm)
  expect_equal(test@x, a_1)
  expect_equal(test@y, a_2)

  a_1 <- c(NA, 1:5, NA)
  a_2 <- c(NA, 6:10, NA)
  a <- c(a_1, a_2)
  b <- as.factor(c(rep(1,7),rep(2,7)))

  data <- data.frame(a,b)
  test <- build_seq_ttest_arguments(input1, y = NULL, data,
                                mu, d, alpha, power, alternative, paired, data_name, na.rm)
  expect_equal(test@x, 1:5)
  #expect_equal(test@y, 6:10)

})

# test_that("", {
#
#
# })

delete_na <-  function(
  x,
  y = NULL,
  one_sample,
  paired = FALSE,
  na.rm = TRUE,
  wanted = "both"
) {
  if (na.rm == FALSE) {
    na_exist_x <- any(is.na(x))
    na_exist_y <- any(is.na(y))
    if (na_exist_x == TRUE || na_exist_y == TRUE) {
      stop("The data contain missing values. Use the argument na.rm = TRUE or
           remove the missings beforehand.")
    }
  } else if (na.rm == TRUE) {
    if (one_sample == FALSE && paired == FALSE) {
      x <- x[!is.na(x)]
      y <- y[!is.na(y)]
    } else if (one_sample == FALSE && paired == TRUE) {
      whichNA <- is.na(x) | is.na(y)
      x <- x[!whichNA]
      y <- y[!whichNA]
    } else if (one_sample == TRUE) {
      x <- x[!is.na(x)]
    }
  }

  if (wanted == "x") {
    x
  } else if (wanted == "y") {
    y
  } else {
    list(x, y)
  }
}

#* @testing delete_na


# delete_na: Correct deletion of the missing values in the data.

test_that("delete_na: with similar NAs in x and y", {
  x <- c(1:5,NA,NA,6:10)
  y <- c(1:5,NA,NA,6:10)
  x_test <- delete_na(x,y,one_sample = F, paired = T, wanted = "x")
  y_test <- delete_na(x,y,one_sample = F, paired = T, wanted = "y")
  y_test <- delete_na(x,y,one_sample = F, paired = T, wanted = "y")
  xy_test <- delete_na(x,y,one_sample = F, paired = T)
  expect_equal(x_test, 1:10)
  expect_equal(y_test, 1:10)
  expect_equal(xy_test, list(1:10, 1:10))

})

test_that("delete_na: with different NAs in x and y", {
  x <- c(1:3,NA,NA,4:10)
  y <- c(1:5,NA,NA,6:10)
  x_test <- delete_na(x,y,one_sample = F, paired = T, wanted = "x")
  y_test <- delete_na(x,y,one_sample = F, paired = T, wanted = "y")
  expect_equal(x_test,y_test)
  expect_equal(x_test, c(1:3,6:10))


  x <- c(1:3,NA,NA,4:10)
  y <- c(1:5,NA,NA,6:10)
  x_test <- delete_na(x,y,one_sample = F, paired = F, wanted = "x")
  y_test <- delete_na(x,y,one_sample = F, paired = F, wanted = "y")
  expect_equal(x_test, y_test)
  expect_equal(x_test, c(1:10))

})

test_that("delete_na: with y = 1", {
  x <- c(1:3,NA,NA,4:10)
  y <- 1
  x_test <- delete_na(x,y,one_sample = T, paired = F, wanted = "x")
  y_test <- delete_na(x,y,one_sample = T, paired = F, wanted = "y")
  expect_equal(x_test,1:10)
  expect_equal(y_test, 1)
})

test_that("delete_na: with y = NULL", {
  x <- c(1:3,NA,NA,4:10)
  y <- NULL
  x_test <- delete_na(x,y,one_sample = T, paired = F, wanted = "x")
  y_test <- delete_na(x,y,one_sample = T, paired = F, wanted = "y")
  expect_equal(x_test,1:10)
  expect_equal(y_test, NULL)
})

test_that("delete_na: with na.rm = FALSE", {
  x <- c(1:3,NA,NA,4:10)
  y <- 1:10
  expect_error(
    delete_na(x,y,one_sample = T, paired = F, na.rm = FALSE),
    "The data contain missing values.")

  x <- 1:10
  y <- c(1:3,NA,NA,4:10)
  expect_error(
    delete_na(x,y,one_sample = T, paired = F, na.rm = FALSE),
    "The data contain missing values.")

  x <- 1:10
  y <- x
  x_test <- delete_na(x,y,one_sample = F, paired = F, na.rm = FALSE, wanted = "x")
  y_test <- delete_na(x,y,one_sample = F, paired = F, na.rm = FALSE, wanted = "y")
  expect_equal(x_test, 1:10)
  expect_equal(y_test, 1:10)


})

calc_F_statistic_ <- function(
  seq_anova_arguments,
  ss_effect,
  ss_residual
) {

  n_levels_factor_A <- length(unique(seq_anova_arguments@data$factor_A))

  df_1 <- n_levels_factor_A - 1
  df_2 <- seq_anova_arguments@total_sample_size - n_levels_factor_A
  F_value <- (ss_effect / df_1) / (ss_residual / df_2)

  return(list(
    F_value = F_value,
    df_1 = df_1,
    df_2 = df_2
  ))
}

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @importFrom lifecycle deprecated
## usethis namespace: end
NULL

#' sprtt: help page.
#'
#' This package provides the implementation of sequential probability
#' ratio tests using t-statistic.
#'
#' @section detailed help pages:
#' For detailed instructions on the `sprtt` package, see:
#' `vignette("sprtt")` or https://meikesteinhilber.github.io/sprtt/"
#'
#' @section sprtt functions:
#' * [`seq_ttest`] Performs one and two sample sequential t-tests on vectors of
#' data.
#' * [`seq_anova`] Performs a sequential one-way ANOVA on vectors of
#' data.
#'
#' @docType package
#' @name sprtt
#'

# t-test -----------------------------------------------------------------------
calc_likelihoods_ttest <- function(
  seq_ttest_arguments,
  t_statistic,
  df,
  non_centrality_parameter,
  wanted = NULL
){
  t_val <- t_statistic["statistic"][[1]][[1]]
  ncp <- non_centrality_parameter

  if (seq_ttest_arguments@alternative == "two.sided") {
    likelihood_1 <- df(t_val^2, df1 = 1, df2 = df, ncp = ncp^2)
    likelihood_0 <- df(t_val^2, df1 = 1, df2 = df)
    likelihood_1_log <- df(t_val^2, df1 = 1, df2 = df, ncp = ncp^2, log = TRUE)
    likelihood_0_log <- df(t_val^2, df1 = 1, df2 = df, log = TRUE)

  } else{
    likelihood_1 <- dt(t_val, df, ncp = ncp)
    likelihood_0 <- dt(t_val, df)
    likelihood_1_log <- dt(t_val, df, ncp = ncp, log = TRUE)
    likelihood_0_log <- dt(t_val, df, log = TRUE)
  }

  ratio <- likelihood_1/likelihood_0
  ratio_log <- likelihood_1_log - likelihood_0_log

  if (likelihood_1 == 0 ||
      likelihood_0 == 0
  ) {
    warning("At least one likelihood is equal to 0.
            The test works with the logarithm of the likelihoods.")
  }

  if (likelihood_1_log == -Inf ||
      likelihood_0_log == -Inf ||
      likelihood_1_log ==  Inf ||
      likelihood_0_log ==  Inf
  ) {
    warning("At least one log-likelihood reached infinity.")
  }

  return(list(
    likelihood_1 = likelihood_1,
    likelihood_0 = likelihood_0,
    ratio = ratio,
    likelihood_1_log = likelihood_1_log,
    likelihood_0_log = likelihood_0_log,
    ratio_log = ratio_log))

}

# ANOVA ------------------------------------------------------------------------

calc_likelihoods_anova <- function(
  seq_anova_arguments,
  non_centrality_parameter,
  F_statistic
) {

  likelihood_1 <- df(F_statistic$F_value, F_statistic$df_1, F_statistic$df_2, non_centrality_parameter)
  likelihood_0 <- df(F_statistic$F_value, F_statistic$df_1, F_statistic$df_2)
  likelihood_1_log <- df(F_statistic$F_value, F_statistic$df_1, F_statistic$df_2, non_centrality_parameter, log = TRUE)
  likelihood_0_log <- df(F_statistic$F_value, F_statistic$df_1, F_statistic$df_2, log = TRUE)
  likelihood_ratio <- likelihood_1 / likelihood_0
  likelihood_ratio_log <- likelihood_1_log - likelihood_0_log

  if (likelihood_1 == 0 ||
      likelihood_0 == 0
  ) {
    warning("At least one likelihood is equal to 0.
            The test works with the logarithm of the likelihoods.")
  }

  if (likelihood_1_log == -Inf ||
      likelihood_0_log == -Inf ||
      likelihood_1_log ==  Inf ||
      likelihood_0_log ==  Inf
  ) {
    warning("At least one log-likelihood reached infinity.")
  }

  return(list(
    likelihood_1 = likelihood_1,
    likelihood_0 = likelihood_0,
    ratio = likelihood_ratio,
    likelihood_1_log = likelihood_1_log,
    likelihood_0_log = likelihood_0_log,
    ratio_log = likelihood_ratio_log))
}

#--- GENERAL SETTINGS ----
#' @importFrom stats rnorm dt df model.frame sd t.test
#' @importFrom methods callNextMethod new validObject
# #' @useDynLib sprtt
# #' @importFrom Rcpp sourceCpp
# NULL
# #> NULL


#---- MAIN FUNCTION DOCUMENTATION----
#' @title Sequential Probability Ratio Test using t-statistic
#'
#' @description
#' Performs one-sample, two-sample, and paired sequential t-tests, which are variants of
#' Sequential Probability Ratio Tests (SPRT). The test allows for continuous
#' monitoring of data collection and provides stopping boundaries based on
#' likelihood ratios, offering efficiency gains over traditional fixed-N designs.
#'
#' The sequential t-test continuously evaluates the likelihood ratio after each
#' observation (or pair of observations), stopping when sufficient evidence
#' accumulates for either H0 or H1.
#'
#' For methodological details, see Schnuerch & Erdfelder (2019)
#' <doi:10.1037/met0000234>. For practical guidance, see
#' `vignette("t_test", package = "sprtt")`.
#'
#' @param x Works with two classes: `numeric` and `formula`. Therefore you can
#' write `"x"` or `"x~y"`.
#' * `"numeric input"`: a (non-empty) numeric vector of data values.
#' * `"formula input"`: a formula of the form lhs ~ rhs where lhs is a numeric
#' variable giving the data values and rhs either 1 for a one-sample test or a
#' factor with two levels giving the corresponding groups.
#' @param y An optional (non-empty) numeric vector of data values.
#'   Only used for two-sample tests when `x` is numeric.
#' @param data An optional data frame containing the variables specified in the formula.
#'   Only used when `x` is a formula.
#' @param mu a number indicating the true value of the mean (or difference in
#' means if you are performing a two sample test).
#' @param d a number indicating the specified (expected) effect size (Cohen's d)
#' @param alpha Type I error rate (alpha level). The probability of rejecting H0
#'   when it is true. Default is 0.05. Must be between 0 and 1.
#' @param power Statistical power (1 - beta), where beta is the Type II error rate.
#'   The probability of correctly rejecting H0 when H1 is true with effect size d.
#'   Default is 0.95. Must be between 0 and 1. Higher values lead to wider
#'   stopping boundaries and potentially larger sample sizes.
#' @param alternative a character string specifying the alternative hypothesis,
#' must be one of `two.sided` (default), `greater` or `less`.
#' You can specify just the initial letter.
#' @param paired Logical indicating whether to perform a paired t-test.
#'   Default is `FALSE`.
#' @param na.rm a logical value indicating whether `NA` values should be
#' removed before the computation proceeds.
#' @param verbose a logical value whether you want a verbose output or not.
#'
#' @return An object of the S4 class [`seq_ttest_results-class`]. Click on the
#' class link to see the full description of the slots.
#' To get access to the object use the
#' `@`-operator or `[]`-brackets instead of `$`.
#' See the examples below.
#'
#' @seealso
#' * [`seq_anova()`] for sequential one-way ANOVA
#' * [`draw_sample_normal()`] for simulating test data
#' * `vignette("t_test", package = "sprtt")` for detailed tutorial
#' * `vignette("usage_sprtt", package = "sprtt")` for package overview
#' * Schnuerch & Erdfelder (2019) <doi:10.1037/met0000234> for theoretical background
#' @export
#'
#' @example inst/examples/seq_ttest.R


#---- MAIN FUNCTION ----
seq_ttest <- function(
  x,
  y = NULL,
  data = NULL,
  mu = 0,
  d,
  alpha = 0.05,
  power = 0.95,
  alternative = "two.sided",
  paired = FALSE,
  na.rm = TRUE,
  verbose = TRUE
){
  # get the original names of the variables
  input1_name <- deparse(substitute(x))
  if (is.numeric(x) &&
      !is.null(y)) {
    y_name <- deparse(substitute(y))
    data_name <- paste(input1_name, "and ", y_name)
  } else{
    data_name <- paste(input1_name)
  }

  # calculate the sequential test
  seq_ttest_arguments <-
    build_seq_ttest_arguments(
      input1 = x,
      y,
      data,
      mu,
      d,
      alpha,
      power,
      alternative,
      paired,
      data_name,
      na.rm
  )
  seq_ttest_results <-
    calc_seq_ttest(
      seq_ttest_arguments,
      verbose
    )
  seq_ttest_results
}

#* @testing seq_ttest
#*

# MARTIN Skript ----------------------------------------------------------

.sprt.lr <- function(x, y, mu, d, type, alt){
  ttest <- switch(type,
                  one.sample = t.test(x, mu = mu),
                  two.sample = t.test(x, y, mu = mu,
                                      var.equal = T),
                  paired = t.test(x, y, mu = mu,
                                  paired = T))
  # print(ttest)
  ncp <- ifelse(type == "two.sample",
                d/sqrt(1/length(x) + 1/length(y)),
                d * sqrt(length(x)))

  tval <- ifelse(alt == "less",
                 -1 * as.vector(ttest$statistic),
                 as.vector(ttest$statistic))

  dof <- as.vector(ttest$parameter)

  # print(paste("ncp", ncp))
  # print(paste("tval", tval))
  # print(paste("dof", dof))

  if(alt=="two.sided"){
    lr <- df(tval^2, df1 = 1, df2 = dof, ncp = ncp^2)/df(tval^2, df1 = 1, df2 = dof)
    # l1 <- df(tval^2, df1 = 1, df2 = dof, ncp = ncp^2)
    # l2 <- df(tval^2, df1 = 1, df2 = dof)
    # # print(l1/l2)
  } else{
    lr <- dt(tval, dof, ncp = ncp)/dt(tval, dof)
  }
  lr
  # print(lr)
}


.sprt.formula <- function(formula, data = NULL,
                          mu = 0, d, alpha = 0.05, power = 0.95,
                          alternative = "two.sided", paired = FALSE){

  ##### CHECK INPUT
  match.arg(alternative, c("two.sided", "less", "greater"))
  if(!(alpha > 0 && alpha < 1))
    stop("Invalid argument <alpha>: Probabilities must be in ]0;1[.")
  if(!(power > 0 && power < 1))
    stop("Invalid argument <power>: Probabilities must be in ]0;1[.")
  if(d <= 0)
    stop("Invalid argument <d>: Must be greater than 0.")
  if(!is.numeric(mu))
    stop("Invalid argument <mu>: Must be numeric.")
  if(!is.logical(paired))
    stop("Invalid argument <paired>: Must be logical.")

  if ((length(formula) != 3L) || (length(formula[[3]]) != 1L))
    stop("'formula' is incorrect. Please specify as 'x~y'.")
  temp <- model.frame(formula, data)
  x <- temp[,1]
  y <- temp[,2]

  whichNA <- is.na(x) | is.na(y)
  x <- x[!whichNA]
  y <- y[!whichNA]

  if(!is.numeric(x))
    stop(paste("Dependent variable",  names(temp)[1], "must be numeric."))
  if(length(unique(y))!=2)
    stop(paste("Grouping factor", names(temp)[2], "must contain exactly two levels."))
  if(paired){
    if(!(table(y)[[1]]==table(y)[[2]]))
      stop("Unequal number of observations per group. Independent samples?")
  }else{
    if(length(x)<3)
      stop("SPRT for two independent samples requires at least 3 observations.")
  }

  sd.check <- tapply(x, INDEX=y, FUN=sd)
  sd.check <- ifelse(is.na(sd.check), 0, sd.check)
  if(max(sd.check) == 0)
    stop("Can't perform SPRT on constant data.")

  ##### RETURN ARGUMENTS
  y <- as.factor(y)
  x.1 <- x[y == levels(y)[1]]
  x.2 <- x[y == levels(y)[2]]
  data.name <- paste(names(temp)[1], "by", names(temp)[2])
  method <- ifelse(paired, "Paired SPRT t-test", "Two-Sample SPRT t-test")
  null.value <- mu
  attr(null.value, "names") <- "difference in means"
  if(paired){
    estimate <- mean(x.1 - x.2)
    attr(estimate, "names") <- "mean of the differences"
  }else{
    estimate <- c(mean(x.1), mean(x.2))
    attr(estimate, "names") <- c("mean in group 1", "mean in group 2")
  }
  arg.list <- list(x = x.1, y = x.2,
                   mu = mu, d = d, alpha = alpha, power = power,
                   type = ifelse(paired, "paired", "two.sample"),
                   alt = alternative)
  printarg.list <- list(estimate = estimate,
                        null.value = null.value,
                        alternative = alternative,
                        effect.size = d,
                        method = method,
                        data.name = data.name)

  result <- do.call(.sprt.result, arg.list)
  output <- c(result, printarg.list)
  class(output) <- "sprt"
  return(output)
}


.sprt.default <- function(x, y = NULL,
                          mu = 0, d, alpha = 0.05, power = 0.95,
                          alternative = "two.sided", paired = FALSE){

  ##### CHECK INPUT
  match.arg(alternative, c("two.sided", "less", "greater"))
  if(!(alpha > 0 && alpha < 1))
    stop("Invalid argument <alpha>: Probabilities must be in ]0;1[.")
  if(!(power > 0 && power < 1))
    stop("Invalid argument <power>: Probabilities must be in ]0;1[.")
  if(d<=0)
    stop("Invalid argument <d>: Must be greater than 0.")
  if(!is.numeric(mu))
    stop("Invalid argument <mu>: Must be numeric.")

  if(!is.null(y)){
    x.name <- deparse(substitute(x))
    y.name <- deparse(substitute(y))
    data.name <- paste(x.name, "and", y.name)

    if(!(is.atomic(x) && is.null(dim(x))))
      warning(paste(x.name, "is not a vector. This might have caused problems."), call. = F)
    if(!(is.atomic(y) && is.null(dim(y))))
      warning(paste(y.name, "is not a vector. This might have caused problems."), call. = F)
    if(is.factor(y))
      stop("Is y a grouping factor? Use formula interface x ~ y.")
    if(!is.numeric(x))
      stop(paste("Invalid argument:",  x.name, "must be numeric."))
    if(!is.numeric(y))
      stop(paste("Invalid argument:",  y.name, "must be numeric."))
    if(!paired && (length(x) + length(y) < 3))
      stop("SPRT for two independent samples requires at least 3 observations.")

    sd.check <- c(sd(x), sd(y))
    sd.check <- ifelse(is.na(sd.check), 0, sd.check)
    if(!(max(sd.check) > 0))
      stop("Can't perform SPRT on constant data.")
    if(!is.logical(paired))
      stop("Invalid argument <paired>: Must be logical.")

    type <- ifelse(paired, "paired", "two.sample")
    method <- ifelse(paired, "Paired SPRT t-test", "Two-Sample SPRT t-test")
    null.value <- mu
    attr(null.value, "names") <- "difference in means"
    if(paired){
      if(length(x) != length(y))
        stop("Unequal number of observations per group. Independent samples?")
      whichNA <- is.na(x) | is.na(y)
      x <- x[!whichNA]
      y <- y[!whichNA]
      estimate <- mean(x - y)
      attr(estimate, "names") <- "mean of the differences"
    }else{
      x <- x[!is.na(x)]
      y <- y[!is.na(y)]
      estimate <- c(mean(x), mean(y))
      attr(estimate, "names") <- c(paste("mean of", x.name), paste("mean of", y.name))
    }
  }else{
    data.name <- deparse(substitute(x))

    x <- x[!is.na(x)]
    if(!is.numeric(x))
      stop(paste("Invalid argument:",  data.name, "must be numeric."))
    sd.check <- ifelse(is.na(sd(x)), 0, sd(x))
    if(!(sd.check > 0))
      stop("Can't perform SPRT on constant data.")

    type <- "one.sample"
    method <- "One-Sample SPRT t-test"
    null.value <- mu
    attr(null.value, "names") <- "mean"
    estimate <- mean(x)
    attr(estimate, "names") <- "mean of x"
  }

  ##### RETURN ARGUMENTS

  arg.list <- list(x = x, y = y,
                   mu = mu, d = d, alpha = alpha, power = power,
                   type = type,
                   alt = alternative)
  printarg.list <- list(estimate = estimate,
                        null.value = null.value,
                        alternative = alternative,
                        effect.size = d,
                        method = method,
                        data.name = data.name)
  result <- do.call(.sprt.result, arg.list)
  output <- c(result, printarg.list)
  class(output) <- "sprt"
  return(output)
}


.sprt.result <- function(x, y, mu, d, alpha, power, type, alt){
  A <- power/alpha
  B <- (1 - power)/(1 - alpha)
  lr <- .sprt.lr(x, y, mu, d, type, alt)
  if(lr >= A){
    decision <- "accept H1"
  }else if(lr <= B){
    decision <- "accept H0"
  }else{
    decision <- "continue sampling"
  }
  attr(lr, "names") <- "likelihood ratio"
  parameters <- c(alpha, power)
  attr(parameters, "names") <- c("Type I error", "Power")
  thresholds <- c(B, A)
  attr(thresholds, "names") <- c("lower", "upper")
  return(list(statistic = lr,
              decision = decision,
              parameters = parameters,
              thresholds = thresholds))
}


print.sprt <- function(x){
  cat("    ", x$method, "\n")
  cat("\ndata:", x$data.name, "\n")
  cat(names(x$statistic), " = ", round(x$statistic, digits = 5), ", decision = ", x$decision, sep="")
  cat("\nSPRT thresholds:\n")
  print(round(x$thresholds, digits = 5))
  cat("alternative hypothesis: true", names(x$null.value), "is",
      ifelse(x$alternative=="two.sided", "not equal to", paste(x$alternative, "than")), x$null.value)
  cat("\neffect size: Cohen's d =", x$effect.size, "\n")
  print(x$parameters)
  cat("sample estimates:\n")
  print(round(x$estimate, digits = 5))
}



sprt.t.test <- function(...) UseMethod(".sprt")
.sprt.result <- function(x, y, mu, d, alpha, power, type, alt){
  A <- power/alpha
  B <- (1 - power)/(1 - alpha)
  lr <- .sprt.lr(x, y, mu, d, type, alt)
  if(lr >= A){
    decision <- "accept H1"
  }else if(lr <= B){
    decision <- "accept H0"
  }else{
    decision <- "continue sampling"
  }
  attr(lr, "names") <- "likelihood ratio"
  parameters <- c(alpha, power)
  attr(parameters, "names") <- c("Type I error", "Power")
  thresholds <- c(B, A)
  attr(thresholds, "names") <- c("lower", "upper")
  return(list(statistic = lr,
              decision = decision,
              parameters = parameters,
              thresholds = thresholds))
}


print.sprt <- function(x){
  cat("    ", x$method, "\n")
  cat("\ndata:", x$data.name, "\n")
  cat(names(x$statistic), " = ", round(x$statistic, digits = 5), ", decision = ", x$decision, sep="")
  cat("\nSPRT thresholds:\n")
  print(round(x$thresholds, digits = 5))
  cat("alternative hypothesis: true", names(x$null.value), "is",
      ifelse(x$alternative=="two.sided", "not equal to", paste(x$alternative, "than")), x$null.value)
  cat("\neffect size: Cohen's d =", x$effect.size, "\n")
  print(x$parameters)
  cat("sample estimates:\n")
  print(round(x$estimate, digits = 5))
}



sprt.t.test <- function(...) UseMethod(".sprt")

# x <- rnorm(50)
# sprt.t.test(x, d = 0.3)
#-----------------------------------------------------------------


test_that("seq_ttest: comparison results with original script from m. schnuerch", {

  x <- rnorm(50)
  d <- 0.8
  results_original <- sprt.t.test(x = x, d = d, power = 0.8)
  results_sprtt <- seq_ttest(x, d = d, power = 0.8)
  expect_true(results_sprtt@likelihood_ratio - results_original$statistic[[1]] < 1e-5)
  expect_equal(results_sprtt@decision,
               results_original$decision)


  x <- rnorm(20)
  y <- as.factor(c(rep(1,10), rep(2,10)))
  d <- 0.95
  results_original <- sprt.t.test(x ~ y, d = d)
  results_sprtt <- seq_ttest(x ~ y, d = d)
  expect_true(results_sprtt@likelihood_ratio - results_original$statistic[[1]] < 1e-5)
  expect_equal(results_sprtt@decision,
               results_original$decision)
  # same data, but different input
  x_1 <- x[1:(length(x) * 0.5)]
  x_2 <- x[(length(x) * 0.5 + 1):length(x)]
  results_sprtt2 <- seq_ttest(x_1, x_2, d = d)
  expect_true(results_sprtt@likelihood_ratio_log - results_sprtt2@likelihood_ratio_log < 1e-5)
  expect_equal(results_sprtt@decision,
              results_sprtt2@decision)

  x <- rnorm(20)
  y <- as.factor(c(rep(1,10), rep(2,10)))
  d <- 0.95
  results_numeric <- seq_ttest(x, d = d)
  results_formula <- seq_ttest(x ~ 1, d = d)
  expect_true(results_formula@likelihood_ratio - results_numeric@likelihood_ratio < 1e-5)
  expect_equal(results_sprtt@decision,
               results_original$decision)

  x_1 <- rnorm(5)
  x_2 <- rnorm(5)
  x <- c(x_1, x_2)
  y <- as.factor(c(rep(1,5),rep(2,5)))
  results_original <- sprt.t.test(x ~ y, d = d)
  results_sprtt <- seq_ttest(x ~ y, d = d)
  expect_true(results_sprtt@likelihood_ratio - results_original$statistic[[1]] < 1e-5)
  expect_equal(results_sprtt@decision,
               results_original$decision)


  x_1 <- rnorm(10)
  x_2 <- rnorm(10)
  x <- c(x_1, x_2)
  y <- as.factor(c(rep(1,10),rep(2,10)))
  results_original <- sprt.t.test(x_1, x_2, d = d)
  results_sprtt <- seq_ttest(x_1, x_2, d = d)
  expect_true(results_sprtt@likelihood_ratio - results_original$statistic[[1]] < 1e-5)
  expect_equal(results_sprtt@decision,
               results_original$decision)
  # same data, but different input
  results_sprtt <- seq_ttest(x_1, x_2, d = 0.5, paired = TRUE)
  results_sprtt2 <- seq_ttest(x ~ y, d = 0.5, paired = TRUE)
  expect_true(results_sprtt@likelihood_ratio_log - results_sprtt2@likelihood_ratio_log < 1e-5)
  expect_equal(results_sprtt@decision,
               results_sprtt2@decision)

  d <- 0.7
  x <- rnorm(30)
  y <- rnorm(30)
  paired = TRUE
  t_test <- t.test(x, y, paired = paired)
  results_original <- sprt.t.test(x, y, d = d, paired = paired, alt = "two.sided")
  results_sprtt <- seq_ttest(x, y, d = d, paired = paired, alt = "two.sided")
  expect_true(results_sprtt@likelihood_ratio - results_original$statistic[[1]] < 1e-5)
  expect_equal(results_sprtt@decision,
               results_original$decision)

  results_original <- sprt.t.test(x, y, d = d, paired = paired, alt = "less")
  results_sprtt <- seq_ttest(x, y, d = d, paired = paired, alt = "less")
  expect_true(results_sprtt@likelihood_ratio - results_original$statistic[[1]] < 1e-5)
  expect_equal(results_sprtt@decision,
               results_original$decision)

  results_original <- sprt.t.test(x, y, d = d, paired = paired, alt = "greater")
  results_sprtt <- seq_ttest(x, y, d = d, paired = paired, alt = "greater")
  expect_true(results_sprtt@likelihood_ratio - results_original$statistic[[1]] < 1e-5)
  expect_equal(results_sprtt@decision,
               results_original$decision)

})


# test_that("", {
#
#
# })

test_that("CLASS_seq_ttest_results: Check access to class", {
  x <- rnorm(20)
  d <- 0.8
  results <- seq_ttest(x, d = d)

  expect_equal(results["likelihood_ratio_log"], results@likelihood_ratio_log)
  expect_equal(results["decision"], results@decision)
  expect_equal(results["A_boundary_log"], results@A_boundary_log)
  expect_equal(results["B_boundary_log"], results@B_boundary_log)
  expect_equal(results["d"], results@d)
  expect_equal(results["mu"], results@mu)
  expect_equal(results["alpha"], results@alpha)
  expect_equal(results["power"], results@power)
  expect_equal(results["likelihood_ratio"], results@likelihood_ratio)
  expect_equal(results["likelihood_1"], results@likelihood_1)
  expect_equal(results["likelihood_0"], results@likelihood_0)
  expect_equal(results["likelihood_1_log"], results@likelihood_1_log)
  expect_equal(results["likelihood_0_log"], results@likelihood_0_log)
  expect_equal(results["non_centrality_parameter"], results@non_centrality_parameter)
  expect_equal(results["t_value"], results@t_value)
  expect_equal(results["p_value"], results@p_value)
  expect_equal(results["df"], results@df)
  expect_equal(results["mean_estimate"], results@mean_estimate)
  expect_equal(results["alternative"], results@alternative)
  expect_equal(results["one_sample"], results@one_sample)
  expect_equal(results["ttest_method"], results@ttest_method)
  expect_equal(results["data_name"], results@data_name)
  expect_equal(results["verbose"], results@verbose)

  expect_error(results["gdjhfgd"], "Wrong slot name:")


})

#
# setGeneric("test_function",
#            function(object, ..., verbose = FALSE) standardGeneric("test_function"),
#            signature = "object"
# )
#
#
# setMethod("test_function",
#           signature = "seq_ttest_arguments",
#           function(object, data_, verbose) {
#             print(object@paired)
#             print(data_)
#             print(verbose)
# })
# setMethod("test_function",
#           signature = "seq_anova_arguments",
#           function(object, verbose) {
#             print(object@f)
#             print(verbose)
#
# })
#
# seq_ttest_arguments <- build_prototype_seq_ttest_arguments()
# seq_anova_arguments <- build_prototype_seq_anova_arguments()
#
# data <- rnorm(10)
# test_function(seq_ttest_arguments, data, )
# test_function(seq_anova_arguments, verbose = "TRUES")
#

# Test messwiederholte ANOVA Daten -----------------------------------------------

# tdef <- defData(varname = "T", dist = "binary", formula = 0.5)
# tdef <- defData(tdef, varname = "Y0", dist = "normal", formula = 10, variance = 1)
# tdef <- defData(tdef, varname = "Y1", dist = "normal", formula = "Y0 + 5 + 5 * T", variance = 1)
# # tdef <- defData(tdef, varname = "Y2", dist = "normal", formula = "Y0 + 10 + 5 * T", variance = 1)
# dtTrial <- genData(5, tdef)
# dtTrial
# dtTime <- addPeriods(dtTrial,
#                      nPeriods = 2, idvars = "id",
#                      timevars = c("Y0", "Y1"), timevarName = "Y"
# )
# dtTime$period <- as.factor(dtTime$period)
#
# results <- sprtt::seq_ttest(Y~period, data = dtTime, d = 0.2, paired = TRUE)
# results

calc_boundaries <- function(power, alpha, log = FALSE){
  A <- power / alpha
  B <- (1 - power) / (1 - alpha)

  if (log == FALSE) {
    list(
      A = A,
      B = B
    )
  } else{
    list(
      A = log(A),
      B = log(B)
    )
  }
}

test_that("calc_boundaries: ", {
  power <- 0.95
  alpha <- 0.05
  boundaries <- calc_boundaries(power, alpha)
  boundaries_log <- calc_boundaries(power, alpha, log = TRUE)
  expect_equal(log(boundaries$A), boundaries_log$A)
  expect_equal(log(boundaries$B), boundaries_log$B)
  expect_equal(boundaries_log$A, boundaries_log$B * -1) #log boundaries ar symmetrical if alpha & beta are equal

  power <- 0.80
  alpha <- 0.10
  boundaries <- calc_boundaries(power, alpha)
  boundaries_log <- calc_boundaries(power, alpha, log = TRUE)
  expect_equal(log(boundaries$A), boundaries_log$A)
  expect_equal(log(boundaries$B), boundaries_log$B)
})

#' @title Draw Samples from a Normal Distribution
#' @description
#'
#' Draws exemplary samples with a certain effect size for the sequential one-oway ANOVA or the sequential t-test, see Steinhilber et al. (2023) <doi:10.31234/osf.io/m64ne>
#' @param k_groups number of groups (levels of factor_A)
#' @param f Cohen's f. The simulated effect size.
#' @param sd vector of standard deviations of the groups. Default value is 1 for each group.
#' @param sample_ratio vector of sample ratios between th groups. Default value is 1 for each group.
#' @param max_n sample size for the groups (total sample size = max_n*k_groups)
#'
#' @return returns a data.frame with the columns y (observations) and x (factor_A).
#'
#' @export
#'
#' @example inst/examples/draw_sample_normal.R



draw_sample_normal <- function(
  k_groups,
  f,
  max_n,
  sd = NULL,
  sample_ratio = NULL
) {

  if (is.null(sd)) {sd <- rep(1, k_groups)}
  if (is.null(sample_ratio)) {sample_ratio <- rep(1, k_groups)}

  if (!is.numeric(k_groups)) {stop("argument k_groups must be numeric.")}
  if (k_groups <= 1) {stop("argument k_groups must be larger than 1.")}

  if (!is.numeric(f)) {stop("argument f must be numeric.")}
  if (f < 0) {stop("argument f must be equal to or larger than 0.")}

  if (!is.numeric(max_n)) {stop("argument max_n must be numeric.")}
  if (max_n <= 0) {stop("argument max_n must be larger than 1.")}

  if (!is.numeric(sd)) {stop("argument sd must be numeric.")}
  if (length(sd) != k_groups) {stop("argument sd must contain as many elements as defined in k_groups argument.")}

  if (!is.numeric(sample_ratio)) {stop("argument sample_ratio must be numeric.")}
  if (length(sample_ratio) != k_groups) {stop("argument sample_ratio must contain as many elements as defined in k_groups argument.")}

  total_sample_size <- max_n*sample_ratio
  if (sum(total_sample_size) <= k_groups) {stop("total sample size must be greater than k_groups. Try to increase max_n argument.")}

  seq_step_increase <- sum(sample_ratio)
  seq_steps <- seq((seq_step_increase), sum(total_sample_size), seq_step_increase)
  # seq_steps <- seq_steps[2:length(seq_steps)]


  raw_means <- rnorm(k_groups)

  f_new <- f*sqrt(
    sum(sd^2*(total_sample_size - 1)) /
      (sum(total_sample_size) - k_groups)
  )

  pop_means = (raw_means - mean(raw_means)) / sd(raw_means) * sqrt(k_groups / (k_groups - 1)) * f_new
  position <- 1
  data <- matrix(double(sum(total_sample_size)*2), ncol = 2)

  for (steps in seq_steps) {

    y <- 1:k_groups %>%
      purrr::map(~rnorm(n = sample_ratio[.x], mean = pop_means[.x], sd = sd[.x])) %>%
      unlist()

    indice <- 1:100
    x <- 1:k_groups %>%
      purrr::map(~rep(indice[.x], sample_ratio[.x])) %>%
      unlist()

    data[position:steps, 1] <- y
    data[position:steps, 2] <- x
    position <- position + seq_step_increase
  }
  data.frame(y = data[, 1], x = as.factor(data[, 2]))
}

calc_seq_anova <- function(seq_anova_arguments) {
  non_centrality_parameter <-
    calc_non_centrality_parameter_anova(
      seq_anova_arguments
    )
  seq_anova_arguments@data <-
    calc_group_means(
      seq_anova_arguments
    )
  ss_effect <-
    calc_ss_effect(
      seq_anova_arguments
    )
  ss_residual <-
    calc_ss_residual(
      seq_anova_arguments
    )
  ss_total <-
    calc_ss_total(
      seq_anova_arguments
    )
  F_statistic <-
    calc_F_statistic_(
      seq_anova_arguments,
      ss_effect,
      ss_residual
    )
  likelihoods <-
    calc_likelihoods_anova(
      seq_anova_arguments,
      non_centrality_parameter,
      F_statistic
    )
  boundaries <-
    calc_boundaries(
      power = seq_anova_arguments@power,
      alpha = seq_anova_arguments@alpha,
      log = TRUE
    )
  decision <-
    get_seq_decision(
      likelihood_ratio = likelihoods$ratio_log,
      boundaries = boundaries
    )

  effect_sizes <- calc_effect_sizes(
    seq_anova_arguments,
    ss_effect,
    ss_total,
    F_statistic,
    decision)

  seq_anova_results <-
    build_seq_anova_results(
      seq_anova_arguments,
      likelihoods,
      boundaries,
      decision,
      non_centrality_parameter,
      ss_effect,
      ss_residual,
      ss_total,
      F_statistic,
      effect_sizes
    )

  seq_anova_results
}

setClassUnion("numericORnull", c("numeric","NULL"))


#' An S4 class to represent the results of a sequential t-test.
#'
#' @slot likelihood_ratio_log the logarithmic test statistic.
#' @slot decision the test decision: "accept H1", "accept H0",
#' or "continue sampling".
#' @slot A_boundary_log the lower logarithmic boundary of the test.
#' @slot B_boundary_log the upper logarithmic boundary of the test.
#' @slot d a number indicating the specified effect size (Cohen's d).
#' @slot mu a number indicating the true value of the mean
#' (or difference in means if you are performing a two sample test).
#' @slot alpha the type I error. A number between 0 and 1.
#' @slot power 1 - beta (beta is the type II error probability). A number
#' between 0 and 1.
#' @slot likelihood_ratio the likelihood ratio of the test without logarithm.
#' @slot likelihood_1 the likelihood of the alternative Hypothesis (H1).
#' @slot likelihood_0 the likelihood of the null Hypothesis (H0).
#' @slot likelihood_1_log the logarithmic likelihood of the alternative
#' Hypothesis (H1).
#' @slot likelihood_0_log the logarithmic likelihood of the null
#' Hypothesis (H0).
#' @slot non_centrality_parameter parameter to calculate the likelihoods
#' @slot t_value the t-value of the t-statistic.
#' @slot p_value the p-value of the t-test.
#' @slot df degrees of freedom.
#' @slot mean_estimate the estimated mean or difference in means depending on
#' whether it was a one-sample test or a two-sample test.
#' @slot alternative a character string specifying the alternative hypothesis:
#'       "two.sided" (default), "greater" or "less".
#' @slot one_sample "true" if it is a one-sample test, "false" if it is a
#' two-sample test.
#' @slot ttest_method a character string indicating what type of t-test was
#' performed.
#' @slot data_name a character string giving the name(s) of the data.
#' @slot verbose a logical value whether you want a verbose output or not.
#'
#' @keywords internal
setClass(
  "seq_ttest_results",
  slots = c(
    likelihood_ratio_log = "numeric",
    decision = "character",
    A_boundary_log = "numeric",
    B_boundary_log = "numeric",
    d = "numeric",
    mu = "numericORnull",
    alpha = "numeric",
    power = "numeric",
    likelihood_ratio = "numeric",
    likelihood_1 = "numeric",
    likelihood_0 = "numeric",
    likelihood_1_log = "numeric",
    likelihood_0_log = "numeric",
    non_centrality_parameter = "numeric",
    t_value = "numeric",
    p_value = "numeric",
    df = "numeric",
    mean_estimate = "list",
    alternative = "character",
    one_sample = "logical",
    ttest_method = "character",
    data_name = "character",
    verbose = "logical",
    plot = "ANY"
  )
)

#' Method to retrieve the contents of a slot of an object of the
#'  [`seq_ttest_results-class`] class.
#'
#' @param seq_ttest_results the corresponding class to this method.
#' @param x the seq_ttest_results object.
#' @param i indices indicating elements to extract.
#' @param j not used.
#' @param drop not used.
#'
#' @return Returns the contents of the specified slot. For more information,
#' see the documentation for the [`seq_ttest_results-class`] class.
#' @export
#'
#' @keywords internal
# #' @examples
setMethod(
  f = "[",
  signature = "seq_ttest_results",
  function(x, i, j, drop){ # must be this names!
    if (i == "likelihood_ratio") {return(x@likelihood_ratio)}
    if (i == "likelihood_ratio_log") {return(x@likelihood_ratio_log)}
    if (i == "decision") {return(x@decision)}
    if (i == "A_boundary_log") {return(x@A_boundary_log)}
    if (i == "B_boundary_log") {return(x@B_boundary_log)}
    if (i == "d") {return(x@d)}
    if (i == "mu") {return(x@mu)}
    if (i == "alpha") {return(x@alpha)}
    if (i == "power") {return(x@power)}
    if (i == "likelihood_1") {return(x@likelihood_1)}
    if (i == "likelihood_0") {return(x@likelihood_0)}
    if (i == "likelihood_1_log") {return(x@likelihood_1_log)}
    if (i == "likelihood_0_log") {return(x@likelihood_0_log)}
    if (i == "non_centrality_parameter") {return(x@non_centrality_parameter)}
    if (i == "t_value") {return(x@t_value)}
    if (i == "p_value") {return(x@p_value)}
    if (i == "df") {return(x@df)}
    if (i == "mean_estimate") {return(x@mean_estimate)}
    if (i == "alternative") {return(x@alternative)}
    if (i == "one_sample") {return(x@one_sample)}
    if (i == "ttest_method") {return(x@ttest_method)}
    if (i == "data_name") {return(x@data_name)}
    if (i == "verbose") {return(x@verbose)}
    stop(paste("Wrong slot name: '", i, "' is not a slot name of the class 'seq_ttest_results'"))
  }
)

digits <- 3

# t-Test -----------------------------------------------------------------------
setMethod(
  "show",
  signature = "seq_ttest_results",
  function(object){
    cat("\n*****  Sequential ", object@ttest_method, " *****\n", sep = "")
    cat("\nformula:", object@data_name)
    cat("\ntest statistic:\n")
    cat(" log-likelihood ratio",
        " = ",
        round(object@likelihood_ratio_log, digits = digits),
        ", decision = ",
        object@decision,
        sep = ""
        )

    cat("\nSPRT thresholds:\n")
    cat(" lower log(B) = ", round(object@B_boundary_log, digits = digits),
        ", ",
        "upper log(A) = ", round(object@A_boundary_log, digits = digits),
        sep = ""
        )

    if (object@verbose == TRUE) {
      cat("\nLog-Likelihood of the:")
      cat("\n alternative hypothesis =", round(object@likelihood_1_log, digits))
      cat("\n null hypothesis =", round(object@likelihood_0_log, digits))

      cat("\nalternative hypothesis: true",
          if (object@one_sample == TRUE) {
            if (object@alternative == "two.sided") {
              paste0("mean is not equal to ", object@mu, ".")
            } else{
              paste0("mean is ",object@alternative , " than ", object@mu, ".")
            }
          } else{
            if (object@alternative == "two.sided") {
              paste0("difference in means is not equal to ",object@mu, ".")
            } else{
              paste0("difference in means is ", object@alternative, " than ", object@mu, ".")
            }
          })

      cat("\nspecified effect size: Cohen's d =", object@d)
      cat("\ndegrees of freedom: df =", object@df)
      cat("\nsample estimates:\n")
      print(round(object@mean_estimate[[1]], 5))
      cat('*Note: to get access to the object of the results use the @ or [] instead of the $ operator.\n')
    }
})

# ANOVA ------------------------------------------------------------------------
setMethod(
  "show",
  signature = "seq_anova_results",
  function(object){
    # cat("\n*****  Sequential ", object@anova_method, " *****\n", sep = "")
    cat("\n*****  Sequential ANOVA *****\n", sep = "")
    cat("\nformula:", object@data_name)
    cat("\ntest statistic:\n")
    cat(" log-likelihood ratio",
        " = ",
        round(object@likelihood_ratio_log, digits = digits),
        ", decision = ",
        object@decision,
        sep = ""
    )

    cat("\nSPRT thresholds:\n")
    cat(" lower log(B) = ", round(object@B_boundary_log, digits = digits),
        ", ",
        "upper log(A) = ", round(object@A_boundary_log, digits = digits),
        sep = ""
    )

    if (object@verbose == TRUE) {
      cat("\nLog-Likelihood of the:")
      cat("\n alternative hypothesis =", round(object@likelihood_1_log, digits))
      cat("\n null hypothesis =", round(object@likelihood_0_log, digits))

      cat("\nalternative hypothesis: true difference in means is not equal to 0.")

      cat("\nspecified effect size: Cohen's f = ", object@f,
          "\nempirical Cohen's f = ", object@effect_sizes$cohens_f,
          ", 95% CI[", object@effect_sizes$ci_cohens_f_lower, ", ",
          object@effect_sizes$ci_cohens_f_upper, "]",
          "\nCohen's f adjusted = ", round(object@effect_sizes$cohens_f_adj, digits),
          sep = "")
      cat("\ndegrees of freedom: df1 = ", object@df_1,
          ", df2 = ", object@df_2,
          sep = "")
      cat("\nSS effect = ", object@ss_effect,
          ", SS residual = ", object@ss_residual,
          ", SS total = ", object@ss_total,
          sep = "")
      # cat("\nsample estimates:\n")
      # print(round(object@mean_estimate[[1]], 5))
      cat('\n*Note: to get access to the object of the results use the @ or [] instead of the $ operator.\n')
    }
  })


# object = seq_anova(y~x, f = 0.25, data = draw_sample_normal())
# object
# seq_anova(y~x, f = 0.25, data = draw_sample_normal(), verbose = FALSE)

build_seq_ttest_results <- function(
  seq_ttest_arguments,
  t_statistic,
  df,
  non_centrality_parameter,
  likelihoods,
  boundaries,
  decision,
  verbose
){
  # # no mean in y if one sample
  # if (length(t_statistic$estimate) > 1) {
  #   mean_y <- t_statistic$estimate[[2]]
  # } else{
  #   mean_y <- NULL
  # }

  seq_ttest_results <-
    new(
      "seq_ttest_results",
      likelihood_ratio = likelihoods["ratio"][[1]],
      likelihood_ratio_log = likelihoods["ratio_log"][[1]],
      decision = decision,
      A_boundary_log = boundaries["A"][[1]],
      B_boundary_log = boundaries["B"][[1]],
      d = seq_ttest_arguments["d"],
      mu = seq_ttest_arguments["mu"],
      alpha = seq_ttest_arguments["alpha"],
      power = seq_ttest_arguments["power"],
      likelihood_1 = likelihoods["likelihood_1"][[1]],
      likelihood_0 = likelihoods["likelihood_0"][[1]],
      likelihood_1_log = likelihoods["likelihood_1_log"][[1]],
      likelihood_0_log = likelihoods["likelihood_0_log"][[1]],
      non_centrality_parameter = non_centrality_parameter,
      t_value = as.vector(t_statistic["statistic"][[1]]),
      p_value = t_statistic["p.value"][[1]],
      df = df,
      mean_estimate = t_statistic["estimate"],
      alternative = seq_ttest_arguments["alternative"],
      one_sample = seq_ttest_arguments["one_sample"],
      ttest_method = t_statistic["method"][[1]],
      data_name = seq_ttest_arguments["data_name"],
      verbose = verbose
    )
  seq_ttest_results
}

#' Get path to cached simulation data directory
#' @keywords internal
get_sprtt_cache_dir <- function() {
  cache_dir <- rappdirs::user_data_dir("sprtt")
  if (!dir.exists(cache_dir)) {
    dir.create(cache_dir, recursive = TRUE)
  }
  cache_dir
}

#' Download simulation data for sample size planning
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Downloads pre-computed simulation results from GitHub releases.
#' Data is cached locally and only needs to be downloaded once.
#'
#' Data is hosted at:
#' \href{https://github.com/MeikeSteinhilber/sprtt_plan_sample_size}{MeikeSteinhilber/sprtt_plan_sample_size}
#'
#' @param force Logical. If TRUE, re-download even if data exists. Default FALSE.
#' @return Invisibly returns the path to the cached data file.
#' @export
#' @examples
#' \dontrun{
#' # Download data (only needed once)
#' download_sample_size_data()
#'
#' # Force re-download (e.g., after data update)
#' download_sample_size_data(force = TRUE)
#' }
#'
download_sample_size_data <- function(force = FALSE) {
  cache_dir <- get_sprtt_cache_dir()
  data_file <- file.path(cache_dir, "sprtt_external_data_plan_sample_size.rds")

  # Return early if already cached and no force re-download
  if (file.exists(data_file) && !force) {
    message("Simulation data already cached at: ", data_file)
    return(invisible(data_file))
  }

  # Ask for consent before downloading in interactive sessions
  auto_consent <- isTRUE(as.logical(Sys.getenv("SPRTT_CONSENT_DOWNLOAD", "false")))

  if (interactive() && !auto_consent) {
    answer <- utils::menu(
      c("Yes", "No"),
      title = "This will download ~150 MB of simulation data to your local cache. Proceed?"
    )
    if (answer != 1) {
      message("Download cancelled.")
      return(invisible(NULL))
    }
  }

  message("Downloading simulation data for sample size planning...")
  message("This is a one-time download (~150 MB).")
  tryCatch({
    piggyback::pb_download(
      file = "sprtt_external_data_plan_sample_size.rds",
      repo = "MeikeSteinhilber/sprtt_plan_sample_size",
      tag  = "latest",
      dest = cache_dir
    )
    message("\u2713 Download complete! Data cached at: ", cache_dir)
    invisible(data_file)
  }, error = function(e) {
    stop(
      "Failed to download simulation data.\n",
      "Please check:\n",
      "  1. Your internet connection\n",
      "  2. The repository 'MeikeSteinhilber/sprtt_plan_sample_size' is public\n",
      "  3. A release with 'sprtt_external_data_plan_sample_size.rds' exists\n",
      "Error message: ", e$message,
      call. = FALSE
    )
  })
}

#' Access sample size simulation data
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Loads pre-computed simulation results for SPRT sample size planning.
#' If not already cached locally, the data (~150 MB) will be downloaded automatically
#' from GitHub releases. Use this function to access the complete dataset for custom
#' analysis and visualization. See the **Data Structure** section below for details
#' on available columns.
#'
#' Data is hosted at:
#' \href{https://github.com/MeikeSteinhilber/sprtt_plan_sample_size}{MeikeSteinhilber/sprtt_plan_sample_size}
#'
#'
#' @return A named list with the following elements:
#' \itemize{
#'   \item \code{description}: Short description of the dataset
#'   \item \code{version}: GitHub release tag of the dataset (e.g., \code{"v0.1.0-data"})
#'   \item \code{created}: Date the dataset was created (as character string)
#'   \item \code{n_rep}: Number of simulation iterations per condition
#'   \item \code{data}: A data frame with simulation results (see **Data Structure**)
#' }
#' @export
#' @section Data Structure:
#' The \code{data} element contains simulation results with the following columns:
#'
#' **Simulation Metadata:**
#' \itemize{
#'   \item \code{batch}: Batch identifier for the simulation run
#'   \item \code{iteration}: Individual simulation iteration within a batch
#'   \item \code{source_file}: Path to the file containing simulation parameters or results
#' }
#'
#' **Input Parameters:**
#' \itemize{
#'   \item \code{f_simulated}: The true effect size used to generate the simulated data
#'   \item \code{f_expected}: The expected effect size specified for the SPRT
#'   \item \code{k_groups}: Number of groups in the design
#'   \item \code{alpha}: Significance level (Type I error rate)
#'   \item \code{power}: Desired statistical power (1 - Type II error rate)
#'   \item \code{distribution}: Data distribution used for simulation
#'   \item \code{sd}: Standard deviation(s) used in data generation in each group
#'   \item \code{sample_ratio}: Ratio of sample sizes between groups (e.g., 1:1, 2:1)
#'   \item \code{n_raw_data}: Total number of raw observations generated in each group
#'   \item \code{fix_n}: Fixed sample size
#' }
#'
#' **Individual Test Results:**
#' \itemize{
#'   \item \code{n}: Actual sample size at which the SPRT terminated
#'   \item \code{decision}: Test decision
#'   \item \code{decision_error}: Whether the decision was erroneous (Type I or Type II error)
#'   \item \code{log_lr}: Log-likelihood ratio at termination
#'   \item \code{f}: Calculated effect size from the data
#'   \item \code{f_adj}: Adjusted effect size
#'   \item \code{f_statistic}: F-statistic from ANOVA test
#' }
#'
#' **Summary Statistics (Aggregated across iterations):**
#' \itemize{
#'   \item \code{decision_error_rate}: Proportion of incorrect decisions
#'   \item \code{mean_n}: Mean sample size across all iterations
#'   \item \code{sd_error_n}: Standard error of the mean sample size (sd(n)/sqrt(n))
#'   \item \code{median_n}: Median sample size (50th percentile)
#'   \item \code{min_n}, \code{max_n}: Minimum and maximum sample sizes observed
#'   \item \code{q25_n}, \code{q50_n}, \code{q75_n}, \code{q90_n}, \code{q95_n}: Sample size quantiles
#'   \item \code{decision_rate_25}, \code{decision_rate_50}, \code{decision_rate_75},
#'         \code{decision_rate_90}, \code{decision_rate_95}, \code{decision_rate_100}:
#'         Cumulative decision rates at various percentages of maximum sample size
#' }
#' @examples
#' \dontrun{
#' # Load data (downloads automatically if needed)
#' loaded <- load_sample_size_data()
#'
#' # Access the simulation data frame
#' head(loaded$data)
#'
#' # Check dataset version
#' loaded$version  # e.g. "v0.1.0-data"
#' loaded$created
#' }

load_sample_size_data <- function() {
  cache_dir <- get_sprtt_cache_dir()
  data_file <- file.path(cache_dir, "sprtt_external_data_plan_sample_size.rds")

  # Download if not cached
  if (!file.exists(data_file)) {
    result <- download_sample_size_data()
    # Return NULL if download was cancelled
    if (is.null(result)) return(invisible(NULL))
  }

  # Load and return
  readRDS(data_file)
}
#' Clear cached simulation data
#'
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Removes locally cached simulation data (~150 MB) used by [`plan_sample_size()`].
#' Data will be automatically re-downloaded on next use of sample size planning functions.
#'
#' This function is useful when:
#' * You want to free up disk space
#' * The cached data may be outdated and you want to force a fresh download
#' * Troubleshooting cache-related issues
#'
#' @return Invisibly returns `TRUE` if cache was cleared, `FALSE` if no cache existed.
#' @export
#' @examples
#' \dontrun{
#' # Clear cache
#' cache_clear()
#' }
cache_clear <- function() {
  cache_dir <- get_sprtt_cache_dir()
  data_file <- file.path(cache_dir, "sprtt_external_data_plan_sample_size.rds")

  if (file.exists(data_file)) {
    unlink(data_file)
    message("Cached simulation data cleared.")
    message("Data will be re-downloaded on next use.")
    return(invisible(TRUE))
  } else {
    message("No cached data found.")
    return(invisible(FALSE))
  }
}

#' Cache information
#' @description
#' `r lifecycle::badge("experimental")`
#'
#' Displays information about cached simulation data (~150 MB) used by [`plan_sample_size()`].
#' Shows the cache directory location, whether data is cached, file size, and dataset
#' version metadata.
#'
#' The simulation data is automatically downloaded on first use of sample size planning
#' functions and stored locally for faster subsequent access.
#'
#' @return Invisibly returns a list with:
#'   * `cache_dir`: Character string with the cache directory path
#'   * `data_cached`: Logical indicating if simulation data is cached
#'   * `file_size_mb`: Numeric file size in MB (or `NA` if not cached)
#'   * `data_version`: GitHub release tag of the cached dataset (or `NA` if not cached)
#'   * `data_created`: Date the dataset was created (or `NA` if not cached)
#'
#' @seealso
#' * [`cache_clear()`] to remove cached data
#' * [`download_sample_size_data()`] to manually download simulation data
#' * [`plan_sample_size()`] which uses the cached data
#'
#' @export
cache_info <- function() {
  cache_dir <- get_sprtt_cache_dir()
  data_file <- file.path(cache_dir, "sprtt_external_data_plan_sample_size.rds")

  data_cached <- file.exists(data_file)

  # Extract version metadata if data is cached
  if (data_cached) {
    loaded       <- readRDS(data_file)
    data_version <- loaded$version
    data_created <- loaded$created
    file_size_mb <- round(file.size(data_file) / 1e6, 2)
  } else {
    data_version <- NA
    data_created <- NA
    file_size_mb <- NA
  }

  info <- list(
    cache_dir    = cache_dir,
    data_cached  = data_cached,
    file_size_mb = file_size_mb,
    data_version = data_version,
    data_created = data_created
  )

  cat("SPRTT Simulation Data Cache\n")
  cat("---------------------------\n")
  cat("Cache directory:", info$cache_dir, "\n")
  cat("Data cached:", info$data_cached, "\n")
  if (info$data_cached) {
    cat("File size:", info$file_size_mb, "MB\n")
    cat("Dataset version:", info$data_version, "\n")
    cat("Dataset created:", info$data_created, "\n")
  }

  invisible(info)
}

test_that("data download works", {
  skip_if_offline()
  testthat::skip_on_cran()

  # Clear cache to test fresh download
  cache_clear()

  # Should download successfully
  expect_message(download_sample_size_data(), "Downloading")

  # Should be cached now
  expect_message(download_sample_size_data(), "already cached")

  # Test that file actually exists
  cache_dir <- get_sprtt_cache_dir()
  data_file <- file.path(cache_dir, "sprtt_external_data_plan_sample_size.rds")
  expect_true(file.exists(data_file))
})

test_that("force download works", {
  skip_if_offline()
  testthat::skip_on_cran()

  # Clear cache
  cache_clear()

  # Force re-download should show downloading message
  expect_message(download_sample_size_data(force = TRUE), "Downloading")

  # Regular call should show cached message
  expect_message(download_sample_size_data(force = FALSE), "already cached")
})

test_that("data loads correctly", {
  skip_if_offline()
  testthat::skip_on_cran()

  loaded <- load_sample_size_data()

  # Check returned object is a list with version metadata
  expect_type(loaded, "list")
  expect_true("data" %in% names(loaded))
  expect_true("version" %in% names(loaded))
  expect_true("created" %in% names(loaded))
  expect_true("n_rep" %in% names(loaded))
  expect_true("description" %in% names(loaded))

  # Check version metadata is non-empty
  expect_type(loaded$version, "character")
  expect_false(is.na(loaded$version))
  expect_type(loaded$created, "character")
  expect_false(is.na(loaded$created))
  expect_type(loaded$n_rep, "double")
  expect_false(is.na(loaded$n_rep))

  # Check data element is a data frame
  df_all <- loaded$data
  expect_s3_class(df_all, "data.frame")
  expect_true(nrow(df_all) > 0)

  # Check for key input parameter columns
  expect_true("f_expected" %in% names(df_all))
  expect_true("f_simulated" %in% names(df_all))
  expect_true("power" %in% names(df_all))
  expect_true("alpha" %in% names(df_all))
  expect_true("k_groups" %in% names(df_all))

  # Check for result columns
  expect_true("n" %in% names(df_all))
  expect_true("decision" %in% names(df_all))
  expect_true("mean_n" %in% names(df_all))

  # Check for summary statistics
  expect_true("median_n" %in% names(df_all))
  expect_true("q25_n" %in% names(df_all))
  expect_true("q75_n" %in% names(df_all))
})

test_that("load triggers download if not cached", {
  skip_if_offline()
  testthat::skip_on_cran()

  # Clear cache
  cache_clear()

  # load_sample_size_data should trigger download
  expect_message(load_sample_size_data(), "Downloading")

  # Second load should be silent (no download message)
  loaded <- suppressMessages(load_sample_size_data())
  expect_type(loaded, "list")
  expect_s3_class(loaded$data, "data.frame")
})

test_that("cache_clear works correctly", {
  skip_if_offline()
  testthat::skip_on_cran()

  # Ensure data exists
  suppressMessages(download_sample_size_data())

  # Clear should return TRUE and show message
  expect_message(result <- cache_clear(), "Cached simulation data cleared")
  expect_true(result)

  # File should no longer exist
  cache_dir <- get_sprtt_cache_dir()
  data_file <- file.path(cache_dir, "sprtt_external_data_plan_sample_size.rds")
  expect_false(file.exists(data_file))

  # Clearing again should return FALSE
  expect_message(result <- cache_clear(), "No cached data found")
  expect_false(result)
})

test_that("cache directory is created if missing", {
  testthat::skip_on_cran()

  cache_dir <- get_sprtt_cache_dir()

  # Remove the cache directory completely
  if (dir.exists(cache_dir)) {
    unlink(cache_dir, recursive = TRUE)
  }

  # Verify it's gone
  expect_false(dir.exists(cache_dir))

  # Call function that should create it
  new_cache_dir <- get_sprtt_cache_dir()

  # Now it should exist
  expect_true(dir.exists(new_cache_dir))
  expect_equal(new_cache_dir, cache_dir)
})

test_that("cache_info returns correct information", {
  testthat::skip_on_cran()

  # Test with cached data
  suppressMessages(download_sample_size_data())
  info <- suppressMessages(cache_info())

  expect_type(info, "list")
  expect_true("cache_dir" %in% names(info))
  expect_true("data_cached" %in% names(info))
  expect_true("file_size_mb" %in% names(info))
  expect_true("data_version" %in% names(info))
  expect_true("data_created" %in% names(info))

  expect_true(info$data_cached)
  expect_true(info$file_size_mb > 0)
  expect_true(dir.exists(info$cache_dir))
  expect_false(is.na(info$data_version))
  expect_false(is.na(info$data_created))

  # Test without cached data
  cache_clear()
  info_empty <- suppressMessages(cache_info())

  expect_false(info_empty$data_cached)
  expect_true(is.na(info_empty$file_size_mb))
  expect_true(is.na(info_empty$data_version))
  expect_true(is.na(info_empty$data_created))
})

test_that("cache directory is created if missing", {
  cache_dir <- get_sprtt_cache_dir()
  expect_true(dir.exists(cache_dir))
})

test_that("download returns file path invisibly", {
  testthat::skip_on_cran()
  cache_clear()

  result <- suppressMessages(download_sample_size_data())
  expect_type(result, "character")
  expect_true(file.exists(result))
  expect_match(result, "sprtt_external_data_plan_sample_size\\.rds$")
})

test_that("cache_info produces expected output", {
  skip_if_offline()
  testthat::skip_on_cran()

  suppressMessages(download_sample_size_data())

  expect_output(cache_info(), "SPRTT Simulation Data Cache")
  expect_output(cache_info(), "Cache directory:")
  expect_output(cache_info(), "Data cached:")
  expect_output(cache_info(), "File size:")
  expect_output(cache_info(), "Dataset version:")
  expect_output(cache_info(), "Dataset created:")
})

test_that("download handles missing internet gracefully", {
  skip_if_offline()
  testthat::skip_on_cran()

  # This test is more conceptual - actual implementation would require mocking
  # Just verify the error message structure exists
  cache_clear()

  # Normal download should work
  expect_message(download_sample_size_data(), "Downloading")
})

test_that("data structure contains all documented columns", {
  skip_if_offline()
  testthat::skip_on_cran()

  loaded <- suppressMessages(load_sample_size_data())
  df <- loaded$data

  # Simulation metadata
  metadata_cols <- c("batch", "iteration", "source_file")
  expect_true(all(metadata_cols %in% names(df)))

  # Input parameters
  input_cols <- c("f_simulated", "f_expected", "k_groups", "alpha", "power",
                  "distribution", "sd", "sample_ratio", "n_raw_data", "fix_n")
  expect_true(all(input_cols %in% names(df)))

  # Individual test results
  result_cols <- c("n", "decision", "decision_error", "log_lr", "f",
                   "f_adj", "f_statistic")
  expect_true(all(result_cols %in% names(df)))

  # Summary statistics
  summary_cols <- c("decision_error_rate", "mean_n", "sd_error_n", "median_n",
                    "min_n", "max_n", "q25_n", "q50_n", "q75_n", "q90_n", "q95_n")
  expect_true(all(summary_cols %in% names(df)))
})

calc_ss_residual <- function(seq_anova_arguments) {
  sum((seq_anova_arguments@data$y - seq_anova_arguments@data$group_mean_A)^2)
}

calc_ss_effect <- function(seq_anova_arguments) {
  # n_group <- dplyr::count(seq_anova_arguments@data, .data$factor_A)$n[1]
  n_group <- table(seq_anova_arguments@data$factor_A)

  grand_mean_factor_A <- mean(seq_anova_arguments@data$y)
  group_mean_A <- unique(seq_anova_arguments@data$group_mean_A)

  sum(n_group * (group_mean_A - grand_mean_factor_A)^2)
}

calc_ss_total <- function(seq_anova_arguments) {
  grand_mean_factor_A <- mean(seq_anova_arguments@data$y)
  sum((seq_anova_arguments@data$y - grand_mean_factor_A)^2)
}

setClassUnion("numericORnull", c("numeric","NULL"))


#' An S4 class to represent the results of a sequential anova.
#'
#' @slot likelihood_ratio_log the logarithmic test statistic.
#' @slot decision the test decision: "accept H1", "accept H0",
#' or "continue sampling".
#' @slot A_boundary_log the lower logarithmic boundary of the test.
#' @slot B_boundary_log the upper logarithmic boundary of the test.
#' @slot f a number indicating the specified effect size (Cohen's f).
#' @slot effect_sizes a list with effect sizes (Cohen's f, eta squared, ...).
#' @slot alpha the type I error. A number between 0 and 1.
#' @slot power 1 - beta (beta is the type II error probability). A number
#' between 0 and 1.
#' @slot likelihood_ratio the likelihood ratio of the test without logarithm.
#' @slot likelihood_1 the likelihood of the alternative Hypothesis (H1).
#' @slot likelihood_0 the likelihood of the null Hypothesis (H0).
#' @slot likelihood_1_log the logarithmic likelihood of the alternative
#' Hypothesis (H1).
#' @slot likelihood_0_log the logarithmic likelihood of the null
#' Hypothesis (H0).
#' @slot non_centrality_parameter parameter to calculate the likelihoods
#' @slot F_value the F-value of the F-statistic.
#' @slot df_1 degrees of freedom.
#' @slot df_2 degrees of freedom.
#' @slot ss_effect ss_effect.
#' @slot ss_residual ss_residual.
#' @slot ss_total ss_total.
#' @slot total_sample_size total sample size.
#' @slot data_name a character string giving the name(s) of the data.
#' @slot verbose a logical value whether you want a verbose output or not.
#' @param plot list with all arguments for the plot_anova() function
#'
#' @keywords internal
setClass(
  "seq_anova_results",
  slots = c(
    likelihood_ratio_log = "numeric",
    decision = "character",
    A_boundary_log = "numeric",
    B_boundary_log = "numeric",
    f = "numeric",
    effect_sizes = "ANY",
    alpha = "numeric",
    power = "numeric",
    likelihood_ratio = "numeric",
    likelihood_1 = "numeric",
    likelihood_0 = "numeric",
    likelihood_1_log = "numeric",
    likelihood_0_log = "numeric",
    non_centrality_parameter = "numeric",
    F_value = "numeric",
    df_1 = "numeric",
    df_2 = "numeric",
    ss_effect = "numeric",
    ss_residual = "numeric",
    ss_total = "numeric",
    total_sample_size = "numeric",
    data_name = "character",
    verbose = "logical",
    plot = "ANY"
  )
)

setMethod(
  f = "initialize",
  signature = "seq_anova_results",
  function(.Object, ...) { # '.Object' is necessary and can not replaced by 'x'
    .Object <- callNextMethod() # necessary line
    # validObject(.Object)
    .Object
  }
)

#' Method to retrieve the contents of a slot of an object of the
#'  [`seq_anova_results-class`] class.
#'
#' @param seq_anova_results the corresponding class to this method.
#' @param x the seq_ttest_results object.
#' @param i indices indicating elements to extract.
#' @param j not used.
#' @param drop not used.
#'
#' @return Returns the contents of the specified slot. For more information,
#' see the documentation for the [`seq_anova_results-class`] class.
#' @export
#'
#' @keywords internal
# #' @examples
setMethod(
  f = "[",
  signature = "seq_anova_results",
  function(x, i, j, drop){ # must be this names!
    if (i == "likelihood_ratio") {return(x@likelihood_ratio)}
    if (i == "likelihood_ratio_log") {return(x@likelihood_ratio_log)}
    if (i == "decision") {return(x@decision)}
    if (i == "A_boundary_log") {return(x@A_boundary_log)}
    if (i == "B_boundary_log") {return(x@B_boundary_log)}
    if (i == "f") {return(x@f)}
    if (i == "effect_sizes") {return(x@effect_sizes)}
    if (i == "alpha") {return(x@alpha)}
    if (i == "power") {return(x@power)}
    if (i == "likelihood_1") {return(x@likelihood_1)}
    if (i == "likelihood_0") {return(x@likelihood_0)}
    if (i == "likelihood_1_log") {return(x@likelihood_1_log)}
    if (i == "likelihood_0_log") {return(x@likelihood_0_log)}
    if (i == "non_centrality_parameter") {return(x@non_centrality_parameter)}
    if (i == "F_value") {return(x@F_value)}
    if (i == "df_1") {return(x@df_1)}
    if (i == "df_2") {return(x@df_2)}
    if (i == "ss_effect") {return(x@ss_effect)}
    if (i == "ss_residual") {return(x@ss_residual)}
    if (i == "ss_total") {return(x@ss_total)}
    if (i == "total_sample_size") {return(x@total_sample_size)}
    if (i == "data_name") {return(x@data_name)}
    if (i == "verbose") {return(x@verbose)}
    stop(paste("Wrong slot name: '", i, "' is not a slot name of the class 'seq_anova_results'"))
  }
)


## @slot f_empiric a number indicating the calculated empirical effect size (Cohen's f).
## @slot eta_squared a number indicating the calculated empirical effect size (eta^2).
# if (i == "f_empiric") {return(x@f_empiric)}
# if (i == "eta_squared") {return(x@eta_squared)}

setClassUnion("numericORnull", c("numeric","NULL"))


setClass(
  Class = "seq_ttest_arguments",
  slots = c(
    x = "numeric",
    y = "numericORnull",
    mu = "numeric",
    d = "numeric",
    alpha = "numeric",
    power = "numeric",
    alternative = "character",
    paired = "logical",
    one_sample = "logical",
    data_name = "character",
    na.rm = "logical"
  )
)
build_prototype_seq_ttest_arguments <- function() {
  new(
    Class = "seq_ttest_arguments",
    x = c(-0.8576829, -1.2119236, -1.2883021,  0.2532647, -1.6749356,
          1.2018224,  0.4220293, -0.3896717,  1.5371334,  0.6254288
          ),
    y = c(1.4016121,  1.2807517,  1.8763097,  0.5280401,  1.7045422,
          1.2462588, -0.2278561, -0.1054774,  0.9764811,  1.8253478
          ),
    mu = 0,
    d = 0.8,
    alpha = .05,
    power = .80,
    alternative = "two.sided",
    paired = FALSE,
    one_sample = FALSE,
    data_name = "x and y",
    na.rm = FALSE
  )
}

setValidity(
  Class = "seq_ttest_arguments",
  function(object) {
  # correct input arguments
  if (object@alternative != "two.sided" &&
     object@alternative != "greater" &&
     object@alternative != "less"
     )
    stop("Invalid argument <alternative>: Must be either 'two.sided', 'greater' or 'less'.")
  if (!(object@alpha > 0 &&
       object@alpha < 1
       ))
    stop("Invalid argument <alpha>: Probabilities must be in ]0;1[.")
  if (!(object@power > 0 &&
       object@power < 1
       ))
    stop("Invalid argument <power>: Probabilities must be in ]0;1[.")
  if (object@d <= 0)
    stop("Invalid argument <d>: Must be greater than 0.")
  if (length(object@one_sample) == 0)
    stop("Invalid argument <one_sample>: Error in class input_arguments.")

  # missing data in x or y
  if (length(object@x) < 2)
    stop("Length of x is less than 2. Length of x must be greater than 2. ")
  if (!is.null(object@y)) {
    if (object@one_sample == FALSE &&
        length(object@y) < 2
        )
      stop("Length of y is less than 2. Length of y must be greater than 2. ")
  }
  TRUE
})
setMethod(
  f = "initialize",
  signature = "seq_ttest_arguments",
  function(.Object, ...) { # '.Object' is necessary and can not replaced by 'x'
    .Object <- callNextMethod() # necessary line
    validObject(.Object)
    .Object
  }
)

#' Method to retrieve the contents of a slot of an object of the
#'  seq_ttest_arguments class.
#'
#' @description This method is only used internally to process
#' the input arguments of the [`seq_ttest`] function. As a normal user,
#' you can ignore this specific documentation.
#'
#' @param seq_ttest_arguments the corresponding class to this method.
#' @param x the seq_ttest_arguments object.
#' @param i indices indicating elements to extract.
#' @param j not used.
#' @param drop not used.
#'
#' @keywords internal
#' @return Returns the contents of the specified slot of an
#' seq_ttest_arguments object. For more information, see the arguments of the
#' [`seq_ttest`] function.
setMethod(
  f = "[",
  signature = "seq_ttest_arguments",
  function(x, i, j, drop){ # must be this names!
    if (i == "x") {return(x@x)}
    if (i == "y") {return(x@y)}
    if (i == "mu") {return(x@mu)}
    if (i == "d") {return(x@d)}
    if (i == "alpha") {return(x@alpha)}
    if (i == "power") {return(x@power)}
    if (i == "alternative") {return(x@alternative)}
    if (i == "paired") {return(x@paired)}
    if (i == "one_sample") {return(x@one_sample)}
    if (i == "data_name") {return(x@data_name)}
    if (i == "na.rm") {return(x@na.rm)}
    stop(paste("Wrong slot name: '", i, "' is not a slot name of the class 'seq_ttest_arguments'"))
  }
)

test_that("CLASS_seq_ttest_arguments: Check getters", {
  arguments <- build_prototype_seq_ttest_arguments()
  expect_equal(arguments["x"], arguments@x)
  expect_equal(arguments["y"], arguments@y)
  expect_equal(arguments["mu"], arguments@mu)
  expect_equal(arguments["d"], arguments@d)
  expect_equal(arguments["alpha"], arguments@alpha)
  expect_equal(arguments["power"], arguments@power)
  expect_equal(arguments["alternative"], arguments@alternative)
  expect_equal(arguments["paired"], arguments@paired)
  expect_equal(arguments["one_sample"], arguments@paired)
  expect_equal(arguments["data_name"], arguments@data_name)
  expect_equal(arguments["na.rm"], arguments@na.rm)

  expect_error(arguments["jewhf"], "Wrong slot name:")
})

calc_t_statistic <- function(seq_ttest_arguments) {

  if (seq_ttest_arguments["one_sample"] == TRUE) {
    t_statistic <- t.test(
      x = seq_ttest_arguments["x"],
      mu = seq_ttest_arguments["mu"]
    )

  } else if (
    seq_ttest_arguments["one_sample"] == FALSE &&
    seq_ttest_arguments@paired == FALSE
  ) {
    t_statistic <- t.test(
      x = seq_ttest_arguments["x"],
      y = seq_ttest_arguments["y"],
      mu = seq_ttest_arguments["mu"],
      var.equal = TRUE
    )

  } else if (
    seq_ttest_arguments["one_sample"] == FALSE &&
    seq_ttest_arguments@paired == TRUE
  ) {
    t_statistic <- t.test(
      x = seq_ttest_arguments["x"],
      y = seq_ttest_arguments["y"],
      mu = seq_ttest_arguments["mu"],
      paired = TRUE
    )
  }

  if (is.na(t_statistic$statistic)) {
    stop("It was not possible to calculate the t-value.")
  }

  if (seq_ttest_arguments["alternative"] == "less") {
    t_statistic$statistic <- t_statistic$statistic * -1
  }

  t_statistic
}

#* @testing
# set.seed(26753)

test_that("calc_t_statistic", {
  arguments <- new("seq_ttest_arguments",
                   x = rnorm(15),
                   y = NULL,
                   mu = 0,
                   d = 0.8,
                   alpha = 0.05,
                   power = 0.95,
                   alternative = "two.sided",
                   paired = FALSE,
                   one_sample = TRUE,
                   data_name = "data name",
                   na.rm = FALSE
                   )
  results_package <- calc_t_statistic(arguments)
  results_ttest <- t.test(x = arguments@x)
  expect_equal(results_package$statistic[[1]] , results_ttest$statistic[[1]])

  arguments@y <- arguments@x # error
  arguments@one_sample <- FALSE
  arguments@paired <- TRUE
  expect_error(calc_t_statistic(arguments),
               "not possible to calculate the t-value")

  arguments@y <- rnorm(15)
  arguments@one_sample <- FALSE
  arguments@paired <- TRUE
  results_package <- calc_t_statistic(arguments)
  results_ttest <- t.test(x = arguments@x, y = arguments@y, paired = TRUE)
  expect_equal(results_package$statistic[[1]] , results_ttest$statistic[[1]])

  arguments@alternative <- "greater"
  results_package <- calc_t_statistic(arguments)
  results_ttest <- t.test(x = arguments@x,
                          y = arguments@y,
                          alternative = arguments@alternative,
                          paired = TRUE)
  expect_equal(results_package$statistic[[1]] , results_ttest$statistic[[1]])

  arguments@alternative <- "less"
  results_package <- calc_t_statistic(arguments)
  results_ttest <- t.test(x = arguments@x,
                          y = arguments@y,
                          alternative = arguments@alternative,
                          paired = TRUE)
  expect_equal(results_package$statistic[[1]] , -1*results_ttest$statistic[[1]])
})

# DOCUMENTATION ----------------------------------------------------------------


# SIMULATION -------------------------------------------------------------------
# seed
set.seed(333)

# number of repititions per parameter combination
S <- 1000				# choose 80 for quick results

##---- PACKAGES AND HELP FUNCTIONS ---------------------------------------------

library(dplyr)
library(foreach)
library(doParallel)
library(sprtt)
library(PropCIs)

# Clopper-Pearson exact CI
conf.int <- Vectorize(function(x, n, a){
  ci <- exactci(x * n, n, conf.level = 1 - a)
  return(as.numeric(ci$conf.int))
}, "x")

# draws the full sample
draw_sample <- function(n, d) {
  matrix(rnorm(n * 2, mean = c(d, 0)),
         nrow = n, byrow = TRUE)
}


##---- SETTINGS OF THE SIMULATION ----------------------------------------------


# settings for parallel processing
cl <- makeCluster(8)
registerDoParallel(cl)

minN <- 2				# sample size per group at start
maxN <- 50000			# maximum n (increase if reached before decision)
ns <- c(minN:9999, seq(10000, maxN, by = 5))
# ns <- minN:maxN

max_s <- S/getDoParWorkers() # number of repetitions per batch

d_true <- c(rep(0, 3), 0.8, 0.5, 0.2) # true effect sizes
d_hyp <- rep(c(.8, .5, .2), 2) # expected effect sizes
d_mat <- data.frame(d_true, d_hyp) # Cohen's d matrix

alpha <- .05
beta <- .05


##---- THE SIMULATION ----------------------------------------------------------

start <- Sys.time()

sim <- foreach(batch=1:getDoParWorkers(), .combine = rbind) %dopar% {

  normal_final <- matrix(
    NA,
    nrow = nrow(d_mat) * length(alpha) * length(beta) * max_s, ncol = 11,
    dimnames = list(
      NULL,
      c("d", "d_hyp", "alpha", "beta",
        "n", "tValue", "A", "B",
        "LR", "Decision", "error")
      )
  )
  normal_final.counter <- 1

  for(d_i in 1:nrow(d_mat)){ # d_i denotes the index of d_mat

    d0 <- d_mat[d_i, "d_true"]
    d1 <- d_mat[d_i, "d_hyp"]

    for (a in alpha) {

      for (b in beta){

        normal_final0 <- matrix(NA, nrow = max_s, ncol = 11,
                                dimnames = dimnames(normal_final))

        A <- log((1 - b) / a)
        B <- log(b / (1 - a))

        for (i in 1:max_s) {

          maxsamp <- draw_sample(maxN, d0)
          normal_finalhit <- FALSE
          error <- as.numeric(NA)
          normal_finaldec <- -1

          for (n in ns) {

            samp <- maxsamp[1:n,]

            test <- sprtt::seq_ttest(samp[,1],
                                     samp[,2],
                                     d = d1,
                                     alpha = a,
                                     power = 1 - b,
                                     alternative = "two.sided"
                    )
            tVal <- test@t_value
            LR <- test@likelihood_ratio_log

            if(test@decision == "accept H0"){
              normal_finaldec <- 0
              normal_finalhit <- TRUE

              if(d0!=0)
                error <- 1
              else
                error <- 0

              break
            } else if(test@decision == "accept H1"){
              normal_finaldec <- 1
              normal_finalhit <- TRUE

              if(d0!=0)
                error <- 0
              else
                error <- 1

              break
            }

          } # end of n-iteration (single sample)

          if (!normal_finalhit) {
            # nmax was reached before the test reached a decision boundary

            sink(paste0(path, "warnings.txt"), append = T)
            print(paste0("Truncated before decision: Repetition = ", i,
                         "; d = ", d0,
                         "; d_hyp = ", d1,
                         "; Alpha = ", a,
                         "; Beta = ", b,
                         "; batch = ", batch))
            sink()

          }

          normal_final0[i, ] <- c(
            d = d0,
            d_hyp = d1,
            alpha = a,
            beta = b,
            n = n,
            tValue = tVal,
            A = A,
            B = B,
            LR = LR,
            Decision = normal_finaldec,
            error = error)

        }  # end of max_s repetitions (samples per parameter combination)

        normal_final[normal_final.counter:(normal_final.counter + max_s - 1), ] <- normal_final0
        normal_final.counter <- normal_final.counter + max_s

      } # end of beta iteration
    } #end of alpha iteration

  } # end of d_i iteration
  return(normal_final)
}

end <- Sys.time()
cat("\nresults_2sample_2sided: ")
print(end - start)


# THE RESULTS ------------------------------------------------------------------

results_2sample_2sided <- sim %>%
  as_tibble() %>%
  group_by(d, d_hyp, alpha, beta) %>%
  dplyr::summarise(error = mean(error),
                   .groups = "drop") %>%
  dplyr::mutate(ci_l = t(conf.int(error, S, .1))[,1],
                alert = case_when(d == 0 & ci_l > alpha ~ 1,
                                  d > 0 & ci_l > beta ~ 1,
                                  TRUE ~ 0)
                )

# SAVE THE RESULTS -------------------------------------------------------------

testthis::use_testdata(results_2sample_2sided, overwrite = TRUE)

##---- UNDO SET.SEED -----------------------------------------------------------

set.seed(NULL)

check_formula_ttest <- function(formula, data, paired) {
  # check structure
  if ((length(formula) != 3L) || (length(formula[[3]]) != 1L))
    stop(
      "'formula' is incorrect. Please specify as 'x~y'.
      If your variables are in a data frame, please use the 'data' argument."
    )

  # quick extraction of the formula for testing y
  data_matrix_formula <- model.frame(formula, data)

  # get the y value
  if (length(data_matrix_formula) == 2) {
    y <- data_matrix_formula[,2]
  } else{
    # correct because model.frame allows only 1 as input
    y <- 1
  }

  # check y
  if (!is.factor(y) && (length(y) != 1)) {
    stop(paste(
      formula[[3]],
      "is not a factor. Are you sure, that you want to use the fomula input?"))
  }
  if (length(unique(y)) != 2 && (length(y) != 1)) {
    stop(paste("Grouping factor must contain exactly two levels."))
  }
  if (paired == TRUE) {
    if (length(y) == 1 & y[1] == 1) {
      stop("Paired test: The second group is missing.")
    }
    if (!(table(y)[[1]] == table(y)[[2]])) {
      stop("Unequal number of observations per group. Independent samples?")
    }
  }
}


check_formula_anova <- function(formula, data) {

  if (!inherits(formula,"formula"))
    stop("The formula argument must be a formula (class(formula) == 'formula'")
  # check structure
  # one-way ANOVA check: formula[[3]] can have only one element
  # if (length(formula[[1]]) != "~")
  #   stop("'formula' is incorrect. Please specify as 'y~factor_A'.")
  if (length(formula[[2]]) != 1L)
    stop("'formula' is incorrect. Please specify as 'y~factor_A'.")
  if (length(formula[[3]]) != 1L)
    stop(
      "one-way ANOVA: 'formula' is incorrect. Please specify as 'y~factor_A'."
    )

  # quick extraction of the formula for testing y
  data_matrix_formula <- model.frame(formula, data)
  factor_A <- data_matrix_formula[, 2]

  # check y
  if (!is.factor(factor_A)) {
    stop(paste(
      formula[[3]],
      "must be a factor."))
  }

  if (length(levels(factor_A)) < 2) {
    stop(paste("Grouping factor must contain at least two levels."))
  }

  # if (paired == TRUE) {
  #   if (length(y) == 1 & y[1] == 1) {
  #     stop("Paired test: The second group is missing.")
  #   }
  #   if (!(table(y)[[1]] == table(y)[[2]])) {
  #     stop("Unequal number of observations per group. Independent samples?")
  #   }
  # }
}

#* @testing check_formula

# t-test -----------------------------------------------------------------------

test_that("check_formula_ttest: Check structure: Correct error messages", {
  x <- 1:10
  y <- as.factor(c(rep(1,5), rep(2,5)))
  data <- data.frame(x, z = x, y)
  formula <- x~y
  paired <- FALSE

  expect_error(
    check_formula_ttest(formula = NULL, data, paired),
    "'formula' is incorrect."
  )
  expect_error(
    check_formula_ttest(formula = x~y+z, data, paired),
    "'formula' is incorrect."
  )
  expect_error(
    check_formula_ttest(formula = data$x ~ data$y, data = NULL, paired = paired),
    "please use the 'data' argument"
  )
  expect_error(
    check_formula_ttest(formula = x ~ 1, data, paired = TRUE),
    "Paired test: The second group is missing."
  )

})

test_that("check_formula_ttest: Check y: Correct messages", {
  x <- rnorm(20, mean = 0, sd = 1)
  y <- c(rep(1, 10), rep(2, 10))
  expect_error(
    seq_ttest(x ~ y, d = 0.8),
    "is not a factor"
  )


  x <- 1:10
  y_2 <- as.factor(c(rep(1,5), rep(2,5)))
  y_3 <- as.factor(c(rep(1,4), rep(2,5), 3))
  formula <- x~y
  paired <- FALSE

  data <- data.frame(x, y = rep(c(2,1), 5))
  expect_error(
    check_formula_ttest(formula, data, paired),
    "is not a factor"
  )

  data <- data.frame(x, y = y_3)
  expect_error(
    check_formula_ttest(formula, data, paired),
    "contain exactly two levels"
  )

  data <- data.frame(x, y = y_2)
  formula <- x ~ 5
  expect_error(
    check_formula_ttest(formula, data, paired)
  )

  x <- 1:10
  y <- as.factor(c(rep(1,3), rep(2,7)))
  expect_error(
    check_formula_ttest(x ~ y, data = data.frame(x,y), paired = TRUE),
    "Unequal number of observations per group"
  )
})

test_that("check_formula_ttest: silent behaviour: no errors occur", {

  x <- 1:10
  y <- as.factor(c(rep(1,5), rep(2,5)))
  z <- 1:10
  data <- data.frame(x, z, y)
  formula <- x~y
  paired <- FALSE

  expect_silent(
    check_formula_ttest(formula = x~y, data, paired)
  )
  expect_silent(
    check_formula_ttest(formula = x ~ 1, data, paired)
  )

  data <- data.frame(a = x, c = z, b = y, paired)
  expect_silent(
    check_formula_ttest(formula = a ~ b, data, paired)
  )

})

# ANOVA ------------------------------------------------------------------------

test_that("check_formula_anova: Check structure: Correct error messages", {
  data <- draw_sample_normal(k_groups = 4, f = 0, max_n = 30)
  formula <- y~x

  expect_error(
    check_formula_anova(formula = "y~x", data),
    "The formula argument must be a formula"
  )
  expect_error(
    check_formula_anova(formula = y+z~x, data),
    "'formula' is incorrect."
  )
  expect_error(
    check_formula_anova(formula = x~y+z, data),
    "one-way ANOVA: 'formula' is incorrect."
  )
  expect_error(
    check_formula_anova(formula = data$x ~ data$y, data = NULL),
    "'formula' is incorrect."
  )
  expect_error(
    check_formula_anova(formula = formula),
    'argument "data" is missing, with no default'
  )
  expect_error(
    check_formula_anova(data = data),
    'argument "formula" is missing, with no default'
  )
  data$x <- as.double(data$x)
  expect_error(
    check_formula_anova(formula, data),
    'x must be a factor'
  )
  data$x <- as.factor(double(length(data$x)))
  expect_error(
    check_formula_anova(formula, data),
    'Grouping factor must contain at least two levels.'
  )
})

test_that("check_formula_anova: silent behaviour: no errors occur", {
  data <- draw_sample_normal(k_groups = 4, f = 0, max_n = 30)
  colnames(data) <- c("observation", "factor")
  formula <- observation ~ factor

  expect_silent(
    check_formula_anova(formula, data)
  )
})

#* @testing get_one_sample

test_that("get_one_sample: Check correct output.", {
  y <- NULL
  expect_equal(
    get_one_sample(y),
    TRUE
  )
  y <- 1
  expect_equal(
    get_one_sample(y),
    TRUE
  )
  y <- 2
  expect_equal(
    get_one_sample(y),
    TRUE
  )
  y <- c(1,2,3)
  expect_equal(
    get_one_sample(y),
    FALSE
  )
  y <- list(1,2,3)
  expect_equal(
    get_one_sample(y),
    FALSE
  )
})

# #* @testing get_one_sided
# test_that("get_one_sided: Check correct output.", {
#
#     expect_equal(
#     get_one_sided(alternative = "two.sided"),
#     FALSE
#   )
#   expect_equal(
#     get_one_sided(alternative = "less"),
#     TRUE
#   )
#   expect_equal(
#     get_one_sided(alternative = "greater"),
#     TRUE
#   )
#   expect_error(
#     get_one_sided(alternative = "test")
#   )
#
# })



# test_that("", {
#
#
# })

#* @testing seq_ttest

test_that("Check error rates", {
  # load packages
  testthat::skip_if_not_installed("testthis")
  library(testthis)
  library(dplyr)

  # skip the tests in special situations
  skip_on_cran() # skip on CRAN checks: necessary!
  skip_on_ci() # skip on continuous platforms like GitHub Actions
  skip("Check error rates. (Takes about 10 minutes)") # skip test

  # test setup: simulate the data
  source_dir(
    "testdata-raw/",
    pattern = "sided\\.[rR]$",
    env = test_env(),
    chdir = TRUE,
    wrap = TRUE
  )

  results_2sample_1sided <- testthis::read_testdata("results_2sample_1sided")
  results_2sample_2sided <- testthis::read_testdata("results_2sample_2sided")
  results_1sample_2sided <- testthis::read_testdata("results_1sample_2sided")

  expect_true(all(results_2sample_1sided$alert == 0))
  expect_true(all(results_2sample_2sided$alert == 0))
  expect_true(all(results_1sample_2sided$alert == 0))

})

---
title: "Developer Guide: sprtt Package"
author: "Meike Snijder-Steinhilber"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Developer Guide: sprtt Package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---
```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

## Overview

This vignette provides technical documentation for developers who want to maintain, extend, or contribute to the `sprtt` package.
It explains the internal architecture, code organization, computational methods, and development workflows.

**Target audience**: R package developers, future maintainers, and contributors

**Prerequisite knowledge**: 

- R package development basics
- S4 object-oriented programming
- Statistical knowledge of Sequential Probability Ratio Tests, see `vignette("sprt")`

## Package Architecture

### Overview

The `sprtt` package follows a modular architecture with clear separation of concerns:

1. **Main User-facing functions** (`seq_ttest()`, `seq_anova()`) handle input validation and provide clean interfaces

5. **Utility User-facing functions** provide supporting functionality (plotting, data generation, caching)

2. **Builder functions** (`build_seq_*_arguments()`) Transform and validate user inputs into a structured S4 argument object. Builder functions act as a processing pipeline between the user-facing interface and the core calculation functions. They handle the complex task of parsing different input formats, validating all parameters, and packaging everything into a type-safe container.

3. **Calculation functions** (`calc_seq_*()`) perform the core computations

4. **Internal utility functions** (e.g., `delete_na()`, `extract_formula()`, `get_seq_decision()`, ...) are small, focused helpers that handle specific tasks across the package. They follow the Single Responsibility Principle - each function does one thing well. This design reduces code duplication, improves testability, and makes the codebase easier to maintain.

5. **Result Classes** - S4 classes that store results in type-safe, structured containers. These classes provide controlled access to result components through standardized methods (`@`, `[]`, and `show()`), ensuring consistency and extensibility.

This separation allows for:

- Easy testing of individual components
- Clear error handling at appropriate levels
- Extensibility for future sequential tests
- Consistent user experience across functions

**Function Naming Convention:**

- `seq_*()`: User-facing main functions
- `build_*()`: Argument preparation functions  
- `calc_*()`: Computational core functions
- `*_class.R`: S4 class definitions

### Function Structure

**Table: Structure of the main `seq_ttest()` function (Level 1)**

| Level 2                          | Level 3                                      | Level 4               |
|:---------------------------------|:---------------------------------------------|:----------------------|
| `build_seq_ttest_arguments()` <br> *Class: seq_ttest_arguments*    | `check_formula()`                            |                       |
|      | `extract_formula()`                          |                       |
|                                  | `get_one_sample()`                           |                       |
|                                  | `delete_na()`                                |                       |
|                                  | `check_data()`                               | `check_constant_data()` |
| `calc_seq_ttest()` <br> *Class: seq_ttest_results*                | `calc_seq_ttest_t_statistic()`               |                       |
|       | `calc_seq_ttest_non_centrality_parameter()`  |                       |
|                                  | `calc_seq_ttest_likelihoods()`               |                       |
|                                  | `calc_seq_ttest_boundaries()`                |                       |
|                                  | `get_seq_ttest_decision()`                   |                       |
|                                  | `build_seq_ttest_results()`                  |                       |

**Table: Structure of the main `seq_anova()` function (Level 1)**

| Level 2                          | Level 3                                      |
|:---------------------------------|:---------------------------------------------|
| `build_seq_anova_arguments()` <br> *Class: seq_anova_arguments*  | `check_formula_anova()`|
|                                  | `extract_formula_anova()`                    |
|                                  | `check_data_anova()`                         |
| `calc_seq_anova()` <br> *Class: seq_anova_results* | `calc_non_centrality_parameter_anova()`|
|                                  | `calc_group_means()`                         |
|                                  | `calc_ss_effect()`                           |
|                                  | `calc_ss_residual()`                         |
|                                  | `calc_ss_total()`                            |
|                                  | `calc_F_statistic_()`                        |
|                                  | `calc_likelihoods_anova()`                   |
|                                  | `calc_boundaries()`                          |
|                                  | `get_seq_decision()`                         |
|                                  | `calc_effect_sizes()`                        |
|                                  | `build_seq_anova_results()`                  |
| `calc_plot_anova()` <br> *Class: seq_anova_results*     |                            |
## Software Testing

### Continuous Integration

## GitHub Structure

## Release Checklist

### Pre-release

- [ ] Update version number in DESCRIPTION
- [ ] Update NEWS.md with all changes
- [ ] Run `devtools::check()` - must pass cleanly
- [ ] Update documentation: `devtools::document()`
- [ ] Rebuild vignettes: `devtools::build_vignettes()`
- [ ] Test on multiple platforms (rhub, win-builder)
- [ ] Update README if needed
- [ ] Check all examples run correctly
- [ ] Verify all URLs in documentation are valid
- [ ] Run code coverage: `covr::package_coverage()`
- [ ] Update copyright year if needed

### CRAN Submission

- [ ] Prepare cran-comments.md
- [ ] Submit to CRAN: `devtools::release()`
- [ ] Monitor email for CRAN response
- [ ] Check CRAN package page after acceptance

### Post-release

- [ ] Create GitHub release with tag
- [ ] Tweet/announce new version
- [ ] Update pkgdown site: `pkgdown::build_site()`
- [ ] Increment development version number
- [ ] Update NEWS.md with "Development version" section


## Contributing

**How to contribute:**

1. Fork the repository
2. Create a feature branch
3. Make changes with tests
4. Submit pull request
5. Respond to review comments

**Contribution guidelines:**

- Follow existing code style
- Add tests for new features
- Update documentation
- Keep commits focused and atomic
- Write clear commit messages

---
title: "Sample Size Planning for Sequential ANOVAs"
author: "Meike Snijder-Steinhilber"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 4
    css: customize.css
description: >
  This vignette describes how to plan sample sizes for sequential ANOVAs using simulation-based approaches.
vignette: >
  %\VignetteIndexEntry{Sample Size Planning for Sequential ANOVAs}
  %\VignetteEncoding{UTF-8}{inputenc}
  %\VignetteEngine{knitr::rmarkdown}
bibliography: references.bib
csl: "apa.csl"
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  echo = TRUE,
  eval = FALSE,
  message = TRUE,
  warning = FALSE,
  collapse = TRUE,
  comment = "#>"
)
```



## Why Sample Size Planning Matters

Sample size planning for sequential tests differs fundamentally from fixed-design studies.
In sequential ANOVA, the final sample size is determined by the evidence in the data itself and consequently remains unknown beforehand -- data collection continues until either the upper or lower decision boundary is reached.

**The challenge:** While this data-driven stopping rule is very efficient, it creates practical difficulties.
Resource planning requires knowing whether you might need 100 observations or 1,000.
Budget constraints, time limitations, and logistical considerations all demand some advance estimate of required resources.

**The solution:** Although the exact final sample size cannot be known in advance, simulation-based planning bridges the gap between statistical theory and practical constraints.
The `sprtt` package provides the `plan_sample_size()` function, which generates HTML reports summarizing simulation results for sequential ANOVAs.
Researchers can obtain guidance on:

- The typical amount of data required to reach a decision ($N_{\text{median}}$)
- The upper limit of resources needed to achieve a specified decision rate ($N_{\text{max}}$)

### Resource Constraints and Decision Rates
The decision boundaries of the sequential ANOVA control Type I ($\alpha$) and Type II ($\beta$) errors in the long run.
However, introducing a maximum sample size $N_{\text{max}}$ for practical resource planning creates an important complication: it reduces the achievable power below the nominal $1-\beta$.

When $N_{\text{max}}$ is reached before a decision boundary is crossed, this results in a **non-decision**.
The non-decision rate depends directly on the chosen maximum sample size.
This introduces a new metric: the **decision rate** (the probability of reaching a decision) given resource limitations.

While non-decisions are undesirable, they represent a crucial conceptual distinction from accepting the null hypothesis.
SPRTs like the sequential ANOVA differentiate between stopping data collection to accept the null hypothesis and stopping because more evidence would be required to make a decision but resources are exhausted.
Importantly, as long as no decision has been reached, data collection can continue if additional resources become available.

## The `plan_sample_size()` Function

The `plan_sample_size()` function generates interactive HTML reports for sample size planning based on a large simulation database.
Reports include recommended maximum sample sizes, expected sample sizes, power curves, and comparisons to traditional ANOVA designs.

### Pre-computed Simulation Database

To make sample size planning fast and accessible, `sprtt` includes access to extensive simulation results.
These simulations were conducted by:

1. Generating 10,000 of datasets for each combination of parameters
2. Running sequential ANOVAs on each dataset
3. Recording when each test stopped
4. Aggregating these results to get key summary statistics that guide the sample size planning

This simulation database is stored externally to keep the package installation size small.
The data are downloaded automatically on first use of `plan_sample_size()` and cached locally for future sessions.

```{r, echo=FALSE, results='asis'}
cat(sprintf(
  'Source code of the simulation database: <a href="https://github.com/MeikeSteinhilber/sprtt_plan_sample_size" target="_blank" rel="noopener"><img src="https://img.shields.io/badge/GitHub-MeikeSteinhilber/sprtt__plan__sample__size-blue?logo=github" alt="GitHub" style="vertical-align:middle;"></a></span></p>'
))
```


## Getting Started


### Your First Sample Size Report
Let's walk through a practical example.
Imagine you're planning a study to compare three groups.
You want to detect medium-sized effects (Cohen's $f = 0.25$) or larger with specific error control.

You set $\alpha = 0.05$ to control Type I errors at the standard 5% level, ensuring that rejections of the null hypothesis are trustworthy in the long run.
To minimize Type II errors, you set $\beta = 0.05$, limiting false acceptances of $H_0$ also to 5%.
However, given limited resources, you're willing to accept a 15% non-decision rate, meaning you'll reach a decision 85% of the time.

Critically, this setup reflects a deliberate trade-off: by keeping both error rates as low as 5%, you accept that a decision will not always be reached — but when it is, it can be trusted.
Non-decisions, by contrast, indicate that the available evidence was insufficient given your error constraints, and more data are required.

Now let's see how to generate a sample size planning report for this scenario:

```{r first-report}
plan_sample_size(
  f_expected = 0.25,   # Expected effect size
  k_groups = 3,        # Number of groups
  beta = 0.05,         # beta error rate
  decision_rate = 0.85 # desired percentage of decisions 
)
```

When you run this code for the first time, several things happen:

1. **Data download:** The simulation database (~70 MB) is downloaded from GitHub and saved to your local cache directory. This is a one-time operation.
2. **Report generation:** An HTML report is created in your temporary directory.
3. **Browser launch:** The report automatically opens in your default web browser (if running interactively).

The entire process typically takes a couple of seconds for the initial download, then just a few seconds more for generating the subsequent report.



### Function Parameters

| Parameter | Type | Default | Description |
|-----------|------|---------|-------------|
| `f_expected` | numeric | *required* | Expected standardized effect size (Cohen's *f*). Must be one of: 0.10, 0.15, 0.20, 0.25, 0.30, 0.35, or 0.40. |
| `k_groups` | integer | *required* | Number of groups to compare. Must be 2, 3, or 4. |
| `beta` | numeric | 0.05 | Beta error rate. Must be 0.20, 0.10, or 0.05. |
| `output_dir` | character | `tempdir()` | Directory where the HTML report will be saved. |
| `output_file` | character | `"sprtt-report-sample-size-planning.html"` | Filename for the generated report. |
| `open` | logical | `interactive()` | Whether to open the report in your browser after generation. Set to `FALSE` for batch processing. |
| `overwrite` | logical | FALSE | Whether to overwrite an existing file with the same name without prompting. |



### Input Validation

The function validates all inputs before generating the report.
If you specify a parameter value that doesn't exist in the simulation database, you'll receive an informative error message listing the available options. For example:

```{r error-example, eval=FALSE}
# This will produce an error:
plan_sample_size(f_expected = 0.22, k_groups = 3)
#> Error: `f_expected` = 0.22 is not available. 
#> Please choose one of 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, or 0.4
```



## Practical Use Cases

### Case 1: Comparing Different Effect Sizes

The expected effect size has a large impact on required sample size.
Here's how to generate reports for different scenarios:

```{r example-effect-sizes}
# report 1
plan_sample_size(f_expected = 0.15, k_groups = 3, beta = 0.05)

# report 2
plan_sample_size(f_expected = 0.35, k_groups = 3, beta = 0.05)
```



### Case 2: Saving Reports to a Specific Location

By default, reports are saved to a temporary directory.
For reports you want to keep, specify a custom location:

```{r example-custom-location}
plan_sample_size(
  f_expected = 0.25,
  k_groups = 4,
  output_dir = "~/Documents/research/sample_size_planning",
  output_file = "study1_anova.html",
  open = TRUE
)
```

This is particularly useful when preparing documentation for grant applications, pre-registrations, or manuscript supplementary materials.

### Case 3: Preparing Multiple Scenarios

When preparing grant applications or pre-registrations, you might want to explore multiple scenarios (e.g., different effect size assumptions):

```{r workflow-scenarios}
# Define scenarios to compare
scenarios <- data.frame(
  effect = c(0.15, 0.20, 0.25),
  label = c("conservative", "expected", "optimistic")
)

# Generate reports for each scenario
for (i in 1:(nrow(scenarios))) {
  plan_sample_size(
    f_expected = scenarios$effect[i],
    k_groups = 3,
    beta = 0.10,
    output_dir = "sample_size_reports",
    output_file = sprintf("plan_sample_size_%s.html", scenarios$label[i]),
    open = FALSE,  # Don't open each one
    overwrite = TRUE
  )
}

message("Generated ", nrow(scenarios), " sample size reports")
```

This approach creates a set of reports that document your planning across different assumptions.


## Managing the Simulation Data

**Downloading Data Explicitly**

While `plan_sample_size()` downloads data automatically when needed, you can also download it explicitly:

```{r download-explicit}
# Download simulation data manually
download_sample_size_data()
```

This is useful if you want to:

- Pre-download data on a fast internet connection before traveling
- Verify the download completed successfully
- Troubleshoot download issues

To force a re-download (for example, after a package update with new simulation data):

```{r download-force}
download_sample_size_data(force = TRUE)
```

**Checking Cache Status**

To see whether data are cached and how much disk space they occupy:

```{r cache-status}
cache_info()
```

This displays:

- The cache directory location on your system
- Whether simulation data are currently cached
- The file size

**Clearing the Cache**

If you need to free up disk space or suspect corrupted data, you can clear the cache:

```{r cache-clear}
cache_clear()
```

The data will be re-downloaded automatically the next time you run `plan_sample_size()`.

**Working with Simulation Data Directly**

Advanced users may want to access the raw simulation data for custom analyses or visualizations.
You can load the data directly into your R session:

```{r load-data-advanced}
# Load the complete simulation dataset (downloads automatically if not yet cached)
loaded <- load_sample_size_data()

# Access the simulation data frame
df_all <- loaded$data

# Check which dataset version this report is based on
loaded$description  # short description
loaded$version      # e.g. "v0.1.0-data"
loaded$created      # date the dataset was created
loaded$n_rep        # number of simulation iterations per condition
```

The data frame `df_all` contains all simulation results and can be filtered, summarized, or visualized using standard R tools. See `?load_sample_size_data` for a full description of all available columns.

---
title: "How to use the sprtt package?"
author: "Meike Snijder-Steinhilber"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 4
description: >
  This vignette describes the sequential t-test and the usage of the
  `sprtt` package.
vignette: >
  %\VignetteIndexEntry{How to use the sprtt package}
  %\VignetteEncoding{UTF-8}{inputenc}
  %\VignetteEngine{knitr::rmarkdown}
bibliography: references.bib
csl: "apa.csl"
---



<style>
/* Use custom class names to avoid Bootstrap conflicts */
.custom-info-box {
  background-color: white !important;
  border: 2px solid #2c3e50 !important;
  border-left: 5px solid #2c3e50 !important;
  color: #2c3e50 !important;
  margin-top: 1.5em !important;
  margin-bottom: 1.5em !important;
  border-radius: 4px !important;
  padding: 15px !important;
  padding-left: 50px !important;
  position: relative !important;
}
.custom-info-box::before {
  content: "🛈";
  position: absolute;
  left: 0;
  top: 0;
  bottom: 0;
  width: 40px;
  background-color: #2c3e50;
  color: white;
  display: flex;
  align-items: center;
  justify-content: center;
  font-size: 1.3em;
  border-radius: 2px 0 0 2px;
}

.custom-warning-box {
  background-color: white !important;
  border: 2px solid #18bc9c !important;
  border-left: 5px solid #18bc9c !important;
  color: #2c3e50 !important;
  margin-top: 1.5em !important;
  margin-bottom: 1.5em !important;
  border-radius: 4px !important;
  padding: 15px !important;
  padding-left: 50px !important;
  position: relative !important;
}
.custom-warning-box::before {
  content: "⚠";
  position: absolute;
  left: 0;
  top: 0;
  bottom: 0;
  width: 40px;
  background-color: #18bc9c;
  color: white;
  display: flex;
  align-items: center;
  justify-content: center;
  font-size: 1.3em;
  border-radius: 2px 0 0 2px;
}

</style>


```{r, echo = FALSE, message = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

The `sprtt` package is a toolbox for **s**equential **p**robability **r**atio **t**ests (SPRTs).
This vignette introduces the core functions and demonstrates a typical analysis workflow.
For more comprehensive guides on specific topics, see the other package vignettes.
If you are unfamiliar with SPRTs, please read first the vignette `vignette("sprt")`.


## Workflow

### 1. Understand the theoretical background of SPRTs

The foundational literature [@wald1945; @wald1947] established the mathematical framework for SPRTs.
While these original papers provide theoretical depth, they require strong mathematical statistics background and focus primarily on abstract theory rather than practical application.

For practical implementation, we strongly recommend starting with the following simulation studies, which demonstrate robustness to assumption violations, explain common pitfalls, and provide actionable guidance:

**Essential reading:**

- **Sequential *t*-tests**: @schnuerch2020 — Demonstrates performance under various assumption violations and provides practical guidance
- **Sequential one-way ANOVA**: @steinhilber2024 — Extends SPRTs to the one-way ANOVA design with comprehensive simulation evidence
- **Understanding questionable research practices in sequential designs**: @steinhilber2025 — Critical for avoiding misuse and ensuring appropriate application


For comparisons of different sequential designs (including SPRTs), see:

- @schnuerch2020 -- Group sequential designs vs Sequential Bayes Factor Test vs SPRT
- @stefan2022 -- Sequential Bayes Factor Test vs SPRT

### 2. When to use SPRTs

Whether a statistical tool is appropriate depends strongly on the research context and intended use.
SPRTs are recommended when:

- Resources are limited and efficiency is a priority
- The research question translates to a clear hypothesis test: Concluding in favor of either $H_0$ (no effect) or $H_1$ (effect) provides useful information for your research question.
- Data can be collected sequentially
- Data can be analyzed sequentially (ideally after each new observation)
- Long-term error rates need to be controlled (both $\alpha$ and $\beta$ error rates)


SPRTs are **not** recommended when:

- Data are collected all at once or only in very large batches, see here fixed designs or Group Sequential Designs
- Multiple hypotheses must be tested on the same dataset (theoretically possible but not yet implemented)
- Groups cannot be sampled somewhat equally over time (e.g., collecting all participants from one group before starting the other)
- Sufficient literature already establishes the existence of the effect and the main goal is precise effect size estimation (see @steinhilber2024)


<!-- ### 3. Plan your resources -->

<!-- The `plan_sample_size()` function helps establish realistic expectations for data requirements and resource planning. -->

<!-- Unlike traditional designs, SPRTs do not require classical a priori power analysis because power ($1-\beta$) is controlled through the stopping boundaries. -->
<!-- This allows starting data collection immediately and stopping once the test reaches a decision. -->
<!-- However, planning necessary resources remains important for practical concerns as it will not always be possible to collect data until a decision is reached. -->
<!-- While the upper and lower boundaries of a SPRT control for the $\alpha$ and $\beta$ error rate in the long run, they do not guarantee that you will have enough resources to collect enough data to reach a decision in the first place. -->

<!-- Thus, two sample sizes are relevant for planning: -->

<!-- - **Expected sample size** ($N_{\text{median}}$): The median sample size required to reach a decision. Most studies will finish near this value. -->
<!-- - **Maximum sample size** ($N_{\text{max}}$): The maximum affordable sample size. Choose settings where $N_{\text{max}}$ yields a decision rate of at least 80% (i.e., the test reaches a decision in at least 80% of cases where the maximum sample is reached). -->

<!-- The `plan_sample_size()`function provides tables and plots to guide. -->

<!-- For more details see the vignette *XXX*. -->

### 3. Plan your resources

The `plan_sample_size()` function helps establish realistic expectations for data requirements and resource planning.

Unlike traditional designs, SPRTs do not require classical a priori power analysis.
Power is controlled through the stopping boundaries, allowing you to start data collection immediately and stop once the test reaches a decision.

However, resource planning remains essential.
While the boundaries control $\alpha$ and $\beta$ error rates in the long run, they cannot guarantee you will collect enough data to reach a decision within your available resources.

Two sample sizes are relevant for planning:

- **Expected sample size** ($N_{\text{median}}$): The median sample size required to reach a decision. Most studies will finish near this value.
- **Maximum sample size** ($N_{\text{max}}$): The maximum affordable sample size. Choose settings where $N_{\text{max}}$ yields a decision rate of at least 80% (i.e., the test reaches a decision in at least 80% of cases where the maximum sample is reached).

The `plan_sample_size()` function provides tables and plots to help you find appropriate design parameters that balance efficiency with feasibility.

For detailed examples and guidance, see the vignette `vignette("plan_sample_size")`.


### 4. Plan the data collection and register your test specifications

A thoughtful data collection plan is highly recommended to keep groups somewhat balanced and to track and balance (or randomize) potential confounders.

::: {.custom-info-box}
**Outliers**

The question of how one should deal with outliers in sequential testing is still an ongoing research topic. Note, that implementing a naive sequential outlier analysis can lead to a inflation of the $\alpha$ error rate, see @steinhilber2025.
:::

As SPRTs are best suited for confirmatory research, preregistering the data collection plan, hypothesis, and test specifications (e.g., effect size of interest, $\alpha$ and $\beta$ levels) is strongly recommended.

Preparing an analysis script in advance enables a smooth process for continuously analyzing incoming data.
This ensures data collection stops immediately once the stopping criterion is reached, avoiding unnecessary additional data collection.

To test the analysis pipeline, use either:

- The example datasets included in the `sprtt` package
- The data generating functions, which allow specifying the true effect size for simulation testing

### 5. Collect the data and apply the SPRTs

The following test are currently implemented:

- sequential *t*-tests, see `vignette("t_test")`
- sequential one-way ANOVA with independent groups, see `vignette("one_way_anova")`


### 5. Reporting of the results

Guidelines for the reporting of SPRTs can be found in the paper of @schubert2025 that also explicitly covers sequential testing and specifically SPRTs.

A complete SPRT report should include: the specific variant of the SPRT used (e.g., sequential *t*-test), the $\alpha$ and $\beta$ levels, the effect size or other parameters specifying the alternative hypothesis, the starting point of the SPRT (the sample size at the first look), the final sample size when data collection was stopped, the final likelihood ratio, a plot showing the full likelihood progression across all looks, and an effect size estimate with confidence interval [@schubert2025].
Note that effect size estimates in sequential designs are often biased and should be interpreted with caution.

**Example report --- decision reached ($H_1$ accepted):**

> We preregistered a sequential *t*-test [@wald1945; @schnuerch2020] with error probabilities $\alpha = .05$ and $\beta = .05$, and a minimum effect size of interest of $d = 0.5$.
The first look took place at $n = 5$ per group.
Data collection stopped at $N = 48$ (24 per group) when the likelihood ratio crossed the upper decision boundary ($LR_{48} = 21.3 > B = 19$), providing sufficient evidence to accept $H_1$.
The estimated effect size was $d = 0.61$ (95% CI [0.21, 1.00]).
A plot of the full likelihood progression is provided in Figure X.
All materials and the preregistration are available at [OSF link].

**Example report --- $N_{max}$ reached (non-decision):**

> We preregistered a sequential *t*-test [@wald1945; @schnuerch2020] with error probabilities $\alpha = .05$ and $\beta = .05$, a minimum effect size of interest of $d = 0.5$, and a maximum sample size of $N_{max} = 200$ (100 per group).
The first look took place at $n = 4$ per group.
Data collection stopped upon reaching $N_{max} = 200$ without the likelihood ratio crossing either decision boundary ($LR_{200} = 3.1$; lower boundary $A = 1/19$, upper boundary $B = 19$).
This constitutes a non-decision: the accumulated evidence was insufficient to accept either $H_0$ or $H_1$ with the prespecified error control.
The results are therefore inconclusive, though the final likelihood ratio 
indicates that the data are 3.1 times more likely under $H_1$ 
than under $H_0$.
Note that a non-decision due to resource depletion does not constitute evidence for $H_0$.
The estimated effect size was $d = 0.21$ (95% CI $[-0.07, 0.49]$).
A plot of the full likelihood progression is provided in Figure X.

## References

---
title: "Simple t-Test Use Case"
author: "Meike Snijder-Steinhilber"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 4
    css: "customize.css"
vignette: >
  %\VignetteIndexEntry{Use Case}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: references.bib
csl: "apa.csl"
---

<style>
/* Use custom class names to avoid Bootstrap conflicts */
.custom-info-box {
  background-color: white !important;
  border: 2px solid #2c3e50 !important;
  border-left: 5px solid #2c3e50 !important;
  color: #2c3e50 !important;
  margin-top: 1.5em !important;
  margin-bottom: 1.5em !important;
  border-radius: 4px !important;
  padding: 15px !important;
  padding-left: 50px !important;
  position: relative !important;
}
.custom-info-box::before {
  content: "🛈";
  position: absolute;
  left: 0;
  top: 0;
  bottom: 0;
  width: 40px;
  background-color: #2c3e50;
  color: white;
  display: flex;
  align-items: center;
  justify-content: center;
  font-size: 1.3em;
  border-radius: 2px 0 0 2px;
}

.custom-warning-box {
  background-color: white !important;
  border: 2px solid #18bc9c !important;
  border-left: 5px solid #18bc9c !important;
  color: #2c3e50 !important;
  margin-top: 1.5em !important;
  margin-bottom: 1.5em !important;
  border-radius: 4px !important;
  padding: 15px !important;
  padding-left: 50px !important;
  position: relative !important;
}
.custom-warning-box::before {
  content: "⚠";
  position: absolute;
  left: 0;
  top: 0;
  bottom: 0;
  width: 40px;
  background-color: #18bc9c;
  color: white;
  display: flex;
  align-items: center;
  justify-content: center;
  font-size: 1.3em;
  border-radius: 2px 0 0 2px;
}

</style>

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(sprtt)
```

::: {.custom-info-box}

-   The data set `df_stress` is included in the `sprtt` package. Thus, the data set is available when the package is loaded.

-   In the R code sections:

    -   `# comment`: is a comment
    -   `function()`: is R code
    -   `#> results of function()`: is console output

:::


This vignette describes an exemplary use case to improve the understanding of the package and the sequential t-test.

## Use Case

A team of researchers wants to know if the stress of the employees increased in the company itFlow AG over the last year. 
The Human Resources department has gathered data from a new self-care tool they implemented for the employees. It is difficult to predict how many employees will participate at the second measurement point (one year after the baseline). 
The team suggests using a sequential *t*-test instead of a Students *t*-test, because the sequential procedure can reduce the required sample size and is stopped right after the decision for one of the hypotheses is made.
Furthermore, the sequential *t*-test is able to gather evidence for both the alternative and the null hypothesis.

### Hypothesis

The researchers know that the company has received more orders than in the year before.
Thus, they expect an increase in self-reported stress.

### Data Analysis

The parameters of the sequential *t*-test are specified as follows:

-   The researchers expect an increase in stress but they have no prior knowledge about the expected effect size.
Thus, they define the lower limit of a substantial effect of interest.

    ```{r, echo=TRUE}
    d <- 0.2
    ```

-   They choose the common $\alpha$ level of .05 and the same level for the $\beta$ error probability, which leads to a power of .95 ($1 - \beta$).

    ```{r, echo=TRUE}
    alpha <- 0.05
    power <- 0.95
    ```

-   The data are repeated measures, therefore the test is a paired sequential *t*-test.

    ```{r, echo=TRUE}
    paired <- TRUE
    ```

-   The researchers expect that the stress will have increased after one year. Thus, they specify that the true difference in means is greater than 0.\
    `(mean(one_year_stress) - mean(baseline_stress)) > 0`

    ```{r, echo=TRUE}
    alternative <- "greater"
    ```

The HR department receives the new data piece by piece and passes it directly on to the researchers.
The test is performed for the first time and starts with the first two data points.

```{r, echo=TRUE}
# first data from the Human Resources department ---
# current sample size
n_person <- 2

# get data
df <- df_stress[1:n_person,]

# print data
df

# sequential t-test
results <- seq_ttest(df$one_year_stress,
                     df$baseline_stress,
                     alpha = alpha,
                     power = power,
                     d = d,
                     paired = paired,
                     alternative = alternative,
                     verbose = FALSE)

# print results: console output
results
```

The decision from the first test is:

```{r, echo=TRUE}
results@decision
```

As a result, the researchers take one more data point and run the test again.

```{r, echo=TRUE}
# new data from the Human Resources department ---
# get one more person
n_person <- n_person + 1
df <- df_stress[1:n_person,]

# print new data
df

# sequential t-test
results <- seq_ttest(df$one_year_stress,
                     df$baseline_stress,
                     alpha = alpha,
                     power = power,
                     d = d,
                     paired = paired,
                     alternative = alternative,
                     verbose = FALSE)

# print results
results@decision
```

This process is repeated until the decision is made for one of the two hypotheses.
To simulate this sequential process, a while-loop embraces the sequential *t*-test function in the code below, which will not stop until one of the hypotheses is accepted or the maximum of the data is reached.

```{r, echo=TRUE}
# define the starting point
decision <- "continue sampling"
n_person <- 3

# simulation of the sequential procedure
while(decision == "continue sampling") {
  # get the current data
  df <- df_stress[1:n_person,]
  
  # run the sequential test and save the results
  results <- seq_ttest(df$one_year_stress,
                       df$baseline_stress,
                       alpha = alpha,
                       power = power,
                       d = d,
                       paired = paired,
                       alternative = alternative)
  # save the current desicion
  decision <- results@decision
  
  # add a new person
  n_person <- n_person + 1
  
  # break if the maximum of the data is reached
  if (n_person > nrow(df_stress)) {
    break
  }
}

# console output
results
```

The while-loop comes to an end after `r n_person - 1` data points.

### Report Results

```{r, echo=TRUE}
# Required results for the report

# likelihood ratio (LR)
LR <- round(results@likelihood_ratio, digits = 2)
LR

# sample size (N) = degrees of freedom +2 (two-samples) or +1 (one-sample & paired)
N <- results@df + 1
N

# baseline stress (M and SD)
mean_t1 <- round(mean(df$baseline_stress), digits = 2)
mean_t1
sd_t1 <- round(sd(df$baseline_stress), digits = 2)
sd_t1

# after one year stress (M and SD)
mean_t2 <- round(mean(df$one_year_stress), digits = 2)
mean_t2
sd_t2 <- round(sd(df$one_year_stress), digits = 2)
sd_t2

# NOT INCLUDED IN THE PACKAGE 

# calculate effect size: Cohen´s d
d_results <- effsize::cohen.d(df$one_year_stress,
                 df$baseline_stress,
                 paired = TRUE)
d <- round(d_results$estimate, digits = 2)
d

# confidence intervall
d_lower <- round(d_results$conf.int[[1]], digits = 2)
d_lower
d_upper <- round(d_results$conf.int[[2]], digits = 2)
d_upper
```

Starting at *N* = 2, the test stops sampling at a total sample size of *N* = `r N` with *LR*~`r N`~ = `r LR`.
This ratio indicates that the data are about 20 times more likely under H~1~ than under H~0~.
Thus, we accept the alternative hypothesis: The perceived stress at the second measurement (*M* = `r mean_t2`, *SD* = `r sd_t2`) is higher than one year ago at the baseline measurement (*M* = `r mean_t1`, *SD* = `r sd_t1`), Cohen's *d* = `r d`, 95% CI [`r d_lower`, `r d_upper`].[^1]^,^[^2]

## References

[^1]: see @schnuerch2020.

[^2]: Note that this estimate of Cohen's d as well as the CI are based on the assumption of a fixed sample size and, thus, might be biased toward an overestimation of the true effect size. See for details @schnuerch2020a.

---
title: 'sprtt: An R Package and Toolbox for Sequential Probability Ratio Tests'
tags:
  - R
  - sequential testing
  - sequential probability ratio test
  - SPRT
  - sequential ANOVA
authors:
  - name: Meike Snijder-Steinhilber
    orcid: 0000-0002-7144-2100
    affiliation: "1" 
  - name: Martin Schnuerch
    orcid: 0000-0001-6531-2265
    affiliation: "2"
  - name: Anna-Lena Schubert
    orcid: 0000-0001-7248-0662
    affiliation: "2"
affiliations:
 - name: University of Mainz, Germany
   index: 1
 - name: University of Mannheim, Germany
   index: 2
date: 28 February 2026
bibliography: references.bib
---

# Summary
[//]: # (*A description of the high-level functionality and purpose of the software for a diverse, non-specialist audience.*)

The `sprtt` package provides a toolbox for Sequential Probability Ratio Tests (SPRTs), implementing modern variants including sequential *t*-tests and sequential ANOVA for applied and methodological research. 
While traditional fixed-sample designs require researchers to commit to a predetermined sample size, SPRTs enable continuous evidence evaluation with predefined stopping rules -- terminating data collection as soon as evidence crosses a threshold for rejecting or accepting the null hypothesis [@wald1945].
Crucially, this flexibility comes without inflating long-run Type I and Type II error rates beyond the levels specified in advance.
For Wald’s original SPRT under simple hypotheses, 50% fewer observations are required compared to a Neyman-Pearson fixed-sample design [@wald1945].
Newer variants extend these methods to composite hypotheses through sequential *t*-tests and sequential ANOVA -- designs standard in fields like psychology and medicine.
For the composite-hypothesis extensions implemented in `sprtt`, efficiency cannot be derived analytically [@wald1945; @cox1952; @kollerstrom1979; @schnuerch2020], but simulation studies have shown that error rates are well-controlled and efficiency gains remain of similar magnitude [@schnuerch2020; @steinhilber2024; @stefan2022].
Despite the long history of SPRTs, the `sprtt` package is the first to provide accessible software implementations for both sequential *t*-tests [@rushton1950; @hajnal1961; @schnuerch2020] and sequential ANOVA [@wetherill1986; @steinhilber2024].
The package implements these validated procedures and additionally provides example datasets, data generating functions, sample size planning, and visualization tools to facilitate the adoption of SPRTs in applied research.


# Statement of need

[//]: # (*A section that clearly illustrates the research purpose of the software and places it in the context of related work. This should clearly state what problems the software is designed to solve, who the target audience is, and its relation to other work.*)

Due to the replication crisis [@opensciencecollaboration2015; @ioannidis2005; @bogdan2025] in empirical fields like psychology and medicine, statistical procedures have been scrutinized, and new alternatives have gained attention [@cumming2014; @lakens2018; @wagenmakers2018].
Sequential testing methods have become increasingly popular in recent years as they directly address pressing demands in empirical research: the need to minimize resource expenditure and participant burden without sacrificing statistical rigor [@schnuerch2020; @steinhilber2024; @ly2025; @lakens2021b; @erdfelder2021].
This is relevant across all empirical research, and particularly vital in clinical settings where continued data collection can carry real ethical costs.

Although SPRTs are well-established in the statistical literature [@wald1947; @siegmund1985; @bartroff2012; @tartakovsky2014], their original formulation relies on simple hypotheses, which are rarely used in applied research: they require researchers to specify nuisance parameters -- such as the variance -- which are rarely known in advance.
As a first step toward practical applicability, variants based on composite hypotheses were developed, namely the sequential *t*-test [@rushton1950; @hajnal1961] and sequential ANOVA [@wetherill1986].
As a second step, these variants were recently validated in simulation studies, establishing their statistical properties under realistic conditions [@schnuerch2020; @steinhilber2024].
As a third step, this methodological progress needed to be matched by accessible software: prior to `sprtt`, the only available implementation was a bare R script provided alongside validation work [@schnuerch2020].
Translating these promising statistical methods into accessible, user-friendly, and open-source software is therefore essential for finally closing the gap between statistical theory and adoption in practice.

# State of the field                                                                                                                  
[//]: # (*A description of how this software compares to other commonly-used packages in the research area. If related tools exist, provide a clear “build vs. contribute” justification explaining your unique scholarly contribution and why existing alternatives are insufficient.*)

The landscape of sequential testing software is sparse.
Beyond R, very few software packages appear to exist, though several major technology companies including Netflix, Uber, and Spotify have either published on sequential testing and SPRT variants or stated their use, suggesting that proprietary implementations may exist in industry [@bibaut2024; @schultzberg2023; @deb2018].
The only Python implementation, the `sprt` package on PyPI [@yu2017], covers Wald's SPRT for Normal, Binomial, and Poisson distributions but has not been updated since its initial release in 2017 and lacks documentation.
No SPRT implementations seem to exist in Julia.
A JavaScript library for sequential generalized likelihood ratio tests `SeGLiR` [@oygard2014] targets browser-based A/B testing and has not been maintained since 2017.
JASP [@love2019] is a free and open-source application that implements sequential Bayesian hypothesis testing [@schonbrodt2017], using a Bayes Factor rather than a likelihood ratio as the monitoring statistic, which requires the specification of prior distributions.
These Bayesian tools address an important but different use case.
The present package is intended for researchers who prefer a frequentist sequential framework, want to control long-run Type I and Type II error rates in familiar Neyman–Pearson terms, or wish to avoid the need to specify prior distributions.
In R, the package `SPRT` [@budihal2025] implements Wald's original sequential tests for simple hypotheses, the `gsDesign` [@anderson2026] package provides a function for truncated binomial SPRTs, and the `MSPRT` [@pramanik2020] and `Sequential` [@silva2025] packages cover a variety of truncated SPRT variants.
Beyond the SPRT, anytime-valid inference has emerged as an alternative sequential testing framework, using e-values to guarantee validity at any sample size [@ramdas2023; @grunwald2023] -- current software implementations include the R package `safestats` [@ly2024; @ly2025] and the Python package `savvi` [@assuncao2024].
To our knowledge, no publicly available software implements sequential *t*-tests or sequential one-way ANOVA as described and validated by @schnuerch2020 and @steinhilber2024.
The `sprtt` package fills this gap directly.

# Research impact

[//]: # (*Evidence of realized impact (publications, external use, integrations) or credible near-term significance (benchmarks, reproducible materials, community-readiness signals). The evidence should be compelling and specific, not aspirational.*)

The `sprtt` package was first published on CRAN in 2021 and has since accumulated close to 13,000 downloads, averaging approximately 200 downloads per month in the 12 months preceding March 2026 [@steinhilber2023].
The package has been used in experimental research [@quevedoputter2022], simulation studies [@steinhilber2024; @steinhilber2025], and has been referenced in methodological work [@schubert2025a; @fischer2025].
The target audience includes applied researchers using SPRT variants in their empirical work, as well as methodologists conducting simulation studies to gain further insights into the properties of SPRTs.

![Monthly CRAN downloads of the `sprtt` package since its first release in August 2021. Dashed vertical lines indicate CRAN release versions. The LOESS trend line with 95% confidence band reflects the overall download trajectory across complete months.](sprtt_downloads.png)


# Software design

[//]: # (*An explanation of the trade-offs you weighed, the design/architecture you chose, and why it matters for your research application. This should demonstrate meaningful design thinking beyond a superficial code structure description.*)

The `sprtt` package is built around two main user-facing functions:
`seq_ttest()` and `seq_anova()`.
The `seq_ttest()` function implements the sequential *t*-test and deliberately mirrors the interface of the `t.test()` function from the `stats` package to ensure familiarity for R users.
The `seq_anova()` function follows a similar design philosophy, maintaining consistency across the package's interface.

## Design principle

The core design principle is modularity: each internal function should perform one task well.
This approach emphasizes simplicity, testability, clear structure, and minimal code repetition.
The internal architecture of the core functions is documented in more detail in the developer vignette of the `sprtt` package.
The package is designed to return interpretable results not only when a stopping boundary is crossed, but also when monitoring remains inconclusive at the current stage of data collection.
More generally, functions perform input validation to catch common issues such as invalid argument types, missing values, or out-of-range parameters.

While the primary focus remains on implementing well-tested SPRT variants with proven efficiency and error rate control, the package continuously expands its functionality to improve user experience.
Supporting features include example datasets, data simulation functions, visualization tools for sequential ANOVA results, and sample size planning for sequential ANOVA.
The `lifecycle` package is used throughout to clearly communicate the maturity status of each function -- an important consideration for research software where interface stability directly affects reproducibility.
The core functions `seq_ttest()`, `seq_anova()`, and the data simulation utilities are stable: we commit to not introducing silent breaking changes to these functions.
Where changes are unavoidable, users will be informed through deprecation warnings and messaging well in advance.
Newer additions, including the visualization tools and the sample size planning function, are marked as experimental, reflecting that their interfaces may still undergo substantial revisions as they mature.

A concrete illustration of why this distinction matters is the plot function for `seq_ttest()`.
Mirroring the `t.test()` interface was a deliberate choice to lower the barrier to adoption, but as the package grew, a complication emerged: the wide variety of input formats accepted by `t.test()` has so far prevented the implementation of a consistent plot function for `seq_ttest()` -- a feature that already exists for `seq_anova()` and is planned for a future release.
Resolving this may require interface adjustments to `seq_ttest()`, which will be handled through the deprecation-with-messaging approach rather than silent breaking changes.


## External data

Sample size planning for the implemented tests cannot be derived analytically and instead requires extensive Monte Carlo simulations to characterize sampling behavior across a wide range of parameter combinations.
The `plan_sample_size()` function addresses this by generating an HTML report based on a pre-computed simulation dataset covering multiple effect sizes, group sizes, and Type II error rates -- each estimated from 10,000 replications per condition, run on a high-performance computing cluster.
Pre-computing this dataset offers several advantages over on-demand simulation: recommendations are returned instantly, all users access identical results ensuring reproducibility, and redundant computation across research groups is avoided.
The trade-off is that the lookup covers only a predefined set of parameter combinations; users with custom scenarios are therefore directed to the simulation functions to generate tailored estimates.

However, the comprehensive nature of these simulations produces a dataset too large to bundle directly with the package under CRAN size constraints.
To resolve this tension, the simulation dataset is maintained in a separate GitHub repository (https://github.com/MeikeSteinhilber/sprtt_plan_sample_size) and downloaded on demand, after which it is cached locally to avoid repeated downloads.
This separation also serves a transparency purpose: the full simulation pipeline including the hierarchical SLURM scripts used for cluster execution is publicly available for inspection and verification.
To give users direct control over this external dependency, the sprtt package includes dedicated helper functions (`download_sample_size_data()`, `cache_info()`, `cache_clear()`) for manually downloading, inspecting, and clearing the locally cached dataset.
The generated HTML report records the package version and the exact version of the downloaded simulation dataset, allowing users to reproduce recommendations even if the external repository is updated later.

# Software documentation

The `sprtt` package is documented through a dedicated website (https://meikesteinhilber.github.io/sprtt/), a README on both the main GitHub repository and the supplementary repository hosting simulation code and results for the `plan_sample_size()` function.
The package further includes a comprehensive set of vignettes.
Introductory vignettes cover general package usage and a recommended workflow and an introduction to SPRTs, complemented by a simple *t*-test use case.
More advanced vignettes provide dedicated guidance on the sequential *t*-test and sequential one-way ANOVA.
Finally, further topics are addressed in vignettes on sample size planning and a developer guide for users who want to contribute to or extend the package.

# AI usage disclosure

[//]: # (*Transparent disclosure of any use of generative AI in the software creation, documentation, or paper authoring. If no AI tools were used, state this explicitly. If AI tools were used, describe how they were used and how the quality and correctness of AI-generated content was verified.*)

The core `sprtt` implementation, all architectural decisions, and the research contributions are original human intellectual work.
Development began in February 2021 and predates the widespread availability of modern AI-assisted programming tools, with the majority of the codebase written without AI assistance (CRAN releases: August 2021 and July 2023).
For the latest release, generative AI (Claude, Anthropic) was used to assist with debugging new code, writing unit tests, and reviewing the package documentation for improvements.
For this manuscript, AI was additionally used to support writing tasks such as improving grammar and spelling, formatting of references, and suggesting manuscript structure.
In all cases, AI served an assistive role only, and all output was thoroughly reviewed and verified by the authors.


# Acknowledgements

We thank the Carl Zeiss Foundation for the generous 5-year funding of SMART-AGE (P2019-01-003; 2021-2026).
Parts of this research were supported by a grant from the German Research Foundation (Deutsche Forschungsgemeinschaft, GRK 2277) to the Research Training Group “Statistical Modeling in Psychology”.
Parts of this research were conducted using the supercomputer Mogon II and services offered by Johannes Gutenberg University Mainz (hpc.uni-mainz.de).

# References

---
title: "Sequential t-Test"
author: "Meike Snijder-Steinhilber"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 4
description: >
  This vignette describes the sequential t-test.
vignette: >
  %\VignetteIndexEntry{Sequential t-Test}
  %\VignetteEncoding{UTF-8}{inputenc}
  %\VignetteEngine{knitr::rmarkdown}
bibliography: references.bib
csl: "apa.csl"
---



<style>
/* Use custom class names to avoid Bootstrap conflicts */
.custom-info-box {
  background-color: white !important;
  border: 2px solid #2c3e50 !important;
  border-left: 5px solid #2c3e50 !important;
  color: #2c3e50 !important;
  margin-top: 1.5em !important;
  margin-bottom: 1.5em !important;
  border-radius: 4px !important;
  padding: 15px !important;
  padding-left: 50px !important;
  position: relative !important;
}
.custom-info-box::before {
  content: "🛈";
  position: absolute;
  left: 0;
  top: 0;
  bottom: 0;
  width: 40px;
  background-color: #2c3e50;
  color: white;
  display: flex;
  align-items: center;
  justify-content: center;
  font-size: 1.3em;
  border-radius: 2px 0 0 2px;
}

.custom-warning-box {
  background-color: white !important;
  border: 2px solid #18bc9c !important;
  border-left: 5px solid #18bc9c !important;
  color: #2c3e50 !important;
  margin-top: 1.5em !important;
  margin-bottom: 1.5em !important;
  border-radius: 4px !important;
  padding: 15px !important;
  padding-left: 50px !important;
  position: relative !important;
}
.custom-warning-box::before {
  content: "⚠";
  position: absolute;
  left: 0;
  top: 0;
  bottom: 0;
  width: 40px;
  background-color: #18bc9c;
  color: white;
  display: flex;
  align-items: center;
  justify-content: center;
  font-size: 1.3em;
  border-radius: 2px 0 0 2px;
}

</style>


```{r, echo = FALSE, message = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE
)
```


<!-- ```{r, child="sprts.Rmd"} -->

<!-- ``` -->


## What is the sequential *t*-test?

The sequential *t*-test is based on the Sequential Probability Ratio Test (SPRT) by Abraham @wald1947, which is a highly efficient sequential hypothesis test.
However, the usage of Wald´s SPRT is limited in the case of normally distributed data, because the variance has to be known or specified in the hypothesis.
Rushton [-@rushton1950; -@rushton1952] and @hajnal1961 have further developed the SPRT using the *t*-statistic.
The basic idea is to transform the sequence of observations (which is dependent on the variance) into a sequence of the associated *t*-statistic (which is *in*dependent of the variance).

In the SPRT the null and alternative hypotheses are defined as follows, with $\theta$ representing the model parameter :

$$
H_0:\ \theta\ =\ \theta_0,
$$

$$
H_1:\ \theta\ =\ \theta_1.
$$

The test statistic of the SPRT is based on a likelihood ratio, which is a measure of the relative evidence in the data for the given hypotheses.
More specifically, it is the ratio of the likelihood of the alternative hypothesis to the likelihood of the null hypothesis at the *m*-th step of the sampling process (LR~m~).

$$
LR_{m} 
= \frac{f(\text{data}_m  | H_1)}{f(\text{data}_m | H_0)}
= \frac{f(x_1,...,x_m | \theta_1)}{f(x_1,...,x_m  | \theta_0)} 
$$

Before the transformation into the *t*-statistic, the model parameter $\theta$ contains the parameters of a normal distribution: the mean ($\mu$) and the standard deviation ($\sigma$).
Therefore, the Wald SPRT requires prior knowledge about the variance ($\sigma^2$) or a specification in the hypotheses.

After the transformation of the observed values into the associated *t*-statistic, the model parameter $\theta$ contains the parameters of the non-central *t*-distribution: the degrees of freedom (*df*) and the non-centrality parameter ($\Delta$).

$$
{f(x_1,...,x_m | \mu,\sigma)} => {f(t_2,...,t_m  | df,\Delta)} 
$$

For the calculation of the degrees of freedom, only the sample size of the group(s) is needed. The non-centrality parameter also requires a specification of the expected effect size in form of Cohen's *d*.

To eventually calculate the LR of the sequential *t*-test, only the current t~*m*~-statistic is necessary. @rushton1950 demonstrated that an SPRT can be performed by simply considering the ratio of probability densities for the most recent t~*m*~ statistic under the alternative and null hypothesis at any *m-*th stage.
Thus, the test statistic for a one and two-sided sequential t-test can be calculated as follows:

$$
LR_{m,\ one-sided\ sequential\ t-test} 
= \frac{f(t_m | \theta_1)} {f(t_m | \theta_0)},
$$

$$
LR_{m,\ two-sided\ sequential\ t-test} 
= \frac{f(t_m^2 | \theta_1)} {f(t_m^2 | \theta_0)}.
$$

To account for the fact that the algebraic sign is unknown in a two-sided test, the *t*-value is squared [@rushton1952].

After the calculation of the test statistic, the decision will be either to continue sampling or to terminate the sampling and accept one of the hypotheses.
@wald1945 defined the following rules for the SPRT:

|    Condition    |                    Decision |
|:---------------:|----------------------------:|
|    LR~m~ ≤ B    | accept H~0~ and reject H~1~ |
| B \< LR~m~ \< A |           continue sampling |
|    LR~m~ ≤ A    | accept H~1~ and reject H~0~ |

The A and B boundaries are calculated with the previously defined error rates $\alpha$ (Type I error) and $\beta$ (Type II error) as follows:

$$
A = \left( \frac{1 - \beta}{\alpha} \right),
$$

$$
B = \left( \frac{\beta}{1 - \alpha} \right).
$$

In summary, three specifications are required to calculate a sequential *t*-test:

-   the $\alpha$ error probability (usually 0.05 or less),

-   the $\beta$ error probability (usually .20 or less), and

-   Cohen´s d (either as the expected effect size or as the lower limit for a effect size of interest).




## How do I use the `seq_ttest()` function?


The `seq_ttest()` function works similarly to the `t.test()` function from the `stats` package if one is familiar with that already.
The table below shows all possible combinations which can be performed with the `sprtt` package.

|                            | Two-sample test | One-sample test |
|----------------------------|-----------------|-----------------|
| Two-sided                  | x               | x               |
| One-sided                  | x               | x               |
| Paired (repeated measures) | x               |                 |


Sequential *t*-tests require some specification from the user:

-   the variables, which contain the data,

-   the error probability `alpha`,

-   the `power` (1 - $\beta$),

-   the effect size Cohen's *d*, which represents the expected effect size or the lowest effect size of interest, and

-   optional arguments to further specify the test.

However, in some cases, it is not necessary to specify all arguments because some of them have default values.
If these values are the ones required, they can be skipped.

| Argument    | Default value | Input option                   |
|-------------|:--------------|:-------------------------------|
| x           | -             | formula or numeric input       |
| y           | NULL          | numeric vector                 |
| data        | NULL          | data frame                     |
| mu          | 0             | numeric value                  |
| d           | -             | numeric value                  |
| alpha       | 0.05          | numeric value between 0 and 1  |
| power       | 0.95          | numeric value between 0 and 1  |
| alternative | "two.sided"   | "two.sided", "greater", "less" |
| paired      | FALSE         | TRUE or FALSE                  |
| na.rm       | TRUE          | TRUE or FALSE                  |
| verbose     | TRUE          | TRUE or FALSE                  |

There are two different ways how the data can be transferred into the function.
The `x` argument takes either `formula` or `numeric` input. Which input option is recommended depends on the structure of the data.

### x argument: formula input

The `formula` input is used when both groups are merged in one variable and there is a second variable that indicates group membership. This input option uses the `x` argument and the `data` argument if the variables are stored in a data frame.

#### *Two-sample test*

```{r}
library(sprtt)
# show data frame --------------------------------------------------------------
head(df_income)

# sequential t-test: data argument ---------------------------------------------
seq_ttest(monthly_income ~ sex,     # x argument 
          data = df_income,
          d = 0.2)
```

#### *One-sample test*

To perform a one-sample test, the right side of the formula has to be 1.
The `mu` argument is also required, which specifies the mean value that one wants to test against.

```{r}
# show data frame --------------------------------------------------------------
head(df_income)

# sequential t-test: data argument ---------------------------------------------
seq_ttest(monthly_income ~ 1,     # x argument 
          mu = 3000,
          d = 0.5,
          data = df_income)
```

### x argument: numeric input

The `numeric` input is used when each group has its own variable.
The variables can either be put in the global environment directly or be stored in a data frame.

#### *Two-sample test*

If one wants to perform a two-sample test, the `y` argument is required in addition to `x`.
If the data are stored in a data frame, the `$` operator is essential to get access to the variables.

```{r}
# show data frame --------------------------------------------------------------
head(df_cancer)

# sequential t-test: $ operator ------------------------------------------------
seq_ttest(df_cancer$treatment_group, # x argument
          df_cancer$control_group,   # y argument
          d = 0.3,                   
          verbose = FALSE)           


# sequential t-test: global variables ------------------------------------------
treatment <- df_cancer$treatment_group
control <- df_cancer$control_group

seq_ttest(treatment,
          control,
          d = 0.3,
          verbose = FALSE)

```

#### *One-sample test*

If one wants to perform a one-sample test there is only one group and therefore only one variable.
If the data are in a data frame, the `$` operator is essential to get access to the variables.
The `mu` argument is additionally required, which specifies the mean which one wants to test against.

```{r}
# show data frame --------------------------------------------------------------
head(df_cancer)

# sequential t-test: $ operator ------------------------------------------------
seq_ttest(df_cancer$treatment_group,     # x argument
          mu = 3.5,            
          d = 0.2,
          verbose = FALSE)


# sequential t-test: global variables ------------------------------------------
treatment <- df_cancer$treatment_group

seq_ttest(treatment,     # x argument
          mu = 3.5, 
          data = df,
          d = 0.2,                  
          verbose = FALSE)

```

### Further arguments

#### Paired

The `paired` argument states if the data are paired.
To perform a paired sequential *t*-test, `paired` has to be set to `TRUE`.

```{r}
# show data frame --------------------------------------------------------------
head(df_stress)

# sequential t-test: $ operator ------------------------------------------------
seq_ttest(df_stress$baseline_stress, # x argument
          df_stress$one_year_stress, # y argument
          d = 0.3,
          paired = TRUE,
          data = df_stress)  
```

#### Alternative

The `alternative` argument states in which way the test is to be performed:

-   two-sided: `"two.sided"` or

-   one-sided: `"less"` or `"greater"`.

```{r}
# show data frame --------------------------------------------------------------
head(df_income)

# sequential t-test: data argument ---------------------------------------------
seq_ttest(monthly_income ~ 1,     # x argument 
          mu = 1000,
          d = 0.3,
          alternative = "two.sided",
          data = df_income)

```

#### Na.rm

The `na.rm` argument defines the handling of missing values. If set to `TRUE` (default value), it will remove all missing values automatically. If this behavior is not wanted, the `na.rm` argument has to be set to `FALSE`. If missing values are discovered, an error is triggered.

#### Verbose

The `verbose` argument defines the extent of the output shown in the console.
If set to `TRUE` (default value), the output will be elaborate, if set to `FALSE` the output will be short.

```{r}
# sequential t-test: verbose FALSE ---------------------------------------------
seq_ttest(df_cancer$treatment_group, # x argument
          df_cancer$control_group,   # y argument
          d = 0.3,
          verbose = FALSE)

# sequential t-test: verbose TRUE ----------------------------------------------
seq_ttest(df_cancer$treatment_group, # x argument
          df_cancer$control_group,   # y argument
          d = 0.3,
          verbose = TRUE) 
```

## How do I get access to the results?

The simplest way to get access to the results is to run the `seq_ttest()` function.
It will print the results automatically in the console. The verbose argument specifies how much information is wished to be shown.

However, the recommended way is to save the results in an object (e.g "results").
Doing so allows running further calculations with it afterward. It is important to keep in mind that the output object will be an S4 class.
Therefore the access operator is the `@` sign or the `[]` brackets.

```{r}
# save the resuts in a object 
results <- seq_ttest(df_cancer$treatment_group,
                     df_cancer$control_group,  
                     d = 0.3)
# access the object with the @ operator 
results@decision

# access the object with the [] brackets
results["decision"]
```


## References

---
title: "SPRTs"
author: "Meike Steinhilber"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 4
description: >
  This vignette describes SPRTs in general.
vignette: >
  %\VignetteIndexEntry{SPRTs}
  %\VignetteEncoding{UTF-8}{inputenc}
  %\VignetteEngine{knitr::rmarkdown}
bibliography: references.bib
csl: "apa.csl"
---


## Overview

The `sprtt` package is a **s**equential **p**robability **r**atio **t**ests **t**oolbox (**sprtt**). This vignette describes the theoretical background of these tests.

Other recommended vignettes cover:

-   a [general guide](https://meikesteinhilber.github.io/sprtt/articles/usage-sprtt.html), how to use the package and

-   an extended [use case](https://meikesteinhilber.github.io/sprtt/articles/use-case.html).


## What is a sequential test procedure?

With a sequential approach, data is continuously collected and an analysis is performed after each data point, which can lead to three different results [@wald1945]:

-   The data collection is *terminated* because enough evidence has been collected for the null hypothesis (H~0~).

-   The data collection is *terminated* because enough evidence has been collected for the alternative hypothesis (H~1~).

-   The data collection *will continue* as there is not yet enough evidence for either of the two hypotheses.

Basically it is not necessary to perform an analysis after each data point --- several data points can also be added at once. However, this affects the sample size (N) and the error rates [@schnuerch2020a].

The efficiency of sequential designs has already been examined. Reductions in the sample by 50% and more were found in comparison to analyses with fixed sample sizes [@wald1945; @schnuerch2020a]. Sequential hypothesis testing is therefore particularly suitable when resources are limited because the required sample size is reduced without compromising predefined error probabilities.

---
title: "Sequential One-Way ANOVA"
author: "Meike Snijder-Steinhilber"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 4
description: >
  This vignette describes the sequential one-way ANOVA.
vignette: >
  %\VignetteIndexEntry{Sequential One-Way ANOVA}
  %\VignetteEncoding{UTF-8}{inputenc}
  %\VignetteEngine{knitr::rmarkdown}
bibliography: references.bib
csl: "apa.csl"
---

```{r, echo = FALSE, message = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE
)
```


The sequential one-way fixed effects ANOVA is a sequential hypothesis test based on the Sequential Probability Ratio Test (SPRT) framework [@wald1947].
It extends SPRTs to the comparison of two or more independent groups and can be used as an efficient alternative to the classical fixed-sample one-way ANOVA.
For detailed information, see @steinhilber2024.
For a general introduction to SPRTs, see the vignette `vignette("sprt")`.

**Note:** The repeated measures ANOVA is not yet implemented in the `sprtt` package.



## Sequential One-Way ANOVA

Analysis of variance (ANOVA) is widely used to compare means across multiple groups.
Traditional fixed-sample ANOVAs require an a priori power analysis to determine the necessary sample size, which often results in large samples, especially when expected effect sizes are small [@steinhilber2024].
Given the prevalence of small to medium effects in psychology [@funder2019; @richard2003], many studies end up underpowered because the required sample sizes exceed available resources [@button2013; @szucs2017].

The sequential one-way ANOVA provides an efficient alternative by applying the SPRT framework to comparisons of $k$ groups.
Instead of collecting a fixed number of observations, data are collected sequentially and the test evaluates evidence after each step until a decision is reached [@steinhilber2024].

### Hypotheses

The sequential one-way ANOVA tests the following hypotheses, specified in terms of Cohen's $f$:


```{=tex}
\begin{aligned}
H_0 &: f = 0 \\
H_1 &: f = f_{\text{exp}}, \quad (f_{\text{exp}} > 0)
\end{aligned}
```



Cohen's $f$ is defined as:

$$f = \frac{\sigma_m}{\sigma}$$

where $\sigma$ is the common within-population standard deviation and $\hat{\sigma}_m = \sqrt{\frac{\sum_{i=1}^{k}(m_i - \bar{m})^2}{k}}$, with $m_i$ being the mean of group $i$ and $\bar{m}$ the overall mean across all $k$ groups. In other words, $f$ captures the spread of group means relative to the common within-group variability.

### The $F$ Statistic

At the $n$-th sequential step, the $F$ statistic is calculated from $k$ groups with $n$ observations each (total $N = k \cdot n$, with $n, k \geq 2$):

$$F_n = \frac{SS_{\text{effect},n} / df_1}{SS_{\text{residual},n} / df_{2,n}}$$

with

$$SS_{\text{effect},n} = \sum_{i=1}^{k} n(\bar{x}_i - \bar{x})^2$$

$$SS_{\text{residual},n} = \sum_{i=1}^{k} \sum_{j=1}^{n} (x_{i,j} - \bar{x}_i)^2$$

$$df_1 = k - 1 \qquad \text{and} \qquad df_{2,n} = N - k$$

where $\bar{x}$ is the overall mean, $\bar{x}_i$ is the mean of group $i$, and $x_{i,j}$ is the $j$-th observation in group $i$ [@wetherill1986; @steinhilber2024].

### The Likelihood Ratio

The likelihood ratio at step $n$ is defined as the ratio of the likelihood under $H_1$ to the likelihood under $H_0$.
Using Cox's theorem [@cox1952], it is sufficient to compute this ratio only for the current $F_n$ statistic rather than the entire sequence of observations:

$$\text{LR}_n = \frac{f(F_n \mid df_1,\; df_{2,n},\; \Delta_{1n})}{f(F_n \mid df_1,\; df_{2,n})}$$

The numerator is the density of a *non-central* $F$ distribution with non-centrality parameter $\Delta_{1n}$, and the denominator is the density of a *central* $F$ distribution (i.e., $\Delta = 0$ under $H_0$).
The non-centrality parameter is linked to Cohen's $f$ via:

$$\Delta_1 = f_{\text{exp}}^2 \cdot N$$

### Decision Boundaries

The sequential ANOVA uses two decision boundaries based on the specified error rates:

- Upper boundary: $A = \frac{1-\beta}{\alpha}$
- Lower boundary: $B = \frac{\beta}{1-\alpha}$

where $\alpha$ is the Type I error rate and $\beta$ is the Type II error rate.

### Decision Rule

At each analysis step, compare the likelihood ratio $\text{LR}_n$ to these boundaries:

- If $\text{LR}_n \geq A$: Stop data collection and accept $H_1$
- If $\text{LR}_n \leq B$: Stop data collection and accept $H_0$
- If $B < \text{LR}_n < A$: Continue collecting data (no decision yet)

### Efficiency and Robustness

Simulations with $k = 4$ groups demonstrate that the sequential ANOVA is substantially more efficient than fixed-sample designs: in 87% of cases the sequential sample was smaller than the fixed sample, with an average sample size reduction of 58%.
Efficiency gains are particularly pronounced for small expected effect sizes [@steinhilber2024].

Robustness analyses showed that the sequential and fixed ANOVA behave similarly under assumption violations.
Both procedures are robust to non-normal data (simulated using Gaussian mixture distributions) and to mildly unequal variances or group sizes.
The most critical scenario is a combination of unbalanced group sizes and unequal variances, specifically when smaller groups have larger variances—in this case, $\alpha$ error rates can be inflated in both the sequential and the fixed ANOVA.

As with all sequential procedures, effect size estimates from individual sequential ANOVAs are conditionally biased (see Section "The Bias-Efficiency Tradeoff").
However, the weighted average across studies closely approximates the true population effect size [@steinhilber2024].

The `seq_anova()` function in the `sprtt` package implements the sequential one-way fixed effects ANOVA described here.

## How to use `seq_anova()`

### Step 1: Simulate or Load Data

First, we simulate data for this tutorial.
In a real-world application, you would use actual data as it arrives from your ongoing data collection.

```{r example-0.1}
set.seed(333)
# Generate data with a medium effect
data <- sprtt::draw_sample_normal(
  k = 3,           # number of groups
  f = 0.25,        # effect size (Cohen's f)
  max_n = 22       # maximum sample size per group
)
```

Let's examine the structure of the simulated data:

```{r example-0.1b}
# View the first few rows
head(data)

# Check the sample sizes per group for the first 6 (2*k) data points
table(data$x[1:6])
```

The data frame contains two columns: `y` (the continuous outcome variable) and `x` (the grouping factor with 3 levels).
<!-- Each group has `r nrow(data)/3` observations, resulting in a total sample size of $N =$ `r nrow(data)`. -->

### Step 2: Initial Sequential Analysis

We can perform the first sequential ANOVA after collecting at least 2 observations per group (minimum $n = 6$ total for 3 groups).

```{r example-0.2}
# Calculate the sequential ANOVA
anova_results <- sprtt::seq_anova(
  y ~ x,
  f = 0.25,
  data = data[1:6, ],
  verbose = FALSE
)

# View results
anova_results

# Access the decision
anova_results@decision
```

The decision indicates we should continue data collection.
In practice, it's optimal to recalculate the sequential test after each new observation (or small batch of observations).

### Step 3: Repeated Sequential Testing

As new data arrives, we *repeatedly recalculate* the sequential ANOVA.
This is the core principle of sequential testing: check the decision criterion after each batch of new observations.

Let's assume we've now collected 20 observations total.
We recalculate the sequential ANOVA with the updated dataset:

```{r example-0.3}
# Calculate sequential ANOVA with larger sample
anova_results <- sprtt::seq_anova(
  y ~ x,
  f = 0.25,
  data = data[1:20, ],
  verbose = FALSE
)

# View results
anova_results

# Check decision
anova_results@decision
```

We still receive the decision to continue data collection.
This means we repeat the process: collect more data, recalculate the sequential test, and check the decision again.
This cycle continues until we reach a definitive decision to accept $H_0$ or $H_1$.

### Step 4: Final Analysis

```{r example-0.4}
# Calculate sequential ANOVA with complete dataset
anova_results <- sprtt::seq_anova(
  y ~ x,
  f = 0.25,
  data = data,
  verbose = TRUE
)

# View full results
anova_results
```

With a total sample size of $N =$ `r anova_results@total_sample_size`, we have reached a decision to `r anova_results@decision`.
Therefore, we stop data collection.

You can access specific components of the results object using the `@` operator:

```{r example-0.5}
# Access the decision
anova_results@decision

# Access the likelihood ratio
anova_results@likelihood_ratio

# Access the total sample size
anova_results@total_sample_size
```


## How to plot the ANOVA results


To visualize the likelihood ratio progression over time, `seq_anova()` must calculate the sequential test at each intermediate sample size.
Since these calculations are only needed for plotting and increase computational time, they are performed only when `plot = TRUE`.

### Scenario 1: Balanced Data with Perfect Sampling Order

When your data are perfectly balanced (equal $n$ per group) and sampled in order, you can set `plot = TRUE` and use either the default `seq_steps = "single"` or explicitly specify `seq_steps = "balanced"`.
Both will produce accurate likelihood ratio trajectories.

See `?seq_anova` for details on the `plot` and `seq_steps` arguments and other options.

```{r example-1}
set.seed(333)
data <- sprtt::draw_sample_normal(3, f = 0.25, max_n = 22)

# calculate the SPRT -----------------------------------------------------------
# Default: plot = TRUE with seq_steps = "single"
anova_results <- sprtt::seq_anova(y~x, f = 0.25,
                                  data = data, plot = TRUE)
# Explicitly specify seq_steps = "single"
anova_results <- sprtt::seq_anova(y~x, f = 0.25,
                                  data = data, plot = TRUE,
                                  seq_steps = "single")
# Use balanced sequential steps
anova_results <- sprtt::seq_anova(y~x, f = 0.25,
                                  data = data, plot = TRUE,
                                  seq_steps = "balanced")

# plot the results -------------------------------------------------------------
sprtt::plot_anova(anova_results)
```


### Scenario 2: Unbalanced Data with Imperfect Sampling Order

In real-world applications, data often arrive with unequal group sizes and in imperfect group order.
This scenario demonstrates how to handle such cases.

**Why custom sequential steps are needed:**

- The `"balanced"` option assumes equal $n$ across groups at each step, which doesn't match our data structure
- The `"single"` option would produce an error because some groups have fewer than 2 observations in the initial first data points
- **Solution:** Define custom sequential steps using a numeric vector

In this example, we start the sequential testing after 12 observations (ensuring each group has sufficient data), then calculate the likelihood ratio after each subsequent observation.

```{r example-2}
set.seed(333)
# Generate unbalanced data with a 1:1:2 sampling ratio -------------------------
data <- sprtt::draw_sample_normal(3, f = 0.25, max_n = 37, sample_ratio = c(1,1,2))
# Randomize the order to get a more realistic data collection
data <- data[sample(nrow(data)),] 

# Calculate the SPRT with custom sequential steps ------------------------------
anova_results <- sprtt::seq_anova(
                          y~x,
                          f = 0.25,
                          data = data,
                          plot = TRUE,
                          # Start at n=12, then test after each observation
                          seq_steps = 12:nrow(data)) 

# Plot the results with custom styling -----------------------------------------
sprtt::plot_anova(anova_results,
                 labels = TRUE,
                 position_labels_x = 0.2,
                 position_labels_y = 0.2,
                 position_lr_x = 100,
                 position_lr_y = 1.8,
                 font_size = 20,
                 line_size = 1,
                 highlight_color = "steelblue"
                 )
```


## References

---
title: "How to use the sprtt package"
author: "Meike Steinhilber"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 4
description: >
  This vignette describes the sequential t-test and the usage of the
  `sprtt` package.
vignette: >
  %\VignetteIndexEntry{How to use the sprtt package}
  %\VignetteEncoding{UTF-8}{inputenc}
  %\VignetteEngine{knitr::rmarkdown}
bibliography: references.bib
csl: "apa.csl"
---

```{r, echo = FALSE, message = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Note

-   The data sets used below are included in the `sprtt` package. Thus, the data sets are available when the package is loaded.

-   In the R code sections:

    -   `# comment`: is a comment
    -   `function()`: is R code
    -   `#> results of function()`: is console output

## Overview

The `sprtt` package is a **s**equential **p**robability **r**atio **t**ests **t**oolbox (**sprtt**). This vignette demonstrates the usage of the package using the `seq_ttest()` function and all arguments with short examples. The theoretical background of the sequential *t*-test will not be covered by this vignette.

The `seq_ttest()` function has arguments to specify the requested sequential *t*-test. The table below shows all possible combinations which can be performed with the package.

|                            | Two-sample test | One-sample test |
|----------------------------|-----------------|-----------------|
| Two-sided                  | x               | x               |
| One-sided                  | x               | x               |
| Paired (repeated measures) | x               |                 |

Other recommended vignettes cover:

-   the [theoretical background](https://meikesteinhilber.github.io/sprtt/articles/sequential_testing.html) and

-   an extended [use case](https://meikesteinhilber.github.io/sprtt/articles/use-case.html).

## Installation

Prior to using the `sprtt` package it must be installed and loaded. The latest release version of the package can either be installed from CRAN or in the latest development version from GitHub. More information for the installation can be found [here](https://meikesteinhilber.github.io/sprtt/#installation).

```{r}
# load the package
library(sprtt)
```

## What does the package contain?

The `sprtt` package contains:

-   `seq_ttest()` : a function which performs sequential *t*-tests

-   `df_income`: a set of data to run the examples given in this vignette

-   `df_stress`: a set of data to run the examples given in this vignette

-   `df_cancer`: a set of data to run the examples given in this vignette

## How do I use the seq_ttest() function?

The `seq_ttest()` function works similarly to the `t.test()` function from the `stats` package if one is familiar with that already.

Sequential *t*-tests require some specification from the user:

-   the variables, which contain the data,

-   the error probability `alpha`,

-   the `power` (1 - 𝛽),

-   the effect size Cohen\`s `d`, which represents the expected effect size or the lowest effect size of interest, and

-   optional arguments to further specify the test.

However, in some cases, it is not necessary to specify all arguments because some of them have default values. If these values are the ones required, they can be skipped.

| Argument    | Default value | Input option                   |
|-------------|:--------------|:-------------------------------|
| x           | \-            | formula or numeric input       |
| y           | NULL          | numeric vector                 |
| data        | NULL          | data frame                     |
| mu          | 0             | numeric value                  |
| d           | \-            | numeric value                  |
| alpha       | 0.05          | numeric value between 0 and 1  |
| power       | 0.95          | numeric value between 0 and 1  |
| alternative | "two.sided"   | "two.sided", "greater", "less" |
| paired      | FALSE         | TRUE or FALSE                  |
| na.rm       | TRUE          | TRUE or FALSE                  |
| verbose     | TRUE          | TRUE or FALSE                  |

There are two different ways how the data can be transferred into the function. The `x` argument takes either `formula` or `numeric` input. Which input option is recommended depends on the structure of the data.

### x argument: formula input

The `formula` input is used when both groups are merged in one variable and there is a second variable that indicates group membership. This input option uses the `x` argument and the `data` argument if the variables are stored in a data frame.

#### *Two-sample test*

```{r}
# show data frame --------------------------------------------------------------
head(df_income)

# sequential t-test: data argument ---------------------------------------------
seq_ttest(monthly_income ~ sex,     # x argument 
          data = df_income,
          d = 0.2)
```

#### *One-sample test*

To perform a one-sample test, the right side of the formula has to be 1. The `mu` argument is also required, which specifies the mean value that one wants to test against.

```{r}
# show data frame --------------------------------------------------------------
head(df_income)

# sequential t-test: data argument ---------------------------------------------
seq_ttest(monthly_income ~ 1,     # x argument 
          mu = 3000,
          d = 0.5,
          data = df_income)
```

### x argument: numeric input

The `numeric` input is used when each group has its own variable. The variables can either be put in the global environment directly or be stored in a data frame.

#### *Two-sample test*

If one wants to perform a two-sample test, the `y` argument is required in addition to `x`. If the data are stored in a data frame, the `$` operator is essential to get access to the variables.

```{r}
# show data frame --------------------------------------------------------------
head(df_cancer)

# sequential t-test: $ operator ------------------------------------------------
seq_ttest(df_cancer$treatment_group, # x argument
          df_cancer$control_group,   # y argument
          d = 0.3,                   
          verbose = FALSE)           


# sequential t-test: global variables ------------------------------------------
treatment <- df_cancer$treatment_group
control <- df_cancer$control_group

seq_ttest(treatment,
          control,
          d = 0.3,
          verbose = FALSE)

```

#### *One-sample test*

If one wants to perform a one-sample test there is only one group and therefore only one variable. If the data are in a data frame, the `$` operator is essential to get access to the variables. The `mu` argument is additionally required, which specifies the mean which one wants to test against.

```{r}
# show data frame --------------------------------------------------------------
head(df_cancer)

# sequential t-test: $ operator ------------------------------------------------
seq_ttest(df_cancer$treatment_group,     # x argument
          mu = 3.5,            
          d = 0.2,
          verbose = FALSE)


# sequential t-test: global variables ------------------------------------------
treatment <- df_cancer$treatment_group

seq_ttest(treatment,     # x argument
          mu = 3.5, 
          data = df,
          d = 0.2,                  
          verbose = FALSE)

```

### Further arguments

#### Paired

The `paired` argument states if the data are paired. To perform a paired sequential *t*-test, `paired` has to be set to `TRUE`.

```{r}
# show data frame --------------------------------------------------------------
head(df_stress)

# sequential t-test: $ operator ------------------------------------------------
seq_ttest(df_stress$baseline_stress, # x argument
          df_stress$one_year_stress, # y argument
          d = 0.3,
          paired = TRUE,
          data = df_stress)  
```

#### Alternative

The `alternative` argument states in which way the test is to be performed:

-   two-sided: `"two.sided"` or

-   one-sided: `"less"` or `"greater"`.

```{r}
# show data frame --------------------------------------------------------------
head(df_income)

# sequential t-test: data argument ---------------------------------------------
seq_ttest(monthly_income ~ 1,     # x argument 
          mu = 1000,
          d = 0.3,
          alternative = "two.sided",
          data = df_income)

```

#### Na.rm

The `na.rm` argument defines the handling of missing values. If set to `TRUE` (default value), it will remove all missing values automatically. If this behavior is not wanted, the `na.rm` argument has to be set to `FALSE`. If missing values are discovered, an error is triggered.

#### Verbose

The `verbose` argument defines the extent of the output shown in the console. If set to `TRUE` (default value), the output will be elaborate, if set to `FALSE` the output will be short.

```{r}
# sequential t-test: verbose FALSE ---------------------------------------------
seq_ttest(df_cancer$treatment_group, # x argument
          df_cancer$control_group,   # y argument
          d = 0.3,
          verbose = FALSE)

# sequential t-test: verbose TRUE ----------------------------------------------
seq_ttest(df_cancer$treatment_group, # x argument
          df_cancer$control_group,   # y argument
          d = 0.3,
          verbose = TRUE) 
```

## How do I get access to the results?

The simplest way to get access to the results is to run the `seq_ttest()` function. It will print the results automatically in the console. The verbose argument specifies how much information is wished to be shown.

However, the recommended way is to save the results in an object (e.g "results"). Doing so allows running further calculations with it afterward. It is important to keep in mind that the output object will be an S4 class. Therefore the access operator is the `@` sign or the `[]` brackets.

```{r}
# save the resuts in a object 
results <- seq_ttest(df_cancer$treatment_group,
                     df_cancer$control_group,  
                     d = 0.3)
# access the object with the @ operator 
results@decision

# access the object with the [] brackets
results["decision"]
```

---
title: "Report: Sample size planning for sequential ANOVAs"
# author: "<i>created by the</i> sprtt <i>package</i> `r print(citation('sprtt'), style = 'text')`"
output:
  html_document:
    theme: flatly
    self_contained: true
    toc: true
    toc_float: 
      collapsed: false
      smooth_scroll: true
    number_sections: false
    toc_depth: 3
    toc_collapsed: false
    # code_folding: hide
    includes:
      in_header: favicon.html
params:
  df_all: NULL
  f_expected: 0.25
  beta: 0.05
  k_groups: 3
  decision_rate: 0.85
  data_version: "v.0.0-test"
  data_created: "0-0-0000"
---


<style>

/* Make the entire page layout wider */
.container-fluid {
  max-width: 1500px !important;  /* Increase from default ~1140px */
  padding-left: 40px !important;
  padding-right: 40px !important;
}

body {
  font-size: 16px;
}

/* Make h3 same size as h2 with more spacing */
h3 {
  font-size: 1.5em !important;
  margin-top: 1.5em !important;
  margin-bottom: 0.8em !important;
}

/* Custom alert box styling */
.alert-info {
  background-color: white !important;
  border: 2px solid #2c3e50 !important;
  border-left: 5px solid #2c3e50 !important;
  color: #2c3e50 !important;
  margin-top: 1.5em !important;
  margin-bottom: 1.5em !important;
  border-radius: 4px !important;
  padding-left: 50px !important;
  position: relative !important;
}

.alert-info::before {
  content: "🛈";
  position: absolute;
  left: 0;
  top: 0;
  bottom: 0;
  width: 40px;
  background-color: #2c3e50;
  color: white;
  display: flex;
  align-items: center;
  justify-content: center;
  font-size: 1.3em;
  border-radius: 2px 0 0 2px;
}

.alert-warning {
  background-color: white !important;
  border: 2px solid #18bc9c !important;
  border-left: 5px solid #18bc9c !important;
  color: #2c3e50 !important;
  margin-top: 1.5em !important;
  margin-bottom: 1.5em !important;
  border-radius: 4px !important;
  padding-left: 50px !important;
  position: relative !important;
}

.alert-warning::before {
  content: "⚠";
  position: absolute;
  left: 0;
  top: 0;
  bottom: 0;
  width: 40px;
  background-color: #18bc9c;
  color: white;
  display: flex;
  align-items: center;
  justify-content: center;
  font-size: 1.3em;
  border-radius: 2px 0 0 2px;
}

/* Make only top-level TOC items bold */
.tocify-header.list-group {
  font-weight: bold !important;
}

/* Make nested items normal weight */
.tocify-subheader .tocify-item {
  font-weight: normal !important;
}


/* TOC styling */
.tocify {
  width: 30% !important;
  max-width: 300px !important;
  min-width: 250px !important;
  top: 0 !important;  /* Start from top since header is inside now */
}

/* Main content area */
.col-md-9 {
  width: 70% !important;
}

/* Responsive breakpoint */
@media (max-width: 1200px) {
  .tocify, .toc-text-top {
    display: none !important;
  }
  
  .col-md-9 {
    width: 100% !important;
  }
}

/* Remove numbering from level 4 headings and deeper */
.section.level4 h4 .header-section-number,
.section.level5 h5 .header-section-number,
.section.level6 h6 .header-section-number {
  display: none;
}

/* Visually hide headings but keep them in layout */
.section.toc-only > h1,
.section.toc-only > h2,
.section.toc-only > h3,
.section.toc-only > h4,
.section.toc-only > h5,
.section.toc-only > h6,
h1.toc-only, h2.toc-only, h3.toc-only, h4.toc-only, h5.toc-only, h6.toc-only {
  position: absolute !important;
  left: -9999px !important;
  width: 1px !important;
  height: 1px !important;
  overflow: hidden !important;
}

.section.toc-only:target::before {
  content: "";
  display: block;
  height: 0;
  margin-top: -8px;
  border-top: 2px solid #bbb;
}
</style>

```{r, include=FALSE, eval=FALSE}
# DO NOT RUN
# This code generates the html.file so the icon in the browser is the sprtt logo

# Read and encode favicon
favicon_base64 <- base64enc::base64encode("pkgdown/favicon/favicon.ico")

# Create the HTML content
html_content <- sprintf(
  '<link rel="shortcut icon" href="data:image/x-icon;base64,%s" type="image/x-icon" />',
  favicon_base64
)

# Write to file in the skeleton folder
writeLines(
  html_content,
  "inst/rmarkdown/templates/report_sample_size/skeleton/favicon.html"
)
```


```{r, include=FALSE}
knitr::opts_chunk$set(warning = FALSE, message = FALSE, echo = FALSE)

library(dplyr)
library(glue)
library(ggplot2)
library(knitr)
library(gt)
library(ggtext)
library(scales)
```

```{r, include=FALSE}
df_all = params$df_all

if (is.null(df_all)) {df_all <- readRDS("../sprtt_plan_sample_size/meta_data/sprt_tool_df_all_10000.rds")}

pick_f_expected =     params$f_expected
pick_power =          1 - params$beta
pick_beta =           params$beta
pick_k_groups =       params$k_groups
pick_decision_rate =  params$decision_rate
data_version <- params$data_version
data_created <- params$data_created

pick_decision_rate_var = case_when(
  pick_decision_rate == 0.75 ~ "q75_n",
  pick_decision_rate == 0.80 ~ "q80_n",
  pick_decision_rate == 0.85 ~ "q85_n",
  pick_decision_rate == 0.90 ~ "q90_n",
  pick_decision_rate == 0.95 ~ "q95_n"
)

n_rep = df_all %>% filter(f_simulated == 0,
                          f_expected == pick_f_expected,
                          k_groups == pick_k_groups,
                          power == pick_power) %>% nrow(.)

```

```{r}
htmltools::tags$script(htmltools::HTML(sprintf("
document.addEventListener('DOMContentLoaded', function() {
  setTimeout(function() {
    var tocify = document.querySelector('.tocify');
    if (tocify) {
      var header = document.createElement('div');
      header.style.padding = '20px';
      header.style.backgroundColor = '#2c3e50';
      header.innerHTML = '<p style=\"font-size: 1.5em; color: white; font-weight: bold; margin: 0;\"><i>f</i><sub>exp</sub> = %s</p>';
      tocify.insertBefore(header, tocify.firstChild);
    }
  }, 100);
});
", pick_f_expected)))
```


```{r, include=FALSE}
# data -------------------------------------------------------------------------
#load(glue("{params$data_path}/sprt_tool_samples_{n_rep}.RData"))
# load(glue("meta_data/sprt_tool_samples_{n_rep}.RData"))

# df <- sprtt:::df
# df_all <- load_sample_size_data()

# df_all <- getFromNamespace("df_all", "sprtt")
# df     <- getFromNamespace("df",     "sprtt")
```


```{r, include=FALSE}
# variables --------------------------------------------------------------------
digits <- 2

palette <- "RdYlBu"  #"Dark2"
light_green <- "#a5d1c7"
green <- "#69b3a2"
dark_green <- "#548f81"
dark2_green <- "#104949"
light_blue <- "#cce5ff" 
grey_blue <- "#bccfe5"
blue <- "#2c7bb6"  #"#404080"
dark_blue <- "#1b3c70"
magenta <- "#e61be5"
rosa <- "#ff0088"
hell_rosa <- "#ffa1c2"
light_red <- "#e72b2e"
red <- "#d7191c"

# red <- 	"#d7191c" #"#ca0020" "#CD2626" "#8B0000"
dark_red <- "#8b1012"
black_red <- "#4b1919" # "#651718"

light_orange <- "#fffcf9"
orange <- "#fdae61"
dark_orange <- "#fc8715"
yellow <- "#ffffbf"

# Define Flatly-compatible accent colors (muted, modern)
light_green       <- "#18BC9C"  # Flatly teal
flatly_yellow     <- "#F1C40F"  # gold
light_red         <- "#E74C3C"  # Flatly red
flatly_red_dark   <- "#C0392B"  # darker red
flatly_blue       <- "#387ba8"  # Accent for lines
flatly_dark       <- "#2c3e50"  # blue of the theme
flatly_purple     <- "#A569BD"  # amethyst
flatly_gray_light <- "#ECF0F1"  # very light gray (bg)
flatly_gray_dark  <- "#7F8C8D"  # dark gray
flatly_border     <- "#BDC3C7"  # subtle borders

fixed_color <- light_red

dpi = 150
base_size <- 30
linewidth = 1.5
theme_set(theme_bw(base_size = base_size))

```


```{r, echo=FALSE, results='asis', warning=FALSE}
# sprtt_cit <- paste(format(citation("sprtt"), style = "text"), collapse = " ")
# sprtt_cit <- gsub("_(.+?)_", "<em>\\1</em>", sprtt_cit, perl = TRUE)
# 
# # Match the last angle-bracket URL by greedily capturing the prefix
# sprtt_cit <- sub("(?s)^(.*)&lt;\\s*(https?://[^&\\s]+)\\s*&gt;(.*)$",
#                  '\\1<br><a href="\\2" target="_blank" rel="noopener">\\2</a>\\3',
#                  sprtt_cit, perl = TRUE)
# sprtt_cit <- sub("(?s)^(.*)<\\s*(https?://[^>\\s]+)\\s*>(.*)$",
#                  '\\1<br><a href="\\2" target="_blank" rel="noopener">\\2</a>\\3',
#                  sprtt_cit, perl = TRUE)
# 
# cat(sprintf(
#   '<p class="author-line"><i>created by the</i> [sprtt](https://meikesteinhilber.github.io/sprtt/) <i>package</i><br>
#    <cite style="font-size:0.85em;color:#555;line-height:1.4;overflow-wrap:anywhere;">%s</cite></p>',
#   sprtt_cit
# ))
# 
sprtt_cit <- paste(format(citation("sprtt"), style = "text"), collapse = " ")
sprtt_cit <- gsub("_(.+?)_", "<em>\\1</em>", sprtt_cit, perl = TRUE)

# Remove URLs in angle brackets (both escaped & literal), plus a trailing dot if present
sprtt_cit <- gsub("&lt;\\s*https?://[^&\\s>]+\\s*&gt;\\.?","", sprtt_cit, perl=TRUE)
sprtt_cit <- gsub("<\\s*https?://[^>\\s]+\\s*>\\.?","", sprtt_cit, perl=TRUE)

# (Optional) also remove bare URLs that might appear without <...>
sprtt_cit <- gsub("\\bhttps?://[^\\s<>]+\\b\\.?","", sprtt_cit, perl=TRUE)

# Clean up leftover spaces/punctuation
sprtt_cit <- gsub("\\s+,", ",", sprtt_cit)
sprtt_cit <- gsub("\\s+\\.", ".", sprtt_cit)
sprtt_cit <- gsub("\\s{2,}", " ", sprtt_cit)
sprtt_cit <- trimws(sprtt_cit)

# cat(sprintf(
#   '<p class="author-line"><i>created by the</i> <a href="https://meikesteinhilber.github.io/sprtt/" target="_blank" rel="noopener">sprtt</a> <i>package</i><br>
#    <cite style="font-size:0.85em;color:#555;line-height:1.4;">%s</cite></p>',
#   sprtt_cit
# ))

cat(sprintf(
  '<p class="author-line"><i>created by the</i> <a href="https://meikesteinhilber.github.io/sprtt/" target="_blank" rel="noopener">sprtt</a> <i>package</i><br>
   <cite style="font-size:0.85em;color:#555;line-height:1.4;">%s</cite><br>
   <span style="font-size:0.85em;color:#555;margin-top:8px;display:inline-block;">Source code of the simulation: <a href="https://github.com/MeikeSteinhilber/sprtt_plan_sample_size" target="_blank" rel="noopener"><img src="https://img.shields.io/badge/GitHub-MeikeSteinhilber/sprtt__plan__sample__size-blue?logo=github" alt="GitHub" style="vertical-align:middle;"></a></span></p>',
  sprtt_cit
))


```
<!-- Source code of the simulation on GitHub: ![MeikeSteinhilber/sprtt_plan_sample_size](https://github.com/MeikeSteinhilber/sprtt_plan_sample_size) -->



<!-- <div class="toc-text-top"> -->
<!-- <p style="font-size: 1.5em; font-weight: bold; margin-top: 1em; margin-bottom: 0.5em; "> -->
<!-- $f_{exp}$ = `r pick_f_expected` -->
<!-- </p> -->
<!-- </div> -->

<br><br>
<br>

## Overview

<!-- ### Overview {#txt-overview .unlisted .unnumbered} -->

This report provides practical guidance for planning studies using **sequential ANOVA** (SPRT), where the final sample size is not fixed in advance.

Unlike traditional sample size analyses, this report cannot predict the exact sample size needed to reach a decision in a specific study.
Instead, it offers evidence-based recommendations derived from simulations, showing what sample sizes are typically required under various scenarios.
Thus, these benchmarks enable realistic resource planning.

::: {.alert .alert-warning}
The actual final sample size in any study may differ substantially from the following recommendations due to sampling variability and departures from the assumed effect structure.
:::


### Abbreviations

| Parameter | Description |
|:---|:-------------|
| $f_{\text{exp}}$ | Expected effect size (target effect) used for planning the sequential design |
| $f_{\text{true}}$ | True effect size in the simulated population |
| $\alpha$ | Type I error rate (probability of false positive) |
| $\beta$ | Type II error rate (probability of false negative) |
| $1-\beta$ | Target power, when no maximum sample size is used. The desired probability of detecting a true effect.  |
| $k$ | Number of groups  |
| $N$ | Sample size  |
| $N_{\text{max}}$ | The maximum sample size when the researcher stops collecting data independent of the test decision. This sample size is only reached in the worst-case scenario. |
| $N_{\text{median}}$ | The median sample size at which a decision is reached across simulations: 50% of simulated studies required fewer observations and 50% required more.|


### Content

**[Chapter I](#chapter-1)** illustrates the consequences of choosing a target effect size $f_{\text{exp}}$ across a range of possible true effects $f_{\text{true}}$. Although not specific to sequential designs, this perspective is essential for selecting a sensible expected effect size.

**[Chapter II](#chapter-2)** presents the core planning results for a sequential study, focusing on two benchmark cases: the true effect equals the specified value ($f_{\text{true}} = f_{\text{exp}}$) and there is no true effect ($f_{\text{true}} = 0$).

**[Chapter III](#chapter-3)** offers extended plots and tables to explore additional scenarios with more varied $f_{\text{true}}$ values.


### Technical details
<!-- ### Technical details {#txt-sim .unlisted .unnumbered} -->

The following results are based on a simulation with `r n_rep` iterations per condition.
Data were generated from a normal distribution, $\mathcal{N}(\mu,\sigma^2)$, under equal variances across groups and strictly balanced group sizes throughout data collection.

The analyses below use the following specifications:

 - $f_{exp}$ = `r pick_f_expected`

 - *k* groups = `r pick_k_groups`
 
 - $\alpha$ = 0.05
 
 - $\beta$ = `r pick_beta`
 
 - decision rate = `r pick_decision_rate*100`%
 
 - Simulation dataset: version `r data_version` (released: `r data_created`, [view on GitHub](https://github.com/MeikeSteinhilber/sprtt_plan_sample_size/releases/tag/`r data_version`))
  
 
<br><br><br>

## I. Influence of $f_{\text{exp}}$ {#chapter-1}

For this report, the effect size was specified as $f_{\text{exp}} = `r pick_f_expected`$.

Choosing $f_{\text{exp}}$ is critical because it determines:

  - which effect sizes can be detected with the target power and
  
  - how much data are typically required to reach a decision.


<!-- ### Power curve -->
<!-- ### Plot: Power curve {#power-curve .toc-only} -->

The plot below shows the power curve across possible true effects in the population, $f_{\text{true}}$.

The red color indicates that the power to detect a true effect of the corresponding size falls below the target power of `r pick_power`, suggesting the design is underpowered for such effect sizes.
The green color indicates that the power meets or exceeds the target power.

This plot serves an educational purpose: it illustrates why you should choose an effect size that reflects your actual research interest.
If detecting that effect requires resources beyond what's feasible, it's worth pausing to consider alternatives.

```{r, fig.height=10, fig.width=15, out.width="100%", fig.align="left"}
plot_cases <- df_all %>% 
  mutate(fill_group = factor(as.integer(f_simulated))) %>% 
  filter(f_expected == pick_f_expected,
         power == pick_power,
         k_groups == pick_k_groups
  ) %>% 
  filter(f_simulated != 0) %>% 
  group_by(
    f_simulated,
    f_expected,
    k_groups,
    alpha,
    power,
    distribution,
    sd,
    sample_ratio) %>% 
  mutate(
    decision_status_50 = if_else(decision != "continue sampling" & n <= q50_n, 1, 0),
    decision_status_75 = if_else(decision != "continue sampling" & n <= q75_n, 1, 0),
    decision_status_90 = if_else(decision != "continue sampling" & n <= q90_n, 1, 0),
    decision_status_95 = if_else(decision != "continue sampling" & n <= q95_n, 1, 0),
    decision_status_100 = if_else(decision != "continue sampling", 1, 0)
  ) %>% 
  summarize(
    error_rate = mean(decision_error),
    mean_n = mean(n),
    median_n = median(n),
    q90_n = quantile(n, 0.90),
  ) %>% mutate(
    simulated_power = if_else(f_simulated != 0, 1 - error_rate, NA),
    .after = power
  )

# 1) Build a dense x-grid and linearly interpolate simulated_power
rng  <- range(plot_cases$f_simulated)
grid <- tibble(f_simulated = seq(rng[1], rng[2], length.out = 1000))
grid$simulated_power <- with(plot_cases,
  approx(x = f_simulated, y = simulated_power, xout = grid$f_simulated,
         method = "linear", rule = 2)$y
)

# 2) Compute above/below and y-bounds for one ribbon layer
shade <- grid %>%
  mutate(
    above  = simulated_power >= pick_power,
    ymin   = pmin(simulated_power, pick_power),
    ymax   = pmax(simulated_power, pick_power)
  )

# 3) Plot: one ribbon (colored by above/below), then your original line/points
ggplot() +
  geom_ribbon(
    data = shade,
    aes(x = f_simulated, ymin = ymin, ymax = ymax, fill = above),
    alpha = 0.25, colour = NA
  ) +
  scale_fill_manual(values = c(`TRUE` = light_green, `FALSE` = light_red), guide = "none") +
  geom_line(data = plot_cases, aes(f_simulated, simulated_power),
            linewidth = 1.2, colour = flatly_dark) +
  geom_point(data = plot_cases, aes(f_simulated, simulated_power),
             size = 3.8, colour = flatly_blue, fill = flatly_blue,
             shape = 21, stroke = 1.2) +
  geom_hline(yintercept = pick_power, linetype = "dashed", colour = flatly_dark) +
  geom_segment(aes(x = pick_f_expected, xend = pick_f_expected, 
                 y = 0, yend = 1),  # Adjust yend to control where line stops
             linetype = "dashed", colour = flatly_dark) +
  annotate(
    "text",
    x = pick_f_expected,
    y = 1,  # Position text just above the line
    label = expression(italic(f)[exp]),
    hjust = 0.5,  # Center horizontally
    vjust = 0,    # Bottom of text at y position
    size = base_size * 0.25,
    colour = flatly_dark
  ) +
  scale_y_continuous(limits = c(0, 1.1), breaks = seq(0, 1, by = 0.25),
                     labels = scales::percent_format(accuracy = 1)) +
  scale_x_continuous(expand = expansion(mult = c(0.02, 0.02)),
                     breaks = unique(plot_cases$f_simulated)) +
  labs(
    title = "Simulated power across true effect sizes",
    x = expression(italic(f)[true]),
    y = "Simulated power",
    subtitle = glue::glue(
      "<i>f</i><sub>exp</sub> = {pick_f_expected}, ",
      "&alpha; = 0.05, &beta; = {pick_beta}, groups = {pick_k_groups}"
    )
  ) +
  theme_minimal(base_size = base_size) +
  theme(
    text = element_text(colour = flatly_dark),   # <-- sets ALL text to blue
    plot.title = element_text(face = "bold", size = base_size, hjust = 0, colour = flatly_dark),
    plot.subtitle = ggtext::element_markdown(size = base_size*0.7, colour = "grey40"),
    axis.title = element_text(face = "bold"),
    axis.text  = element_text(colour = "grey20"),
    panel.grid.minor = element_blank(),
    panel.grid.major.x = element_blank(),
    legend.position = "none",
    plot.title.position = "plot"
  )  -> plot


if (pick_f_expected == 0.10){
  plot + annotate(
    "text",
    x = min(plot_cases$f_simulated),
    y = pick_power,
    label = paste0(" target power = ", percent(pick_power)),
    hjust = 0,
    vjust = 2,
    size = base_size * 0.25,
    colour = flatly_dark
  )
    
} else if (pick_f_expected == 0.15) {
    plot + annotate(
    "text",
    x = max(plot_cases$f_simulated) - 0.07,
    y = pick_power,
    label = paste0(" target power = ", percent(pick_power)),
    hjust = 0,
    vjust = 2,
    size = base_size * 0.25,
    colour = flatly_dark
  )
} else{
    plot + annotate(
    "text",
    x = min(plot_cases$f_simulated),
    y = pick_power + 0.1,
    label = paste0(" target power = ", percent(pick_power)),
    hjust = 0,
    vjust = 2,
    size = base_size * 0.25,
    colour = flatly_dark
  )
}

```

::: {.alert .alert-info}
It is recommended to set $f_{\text{exp}}$ to the **minimum effect of interest**: power declines quickly for effects smaller than $f_{\text{exp}}$, whereas choosing a much larger $f_{\text{exp}}$ may shorten data collection but risks missing practically relevant effects just below that value.
:::



<br><br><br>

## II. Sample size recommendations {#chapter-2}



```{r}
# pick_f_simulated <- c(0, pick_f_expected)

pick_f_simulated = c(0,
                     pick_f_expected - 0.05,
                     pick_f_expected,
                     pick_f_expected + 0.05)

table_cases <- df_all %>% 
  mutate(
    f_simulated = factor(f_simulated, levels = sort(unique(pick_f_simulated))),
    fill_group = factor(as.integer(f_simulated))
  ) %>% 
  filter(
    f_expected == pick_f_expected,
    power == pick_power,
    k_groups == pick_k_groups,
    f_simulated %in% pick_f_simulated
  ) %>% 
  mutate(grand_median_n = as.integer(median(n))) %>% 
  group_by(
    f_simulated,
    f_expected,
    k_groups,
    alpha,
    power,
    distribution,
    sd,
    sample_ratio
  ) %>% 
  mutate(
    decision_status_50 = if_else(decision != "continue sampling" & n <= quantile(n, 0.50), 1, 0),
    decision_status_75 = if_else(decision != "continue sampling" & n <= quantile(n, 0.75), 1, 0),
    decision_status_90 = if_else(decision != "continue sampling" & n <= quantile(n, 0.90), 1, 0),
    decision_status_95 = if_else(decision != "continue sampling" & n <= quantile(n, 0.95), 1, 0),
    decision_status_100 = if_else(decision != "continue sampling", 1, 0)
  ) %>% 
  group_by(f_simulated, f_expected, power) %>%
  summarize(
    decision_error_rate = mean(decision_error),
    fix_n = as.integer(mean(fix_n)),
    mean_n = as.integer(round(mean(n))),
    median_n = as.integer(median(n)),
    grand_median_n = grand_median_n[1],
    min_n = as.integer(min(n)),
    max_n = max(n),
    decision_rate_100 = sum(decision_status_100) / n(),
    q25_n = as.integer(round(quantile(n, 0.25))),
    q50_n = as.integer(round(quantile(n, 0.50))),
    q75_n = as.integer(round(quantile(n, 0.75))),
    q80_n = as.integer(round(quantile(n, 0.80))),
    q85_n = as.integer(round(quantile(n, 0.85))),
    q90_n = as.integer(round(quantile(n, 0.90))),
    q95_n = as.integer(round(quantile(n, 0.95))),
    decision_rate_50 = sum(decision_status_50) / n(),
    decision_rate_75 = sum(decision_status_75) / n(),
    decision_rate_90 = sum(decision_status_90) / n(),
    decision_rate_95 = sum(decision_status_95) / n(),
    .groups = "drop"
  ) %>% 
  mutate(decision_error_rate = round(decision_error_rate*100,1)) %>% 
  select(-power) %>% 
  ungroup()

recommendation <- table_cases %>%
  filter(f_simulated %in% c(0, pick_f_expected)) %>%
  slice_max(order_by = .data[[pick_decision_rate_var]], n = 1, with_ties = FALSE)

n_mean <- round(mean(table_cases$mean_n), 0)
n_median_grand <- table_cases$grand_median_n[1]
n_max <- recommendation %>% pull(.data[[pick_decision_rate_var]])

simulated_power_data <- df_all %>% 
  mutate(
    f_simulated = factor(f_simulated, levels = sort(unique(pick_f_simulated)))
  ) %>%
  filter(
    f_expected == pick_f_expected,
    power == pick_power,
    k_groups == pick_k_groups,
    f_simulated %in% pick_f_simulated
  ) %>%
  group_by(f_simulated, f_expected, power) %>%
  summarize(
    decision_error_rate_nmax = mean(if_else(decision_error == 1 & n <= n_max, 1, 0)),
    sd_error_n = sd(n) / sqrt(length(n)),
    simulated_power_nmax = mean(if_else(f_simulated != "0" & decision == "accept H1" & n <= n_max, 1, 0)),
    .groups = "drop"
  ) %>%
  mutate(decision_error_rate_nmax = round(decision_error_rate_nmax*100,1)) %>% 
  mutate(simulated_power_nmax = if_else(simulated_power_nmax == 0, "-", as.character(round(simulated_power_nmax*100,1)))) %>%
  select(-power)

table_cases <- table_cases %>%
  left_join(simulated_power_data, by = c("f_simulated", "f_expected")) %>%
  select(
    f_simulated,
    decision_error_rate,
    decision_error_rate_nmax,
    simulated_power_nmax,
    mean_n,
    median_n,
    grand_median_n,
    min_n,
    max_n,
    q75_n,
    q80_n,
    q85_n,
    q90_n,
    q95_n,
    fix_n
  )
power_nmax <- table_cases$simulated_power_nmax[table_cases$f_simulated == pick_f_expected]
```


In the following plot you can see a visualization of the simulated samples, where the effect in the data is either as expected or zero. For a extended version of these results see **[Chapter III](#chapter-3)**.

This plot visualizes the distribution of sample sizes required to reach a decision in sequential and fixed testing, comparing two scenarios: when the true effect matches your expectation ($f_{\text{true}} =$ `r pick_f_expected`) versus when there's no effect ($f_{\text{true}} =$ 0).
The violin plots reveal substantial variability, while many sequential studies terminate early (note the low median sample size), some require substantially more data.
The red diamonds show the fixed sample size from traditional a priori power analysis as a point or reference.

In most cases, the final sample size will be close to $N_{\text{median}}$, however for the worst case one should have the resources for $N_{\text{max}}$ to have a `r pick_decision_rate*100`% chance to reach a decision.


<!-- ### Plot: Sample size requirement {#plot .toc-only} -->


```{r, fig.height=10, fig.width=15, out.width="100%", fig.align="left"}
pick_f_simulated <- c(0, pick_f_expected)

# titles
title_html <- glue("<i>f</i><sub>exp</sub> = {pick_f_expected}")
caption_html <- glue(
  # "<i>f</i><sub>true</sub> = {pick_f_expected},<br>",
  "&beta; = {pick_beta}, &alpha; = 0.05, decision rate = {pick_decision_rate*100}%, groups = {pick_k_groups}"
)

df_all %>% 
  mutate(f_simulated = factor(f_simulated, levels = sort(unique(pick_f_simulated))),
         fill_group = factor(as.integer(f_simulated))) %>% 
  filter(f_expected == pick_f_expected,
         power == pick_power,
         k_groups == pick_k_groups
  ) %>% 
  filter(f_simulated %in% pick_f_simulated) %>%
  {. ->> cases} %>% 
  ggplot(aes(x = f_simulated, y = n)) +
  geom_violin(trim = FALSE, fill = light_green,
              scale = "count",
              linewidth = 0.3,
              alpha = 0.25) +
  coord_cartesian(ylim = c(0, cases$fix_n[1] * 2.5)) +
  geom_hline(yintercept = n_median_grand, linetype = "dashed", color = flatly_dark, linewidth = 0.8) +
  geom_hline(yintercept = n_max, linetype = "dashed", color = flatly_dark, linewidth = 0.8) +
  annotate("text", x = 2.45, y = n_median_grand + 5, 
           label = "bar(italic(N))[median]", parse = TRUE, vjust = -0.1,
           size = base_size * 0.25, color = flatly_dark) +
  annotate("text", x = 2.45, y = n_max + 5, 
           label = "italic(N)[max]", parse = TRUE, vjust = -0.1,
           size = base_size * 0.25, color = flatly_dark) +
  geom_point(aes(x = f_simulated, y = fix_n, color = "Fixed sample size"),
             position = position_dodge(0.9),
             shape = 18,
             size = 11) +
  geom_point(aes(x = f_simulated, y = median_n, color = "Median sample size"),
             position = position_dodge(0.9),
             shape = 15,
             size = 8) + 
  
  labs(
    x = "<i>f</i><sub>true</sub>",
    y = "<i>N</i>",
    title = "Required sample sizes to reach a decision",
    subtitle = glue::glue(
      "<i>f</i><sub>exp</sub> = {pick_f_expected}, ",
      "&alpha; = 0.05, &beta; = {pick_beta}, groups = {pick_k_groups}"
    )) +
  scale_color_manual(
    name   = "Legend",
    values = c("Fixed sample size" = fixed_color,
               "Median sample size"  = flatly_dark),
    labels = c("Fixed sample size" = "a priori power analysis",
               "Median sample size"  = "median sample size")
  ) +
  guides(
    color = guide_legend(override.aes = list(size = 5))
  ) +
  theme_minimal(base_size = base_size) +
  theme(text = element_text(colour = flatly_dark),
        plot.title = element_text(face = "bold", size = base_size,
                                  hjust = 0, colour = flatly_dark),
        plot.subtitle = ggtext::element_markdown(size = base_size*0.7, colour = "grey40"),
        plot.caption = element_markdown(),
        axis.title.x = element_markdown(),
        axis.title.y = element_markdown(angle = 180, vjust = 0.5),
        panel.grid.minor.y = element_blank(),
        panel.grid.major.x = element_blank(),
        legend.position = "right",
        legend.text = element_text(size = base_size * 0.7)
        ) -> p_violin
p_violin
```

<!-- ### Summary table {#table .toc-only} -->

::: {.alert .alert-info}
In short, to achieve a decision rate of `r pick_decision_rate*100`%, 
<span style="color: #2c3e50; font-weight: bold;">
plan for a maximum sample size of $N_{\text{max}} = `r n_max`$.
</span><br>
While you must plan for this worst-case maximum, 
<span style="color: #2c3e50; font-weight: bold;">the median sample size is only $\bar{N}_{\text{median}} =$ `r n_median_grand`</span>, meaning half of the time studies will terminate before reaching $N_{\text{median}}$.
Using $N_{\text{max}} = `r n_max`$ will yield a power of $\text{Power}_{N_{\text{max}}}=$ `r power_nmax`%, accounting for both Type II errors and non-decisions.
:::

Extended numeric details are shown in the table below.
Here you can also see the influence of the required sample sizes for different decision rates.

```{r}
table_cases %>%
  filter(f_simulated %in% pick_f_simulated) %>%
  select(-grand_median_n, -simulated_power_nmax, -decision_error_rate_nmax) %>% 
  ungroup() %>% 
  gt() %>%
  tab_header(
    title = gt::html("<i>Summary table</i>"),
    subtitle = gt::html(glue("<i>f</i><sub>exp</sub> = {pick_f_expected}, &alpha; = 0.05, &beta; = {pick_beta}, decision rate = {pick_decision_rate*100}%, groups = {pick_k_groups}"))
  ) %>%
  # center all data columns
  cols_align(align = "center", columns = everything()) %>%
  # center column labels and the heading
  tab_options(
    heading.align = "left"
  ) %>% 
  cols_label(
    f_simulated         = gt::html("<i>f</i><sub>true</sub>"),
    # f_expected          = gt::html("<i>f</i><sub>exp</sub>"),
    decision_error_rate = "Error rate %",
    # decision_error_rate_nmax = gt::html("Error rate<sub>N<sub>max</sub></sub>%"),
    # simulated_power_nmax     = gt::html("Power<sub>N<sub>max</sub></sub>%"),
    fix_n               = html("Reference <i>N</i>:<br>fixed design"),
    mean_n              = "Mean",
    median_n            = "Median",
    min_n               = "Min",
    max_n               = "Max",
    # decision_rate_90    = "90% decision rate",
    q75_n               = "75%",
    q80_n               = "80%",
    q85_n               = "85%",
    q90_n               = "90%",
    q95_n               = "95%"
    # decision_rate_95    = "95% decision rate"
  ) %>%
   # high-level headers (spanners)
  tab_spanner(
    label   = "Sample size (N)",
    columns = c(mean_n, median_n, min_n, max_n)
  ) %>%
  tab_spanner(
    label   = html("<i>N</i> to reach decision rate (%)"),
    columns = c(q75_n, q80_n, q85_n, q90_n, q95_n)
  ) %>%
  # tab_spanner(
  #   label   = html("Reference fixed design N"),
  #   columns = c(fix_n)
  # ) %>%
  data_color(
    columns = median_n,
    palette = grey_blue
  ) %>%
  data_color(
    columns = !!pick_decision_rate_var,
    palette = grey_blue
  ) %>%
  data_color(
    columns = fix_n,
    palette = "darkgrey"
  ) %>%
  fmt_number(
    columns = where(is.double),
    decimals = 2
  ) %>% 
   tab_footnote(
    footnote = html(glue("simulated error rates when sampling continous until a decision is reached")),
    locations = cells_column_labels(columns = decision_error_rate)
  ) %>% 
  #  tab_footnote(
  #   footnote = html(glue("simulated error rates with <i>N</i><sub>max</sub> = {n_max}")),
  #   locations = cells_column_labels(columns = decision_error_rate_nmax)
  # ) %>% 
  #  tab_footnote(
  #   footnote = html(glue("simulated power to detect an effect with <i>N</i><sub>max</sub> = {n_max}")),
  #   locations = cells_column_labels(columns = simulated_power_nmax)
  # ) %>% 
   tab_footnote(
    footnote = html(glue("grand median: <i>N&#x0305;</i><sub>median</sub> = {n_median_grand}")),
    locations = cells_column_labels(columns = median_n)
  ) %>% 
   tab_footnote(
    footnote = html(glue("recommended <i>N</i><sub>max</sub> = {n_max}")),
    locations = cells_column_labels(columns = !!pick_decision_rate_var)
  )
```

For SPRTs, the power is $1-\beta$ when no maximum sample size is imposed, which is the theoretical setup for an SPRT.
However, for practical resource planning, we introduce $N_{\text{max}}$, which decreases the achievable power.
Therefore, we define $\text{Power}_{N_{\text{max}}}$ as the power calculated from simulated data, accounting for both Type II errors and non-decisions.

We generally recommend maintaining a low $\beta$ and accepting a lower decision rate to manage resource constraints, rather than relaxing the Type II error rate.
Low error rates combined with a lower decision rate ensure that when the procedure reaches a decision, the result is trustworthy due to strict long-run error control.
Importantly, non-decisions preserve flexibility: as long as no decision has been reached, data collection can continue if additional resources become available.

<br><br><br>

## III. Extended insights {#chapter-3}

### Plot: Sample size and varing $f_{\text{true}}$ {#plot .toc-only}

```{r}
pick_f_simulated = c(0,
                     pick_f_expected - 0.05,
                     pick_f_expected,
                     pick_f_expected + 0.05)
```


<!-- ### Required sample sizes {#extended-violin .toc-only} -->

This violin plot shows the distribution of required sample sizes $n$ across simulations for different assumed true effects $f_{\text{true}}$.
The violin colors indicate whether the test controls the error rates reliably at a given setting.

  - **Green** marks settings where $f_{\text{true}} \geq f_{\text{exp}}$ (or $f_{\text{true}} = 0$), at which the test reliably controls the error rates.

  - **Red** marks settings where the $f_{\text{true}} < f_{\text{exp}}$ (and $f_{\text{true}} \neq 0$), at which the test no longer reliably controls the error rates and such effects would often go undetected. This is analogous to an underpowered fixed design: the test was not designed for effects this small, and it is not clear what the correct decision would be in such cases. (If `f_expected = 0.1` is selected, no red violin is shown in the plot below as it is outside of the simulated scope.)

The red diamond marks the fixed sample size from an a priori power analysis, and the blue square shows the mean sample size observed across simulations.


```{r, fig.height=10, fig.width=15, out.width="100%", fig.align="left"}
# violin_colors <- c(light_blue, hell_rosa, light_blue, light_blue)

# title_html <- glue("<i>f</i><sub>exp</sub> = {pick_f_expected}")
caption_html <- glue(
  "&beta; = {pick_beta}, &alpha; = 0.05,<br>",
  "groups = {pick_k_groups}"
)


p_violin <- df_all %>%
  mutate(
    f_simulated = factor(f_simulated, levels = sort(unique(pick_f_simulated))),
    fill_label = ifelse(f_simulated == pick_f_expected - 0.05,
                        "error rates not controlled",
                        "error rates controlled")
  ) %>%
  filter(
    f_expected == pick_f_expected,
    power == pick_power,
    k_groups == pick_k_groups
  ) %>%
  filter(f_simulated %in% pick_f_simulated) %>%
  ggplot(aes(x = as.factor(f_simulated), y = n, fill = fill_label)) +
  
  # violins (fill legend only)
  geom_violin(
    trim = FALSE,
    scale = "count",
    # draw_quantiles = c(0.25, 0.50, 0.75),
    # linewidth = 0.1,
    color = "lightgrey",
    alpha = 0.25
  ) +
  scale_fill_manual(
    name = "Power",
    values = c("error rates controlled" = light_green,
               "error rates not controlled" = light_red)
  ) +
  
  # red diamond = fixed sample size
  geom_point(
    aes(y = fix_n, color = "Fixed sample size"),
    position = position_dodge(0.9),
    shape = 18,     # diamond, no border
    size  = 7,
    stroke = 0      # ensures no black outline
  ) +
  
  # blue square = mean sample size
  geom_point(
    aes(y = median_n, color = "Median sample size"),
    position = position_dodge(0.9),
    shape = 15,     # square, no border
    size  = 6,
    stroke = 0      # ensures no black outline
  ) +
scale_color_manual(
  name   = "Legend",
  values = c("Fixed sample size" = fixed_color,
             "Median sample size"  = flatly_dark),
  labels = c("Fixed sample size" = "a priori power analysis",
             "Median sample size"  = "median sample size")
) +
  labs(
    x = "<i>f</i><sub>true</sub>",
    y = "<i>N</i>",
    title = "Required sample sizes to reach a decision",
    subtitle = glue::glue(
      "<i>f</i><sub>exp</sub> = {pick_f_expected}, ",
      "&alpha; = 0.05, &beta; = {pick_beta}, groups = {pick_k_groups}"
    )
  ) +
  
  # clean legend appearance
  guides(
    fill  = guide_legend(override.aes = list(shape = NA)),  # patches only
    color = guide_legend(override.aes = list(stroke = 0, size = 5)) # points only
  ) +
  theme_minimal(base_size = base_size) +
  theme(
    text = element_text(colour = flatly_dark),
    legend.key = element_blank(),  # no background box around keys
    legend.text = element_text(size = base_size * 0.7),
    # plot.title = element_markdown(),
    # plot.subtitle = element_markdown(),
    plot.caption = element_markdown(),
    axis.title.x = element_markdown(),
    axis.title.y = element_markdown(angle = 180, vjust = 0.5),
    # panel.grid.minor.y = element_blank(),
    panel.grid.minor.x = element_blank(),
    panel.grid.major.x = element_blank(),
    plot.title = element_text(face = "bold", size = base_size, hjust = 0, colour = flatly_dark),
    plot.subtitle = ggtext::element_markdown(size = base_size*0.7, colour = "grey40")
  ) 

p_violin


```


<!-- ### Summary table {#extended-table .toc-only} -->

Extended numeric details are shown in the table below.
Here you can also see the influence of the required sample sizes for different decision rates.

```{r}
num_cols <- c("f_simulated", "decision_error_rate")

table_cases %>%
  select(-grand_median_n, -simulated_power_nmax, -decision_error_rate_nmax) %>% 
  ungroup() %>% 
  gt() %>%
  tab_header(
    title = gt::html("<i>Summary table</i>"),
    subtitle = gt::html(glue("<i>f</i><sub>exp</sub> = {pick_f_expected}, &alpha; = 0.05, &beta; = {pick_beta}, decision rate = {pick_decision_rate*100}%, groups = {pick_k_groups}"))
  ) %>%
  cols_align(align = "center", columns = everything()) %>%
  tab_options(
    heading.align = "left"
  ) %>% 
  cols_label(
    f_simulated         = gt::html("<i>f</i><sub>true</sub>"),
    decision_error_rate = "Error rate %",
    fix_n               = html("Reference <i>N</i>:<br>fixed design"),
    mean_n              = "Mean",
    median_n            = "Median",
    min_n               = "Min",
    max_n               = "Max",
    q75_n               = "75%",
    q80_n               = "80%",
    q85_n               = "85%",
    q90_n               = "90%",
    q95_n               = "95%"
  ) %>%
  tab_spanner(
    label   = "Sample size (N)",
    columns = c(mean_n, median_n, min_n, max_n)
  ) %>%
  tab_spanner(
    label   = html("<i>N</i> to reach decision rate (%)"),
    columns = c(q75_n, q80_n, q85_n, q90_n, q95_n)
  ) %>%
  data_color(
    columns = all_of(num_cols),
    rows    = as.numeric(as.character(f_simulated)) < pick_f_expected &
              as.numeric(as.character(f_simulated)) != 0,
    colors  = light_red, 
    alpha = 0.25,
    autocolor_text = FALSE
  ) %>% 
  data_color(
    columns = median_n,
    palette = grey_blue
  ) %>%
  data_color(
    columns = !!pick_decision_rate_var,
    palette = grey_blue
  ) %>%
  data_color(
    columns = fix_n,
    palette = "darkgrey"
  ) %>%
  fmt_number(
    columns = where(is.double),
    decimals = 2
  ) %>% 
  tab_footnote(
    footnote = html(glue("simulated error rates when sampling continuous until a decision is reached")),
    locations = cells_column_labels(columns = decision_error_rate)
  )
```

<!-- ### Sequential vs fixed design -->

<br><br><br><br>

### Plot: Sample size and varing $f_{\text{exp}}$ {#sample-size-diff .toc-only}

The next plot below visualizes the required sample sizes, when the true value is set and one uses different specification of $f_{\text{exp}}$.
Here you can see how the specification of $f_{\text{exp}}$ influences the sample sizes also in direct comparison to a fixed sample plan.
Sequential designs maintain efficiency despite underestimating $f_{\text{true}}$, whereas fixed designs are locked into their initial specification -- a key advantage demonstrated in this plot

```{r, fig.height=10, fig.width=15, out.width="100%", fig.align="left"}
pick_f_simulated = pick_f_expected

# titles
title_html <- glue("<i>f</i><sub>true</sub> = {pick_f_simulated}")
subtitle_html <- glue(
  # "<i>f</i><sub>true</sub> = {pick_f_expected},<br>",
  "&beta; = {pick_beta}, &alpha; = 0.05,<br> groups = {pick_k_groups}"
)

df_all %>% 
  mutate(f_simulated = factor(f_simulated, levels = sort(unique(pick_f_simulated))),
         fill_group = factor(as.integer(f_simulated))
         ) %>% 
  filter(f_simulated == pick_f_simulated,
         f_expected <= pick_f_simulated,
         power == pick_power,
         k_groups == pick_k_groups
  ) %>%
  mutate(violin_color = if_else(f_expected < pick_f_expected & 
                                f_expected != 0, 
                                light_green, light_green),
         f_expected = factor(f_expected)) %>% 
  group_by(f_expected) %>% 
  {. ->> cases} %>% 
  ggplot(aes(x = f_expected, y = n)) +
  geom_violin(aes(fill = violin_color), trim = FALSE,
              scale = "count",
              linewidth = 0.3,
              alpha = 0.25) +
  scale_fill_identity() +
  geom_point(aes(x = f_expected, y = fix_n, color = "Fixed sample size"),
             position = position_dodge(0.9),
             shape = 18,
             size = 9) +
  geom_point(aes(x = f_expected, y = median_n, color = "Median sample size"),
             position = position_dodge(0.9),
             shape = 15,
             size = 6) + 
  labs(
    x = "<i>f</i><sub>exp</sub>",
    y = "<i>N</i>",
    title = "Required sample sizes to reach a decision",
    subtitle = glue::glue(
      "<i>f</i><sub>true</sub> = {pick_f_simulated}, ",
      "&alpha; = 0.05, &beta; = {pick_beta}, groups = {pick_k_groups}"
    )) +
  scale_color_manual(
    name   = "Legend",
    values = c("Fixed sample size" = fixed_color,
               "Median sample size"  = flatly_dark),
    labels = c("Fixed sample size" = "a priori power analysis",
               "Median sample size"  = "median sample size")
  ) +
  guides(
    color = guide_legend(override.aes = list(size = 5))
  ) +
  theme_minimal(base_size = base_size) +
  theme(text = element_text(colour = flatly_dark),
        plot.title = element_text(face = "bold", size = base_size, hjust = 0,
                                colour = flatly_dark),
        plot.subtitle = ggtext::element_markdown(size = base_size*0.7, colour = "grey40"),
        plot.caption = element_markdown(),
        axis.title.x = element_markdown(),
        axis.title.y = element_markdown(angle = 180, vjust = 0.5),
        panel.grid.minor.y = element_blank(),
        panel.grid.major.x = element_blank(),
        legend.position = "right",
        legend.text = element_text(size = base_size * 0.7)
        ) -> p_violin 
p_violin
```


<br><br><br><br>

### Plot: Cumulative decisions {#extended-cumulative .toc-only}

This plot illustrates how decision rate accumulates with increasing sample size for different true effects.
Larger effects reach the 90% decision threshold much earlier, while smaller effects require substantially more data.
The vertical line marks the fixed sample size.
Notice how efficiency varies depending on whether the true effect is larger, equal to, or smaller than your expected effect.

```{r, fig.height=10, fig.width=15, out.width="100%", fig.align="left"}
df_cum <- df_all %>%
  filter(f_expected == pick_f_expected,
         power      == pick_power,
         k_groups   == pick_k_groups) %>% 
  filter(decision != "continue sampling") %>%
  transmute(f_simulated, fix_n, n_decision = n) %>% 
  group_by(f_simulated) %>%
  arrange(n_decision, .by_group = TRUE) %>%
  mutate(cum_prop = row_number() / n()) %>%
  ungroup() %>%
  mutate(
    is_expected = f_simulated == pick_f_expected,
    f_label = ifelse(is_expected, 
                     paste0(f_simulated, " (=expected)"), 
                     as.character(f_simulated))
  )

# Find the minimum n that achieves at least 90% decided
cut90 <- df_cum %>%
  group_by(f_simulated) %>%
  summarise(n_at_90 = min(n_decision[cum_prop >= 0.90]), .groups = "drop")

# Get Set2 palette colors
set2_colors <- RColorBrewer::brewer.pal(n = length(unique(df_cum$f_label)), name = "Set2")

# Create named vector with flatly_dark for expected line
color_values <- setNames(set2_colors, sort(unique(df_cum$f_label)))
expected_label <- paste0(pick_f_expected, " (= expected)")
color_values[expected_label] <- flatly_dark

# Plot with unified legend
ggplot(df_cum, aes(x = n_decision, y = cum_prop, 
                   colour = f_label, 
                   group = f_simulated,
                   linewidth = is_expected,
                   alpha = is_expected)) +
  geom_step() +
  scale_linewidth_manual(values = c("FALSE" = 0.9, "TRUE" = 1.8), guide = "none") +
  scale_alpha_manual(values = c("FALSE" = 0.5, "TRUE" = 1.0), guide = "none") +
  # geom_hline(yintercept = 0.90, linetype = "dotted", colour = flatly_dark) +
  geom_vline(xintercept = unique(df_cum$fix_n), color = flatly_dark, linetype = "dashed") +
  annotate(
    "text",
    x = unique(df_cum$fix_n) + (unique(df_cum$fix_n) * 0.07),
    y = 0.05,
    label = "fixed sample size",
    angle = 90,
    hjust = 0.25,
    vjust = 1.3,
    size = base_size * 0.25,
    colour = flatly_dark
  ) +
  scale_y_continuous(
    limits = c(0, 1),
    breaks = c(0, 0.25, 0.5, 0.75, 0.9, 1),
    labels = percent_format(accuracy = 1)
  ) +
  scale_color_manual(values = color_values,
                     name = expression(italic(f)[true])) +
  labs(
    x = "<i>N</i>",
    y = "",
    title = "Cumulative decision proportions across simulated\neffect sizes",
    subtitle = glue::glue(
      "<i>f</i><sub>exp</sub> = {pick_f_expected}, ",
      "&alpha; = 0.05, &beta; = {pick_beta}, groups = {pick_k_groups}"
    )
  ) +
  theme_minimal(base_size = base_size) +
  theme(
    plot.title = element_text(face = "bold", size = base_size, hjust = 0, colour = flatly_dark),
    plot.subtitle = ggtext::element_markdown(size = base_size*0.7, colour = "grey40"),
    panel.grid.minor = element_blank(),
    panel.grid.major.x = element_blank(),
    legend.title = element_text(face = "bold"),
    axis.title = element_text(face = "bold"),
    plot.title.position = "plot",
    axis.title.x = ggtext::element_markdown()
  )
```

---
title: "Introduction to SPRTs"
author: "Meike Snijder-Steinhilber"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 4
description: >
  This vignette describes SPRTs in general.
vignette: >
  %\VignetteIndexEntry{SPRTs}
  %\VignetteEncoding{UTF-8}{inputenc}
  %\VignetteEngine{knitr::rmarkdown}
bibliography: references.bib
csl: "apa.csl"
---


The `sprtt` package is a **s**equential **p**robability **r**atio **t**ests **t**oolbox (**sprtt**).
This vignette describes the theoretical background of these tests.

If you are interested in a workflow to use the `sprtt` package, see the vignette `vignette("workflow_sprtt")`.
For a simple use case of a sequential *t*-test see the vignette `vignette("use_case")`.

### The Sequential Testing Principle

Sequential Probability Ratio Tests (SPRTs) fundamentally differ from fixed-sample designs by continuously evaluating evidence as data accumulates [@wald1945].
After collecting each data point (or batch of data points), the test leads to one of three outcomes:

- **Continue sampling**: Evidence remains inconclusive

- **Stop and accept $H_0$** (no effect): Sufficient evidence accumulated against an effect

- **Stop and accept $H_1$** (effect): Sufficient evidence accumulated for an effect  

This approach allows researchers to stop data collection as soon as sufficient evidence has been obtained, leading to substantial efficiency gains compared to fixed-sample designs.

### The Likelihood Ratio and Decision Boundaries

At the core of SPRTs is the *likelihood ratio* $\text{LR}_n$, which quantifies the relative evidence for $H_1$ versus $H_0$ after $n$ observations:

$$\text{LR}_n = \frac{\text{L}_n(H_1)}{\text{L}_n(H_0)} = \frac{f(\text{data}_n \mid H_1)}{f(\text{data}_n \mid H_0)}$$

If you are unfamiliar with the concept of likelihood, we recommend the paper by @etz2018.

The SPRT compares the likelihood ratio to two boundaries ($A$ and $B$), and the following rules apply:

| Condition | Data collection | Hypothesis |
|:---------------:|----------------------------:|----------------------------:|
| $LR_m \leq B$  | Stop data collection | Accept $H_0$ and reject $H_1$ |
| $B < LR_m < A$ | Continue sampling    | No decision is made (yet)        |
| $LR_m \geq A$  | Stop data collection | Accept $H_1$ and reject $H_0$ |


<!-- - **Continue sampling**: If $B < \text{LR}_n < A$, collect more data -->

<!-- - **Upper boundary** $A$: If $\text{LR}_n \geq A$, stop and accept $H_1$ -->

<!-- - **Lower boundary** $B$: If $\text{LR}_n \leq B$, stop and accept $H_0$ -->

These boundaries are determined by the desired Type I ($\alpha$) and Type II ($\beta$) error rates:

$$A = \frac{1-\beta}{\alpha} \quad \text{and} \quad B = \frac{\beta}{1-\alpha}$$

In practice, it is often more convenient to work with the log-likelihood ratio $\text{LLR}_n = \log(\text{LR}_n)$.
The logarithm transforms products of likelihoods into sums, which improves numerical stability (avoiding very small or very large numbers) and simplifies computation.

The corresponding log-boundaries are:

$$\text{LLR}_n \geq \log(A) \rightarrow \text{accept } H_1$$

$$\text{LLR}_n \leq \log(B) \rightarrow \text{accept } H_0$$

The `sprtt` package uses the $\text{LLR}_n$ for the internal calculations.

### Random Sample Size

Unlike fixed-sample designs where $N$ is predetermined, the sample size in SPRTs is a random variable.
You don't know beforehand when the test will stop.
When the SPRT will stop depends on:

- Random variation in the observed data
- The true effect size in the population
- The effect size specified under $H_1$
- The specified error rates ($\alpha$, $\beta$)

When the true effect matches $H_1$ (or $H_0$), the test tends to stop quickly.
When the truth lies between the hypotheses, stopping may take longer.
This randomness is a feature, not a bug -- it's what enables the efficiency gains.

### Why SPRTs Always Stop

A crucial theoretical property of SPRTs is that they are *guaranteed to terminate* with probability 1 under both $H_0$ and $H_1$ [@wald1947].
This means that if you continue collecting data, the likelihood ratio will eventually cross one of the boundaries -- you won't collect data indefinitely.

The mathematical proof relies on the law of large numbers and properties of random walks.
However, while termination is guaranteed asymptotically, practical constraints (budget, time) may require setting a maximum sample size $N_{\text{max}}$ based on available resources.
The `plan_sample_size()` function helps determine the required $N_{\text{max}}$ for a given design, ensuring a desired decision rate (e.g., 80%) is achieved, that is, the test reaches a decision before exhausting resources in at least 80% of cases.

### Efficiency

SPRTs achieve remarkable efficiency compared to fixed-sample designs [@wald1945].
On average, SPRTs require approximately *58% fewer observations* to reach the same decision with the same error rates [@steinhilber2024].

Additionally, SPRTs are especially efficient compare to fixed-sample designs, when the expected effect (or effect size of interest) is small and when the expected effect is smaller than the true effect in the data [@steinhilber2024].


### The Bias-Efficiency Tradeoff

While SPRTs are highly efficient, they come with an important caveat: effect size estimates are conditionally biased [@schnuerch2020; @steinhilber2024].

**Small samples** (early stops): lead to *overestimation* of effect sizes in single studies.

**Large samples** (late stops): lead to *underestimation* of effect sizes in single studies.

However, across multiple studies, the weighted average of effect size estimates is close to the true population parameter.

This means that effect size estimates from individual SPRT studies should be interpreted with caution.
It is important to note that conditional bias of sequential samples is not specific to SPRTs but a general phenomenon of sequential testing procedures [@nardini2013; @fan2004; @whitehead1986].

**Practical implications:**

- Use SPRTs for hypothesis testing (accept/reject decisions)
- Be cautious when interpreting point estimates from individual sequential studies
- For precise effect size estimation, consider fixed-sample designs with large enough samples or bias-corrected estimators, where available

### Practical Considerations

While SPRTs theoretically continue until a boundary is crossed, practical constraints require planning.
The `sprtt` package provides tools to explore these tradeoffs through the `plan_sample_size()` function, helping you find designs that balance efficiency with feasibility.

### References

---
title: "Report: Sample size planning for SPRTs"
author: "created by the sprtt package"
output:
  html_document:
    theme: flatly
    toc: true
    toc_float: true
    number_sections: false
    code_folding: hide
params:
  # report controls
  n_rep: 100
  pick_f_expected: 0.25
  pick_power: 0.8
  pick_k_groups: 3
  # data wiring
  data_path: "data/sprt_tool_samples_100.RData"
---



```{r, include=FALSE}
knitr::opts_chunk$set(warning = FALSE, message = FALSE)
```


```{r, include=FALSE}
n_rep = params$n_rep
pick_f_expected = params$pick_f_expected
pick_power = params$pick_power
pick_k_groups = params$pick_k_groups

library(dplyr)
# library(tidyr)
library(kableExtra)
# library(gridExtra)
library(glue)
# library(here)
library(ggplot2)
library(knitr)
# library(kableExtra)
# library(gt)
library(ggtext)
library(scales)



# library(latex2exp)
```

```{r, include=FALSE}
load(params$data_path)
# load(glue("meta_data/sprt_tool_samples_{n_rep}.RData"))

# variables -------------------------------------------------------------------
digits <- 2

palette <- "RdYlBu"  #"Dark2"
light_green <- "#a5d1c7"
green <- "#69b3a2"
dark_green <- "#548f81"
dark2_green <- "#104949"
light_blue <- "#cce5ff" 
grey_blue <- "#bccfe5"
blue <- "#2c7bb6"  #"#404080"
dark_blue <- "#1b3c70"
magenta <- "#e61be5"
rosa <- "#ff0088"
hell_rosa <- "#ffa1c2"
light_red <- "#e72b2e"
red <- "#e72b2e"
# red <- 	"#d7191c" #"#ca0020" "#CD2626" "#8B0000"
dark_red <- "#8b1012"
black_red <- "#4b1919" # "#651718"

light_orange <- "#fffcf9"
orange <- "#fdae61"
dark_orange <- "#fc8715"
yellow <- "#ffffbf"

dpi = 150
base_size <- 20
linewidth = 1.5
theme_set(theme_bw(base_size = base_size))
```


<br><br>
These results are based on a simulation with `r n_rep` iterations per condition.

See XXX for the simulation code.

<br>

## Core Insights

### Power curve

### Influence of f expected

### violin plot

```{r, fig.height=8, fig.width=5, out.width="30%", fig.align="left"}
# pick_f_simulated = c(0, pick_f_expected)
pick_f_simulated = pick_f_expected

# titles
title_html <- glue("<i>f</i><sub>exp</sub> = {pick_f_expected}")
subtitle_html <- glue(
  "<i>f</i><sub>true</sub> = {pick_f_expected},<br>",
  "1 - &beta; = {pick_power}, &alpha; = 0.05,<br> groups = {pick_k_groups}"
)

df_all %>% 
  mutate(f_simulated = factor(f_simulated, levels = sort(unique(pick_f_simulated))),
         fill_group = factor(as.integer(f_simulated))) %>% 
  filter(f_expected == pick_f_expected,
         power == pick_power,
         k_groups == pick_k_groups
  ) %>% 
  filter(f_simulated %in% pick_f_simulated) %>%
  {. ->> cases} %>% 
  ggplot(aes(x = as.factor(pick_k_groups), y = n)) +
  geom_violin(trim = FALSE, fill = light_blue,
              scale = "count",
              #draw_quantiles = c(0.25, 0.50, 0.75),
              linewidth = 0.3) +
  # scale_fill_manual(values = light_blue) +
  geom_point(aes(x = as.factor(pick_k_groups), y = fix_n),
             position = position_dodge(0.9),
             shape = 18,
             color = red,
             size = 11) +
  geom_point(aes(x = as.factor(pick_k_groups), y = mean(mean_n)),
             position = position_dodge(0.9),
             shape = 15,
             color = dark_blue,
             size = 8) + 
  labs(
    x = "<i>f</i><sub>true</sub>",
    title = title_html,
    caption = subtitle_html  ) +
  theme(plot.title = element_markdown(),
        plot.caption = element_markdown(),
        axis.title.x = element_markdown(),
        panel.grid.minor.y = element_blank()
        ) -> p_violin 
p_violin
```

```{r}
table_cases <- df_all %>%
  mutate(f_simulated = factor(f_simulated, levels = sort(unique(pick_f_simulated))),
         fill_group = factor(as.integer(f_simulated))) %>%
  filter(f_expected == pick_f_expected,
         power == pick_power,
         k_groups == pick_k_groups
  ) %>%
  filter(f_simulated %in% pick_f_simulated) %>%
  ungroup() %>% 
  mutate(
    decision_status_50 = if_else(decision != "continue sampling" & n <= q50_n, 1, 0),
    decision_status_75 = if_else(decision != "continue sampling" & n <= q75_n, 1, 0),
    decision_status_90 = if_else(decision != "continue sampling" & n <= q90_n, 1, 0),
    decision_status_95 = if_else(decision != "continue sampling" & n <= q95_n, 1, 0),
    decision_status_100 = if_else(decision != "continue sampling", 1, 0)
  ) %>% 
  summarize(
    alpha_error_rate = mean(decision_error[f_simulated == 0], na.rm = TRUE),
    beta_error_rate = mean(decision_error[f_simulated != 0], na.rm = TRUE),
    mean_n = mean(n),
    median_n = median(n),
    max_n = max(n),
    min_n = min(n),
    decision_rate_100 = sum(decision_status_100)/n(),
    q25_n = quantile(n, 0.25),
    q50_n = quantile(n, 0.50),
    q75_n = quantile(n, 0.75),
    q90_n = quantile(n, 0.90),
    q95_n = quantile(n, 0.95),
    sd_error_n  = sd(n)/sqrt(length((n))),
    decision_rate_50 = sum(decision_status_50)/n(),
    decision_rate_75 = sum(decision_status_75)/n(),
    decision_rate_90 = sum(decision_status_90)/n(),
    decision_rate_95 = sum(decision_status_95)/n()
  )

```


## Extended Results

```{r}
pick_f_simulated = c(0,
                     pick_f_expected - 0.05,
                     pick_f_expected,
                     pick_f_expected + 0.15)
```


### violin

```{r, fig.height=8, fig.width=10, out.width="80%", fig.align="left"}
# violin_colors <- c(light_blue, hell_rosa, light_blue, light_blue)

title_html <- glue("<i>f</i><sub>exp</sub> = {pick_f_expected}")
subtitle_html <- glue(
  # "<i>f</i><sub>true</sub> = {pick_f_expected},<br>",
  "1 - &beta; = {pick_power}, &alpha; = 0.05,<br>",
  "groups = {pick_k_groups}"
)


p_violin <- df_all %>%
  mutate(
    f_simulated = factor(f_simulated, levels = sort(unique(pick_f_simulated))),
    fill_label = ifelse(f_simulated == pick_f_expected - 0.05,
                        "Reference scenario",
                        "Other scenarios")
  ) %>%
  filter(
    f_expected == pick_f_expected,
    power == pick_power,
    k_groups == pick_k_groups
  ) %>%
  filter(f_simulated %in% pick_f_simulated) %>%
  ggplot(aes(x = as.factor(f_simulated), y = n, fill = fill_label)) +
  
  # violins (fill legend only)
  geom_violin(
    trim = FALSE,
    scale = "count",
    draw_quantiles = c(0.25, 0.50, 0.75),
    linewidth = 0.1,
    color = "lightgrey"
  ) +
  scale_fill_manual(
    name = "Scenario",
    values = c("Other scenarios" = light_blue,
               "Reference scenario" = hell_rosa)
  ) +
  
  # red diamond = fixed sample size
  geom_point(
    aes(y = fix_n, color = "Fixed sample size"),
    position = position_dodge(0.9),
    shape = 18,     # diamond, no border
    size  = 7,
    stroke = 0      # ensures no black outline
  ) +
  
  # blue square = mean sample size
  geom_point(
    aes(y = mean_n, color = "Mean sample size"),
    position = position_dodge(0.9),
    shape = 15,     # square, no border
    size  = 6,
    stroke = 0      # ensures no black outline
  ) +
scale_color_manual(
  name   = "Estimates",
  values = c("Fixed sample size" = red,
             "Mean sample size"  = dark_blue),
  labels = c("Fixed sample size" = "Fixed design",
             "Mean sample size"  = "Sequential designs")
) +
  labs(
    x = "<i>f</i><sub>true</sub>",
    title = title_html,
    caption = subtitle_html  
  ) +
  
  # clean legend appearance
  guides(
    fill  = guide_legend(override.aes = list(shape = NA)),  # patches only
    color = guide_legend(override.aes = list(stroke = 0, size = 5)) # points only
  ) +
  theme(
    legend.key = element_blank(),  # no background box around keys
    plot.title = element_markdown(),
    plot.caption = element_markdown(),
    axis.title.x = element_markdown(),
    panel.grid.minor.y = element_blank()
  )

p_violin


```


### table

```{r}
table_cases <- df_all %>% 
  mutate(f_simulated = factor(f_simulated, levels = sort(unique(pick_f_simulated))),
         fill_group = factor(as.integer(f_simulated))) %>% 
  filter(f_expected == pick_f_expected,
         power == pick_power,
         k_groups == pick_k_groups
  ) %>% 
  filter(f_simulated %in% pick_f_simulated) %>% 
    group_by(
    f_simulated,
    f_expected,
    k_groups,
    alpha,
    power,
    distribution,
    sd,
    sample_ratio) %>% 
  mutate(
    decision_status_50 = if_else(decision != "continue sampling" & n <= q50_n, 1, 0),
    decision_status_75 = if_else(decision != "continue sampling" & n <= q75_n, 1, 0),
    decision_status_90 = if_else(decision != "continue sampling" & n <= q90_n, 1, 0),
    decision_status_95 = if_else(decision != "continue sampling" & n <= q95_n, 1, 0),
    decision_status_100 = if_else(decision != "continue sampling", 1, 0)
  ) %>% 
  group_by(f_simulated, f_expected, power) %>%
  summarize(
    decision_error_rate = mean(decision_error),
    fix_n = as.integer(mean(fix_n)),
    mean_n = mean(n),
    median_n = as.integer(median(n)),
    min_n = as.integer(min(n)),
    max_n = max(n),
    decision_rate_100 = sum(decision_status_100)/n(),
    q25_n = quantile(n, 0.25),
    q50_n = quantile(n, 0.50),
    q75_n = quantile(n, 0.75),
    q90_n = quantile(n, 0.90),
    q95_n = quantile(n, 0.95),
    sd_error_n  = sd(n)/sqrt(length((n))),
    decision_rate_50 = sum(decision_status_50)/n(),
    decision_rate_75 = sum(decision_status_75)/n(),
    decision_rate_90 = sum(decision_status_90)/n(),
    decision_rate_95 = sum(decision_status_95)/n()
  ) %>% mutate(
    simulated_power = if_else(f_simulated != 0, 1 - decision_error_rate, NA),
    .after = power
  ) %>% 
  select(-power) %>% 
    ungroup() %>%
    select(
    f_simulated,
    f_expected,
    simulated_power,
    decision_error_rate,
    fix_n,
    mean_n,
    median_n,
    min_n,
    max_n,
    decision_rate_90,
    q95_n,
    decision_rate_95
  )# %>% 
  # mutate(f_simulated = as.double(f_simulated))

table_cases %>%
  ungroup() %>% 
  mutate(decision_error_rate = decision_error_rate) %>% 
  gt() %>%
  tab_header(
    title = gt::html("<i>Summary Table</i>")
  ) %>%
  cols_label(
    f_simulated         = gt::html("<i>f</i><sub>true</sub>"),
    f_expected          = gt::html("<i>f</i><sub>exp</sub>"),
    simulated_power     = "simulated power",
    decision_error_rate = "Error rate %",
    fix_n               = "Fixed n",
    mean_n              = "Mean n",
    median_n            = "Median n",
    min_n               = "Min",
    max_n               = "Max",
    decision_rate_90    = "90% decision rate",
    q95_n               = "95% n",
    decision_rate_95    = "95% decision rate"
  ) %>%
  fmt_number(
    columns = where(is.double),
    decimals = 2
  )


```


### Power curve


```{r}
plot_cases <- df_all %>% 
  mutate(fill_group = factor(as.integer(f_simulated))) %>% 
  filter(f_expected == pick_f_expected,
         power == pick_power,
         k_groups == pick_k_groups
  ) %>% 
  filter(f_simulated != 0) %>% 
  group_by(
    f_simulated,
    f_expected,
    k_groups,
    alpha,
    power,
    distribution,
    sd,
    sample_ratio) %>% 
  mutate(
    decision_status_50 = if_else(decision != "continue sampling" & n <= q50_n, 1, 0),
    decision_status_75 = if_else(decision != "continue sampling" & n <= q75_n, 1, 0),
    decision_status_90 = if_else(decision != "continue sampling" & n <= q90_n, 1, 0),
    decision_status_95 = if_else(decision != "continue sampling" & n <= q95_n, 1, 0),
    decision_status_100 = if_else(decision != "continue sampling", 1, 0)
  ) %>% 
  summarize(
    error_rate = mean(decision_error),
    mean_n = mean(n),
    median_n = median(n),
    q90_n = quantile(n, 0.90),
  ) %>% mutate(
    simulated_power = if_else(f_simulated != 0, 1 - error_rate, NA),
    .after = power
  )


plot_cases %>% 
ggplot(aes(x = f_simulated, y = simulated_power)) +
  # shaded rectangle
  # annotate("rect",
  #          xmin = pick_f_expected, xmax = Inf,
  #          ymin = pick_power, ymax = Inf,
  #          fill = "lightblue", alpha = 0.2) +
  # shaded area under the curve, starting at f_expected
  geom_ribbon(
    data = subset(plot_cases, f_simulated >= pick_f_expected),
    aes(ymin = pick_power, ymax = simulated_power),
    fill = light_green, alpha = 0.7
  ) +
  geom_ribbon(
    data = subset(plot_cases, f_simulated < pick_f_expected),
    aes(ymin = simulated_power, ymax = pick_power),
    fill = light_red, alpha = 0.5
  ) +
  geom_line(linewidth = 1) +
  geom_point(size = 4) +
  geom_hline(yintercept = pick_power, linetype = "dashed") +
  annotate("text", x = min(plot_cases$f_simulated), y = pick_power,
           label = paste0(" target power = ", percent(pick_power)),
           hjust = 0, vjust = 2, size = base_size*0.3) +
  scale_y_continuous(
    limits = c(0, 1),
    breaks = seq(0, 1, by = 0.1),
    labels = percent_format(accuracy = 1)
  ) +
  scale_x_continuous(expand = expansion(mult = c(0.02, 0.02)),
                     breaks = unique(plot_cases$f_simulated)) +
  labs(
    title    = "Simulated power across true effect sizes",
    x = expression(italic(f)[true]),
    y = "Simulated power",
    subtitle = glue("<i>f</i><sub>exp</sub> = {pick_f_expected},",
                    "1 - &beta; = {pick_power}, &alpha; = 0.05, groups = {pick_k_groups}")
  ) +
  theme(
    panel.grid.minor = element_blank(),
    plot.title.position = "plot",
    plot.subtitle = element_markdown()
  )

```

### cumultative

```{r}
df_cum <- df_all %>%
  filter(f_expected == pick_f_expected,
         power      == pick_power,
         k_groups   == pick_k_groups) %>% 
  filter(decision != "continue sampling") %>%
  transmute(f_simulated, n_decision = n) %>% 
  group_by(f_simulated) %>%
  arrange(n_decision, .by_group = TRUE) %>%
  mutate(cum_prop = row_number() / n()) %>%
  ungroup()

# 4) Find the minimum n that achieves at least 90% decided
cut90 <- df_cum %>%
  group_by(f_simulated) %>%
  summarise(n_at_90 = min(n_decision[cum_prop >= 0.90]), .groups = "drop")

# 5) Plot: ECDF of decision n, with a 90% line and vertical markers at n_at_90
ggplot(df_cum, aes(x = n_decision, y = cum_prop, colour = factor(f_simulated), group = f_simulated)) +
  geom_step(linewidth = 1.2) +
  geom_hline(yintercept = 0.90) +
  labs(
    x = "Sample size n",
    y = "Cumulative proportion with a decision",
    colour = "f_simulated"
  ) +
  scale_y_continuous(limits = c(0, 1)) 
```

# calc_likelihoods_anova: Check warnings & messages

    Code
      calc_likelihoods_anova(seq_anova_arguments, non_centrality_parameter,
        F_statistic)
    Output
      $likelihood_1
      [1] 0.05244313
      
      $likelihood_0
      [1] 4.794529e-09
      
      $ratio
      [1] 10938119
      
      $likelihood_1_log
      [1] -2.948026
      
      $likelihood_0_log
      [1] -19.15579
      
      $ratio_log
      [1] 16.20776
      

# snaphot: effect_size

    Code
      effect_sizes(formula, df_no_effect)
    Output
      $cohens_f
      [1] 0.02783414
      
      $ci_cohens_f_lower
      [1] NA
      
      $ci_cohens_f_upper
      [1] 0.04623125
      
      $cohens_f_adj
      [1] 0.01997895
      
      $cohens_f_median
      [1] 0.02269439
      
      $eta_squared
      [1] 0.0007741396
      
      $partial_eta_squared
      [1] 0.0007741396
      
      $adjusted_eta_squared
      [1] 0.0003992425
      
      $ci_ncp_lower
      [1] NA
      
      $ci_ncp_upper
      [1] 17.09863
      

---

    Code
      effect_sizes(formula, df_no_effect_extreme)
    Output
      $cohens_f
      [1] 0.007884077
      
      $ci_cohens_f_lower
      [1] NA
      
      $ci_cohens_f_upper
      [1] 0.01359347
      
      $cohens_f_adj
      [1] 0.00496528
      
      $cohens_f_median
      [1] 0.005989646
      
      $eta_squared
      [1] 6.215481e-05
      
      $partial_eta_squared
      [1] 6.215481e-05
      
      $adjusted_eta_squared
      [1] 2.465526e-05
      
      $ci_ncp_lower
      [1] NA
      
      $ci_ncp_upper
      [1] 14.7826
      

---

    Code
      seq_anova(formula, 1, data = df_no_effect_extreme)
    Condition
      Warning in `calc_likelihoods_anova()`:
      At least one likelihood is equal to 0.
                  The test works with the logarithm of the likelihoods.
    Output
      
      *****  Sequential ANOVA *****
      
      formula: formula
      test statistic:
       log-likelihood ratio = -39375.2, decision = accept H0
      SPRT thresholds:
       lower log(B) = -2.944, upper log(A) = 2.944
      Log-Likelihood of the:
       alternative hypothesis = -39376.7
       null hypothesis = -1.505
      alternative hypothesis: true difference in means is not equal to 0.
      specified effect size: Cohen's f = 1
      empirical Cohen's f = 0.007884077, 95% CI[0, 0.01359347]
      Cohen's f adjusted = 0.005
      degrees of freedom: df1 = 3, df2 = 79996
      SS effect = 4.9679, SS residual = 79922.87, SS total = 79927.84
      *Note: to get access to the object of the results use the @ or [] instead of the $ operator.

---

    Code
      seq_anova(formula, 0.01, data = df_0.01_effect)
    Output
      
      *****  Sequential ANOVA *****
      
      formula: formula
      test statistic:
       log-likelihood ratio = -0.052, decision = continue sampling
      SPRT thresholds:
       lower log(B) = -2.944, upper log(A) = 2.944
      Log-Likelihood of the:
       alternative hypothesis = -0.412
       null hypothesis = -0.36
      alternative hypothesis: true difference in means is not equal to 0.
      specified effect size: Cohen's f = 0.01
      empirical Cohen's f = 0.02693014, 95% CI[NA, 0.05742471]
      Cohen's f adjusted = 0
      degrees of freedom: df1 = 3, df2 = 1996
      SS effect = 1.405763, SS residual = 1938.363, SS total = 1939.768
      *Note: to get access to the object of the results use the @ or [] instead of the $ operator.

---

    Code
      effect_sizes(formula, df_0.01_effect)
    Output
      $cohens_f
      [1] 0.02693014
      
      $ci_cohens_f_lower
      [1] NA
      
      $ci_cohens_f_upper
      [1] 0.05742471
      
      $cohens_f_adj
      [1] 0
      
      $cohens_f_median
      [1] 0
      
      $eta_squared
      [1] 0.0007247067
      
      $partial_eta_squared
      [1] 0.0007247067
      
      $adjusted_eta_squared
      [1] -0.00077721
      
      $ci_ncp_lower
      [1] NA
      
      $ci_ncp_upper
      [1] 6.595196
      

---

    Code
      effect_sizes(formula, df_0.10_effect)
    Output
      $cohens_f
      [1] 0.08548456
      
      $ci_cohens_f_lower
      [1] NA
      
      $ci_cohens_f_upper
      [1] 0.1432724
      
      $cohens_f_adj
      [1] 0.05918448
      
      $cohens_f_median
      [1] 0.06829665
      
      $eta_squared
      [1] 0.007254596
      
      $partial_eta_squared
      [1] 0.007254596
      
      $adjusted_eta_squared
      [1] 0.003513093
      
      $ci_ncp_lower
      [1] NA
      
      $ci_ncp_upper
      [1] 16.42159
      

---

    Code
      effect_sizes(formula, df_0.25_effect)
    Output
      $cohens_f
      [1] 0.3951891
      
      $ci_cohens_f_lower
      [1] 0.212692
      
      $ci_cohens_f_upper
      [1] 0.5496487
      
      $cohens_f_adj
      [1] 0.3709923
      
      $cohens_f_median
      [1] 0.3818288
      
      $eta_squared
      [1] 0.1350786
      
      $partial_eta_squared
      [1] 0.1350786
      
      $adjusted_eta_squared
      [1] 0.123311
      
      $ci_ncp_lower
      [1] 6.785684
      
      $ci_ncp_upper
      [1] 45.31705
      

---

    Code
      effect_sizes(formula, df_0.40_effect)
    Output
      $cohens_f
      [1] 0.5759504
      
      $ci_cohens_f_lower
      [1] 0.1342421
      
      $ci_cohens_f_upper
      [1] 0.7988566
      
      $cohens_f_adj
      [1] 0.4530764
      
      $cohens_f_median
      [1] 0.4838036
      
      $eta_squared
      [1] 0.2490908
      
      $partial_eta_squared
      [1] 0.2490908
      
      $adjusted_eta_squared
      [1] 0.1823433
      
      $ci_ncp_lower
      [1] 0.9010465
      
      $ci_ncp_upper
      [1] 31.90859
      

---

    Code
      effect_sizes(formula, df_2_effect)
    Output
      $cohens_f
      [1] 1.710485
      
      $ci_cohens_f_lower
      [1] 0.928781
      
      $ci_cohens_f_upper
      [1] 2.297912
      
      $cohens_f_adj
      [1] 1.513475
      
      $cohens_f_median
      [1] 1.595753
      
      $eta_squared
      [1] 0.7452722
      
      $partial_eta_squared
      [1] 0.7452722
      
      $adjusted_eta_squared
      [1] 0.7311207
      
      $ci_ncp_lower
      [1] 17.25268
      
      $ci_ncp_upper
      [1] 105.608
      

---

    Code
      effect_sizes(formula, df_no_effect)
    Output
      $cohens_f
      [1] 0.02783414
      
      $ci_cohens_f_lower
      [1] NA
      
      $ci_cohens_f_upper
      [1] 0.04623125
      
      $cohens_f_adj
      [1] 0.01997895
      
      $cohens_f_median
      [1] 0.02269439
      
      $eta_squared
      [1] 0.0007741396
      
      $partial_eta_squared
      [1] 0.0007741396
      
      $adjusted_eta_squared
      [1] 0.0003992425
      
      $ci_ncp_lower
      [1] NA
      
      $ci_ncp_upper
      [1] 17.09863
      

---

    Code
      effect_sizes(formula, df_0.01_effect)
    Output
      $cohens_f
      [1] 0.02693014
      
      $ci_cohens_f_lower
      [1] NA
      
      $ci_cohens_f_upper
      [1] 0.05742471
      
      $cohens_f_adj
      [1] 0
      
      $cohens_f_median
      [1] 0
      
      $eta_squared
      [1] 0.0007247067
      
      $partial_eta_squared
      [1] 0.0007247067
      
      $adjusted_eta_squared
      [1] -0.00077721
      
      $ci_ncp_lower
      [1] NA
      
      $ci_ncp_upper
      [1] 6.595196
      

---

    Code
      effect_sizes(formula, df_0.10_effect)
    Output
      $cohens_f
      [1] 0.08548456
      
      $ci_cohens_f_lower
      [1] NA
      
      $ci_cohens_f_upper
      [1] 0.1432724
      
      $cohens_f_adj
      [1] 0.05918448
      
      $cohens_f_median
      [1] 0.06829665
      
      $eta_squared
      [1] 0.007254596
      
      $partial_eta_squared
      [1] 0.007254596
      
      $adjusted_eta_squared
      [1] 0.003513093
      
      $ci_ncp_lower
      [1] NA
      
      $ci_ncp_upper
      [1] 16.42159
      

---

    Code
      effect_sizes(formula, df_0.25_effect)
    Output
      $cohens_f
      [1] 0.3951891
      
      $ci_cohens_f_lower
      [1] 0.212692
      
      $ci_cohens_f_upper
      [1] 0.5496487
      
      $cohens_f_adj
      [1] 0.3709923
      
      $cohens_f_median
      [1] 0.3818288
      
      $eta_squared
      [1] 0.1350786
      
      $partial_eta_squared
      [1] 0.1350786
      
      $adjusted_eta_squared
      [1] 0.123311
      
      $ci_ncp_lower
      [1] 6.785684
      
      $ci_ncp_upper
      [1] 45.31705
      

---

    Code
      effect_sizes(formula, df_0.40_effect)
    Output
      $cohens_f
      [1] 0.5759504
      
      $ci_cohens_f_lower
      [1] 0.1342421
      
      $ci_cohens_f_upper
      [1] 0.7988566
      
      $cohens_f_adj
      [1] 0.4530764
      
      $cohens_f_median
      [1] 0.4838036
      
      $eta_squared
      [1] 0.2490908
      
      $partial_eta_squared
      [1] 0.2490908
      
      $adjusted_eta_squared
      [1] 0.1823433
      
      $ci_ncp_lower
      [1] 0.9010465
      
      $ci_ncp_upper
      [1] 31.90859
      

---

    Code
      effect_sizes(formula, df_2_effect)
    Output
      $cohens_f
      [1] 1.710485
      
      $ci_cohens_f_lower
      [1] 0.928781
      
      $ci_cohens_f_upper
      [1] 2.297912
      
      $cohens_f_adj
      [1] 1.513475
      
      $cohens_f_median
      [1] 1.595753
      
      $eta_squared
      [1] 0.7452722
      
      $partial_eta_squared
      [1] 0.7452722
      
      $adjusted_eta_squared
      [1] 0.7311207
      
      $ci_ncp_lower
      [1] 17.25268
      
      $ci_ncp_upper
      [1] 105.608
      

# seq_anova: snapshots

    Code
      seq_anova(y ~ x, f = 0.4, data = data)
    Output
      
      *****  Sequential ANOVA *****
      
      formula: y ~ x
      test statistic:
       log-likelihood ratio = 8.483, decision = accept H1
      SPRT thresholds:
       lower log(B) = -2.944, upper log(A) = 2.944
      Log-Likelihood of the:
       alternative hypothesis = -2.219
       null hypothesis = -10.702
      alternative hypothesis: true difference in means is not equal to 0.
      specified effect size: Cohen's f = 0.4
      empirical Cohen's f = 0.444537, 95% CI[0.2414272, 0.5972637]
      Cohen's f adjusted = 0.41
      degrees of freedom: df1 = 3, df2 = 136
      SS effect = 233.3018, SS residual = 1180.598, SS total = 1413.9
      *Note: to get access to the object of the results use the @ or [] instead of the $ operator.

---

    Code
      seq_anova(age ~ sex, f = 0.4, data = df_age)
    Output
      
      *****  Sequential ANOVA *****
      
      formula: age ~ sex
      test statistic:
       log-likelihood ratio = 3.114, decision = accept H1
      SPRT thresholds:
       lower log(B) = -2.944, upper log(A) = 2.944
      Log-Likelihood of the:
       alternative hypothesis = -2.466
       null hypothesis = -5.58
      alternative hypothesis: true difference in means is not equal to 0.
      specified effect size: Cohen's f = 0.4
      empirical Cohen's f = 0.3281472, 95% CI[0.1159147, 0.474829]
      Cohen's f adjusted = 0.286
      degrees of freedom: df1 = 3, df2 = 136
      SS effect = 138.001, SS residual = 1281.577, SS total = 1419.578
      *Note: to get access to the object of the results use the @ or [] instead of the $ operator.

# calc_non_centrality_parameter

    Code
      show(calc_non_centrality_parameter_ttest(seq_ttest_arguments))
    Output
      [1] 1.788854

---

    Code
      show(calc_non_centrality_parameter_anova(seq_anova_arguments))
    Output
      [1] 32

# show: print output ttest?

    Code
      show(seq_ttest(rnorm(20), d = 0.8))
    Output
      
      *****  Sequential One Sample t-test *****
      
      formula: rnorm(20)
      test statistic:
       log-likelihood ratio = -5.783, decision = accept H0
      SPRT thresholds:
       lower log(B) = -2.944, upper log(A) = 2.944
      Log-Likelihood of the:
       alternative hypothesis = -5.676
       null hypothesis = 0.107
      alternative hypothesis: true mean is not equal to 0.
      specified effect size: Cohen's d = 0.8
      degrees of freedom: df = 19
      sample estimates:
      mean of x 
        0.05238 
      *Note: to get access to the object of the results use the @ or [] instead of the $ operator.

---

    Code
      show(seq_ttest(x_special_name, d = 0.8, alternative = "less", mu = 2))
    Output
      
      *****  Sequential One Sample t-test *****
      
      formula: x_special_name
      test statistic:
       log-likelihood ratio = 12.358, decision = accept H1
      SPRT thresholds:
       lower log(B) = -2.944, upper log(A) = 2.944
      Log-Likelihood of the:
       alternative hypothesis = -17.916
       null hypothesis = -30.275
      alternative hypothesis: true mean is less than 2.
      specified effect size: Cohen's d = 0.8
      degrees of freedom: df = 19
      sample estimates:
      mean of x 
       -0.50342 
      *Note: to get access to the object of the results use the @ or [] instead of the $ operator.

---

    Code
      show(seq_ttest(x, d = 0.8, alternative = "greater"))
    Output
      
      *****  Sequential One Sample t-test *****
      
      formula: x
      test statistic:
       log-likelihood ratio = -9.385, decision = accept H0
      SPRT thresholds:
       lower log(B) = -2.944, upper log(A) = 2.944
      Log-Likelihood of the:
       alternative hypothesis = -10.711
       null hypothesis = -1.326
      alternative hypothesis: true mean is greater than 0.
      specified effect size: Cohen's d = 0.8
      degrees of freedom: df = 19
      sample estimates:
      mean of x 
       -0.20286 
      *Note: to get access to the object of the results use the @ or [] instead of the $ operator.

---

    Code
      show(seq_ttest(x_special_name, y_secial_name, d = 0.8))
    Output
      
      *****  Sequential  Two Sample t-test *****
      
      formula: x_special_name and  y_secial_name
      test statistic:
       log-likelihood ratio = -3.2, decision = accept H0
      SPRT thresholds:
       lower log(B) = -2.944, upper log(A) = 2.944
      Log-Likelihood of the:
       alternative hypothesis = 3.455
       null hypothesis = 6.655
      alternative hypothesis: true difference in means is not equal to 0.
      specified effect size: Cohen's d = 0.8
      degrees of freedom: df = 38
      sample estimates:
      mean of x mean of y 
        0.14016   0.13999 
      *Note: to get access to the object of the results use the @ or [] instead of the $ operator.

---

    Code
      show(seq_ttest(x, y, d = 0.8, alternative = "less"))
    Output
      
      *****  Sequential  Two Sample t-test *****
      
      formula: x and  y
      test statistic:
       log-likelihood ratio = -5.225, decision = accept H0
      SPRT thresholds:
       lower log(B) = -2.944, upper log(A) = 2.944
      Log-Likelihood of the:
       alternative hypothesis = -6.486
       null hypothesis = -1.261
      alternative hypothesis: true difference in means is less than 0.
      specified effect size: Cohen's d = 0.8
      degrees of freedom: df = 38
      sample estimates:
      mean of x mean of y 
        0.11723  -0.14143 
      *Note: to get access to the object of the results use the @ or [] instead of the $ operator.

---

    Code
      show(seq_ttest(x ~ y, d = 0.8))
    Output
      
      *****  Sequential  Two Sample t-test *****
      
      formula: x ~ y
      test statistic:
       log-likelihood ratio = -1.596, decision = continue sampling
      SPRT thresholds:
       lower log(B) = -2.944, upper log(A) = 2.944
      Log-Likelihood of the:
       alternative hypothesis = 0.499
       null hypothesis = 2.095
      alternative hypothesis: true difference in means is not equal to 0.
      specified effect size: Cohen's d = 0.8
      degrees of freedom: df = 18
      sample estimates:
      mean of x mean of y 
        0.18073   0.16120 
      *Note: to get access to the object of the results use the @ or [] instead of the $ operator.

---

    Code
      show(seq_ttest(y ~ x, d = 0.6, data = data, alternative = "less", verbose = TRUE))
    Output
      
      *****  Sequential  Two Sample t-test *****
      
      formula: y ~ x
      test statistic:
       log-likelihood ratio = 10.441, decision = accept H1
      SPRT thresholds:
       lower log(B) = -2.944, upper log(A) = 2.944
      Log-Likelihood of the:
       alternative hypothesis = -5.72
       null hypothesis = -16.161
      alternative hypothesis: true difference in means is less than 0.
      specified effect size: Cohen's d = 0.6
      degrees of freedom: df = 78
      sample estimates:
      mean of x mean of y 
       -0.61169   0.80048 
      *Note: to get access to the object of the results use the @ or [] instead of the $ operator.

---

    Code
      show(seq_ttest(y ~ x, d = 0.6, data = data, alternative = "greater", verbose = TRUE))
    Output
      
      *****  Sequential  Two Sample t-test *****
      
      formula: y ~ x
      test statistic:
       log-likelihood ratio = 10.458, decision = accept H1
      SPRT thresholds:
       lower log(B) = -2.944, upper log(A) = 2.944
      Log-Likelihood of the:
       alternative hypothesis = -5.745
       null hypothesis = -16.202
      alternative hypothesis: true difference in means is greater than 0.
      specified effect size: Cohen's d = 0.6
      degrees of freedom: df = 78
      sample estimates:
      mean of x mean of y 
        0.81327  -0.47317 
      *Note: to get access to the object of the results use the @ or [] instead of the $ operator.

---

    Code
      show(seq_ttest(rnorm(50), mu = 1, d = 0.4, verbose = TRUE, alternative = "greater"))
    Output
      
      *****  Sequential One Sample t-test *****
      
      formula: rnorm(50)
      test statistic:
       log-likelihood ratio = -11.078, decision = accept H0
      SPRT thresholds:
       lower log(B) = -2.944, upper log(A) = 2.944
      Log-Likelihood of the:
       alternative hypothesis = -29.296
       null hypothesis = -18.219
      alternative hypothesis: true mean is greater than 1.
      specified effect size: Cohen's d = 0.4
      degrees of freedom: df = 49
      sample estimates:
      mean of x 
        0.01173 
      *Note: to get access to the object of the results use the @ or [] instead of the $ operator.

# show: verbose

    Code
      show(calc_seq_ttest(build_prototype_seq_ttest_arguments()))
    Output
      
      *****  Sequential  Two Sample t-test *****
      
      formula: x and y
      test statistic:
       log-likelihood ratio = 2.193, decision = continue sampling
      SPRT thresholds:
       lower log(B) = -1.558, upper log(A) = 2.773
      Log-Likelihood of the:
       alternative hypothesis = -3.201
       null hypothesis = -5.393
      alternative hypothesis: true difference in means is not equal to 0.
      specified effect size: Cohen's d = 0.8
      degrees of freedom: df = 18
      sample estimates:
      mean of x mean of y 
       -0.13828   1.05060 
      *Note: to get access to the object of the results use the @ or [] instead of the $ operator.

---

    Code
      show(calc_seq_ttest(build_prototype_seq_ttest_arguments(), verbose = FALSE))
    Output
      
      *****  Sequential  Two Sample t-test *****
      
      formula: x and y
      test statistic:
       log-likelihood ratio = 2.193, decision = continue sampling
      SPRT thresholds:
       lower log(B) = -1.558, upper log(A) = 2.773

# show: print output anova?

    Code
      show(seq_anova(y ~ x, f = 0.25, data = data))
    Output
      
      *****  Sequential ANOVA *****
      
      formula: y ~ x
      test statistic:
       log-likelihood ratio = 5.579, decision = accept H1
      SPRT thresholds:
       lower log(B) = -2.944, upper log(A) = 2.944
      Log-Likelihood of the:
       alternative hypothesis = -1.768
       null hypothesis = -7.348
      alternative hypothesis: true difference in means is not equal to 0.
      specified effect size: Cohen's f = 0.25
      empirical Cohen's f = 0.255582, 95% CI[0.1175887, 0.3478401]
      Cohen's f adjusted = 0.228
      degrees of freedom: df1 = 4, df2 = 325
      SS effect = 20.49968, SS residual = 313.8243, SS total = 334.324
      *Note: to get access to the object of the results use the @ or [] instead of the $ operator.

---

    Code
      show(seq_anova(happiness ~ job_satisfaction, f = 0.25, data = df_job))
    Output
      
      *****  Sequential ANOVA *****
      
      formula: happiness ~ job_satisfaction
      test statistic:
       log-likelihood ratio = 4.66, decision = accept H1
      SPRT thresholds:
       lower log(B) = -2.944, upper log(A) = 2.944
      Log-Likelihood of the:
       alternative hypothesis = -6.628
       null hypothesis = -11.288
      alternative hypothesis: true difference in means is not equal to 0.
      specified effect size: Cohen's f = 0.25
      empirical Cohen's f = 0.6758475, 95% CI[0.3452897, 0.9720489]
      Cohen's f adjusted = 0.631
      degrees of freedom: df1 = 1, df2 = 46
      SS effect = 19.50919, SS residual = 42.71121, SS total = 62.2204
      *Note: to get access to the object of the results use the @ or [] instead of the $ operator.

---

    Code
      show(seq_anova(y ~ x, f = 0.1, data = data))
    Output
      
      *****  Sequential ANOVA *****
      
      formula: y ~ x
      test statistic:
       log-likelihood ratio = -0.036, decision = continue sampling
      SPRT thresholds:
       lower log(B) = -2.944, upper log(A) = 2.944
      Log-Likelihood of the:
       alternative hypothesis = -0.799
       null hypothesis = -0.763
      alternative hypothesis: true difference in means is not equal to 0.
      specified effect size: Cohen's f = 0.1
      empirical Cohen's f = 0.1958652, 95% CI[NA, 0.3634934]
      Cohen's f adjusted = 0
      degrees of freedom: df1 = 3, df2 = 76
      SS effect = 3.948363, SS residual = 102.9206, SS total = 106.869
      *Note: to get access to the object of the results use the @ or [] instead of the $ operator.

---

    Code
      show(seq_anova(y ~ x, f = 0.25, alpha = 0.3, power = 0.95, data = data))
    Output
      
      *****  Sequential ANOVA *****
      
      formula: y ~ x
      test statistic:
       log-likelihood ratio = 7.792, decision = accept H1
      SPRT thresholds:
       lower log(B) = -2.639, upper log(A) = 1.153
      Log-Likelihood of the:
       alternative hypothesis = -5.07
       null hypothesis = -12.862
      alternative hypothesis: true difference in means is not equal to 0.
      specified effect size: Cohen's f = 0.25
      empirical Cohen's f = 0.4366689, 95% CI[0.2449371, 0.6194111]
      Cohen's f adjusted = 0.42
      degrees of freedom: df1 = 1, df2 = 118
      SS effect = 25.09296, SS residual = 131.5974, SS total = 156.6904
      *Note: to get access to the object of the results use the @ or [] instead of the $ operator.

---

    Code
      show(seq_anova(y ~ x, f = 0.25, data = data, verbose = FALSE))
    Output
      
      *****  Sequential ANOVA *****
      
      formula: y ~ x
      test statistic:
       log-likelihood ratio = 10.2, decision = accept H1
      SPRT thresholds:
       lower log(B) = -2.944, upper log(A) = 2.944

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_data_external.R
\name{load_sample_size_data}
\alias{load_sample_size_data}
\title{Access sample size simulation data}
\usage{
load_sample_size_data()
}
\value{
A named list with the following elements:
\itemize{
\item \code{description}: Short description of the dataset
\item \code{version}: GitHub release tag of the dataset (e.g., \code{"v0.1.0-data"})
\item \code{created}: Date the dataset was created (as character string)
\item \code{n_rep}: Number of simulation iterations per condition
\item \code{data}: A data frame with simulation results (see \strong{Data Structure})
}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Loads pre-computed simulation results for SPRT sample size planning.
If not already cached locally, the data (~150 MB) will be downloaded automatically
from GitHub releases. Use this function to access the complete dataset for custom
analysis and visualization. See the \strong{Data Structure} section below for details
on available columns.

Data is hosted at:
\href{https://github.com/MeikeSteinhilber/sprtt_plan_sample_size}{MeikeSteinhilber/sprtt_plan_sample_size}
}
\section{Data Structure}{

The \code{data} element contains simulation results with the following columns:

\strong{Simulation Metadata:}
\itemize{
\item \code{batch}: Batch identifier for the simulation run
\item \code{iteration}: Individual simulation iteration within a batch
\item \code{source_file}: Path to the file containing simulation parameters or results
}

\strong{Input Parameters:}
\itemize{
\item \code{f_simulated}: The true effect size used to generate the simulated data
\item \code{f_expected}: The expected effect size specified for the SPRT
\item \code{k_groups}: Number of groups in the design
\item \code{alpha}: Significance level (Type I error rate)
\item \code{power}: Desired statistical power (1 - Type II error rate)
\item \code{distribution}: Data distribution used for simulation
\item \code{sd}: Standard deviation(s) used in data generation in each group
\item \code{sample_ratio}: Ratio of sample sizes between groups (e.g., 1:1, 2:1)
\item \code{n_raw_data}: Total number of raw observations generated in each group
\item \code{fix_n}: Fixed sample size
}

\strong{Individual Test Results:}
\itemize{
\item \code{n}: Actual sample size at which the SPRT terminated
\item \code{decision}: Test decision
\item \code{decision_error}: Whether the decision was erroneous (Type I or Type II error)
\item \code{log_lr}: Log-likelihood ratio at termination
\item \code{f}: Calculated effect size from the data
\item \code{f_adj}: Adjusted effect size
\item \code{f_statistic}: F-statistic from ANOVA test
}

\strong{Summary Statistics (Aggregated across iterations):}
\itemize{
\item \code{decision_error_rate}: Proportion of incorrect decisions
\item \code{mean_n}: Mean sample size across all iterations
\item \code{sd_error_n}: Standard error of the mean sample size (sd(n)/sqrt(n))
\item \code{median_n}: Median sample size (50th percentile)
\item \code{min_n}, \code{max_n}: Minimum and maximum sample sizes observed
\item \code{q25_n}, \code{q50_n}, \code{q75_n}, \code{q90_n}, \code{q95_n}: Sample size quantiles
\item \code{decision_rate_25}, \code{decision_rate_50}, \code{decision_rate_75},
\code{decision_rate_90}, \code{decision_rate_95}, \code{decision_rate_100}:
Cumulative decision rates at various percentages of maximum sample size
}
}

\examples{
\dontrun{
# Load data (downloads automatically if needed)
loaded <- load_sample_size_data()

# Access the simulation data frame
head(loaded$data)

# Check dataset version
loaded$version  # e.g. "v0.1.0-data"
loaded$created
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_data_external.R
\name{cache_clear}
\alias{cache_clear}
\title{Clear cached simulation data}
\usage{
cache_clear()
}
\value{
Invisibly returns \code{TRUE} if cache was cleared, \code{FALSE} if no cache existed.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Removes locally cached simulation data (~150 MB) used by \code{\link[=plan_sample_size]{plan_sample_size()}}.
Data will be automatically re-downloaded on next use of sample size planning functions.

This function is useful when:
\itemize{
\item You want to free up disk space
\item The cached data may be outdated and you want to force a fresh download
\item Troubleshooting cache-related issues
}
}
\examples{
\dontrun{
# Clear cache
cache_clear()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_data_external.R
\name{download_sample_size_data}
\alias{download_sample_size_data}
\title{Download simulation data for sample size planning}
\usage{
download_sample_size_data(force = FALSE)
}
\arguments{
\item{force}{Logical. If TRUE, re-download even if data exists. Default FALSE.}
}
\value{
Invisibly returns the path to the cached data file.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Downloads pre-computed simulation results from GitHub releases.
Data is cached locally and only needs to be downloaded once.

Data is hosted at:
\href{https://github.com/MeikeSteinhilber/sprtt_plan_sample_size}{MeikeSteinhilber/sprtt_plan_sample_size}
}
\examples{
\dontrun{
# Download data (only needed once)
download_sample_size_data()

# Force re-download (e.g., after data update)
download_sample_size_data(force = TRUE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CLASS_seq_anova_results.R
\name{[,seq_anova_results-method}
\alias{[,seq_anova_results-method}
\title{Method to retrieve the contents of a slot of an object of the
\code{\linkS4class{seq_anova_results}} class.}
\usage{
\S4method{[}{seq_anova_results}(x, i, j, drop)
}
\arguments{
\item{x}{the seq_ttest_results object.}

\item{i}{indices indicating elements to extract.}

\item{j}{not used.}

\item{drop}{not used.}

\item{seq_anova_results}{the corresponding class to this method.}
}
\value{
Returns the contents of the specified slot. For more information,
see the documentation for the \code{\linkS4class{seq_anova_results}} class.
}
\description{
Method to retrieve the contents of a slot of an object of the
\code{\linkS4class{seq_anova_results}} class.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CLASS_seq_ttest_arguments.R
\name{[,seq_ttest_arguments-method}
\alias{[,seq_ttest_arguments-method}
\title{Method to retrieve the contents of a slot of an object of the
seq_ttest_arguments class.}
\usage{
\S4method{[}{seq_ttest_arguments}(x, i, j, drop)
}
\arguments{
\item{x}{the seq_ttest_arguments object.}

\item{i}{indices indicating elements to extract.}

\item{j}{not used.}

\item{drop}{not used.}

\item{seq_ttest_arguments}{the corresponding class to this method.}
}
\value{
Returns the contents of the specified slot of an
seq_ttest_arguments object. For more information, see the arguments of the
\code{\link{seq_ttest}} function.
}
\description{
This method is only used internally to process
the input arguments of the \code{\link{seq_ttest}} function. As a normal user,
you can ignore this specific documentation.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/draw_sample_mixture.R
\name{draw_sample_mixture}
\alias{draw_sample_mixture}
\title{Draw Samples from a Gaussian Mixture Distribution}
\usage{
draw_sample_mixture(k_groups, f, max_n, counter_n = 100, verbose = FALSE)
}
\arguments{
\item{k_groups}{number of groups (levels of factor_A)}

\item{f}{Cohen's f. The simulated effect size.}

\item{max_n}{sample size for the groups (total sample size = max_n*k_groups)}

\item{counter_n}{number of times the function tries to find a possible parameter combination for the distribution. Default value is set to 100.}

\item{verbose}{\code{TRUE} or \code{FALSE.} Print out more information about the internal process of sampling the parameters
(the internal counter that was reached, some additional hints and the drawn parameters for the Gaussian Mixture distributions.)}
}
\value{
returns a data.frame with the columns y (observations) and x (factor_A).
}
\description{
Draws exemplary samples with a certain effect size for the sequential one-oway ANOVA or the sequential t-test, see Steinhilber et al. (2023) \url{doi:10.31234/osf.io/m64ne}
}
\examples{
set.seed(333)

data <- sprtt::draw_sample_mixture(
  k_groups = 2,
  f = 0.40,
  max_n = 2
)
data

data <- sprtt::draw_sample_mixture(
  k_groups = 4,
  f = 1.2, # very large effect size
  max_n = 4,
  counter_n = 1000, # increase of counter is necessary
  verbose = TRUE # prints more information to the console
)
data
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sprtt.R
\docType{package}
\name{sprtt-package}
\alias{sprtt}
\alias{sprtt-package}
\title{sprtt: Sequential Probability Ratio Tests Toolbox}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

A toolbox for Sequential Probability Ratio Tests (SPRT) based on Wald (1945) \doi{10.2134/agronj1947.00021962003900070011x}. SPRTs are applied during the sampling process, ideally after each observation, and at every stage return a decision to either continue sampling or terminate and accept one of the specified hypotheses. The `seq_ttest()` function performs one-sample, two-sample, and paired t-tests for one- and two-sided hypotheses (Schnuerch & Erdfelder (2019) \doi{10.1037/met0000234}). The `seq_anova()` function performs a sequential one-way fixed effects ANOVA (Steinhilber et al. (2024) \doi{10.1037/met0000677}). The `plan_sample_size()` function helps plan sequential studies by simulating required sample sizes across a range of effect sizes. For more information, see the vignettes browseVignettes(package = "sprtt") or the package website \url{https://meikesteinhilber.github.io/sprtt/}.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://meikesteinhilber.github.io/sprtt/}
  \item Report bugs at \url{https://github.com/MeikeSteinhilber/sprtt/issues}
}

}
\author{
\strong{Maintainer}: Meike Snijder-Steinhilber \email{Meike.Steinhilber@aol.com} (\href{https://orcid.org/0000-0002-7144-2100}{ORCID})

Authors:
\itemize{
  \item Martin Schnuerch (\href{https://orcid.org/0000-0001-6531-2265}{ORCID}) [thesis advisor]
  \item Anna-Lena Schubert (\href{https://orcid.org/0000-0001-7248-0662}{ORCID}) [thesis advisor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seq_ttest.R
\name{seq_ttest}
\alias{seq_ttest}
\title{Sequential Probability Ratio Test using t-statistic}
\usage{
seq_ttest(
  x,
  y = NULL,
  data = NULL,
  mu = 0,
  d,
  alpha = 0.05,
  power = 0.95,
  alternative = "two.sided",
  paired = FALSE,
  na.rm = TRUE,
  verbose = TRUE
)
}
\arguments{
\item{x}{Works with two classes: \code{numeric} and \code{formula}. Therefore you can
write \code{"x"} or \code{"x~y"}.
\itemize{
\item \code{"numeric input"}: a (non-empty) numeric vector of data values.
\item \code{"formula input"}: a formula of the form lhs ~ rhs where lhs is a numeric
variable giving the data values and rhs either 1 for a one-sample test or a
factor with two levels giving the corresponding groups.
}}

\item{y}{An optional (non-empty) numeric vector of data values.
Only used for two-sample tests when \code{x} is numeric.}

\item{data}{An optional data frame containing the variables specified in the formula.
Only used when \code{x} is a formula.}

\item{mu}{a number indicating the true value of the mean (or difference in
means if you are performing a two sample test).}

\item{d}{a number indicating the specified (expected) effect size (Cohen's d)}

\item{alpha}{Type I error rate (alpha level). The probability of rejecting H0
when it is true. Default is 0.05. Must be between 0 and 1.}

\item{power}{Statistical power (1 - beta), where beta is the Type II error rate.
The probability of correctly rejecting H0 when H1 is true with effect size d.
Default is 0.95. Must be between 0 and 1. Higher values lead to wider
stopping boundaries and potentially larger sample sizes.}

\item{alternative}{a character string specifying the alternative hypothesis,
must be one of \code{two.sided} (default), \code{greater} or \code{less}.
You can specify just the initial letter.}

\item{paired}{Logical indicating whether to perform a paired t-test.
Default is \code{FALSE}.}

\item{na.rm}{a logical value indicating whether \code{NA} values should be
removed before the computation proceeds.}

\item{verbose}{a logical value whether you want a verbose output or not.}
}
\value{
An object of the S4 class \code{\linkS4class{seq_ttest_results}}. Click on the
class link to see the full description of the slots.
To get access to the object use the
\code{@}-operator or \verb{[]}-brackets instead of \code{$}.
See the examples below.
}
\description{
Performs one-sample, two-sample, and paired sequential t-tests, which are variants of
Sequential Probability Ratio Tests (SPRT). The test allows for continuous
monitoring of data collection and provides stopping boundaries based on
likelihood ratios, offering efficiency gains over traditional fixed-N designs.

The sequential t-test continuously evaluates the likelihood ratio after each
observation (or pair of observations), stopping when sufficient evidence
accumulates for either H0 or H1.

For methodological details, see Schnuerch & Erdfelder (2019)
\url{doi:10.1037/met0000234}. For practical guidance, see
\code{vignette("t_test", package = "sprtt")}.
}
\examples{
# set seed --------------------------------------------------------------------
set.seed(333)

# load library ----------------------------------------------------------------
library(sprtt)

# one sample: numeric input ---------------------------------------------------
treatment_group <- rnorm(20, mean = 0, sd = 1)
results <- seq_ttest(treatment_group, mu = 1, d = 0.8)

# get access to the slots -----------------------------------------------------
# @ Operator
results@likelihood_ratio

# [] Operator
results["likelihood_ratio"]

# two sample: numeric input----------------------------------------------------
treatment_group <- stats::rnorm(20, mean = 0, sd = 1)
control_group <- stats::rnorm(20, mean = 1, sd = 1)
seq_ttest(treatment_group, control_group, d = 0.8)

# two sample: formula input ---------------------------------------------------
stress_level <- stats::rnorm(20, mean = 0, sd = 1)
sex <- as.factor(c(rep(1, 10), rep(2, 10)))
seq_ttest(stress_level ~ sex, d = 0.8)

# NA in the data --------------------------------------------------------------
stress_level <- c(NA, stats::rnorm(20, mean = 0, sd = 2), NA)
sex <- as.factor(c(rep(1, 11), rep(2, 11)))
seq_ttest(stress_level ~ sex, d = 0.8, na.rm = TRUE)

# work with dataset (data are in the package included) ------------------------
seq_ttest(monthly_income ~ sex, data = df_income, d = 0.8)
}
\seealso{
\itemize{
\item \code{\link[=seq_anova]{seq_anova()}} for sequential one-way ANOVA
\item \code{\link[=draw_sample_normal]{draw_sample_normal()}} for simulating test data
\item \code{vignette("t_test", package = "sprtt")} for detailed tutorial
\item \code{vignette("usage_sprtt", package = "sprtt")} for package overview
\item Schnuerch & Erdfelder (2019) \url{doi:10.1037/met0000234} for theoretical background
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_data_external.R
\name{cache_info}
\alias{cache_info}
\title{Cache information}
\usage{
cache_info()
}
\value{
Invisibly returns a list with:
\itemize{
\item \code{cache_dir}: Character string with the cache directory path
\item \code{data_cached}: Logical indicating if simulation data is cached
\item \code{file_size_mb}: Numeric file size in MB (or \code{NA} if not cached)
\item \code{data_version}: GitHub release tag of the cached dataset (or \code{NA} if not cached)
\item \code{data_created}: Date the dataset was created (or \code{NA} if not cached)
}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Displays information about cached simulation data (~150 MB) used by \code{\link[=plan_sample_size]{plan_sample_size()}}.
Shows the cache directory location, whether data is cached, file size, and dataset
version metadata.

The simulation data is automatically downloaded on first use of sample size planning
functions and stored locally for faster subsequent access.
}
\seealso{
\itemize{
\item \code{\link[=cache_clear]{cache_clear()}} to remove cached data
\item \code{\link[=download_sample_size_data]{download_sample_size_data()}} to manually download simulation data
\item \code{\link[=plan_sample_size]{plan_sample_size()}} which uses the cached data
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/df.R
\docType{data}
\name{df_income}
\alias{df_income}
\title{Test data to run the examples}
\format{
A data frame with 2 variables:
\describe{
\item{monthly_income}{ }
\item{sex}{ }
}
}
\usage{
df_income
}
\description{
A dataset that includes 120 individuals with sex gender and monthly income.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plan_sample_size.R
\name{plan_sample_size}
\alias{plan_sample_size}
\title{Generates HTML reports for sample size planning for sequential ANOVAs.}
\usage{
plan_sample_size(
  f_expected,
  k_groups,
  beta = 0.05,
  decision_rate = 0.85,
  output_dir = tempdir(),
  output_file = "sprtt-report-sample-size-planning.html",
  open = interactive(),
  overwrite = FALSE
)
}
\arguments{
\item{f_expected}{Numeric scalar. The expected standardized effect size (e.g., Cohen's f).
Must be between 0.1 and 0.4 (increments of 0.05).}

\item{k_groups}{Integer scalar. The number of groups to compare. Must be between 2 and 4.}

\item{beta}{Numeric scalar (default = 0.05). Desired beta error rate (Type II error).
Possible values are 0.20, 0.10, and 0.05.}

\item{decision_rate}{Numeric scalar (default = 0.85). Desired chance to reach a decision.
Must be between 0.75 and 0.95 (increments of 0.05).}

\item{output_dir}{Character string. Directory in which to save the rendered HTML report.
Defaults to a temporary directory (\code{tempdir()}).}

\item{output_file}{Character string. File name of the generated HTML report.
Defaults to \code{"sprtt-report-sample-size-planning.html"}.}

\item{open}{Logical (default = \code{interactive()}). If \code{TRUE}, the generated report is opened
in the system's default web browser after rendering.}

\item{overwrite}{Logical (default = \code{FALSE}). If \code{FALSE} and the target file already exists,
the user is prompted interactively whether to overwrite it. In non-interactive sessions,
an error is raised unless \code{overwrite = TRUE}.}
}
\value{
Invisibly returns the path to the rendered HTML file (character string).
The report is optionally opened in the default browser.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Renders a parameterized R Markdown report that helps plan sample size for the sequential ANOVA.
The function takes expected effect size (\code{f_expected}), number of groups (\code{k_groups}),
the power, and decision rate, then generates a reproducible HTML report summarizing the simulation-based
sample size recommendations. The alpha level is always 0.05.

The template is located under:
\code{inst/rmarkdown/templates/report_sample_size/skeleton/skeleton.Rmd}.
}
\details{
This function is a front-end utility for rendering a pre-defined R Markdown report using
\code{rmarkdown::render()}.
}
\section{File Overwrite Behavior}{

\itemize{
\item If the specified output file already exists:
\itemize{
\item and \code{overwrite = FALSE}, the user is asked whether to overwrite (in interactive sessions);
otherwise, an error is thrown.
\item If \code{overwrite = TRUE}, the file is replaced silently.
}
}
}

\examples{
\dontrun{
# Generate and open an SPRT sample size planning report:
plan_sample_size(
  f_expected = 0.25,
  k_groups = 3,
  decision_rate = 0.9
)

# Prevent overwriting an existing file:
plan_sample_size(0.25, 3, overwrite = FALSE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CLASS_seq_ttest_results.R
\docType{class}
\name{seq_ttest_results-class}
\alias{seq_ttest_results-class}
\title{An S4 class to represent the results of a sequential t-test.}
\description{
An S4 class to represent the results of a sequential t-test.
}
\section{Slots}{

\describe{
\item{\code{likelihood_ratio_log}}{the logarithmic test statistic.}

\item{\code{decision}}{the test decision: "accept H1", "accept H0",
or "continue sampling".}

\item{\code{A_boundary_log}}{the lower logarithmic boundary of the test.}

\item{\code{B_boundary_log}}{the upper logarithmic boundary of the test.}

\item{\code{d}}{a number indicating the specified effect size (Cohen's d).}

\item{\code{mu}}{a number indicating the true value of the mean
(or difference in means if you are performing a two sample test).}

\item{\code{alpha}}{the type I error. A number between 0 and 1.}

\item{\code{power}}{1 - beta (beta is the type II error probability). A number
between 0 and 1.}

\item{\code{likelihood_ratio}}{the likelihood ratio of the test without logarithm.}

\item{\code{likelihood_1}}{the likelihood of the alternative Hypothesis (H1).}

\item{\code{likelihood_0}}{the likelihood of the null Hypothesis (H0).}

\item{\code{likelihood_1_log}}{the logarithmic likelihood of the alternative
Hypothesis (H1).}

\item{\code{likelihood_0_log}}{the logarithmic likelihood of the null
Hypothesis (H0).}

\item{\code{non_centrality_parameter}}{parameter to calculate the likelihoods}

\item{\code{t_value}}{the t-value of the t-statistic.}

\item{\code{p_value}}{the p-value of the t-test.}

\item{\code{df}}{degrees of freedom.}

\item{\code{mean_estimate}}{the estimated mean or difference in means depending on
whether it was a one-sample test or a two-sample test.}

\item{\code{alternative}}{a character string specifying the alternative hypothesis:
"two.sided" (default), "greater" or "less".}

\item{\code{one_sample}}{"true" if it is a one-sample test, "false" if it is a
two-sample test.}

\item{\code{ttest_method}}{a character string indicating what type of t-test was
performed.}

\item{\code{data_name}}{a character string giving the name(s) of the data.}

\item{\code{verbose}}{a logical value whether you want a verbose output or not.}
}}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/draw_sample_normal.R
\name{draw_sample_normal}
\alias{draw_sample_normal}
\title{Draw Samples from a Normal Distribution}
\usage{
draw_sample_normal(k_groups, f, max_n, sd = NULL, sample_ratio = NULL)
}
\arguments{
\item{k_groups}{number of groups (levels of factor_A)}

\item{f}{Cohen's f. The simulated effect size.}

\item{max_n}{sample size for the groups (total sample size = max_n*k_groups)}

\item{sd}{vector of standard deviations of the groups. Default value is 1 for each group.}

\item{sample_ratio}{vector of sample ratios between th groups. Default value is 1 for each group.}
}
\value{
returns a data.frame with the columns y (observations) and x (factor_A).
}
\description{
Draws exemplary samples with a certain effect size for the sequential one-oway ANOVA or the sequential t-test, see Steinhilber et al. (2023) \url{doi:10.31234/osf.io/m64ne}
}
\examples{
set.seed(333)

data <- sprtt::draw_sample_normal(
  k_groups = 2,
  f = 0.20,
  max_n = 2
)
data

data <- sprtt::draw_sample_normal(
  k_groups = 4,
  f = 0,
  max_n = 2,
  sd = c(1, 2, 1, 8)
)
data

data <- sprtt::draw_sample_normal(
  k_groups = 3,
  f = 0.40,
  max_n = 2,
  sd = c(1, 0.8, 1),
  sample_ratio = c(1, 2, 3)
)
data
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seq_anova.R
\name{seq_anova}
\alias{seq_anova}
\title{Sequential Analysis of Variance}
\usage{
seq_anova(
  formula,
  f,
  alpha = 0.05,
  power = 0.95,
  data,
  verbose = TRUE,
  plot = FALSE,
  seq_steps = "single"
)
}
\arguments{
\item{formula}{A formula specifying the model (e.g., \code{outcome ~ group}).
The response variable should be on the left side and the grouping factor
on the right side. Currently only supports one-way designs.}

\item{f}{Cohen's f (expected minimal effect size or effect size of interest), that defines the H1.}

\item{alpha}{Type I error rate (alpha level). The probability of rejecting H0
when it is true. Default is 0.05. Must be between 0 and 1.}

\item{power}{Statistical power (1 - beta), where beta is the Type II error rate.
The probability of correctly rejecting H0 when H1 is true with effect size f.
Default is 0.95. Must be between 0 and 1. Higher values lead to wider
stopping boundaries and potentially larger sample sizes.}

\item{data}{A data frame containing the variables specified in the formula.
Missing values (NA) will be removed with a warning}

\item{verbose}{a logical value whether you want a verbose output or not.}

\item{plot}{Logical. If \code{TRUE}, stores sequential test statistics at each sequential step
for visualization with \code{plot_anova()}. This enables retrospective examination
of the decision process but increases computation time for large datasets.
Default is \code{FALSE}.}

\item{seq_steps}{Specifies when to calculate test statistics during sequential
testing (only relevant when \code{plot = TRUE}). Options:
\itemize{
\item Vector of integers: Calculate at specific sample sizes (e.g., \code{c(10, 20, 30)})
\item \code{"single"}: Calculate after each observation (step size = 1).
Note: Starts at \code{k_groups * 2} to ensure sufficient data
\item \code{"balanced"}: Calculate when each group gains one observation
(step size = number of groups). Note: Starts at \code{k_groups * 2}
}

For unbalanced designs or non-standard sequences, specify custom steps as a vector.}
}
\value{
An object of the S4 class \code{\linkS4class{seq_anova_results}}. Click on the
class link to see the full description of the slots.
To get access to the object use the
\code{@}-operator or \verb{[]}-brackets instead of \code{$}.
See the examples below.
}
\description{
Performs a sequential one-way fixed effects ANOVA, which is a variant of a Sequential
Probability Ratio Test (SPRT). The test allows for continuous
monitoring of data collection and provides stopping boundaries based on
likelihood ratios, offering efficiency gains over traditional fixed-N designs.

The sequential ANOVA continuously evaluates the likelihood ratio after each
observation (or group of observations), stopping when sufficient evidence
accumulates for either H0 or H1.

For methodological details, see Steinhilber et al. (2024)
\url{doi:10.1037/met0000677}. For practical guidance, see
\code{vignette("one_way_anova", package = "sprtt")}.
}
\section{Limitations}{

\itemize{
\item Only one-way fixed effects ANOVA is currently supported
\item Repeated measures ANOVA is not yet implemented
}
}

\examples{
# simulate data ----------------------------------------------------------------
set.seed(333)
data <- sprtt::draw_sample_normal(k_groups = 3,
                    f = 0.25,
                    sd = c(1, 1, 1),
                    max_n = 50)


# calculate sequential ANOVA ---------------------------------------------------
results <- sprtt::seq_anova(y ~ x, f = 0.25, data = data)
# test decision
results@decision
# test results
results

# calculate sequential ANOVA ---------------------------------------------------
results <- sprtt::seq_anova(y ~ x,
                            f = 0.25,
                            data = data,
                            alpha = 0.01,
                            power = .80,
                            verbose = TRUE)
results

# calculate sequential ANOVA ---------------------------------------------------
results <- sprtt::seq_anova(y ~ x,
                            f = 0.15,
                            data = data,
                            alpha = 0.05,
                            power = .80,
                            verbose = FALSE)
results
}
\seealso{
\itemize{
\item \code{\link[=plot_anova]{plot_anova()}} for visualizing sequential ANOVA results
\item \code{\link[=plan_sample_size]{plan_sample_size()}} for sample size planning
\item \code{\link[=seq_ttest]{seq_ttest()}} for sequential t-tests
\item \code{vignette("one_way_anova", package = "sprtt")} for detailed tutorial
\item \code{vignette("plan_sample_size", package = "sprtt")} for planning guidance
\item Steinhilber et al. (2023) for theoretical background
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CLASS_seq_anova_results.R
\docType{class}
\name{seq_anova_results-class}
\alias{seq_anova_results-class}
\title{An S4 class to represent the results of a sequential anova.}
\arguments{
\item{plot}{list with all arguments for the plot_anova() function}
}
\description{
An S4 class to represent the results of a sequential anova.
}
\section{Slots}{

\describe{
\item{\code{likelihood_ratio_log}}{the logarithmic test statistic.}

\item{\code{decision}}{the test decision: "accept H1", "accept H0",
or "continue sampling".}

\item{\code{A_boundary_log}}{the lower logarithmic boundary of the test.}

\item{\code{B_boundary_log}}{the upper logarithmic boundary of the test.}

\item{\code{f}}{a number indicating the specified effect size (Cohen's f).}

\item{\code{effect_sizes}}{a list with effect sizes (Cohen's f, eta squared, ...).}

\item{\code{alpha}}{the type I error. A number between 0 and 1.}

\item{\code{power}}{1 - beta (beta is the type II error probability). A number
between 0 and 1.}

\item{\code{likelihood_ratio}}{the likelihood ratio of the test without logarithm.}

\item{\code{likelihood_1}}{the likelihood of the alternative Hypothesis (H1).}

\item{\code{likelihood_0}}{the likelihood of the null Hypothesis (H0).}

\item{\code{likelihood_1_log}}{the logarithmic likelihood of the alternative
Hypothesis (H1).}

\item{\code{likelihood_0_log}}{the logarithmic likelihood of the null
Hypothesis (H0).}

\item{\code{non_centrality_parameter}}{parameter to calculate the likelihoods}

\item{\code{F_value}}{the F-value of the F-statistic.}

\item{\code{df_1}}{degrees of freedom.}

\item{\code{df_2}}{degrees of freedom.}

\item{\code{ss_effect}}{ss_effect.}

\item{\code{ss_residual}}{ss_residual.}

\item{\code{ss_total}}{ss_total.}

\item{\code{total_sample_size}}{total sample size.}

\item{\code{data_name}}{a character string giving the name(s) of the data.}

\item{\code{verbose}}{a logical value whether you want a verbose output or not.}
}}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_sprt.R
\name{plot_anova}
\alias{plot_anova}
\title{Plot Sequential ANOVA Results}
\usage{
plot_anova(
  anova_results,
  labels = TRUE,
  position_labels_x = 0.15,
  position_labels_y = 0.1,
  position_lr_x = NULL,
  position_lr_y = NULL,
  font_size = 15,
  line_size = 1,
  highlight_color = "#CD2626"
)
}
\arguments{
\item{anova_results}{A \code{seq_anova_results} object from \code{\link[=seq_anova]{seq_anova()}}.
\strong{Important:} The \code{seq_anova()} function must be called with \code{plot = TRUE}
to generate the necessary data for plotting.}

\item{labels}{Logical. If \code{TRUE} (default), display decision labels
("Accept H0" / "Accept H1") and the likelihood ratio at the decision point.}

\item{position_labels_x}{Numeric value between 0 and 1 controlling the
horizontal position of decision labels as a proportion of maximum sample
size. Default is \code{0.15} (left side); \code{0.5} centers the labels.}

\item{position_labels_y}{Numeric value controlling the vertical spacing
between decision boundaries and their labels. The value is multiplied by
\verb{max(|log-likelihood ratio|)} to determine spacing. Larger values move
labels further from boundaries. Default is \code{0.075}.}

\item{position_lr_x}{Optional numeric value for the x-coordinate (sample size)
of the likelihood ratio label. If \code{NULL} (default), positioned at the
decision point or final sample size.}

\item{position_lr_y}{Optional numeric value for the y-coordinate
(log-likelihood ratio) of the likelihood ratio label. If \code{NULL} (default),
positioned at \code{y = 0} for early decisions, or slightly offset for
continuing sampling scenarios.}

\item{font_size}{Numeric. Base font size for plot text. Default is \code{20}.}

\item{line_size}{Numeric. Line width for the trajectory and boundaries.
Default is \code{1.5}.}

\item{highlight_color}{Character string. Color for highlighting the decision
point or final sample. Default is \code{"#CD2626"} (red).}
}
\value{
A \code{\link[ggplot2:ggplot]{ggplot2::ggplot()}} object showing:
\itemize{
\item Log-likelihood ratio trajectory across sample sizes
\item Dashed horizontal lines indicating decision boundaries
\item Highlighted point showing where decision was reached (or final sample)
\item Optional labels for decision regions and likelihood ratio value
}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Creates a visualization of the sequential probability ratio test (SPRT) for
ANOVA results, showing the log-likelihood ratio trajectory across sample sizes
and decision boundaries.
}
\examples{
# simulate data for the example ------------------------------------------------
set.seed(3)
data <- sprtt::draw_sample_normal(3, f = 0.25, max_n = 50)

# calculate the SPRT -----------------------------------------------------------
anova_results <- sprtt::seq_anova(y~x, f = 0.25, data = data, plot = TRUE)

# plot the results -------------------------------------------------------------
# default settings
sprtt::plot_anova(anova_results)
# variant 1
sprtt::plot_anova(anova_results,
                 labels = TRUE,
                 position_labels_x = 0.05,
                 position_lr_x = 150,
                 position_lr_y = 0,
                 highlight_color = "green"
                 )
# variant 2
sprtt::plot_anova(anova_results,
                  labels = TRUE,
                  position_labels_x = 0.15,
                  position_labels_y = 0.2,
                  position_lr_x = 60,
                  position_lr_y = 1,
                  font_size = 25,
                  line_size = 2,
                  highlight_color = "darkred"
)
# no labels
sprtt::plot_anova(anova_results,
                 labels = FALSE
                 )
# custom additions
sprtt::plot_anova(anova_results) +
  ggplot2::geom_vline(xintercept = 66, linewidth = 1, linetype = "dashed")

# further information ----------------------------------------------------------
# run this code:
vignette("one_way_anova", package = "sprtt")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_data_external.R
\name{get_sprtt_cache_dir}
\alias{get_sprtt_cache_dir}
\title{Get path to cached simulation data directory}
\usage{
get_sprtt_cache_dir()
}
\description{
Get path to cached simulation data directory
}
\keyword{internal}
